/**
 * Copyright (c) 2013 - 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implements autocomplete functionality including optional suggestions
 * via simple array, array or objects, restapi or custom functions.
 *
 * @author Joe McClintock
 */

/**
 * The luxautocompleteinput widget
 * @module luxautocompleteinput
 * @jqbase
 * @requires ojet
 * @requires knockout
 * @requires jquery
 * @requires jqueryui
 * @requires underscore
 * @requires luxautocomplete
 */
define(['ojs/ojcore',
    'knockout',
    'jquery',
    'underscore',
    'ojs/ojinputtext',
    'ojs/ojmenu',
    'ojs/ojselectcombobox',
    'luxautocomplete',
    'jqueryui-amd/widget'],
function(oj, ko, $, _) {

    'use strict';
    (function() {
        $.widget('GBU.luxautocompleteinput', {
            version:           '5.1.0',
            defaultElement:    '<div>',
            widgetEventPrefix: 'lux',
            /**
             * Options
             * <dl>
             *     <dt>cssClass</dt>
             *     <dd>
             *         CSS class applied to the input field wrapper element. <br />
             *         Default: 'autocompleteInput'<br />
             *     </dd>
             *     <dt>textInputId</dt>
             *     <dd>
             *         Id value applied to the text input element. (does not typically need to be overridden)<br />
             *         Type: String<br />
             *         Default: 'lux-autocomplete-inputId'<br />
             *         Example:<br />
             *         <code>textInputId: 'my-autocomplete-input'</code>
             *     </dd>
             *     <dt>autocompleteInputLabelId</dt>
             *     <dd>
             *         Id value applied to the text input label element. (does not typically need to be overridden)<br />
             *         Type: String<br />
             *         Default: 'lux-autocomplete-input-label'<br />
             *         Example:<br />
             *         <code>autocompleteInputLabelId: 'my_input_label'</code>
             *     </dd>
             *     <dt>autocompleteFieldLabel</dt>
             *     <dd>
             *         Text input element label<br />
             *         Type: String<br />
             *         Default: 'Input: '<br />
             *         Example:<br />
             *         <code>autocompleteFieldLabel: 'Text input: '</code>
             *     </dd>
             *     <dt>placeholderText</dt>
             *     <dd>
             *         Optional placeholder text inside the text box.<br />
             *         Type: String<br />
             *         Default: null<br />
             *         Example:<br />
             *         <code>placeholderText: 'Enter your term here'</code>
             *     </dd>
             *     <dt>suggestionScopeTitle</dt>
             *     <dd>
             *         Text for category selection drop down title.
             *         Type: String<br />
             *         Default: 'Select suggestion scope'<br />
             *         Example:<br />
             *         <code>suggestionScopeTitle: 'Your suggestions'</code>
             *     </dd>
             *     <dt>searchCaseSensitive</dt>
             *     <dd>
             *         Should searches and suggestions be case-sensitive?<br />
             *         Default: false<br />
             *         Applies only when literal array is provided as data source. Server-side or custom function determines this in the other cases
             *     </dd>
             *     <dt>suggestionsEnabled</dt>
             *     <dd>
             *         Use suggestions from drop down list to narrow the auto-complete<br />
             *         Default: false<br />
             *         NOTE: only works if categories/sources supplied for suggestions source.
             *     </dd>
             *     <dt>suggestionsLimit</dt>
             *     <dd>
             *         Maximum number of suggestions shown.<br />
             *         Default: 15<br />
             *         Note: limit is across all sources, not per suggestion source
             *     </dd>
             *     <dt>suggestionsSeparateByCategory</dt>
             *     <dd>
             *         List suggestions separately by category?<br />
             *         Default: true<br />
             *         When set to false, suggestions are combined according to sort logic
             *     </dd>
             *     <dt>suggestionsSortOnClient</dt>
             *     <dd>
             *         Alphabetically sort the suggestions on the client after they come back from server?<br />
             *         Default: false<br />
             *         (assume server knows more about what's most relevant/important)
             *     </dd>
             *     <dt>suggestionsEnableDirectLink</dt>
             *     <dd>
             *         Can the user navigate directly to a suggested item?<br />
             *         Default: true<br />
             *         Direct navigation depends on having suggestion source(s) that populate the url property of the suggestion
             *         When set to false, selecting the suggestion simply places its value in the search box
             *     </dd>
             *     <dt>suggestionsSources</dt>
             *     <dd>
             *         Array of suggestion sources and which categories they supply.
             *         Categories drop down is populated by the total unique list.
             *         Different types of sources can be mixed together in the array
             *         Each source can have the following properties:<br />
    <pre class="prettyprint" style="overflow-x:visible;">
    format 1:
    value                   : property name holding the value
    label                   : property name holding the display label
    categories              : boolean, indicates node is a category.
    children                : array, contains value, label attributes and can have category and children attributes

    format 2:
    categories (required)   : string array
    source (required)       : array, function, or REST url (string). See code examples below.
    sourceParam             : function to invoke to generate a string to append to end of the REST url based on the search values/category
    responseMap             : used to map JSON responses to search suggestions when the REST service does not return items with value and
                                     label properties
       collection (optional): top-level property of the JSON response that holds the collection of suggestions
       value                : string property name holding the value. Or, function to execute on the JSON item to generate it.
       label                : string property name holding the display label. Or, function to execute on the JSON item to generate it.
       url (optional)       : string property name holding the url for navigating to this searches suggestion. Or, function to execute
                                on the JSON item to generate the value.
       icon (optional)      : string property name holding the url of the icon for this suggestion. Or, function to execute on the
                                JSON item to generate the value.
       category (optional)  : string property name holding the category of this suggestion. Or, function to execute on the JSON
                                 item to generate the value.
    </pre>
             *     <dt>Inline</dt>
             *     <dd>
             *         Lines up the form fields vertically &quote;true&quote; or horizontally &quote;false&quote;.
             *         Default is true
             *     </dd>
             *     <dt>multipleSuggestions</dt>
             *     <dd>
             *         Allows for selection of multiple suggestion.<br />
             *         Default: false
             *     </dd>
             *     <dt>categoryOptionChange</dt>
             *     <dd>
             *         Developer supplied callback function that takes as input the value from category drop down<br />
             *         Default: null<br />
             *         Example: <br />
        <pre class="prettyprint" style="overflow-x:visible;">
                    categoryOptionChange: function(event, data)
                            console.log('category selected: ' + data.value);
                    }
        </pre>
             *     </dd>
             *     <dt>inputChangeCallback</dt>
             *     <dd>
             *         Developer supplied callback function that takes as input the value from input field<br />
             *         Default: null<br />
             *         Example: <br />
        <pre class="prettyprint" style="overflow-x:visible;">
                    inputChangeCallback: function(event, data)
                            console.log('input value: ' + data.value);
                    }
        </pre>
             *     </dd>
             * </dl>
             * @example
             * <caption>Arrays suggestion source, format 1</caption>
                DeepSearchCategories:[{
                    value: "Birds",
                    label: "Birds",
                    categories: true,
                        children: [{
                            value: "Eagle",
                            label: "Eagle"
                        }, {
                            value: "Heron",
                            label: "Heron"
                        }, {
                            value: "Robin",
                            label: "Robin"
                        }, {
                            value: "Mockingbird",
                            label: "Mockingbird"
                        }, {
                            value: "Bluejay",
                            label: "Bluejay"
                        }, {
                            value: "Hawk",
                            label: "Hawk"
                        }]
                    }, {
                    value: "Fish",
                    label: "Fish",
                    categories: true,
                    children: [{
                        value: "Bass",
                        label: "Bass"
                    }, {
                        value: "Pike",
                        label: "Pike"
                    }, {
                        value: "Trout",
                        label: "Trout"
                    }, {
                        value: "Salmon",
                        label: "Salmon"
                    }, {
                        value: "Bluegill",
                        label: "Bluegill"
                    }, {
                        value: "Halibut",
                        label: "Halibut"
                    }, ]
                }],

             * @example
             * <caption>Arrays suggestion source, form 2</caption>
             * [ { categories: ['Birds'], source: ['Eagle', 'Heron', 'Robin', 'Mockingbird', 'Bluejay', 'Hawk'] },
             * { categories: ['Fish'],  source: ['Bass', 'Pike', 'Trout', 'Salmon', 'Bluegill', 'Halibut' ] } ]
             *
             * @example
             * <caption>Example using REST source</caption>
             * [ {
             *   categories: ['Sponsors'],
             *   source: 'webresources/entities.sponsor',
             *   sourceParam: function(req) { return '?beginsWith=' + req.term; },
             * } ]
             *
             * @example
             * <caption>using REST source with complex response mapping</caption>
             * [ {
             *   categories: ['Cities'],
             *   source: 'http://ws.geonames.org/searchJSON',
             *   sourceParam: function(req) { return '?featureClass=P&style=full&maxRows=10&name_startsWith=' + req.term; },
             *   responseMap: {
             *      collection: 'geonames',
             *      value: 'name',
             *      label: function(item) { return item.name + (item.adminName1 ? ", " + item.adminName1 : "") + ", " + item.countryName; },
             *      url: function(item) { return 'https://www.google.com/search?q=' + item.name; },
             *      icon: function(item) { return 'styles/img/confirmation.png?' + item.name; }
             *      category: function(item) { return item.countryName; }
             *   }
             * } ]
             *
             * @example
             * <caption>Example using custom function</caption>
             *  [ {
             *    categories: ['My Category'],
             *    source: function(request, response) {
             *        var arr = [
             *            {value: 'One', label: 'One' },
             *            {value: 'Two', label: 'Two' },
             *            {value: 'Demo', label: 'Demo Custom' }
             *        ];
             *        response($.grep(arr, function (s) { return s.value.indexOf(request.term) === 0; }));
             *    }
             *  } ]
             */
            options:           {
                cssClass:                       'autocompleteInput',
                autocompleteFieldLabel:         null,
                textInputId:                    'lux-autocomplete-inputId',
                textInputTitle:                 'Text Input',
                autocompleteInputLabelId:       'lux-autocomplete-input-label',
                searchCategoriesId:             'lux-autocomplete-categories',
                inline:                         true,
                suggestAllCategories:           'All',
                searchCaseSensitive:            false,
                suggestionsEnabled:             false,
                suggestionsEnableDirectLink:    true,
                suggestionsLimit:               15,
                suggestionsSeparateByCategory:  true,
                suggestionsSortOnClient:        false,
                suggestionsSources:             [],
                suggestionsTimeoutMilliseconds: 1500,
                suggestionScopeTitle:           'Select suggestion scope',
                multipleSuggestions:            false,
                categoryOptionChange:           null,
                placeholderText:                null,
                formContainer:                  null,
                inputChangeCallback:            null,
                inputContainer:                 null,
                labelEl:                        null,
                resultsModel:                   null,
                cssRow:                         'oj-flex',
                cssCol:                         'oj-flex-item'
            },
            _allCategories:      [],
            _observablesMap:     {},
            _includeSuggestions: false,

            _create: function() {
                if (this.options.formContainer) {
                    this._containingForm = this.options.formContainer;
                    this._refresh();
                } else {
                    this._containingForm = $(document).find('.' + this.options.cssClass);
                    if (this._containingForm.length === 0) {
                        this._containingForm = $('<div/>', {'class': this.options.cssClass}).appendTo(this.element);
                    }
                    this._containingForm.addClass('oj-form oj-md-odd-cols-2 oj-md-labels-inline lux-autocomplete');
                    this._refresh();
                }
            },

            _refresh: function() {
                var self = this,
                    allText =  {label: self.options.suggestAllCategories, value: self.options.suggestAllCategories},
                    inputContainer = this.options.inputContainer ? this.options.inputContainer :
                        $(document.createElement('div')).addClass(this.options.cssCol)
                            .attr('aria-labelledby', self.options.autocompleteInputLabelId)
                            .attr('role', 'group'),
                    labelEl = this.options.labelEl ? this.options.labelEl :
                        $('<label/>', {'for': self.options.textInputId, text: this.options.autocompleteFieldLabel})
                            .attr('id', self.options.autocompleteInputLabelId),
                    suggestionElement = this.options.multipleSuggestions ? 'ojCombobox' : 'ojSelect',
                    suggestionTag = this.options.multipleSuggestions ? '<input/>' : '<select/>';

                this._resultsModel = this.options.resultsModel ? this.options.resultsModel : null;

                this.suggestionElement = suggestionElement;

                this._observablesMap.autocompleteFieldLabel = labelEl;

                // determine if its a simple autocomplete or not. suggestionsEnable create a select drop down
                // for the autocomplete to search from
                self._includeSuggestions = (this.options.suggestionsEnabled && this.options.suggestionsSources) ? true : false;

                if (self._includeSuggestions) {
                    this._preProcessCategories();
                    labelEl.attr('for', this.options.searchCategoriesId);
                }

                // display label if specified
                if (this.options.autocompleteFieldLabel) {
                    // if inline is specified (default):
                    if (!this.options.inline) {
                        // create row div
                        this._containingForm = $('<div />', {'class': this.options.cssRow}).appendTo(this._containingForm);
                        // create row div, add the row class to the div, append it the containing form div
                        $('<div />', {'class': this.options.cssCol}).append(labelEl.addClass('oj-label-inline')).appendTo(this._containingForm);
                        // now we have it all setup so ojet can add manipulate the label and wrap it's markup around the label.
                    } else {
                        // otherwise: append the label element directly to the form div
                        $('<div/>', {'class': 'lux-autocomplete-vertical-label'}).append(labelEl).appendTo(this._containingForm);
                        // remove the col related classes from the divs
                        inputContainer.removeClass(this.options.cssCol);
                        this._containingForm.removeClass('oj-md-odd-cols-2 oj-md-labels-inline');
                    }
                }

                if (this._allCategories && this._allCategories.length > 1) {
                    this._categoriesElement = $(suggestionTag).attr('id', self.options.searchCategoriesId);
                    this._categoriesElement.appendTo(inputContainer);
                    if (!this.options.multipleSuggestions) {
                        this._allCategories.unshift(allText); // add the "All" option.
                    }
                    this._categoriesElement[suggestionElement]({
                        multiple:     this.options.multipleSuggestions,
                        options:      this._allCategories,
                        value:        '',
                        optionChange: this.options.categoryOptionChange,
                        title:        self.options.suggestionScopeTitle
                    });

                    this._observablesMap.suggestionScopeTitle = this._categoriesElement;
                    this._observablesMap.suggestAllCategories = this._categoriesElement;

                    this._categoriesElement.on({'ojoptionchange': function(event, data) {
                        if (data.option === 'value') {
                            self._selectCategory = data.value[0];
                            if (self._resultsModel) {
                                self._resultsModel.set({'category': data.value[0]});
                            }
                        }
                    }
                    });

                    if (!this.options.multipleSuggestions) {
                        // force a option change event so the view gets default option.
                        this._categoriesElement[suggestionElement]('option', 'value', allText.value);
                    }
                }

                // if not horizontal break so the form fields are stacked.
                if (this.options.inline) {
                    $('<br/>').appendTo(inputContainer);
                }

                this._inputElement = $('<input/>').attr('id', self.options.textInputId);
                if (this.options.autocompleteFieldLabel) {
                    this._inputElement.attr('aria-label', this.options.autocompleteFieldLabel);
                }
                inputContainer.append(this._inputElement);
                this._inputElement.ojInputText({
                    type:         'text',
                    title:        this.options.textInputTitle,
                    placeholder:  this.options.placeholderText,
                    optionChange: this.options.inputChangeCallback
                });

                if (self._resultsModel) {
                    this._inputElement.on({'ojoptionchange': function(event, data) {
                        if (data.option === 'value') {
                            self._resultsModel.set({'term': data.value});
                        }
                    }
                    });
                }

                this._observablesMap.placeholderText = this._inputElement;
                this._observablesMap.textInputTitle = this._inputElement;

                if (self._includeSuggestions) {
                    this._inputElement.luxautocomplete({
                        includeCategory: this.options.suggestionsSeparateByCategory,
                        source:          _.bind(self.getSuggestions, self),
                        open:            function() {
                            self._inputElement.luxautocomplete('widget');
                        },
                        focus: function(e, ui) {
                            self._inputElement.val(ui.item.label);
                            self._inputElement.ojInputText('option', 'value', ui.item.label);
                            return false;
                        },
                        select: function(e, ui) {
                            self._inputElement.val(ui.item.label);
                            if (self.options.suggestionsEnableDirectLink && ui.item.url) {
                                if ($.isFunction(ui.item.url)) {
                                    ui.item.url();
                                } else {
                                    window.location = ui.item.url;
                                }
                            }
                            return false;
                        }
                    });
                } else {
                    this._inputElement.luxautocomplete({source: self.options.suggestionsSources});
                }

                inputContainer.appendTo(this._containingForm);

                self._inputContainer = inputContainer;
            },

            _preProcessCategories: function() {
                var mapFuncs,
                    sources = this.options.suggestionsSources,
                    sourceLength = sources.length,
                    legacyCompatibility = false,
                    categoryFunctionBuilder = function(categoryName) {
                        return function() {
                            return categoryName;
                        };
                    },
                    restSourceMapper = function(val, source, mapFuncs) {
                        if (!mapFuncs[val]) {
                            // default is to use myjsonItem.value, myjsonItem.label, myjsonItem.url,
                            mapFuncs[val] = function(item) {
                                return item[val];
                            };
                        }
                        // but if they provided a map for this property,
                        if (source.responseMap && source.responseMap[val]) {
                            // assume it's a function
                            mapFuncs[val] = source.responseMap[val];
                            // but if it's a string, build a function based on that string key
                            if ($.type(source.responseMap[val]) === 'string') {
                                mapFuncs[val] = function(key) {
                                    return function(item) {
                                        return item[key];
                                    };
                                }(source.responseMap[val]);
                            }
                        }
                    },
                    buildSuggestedCategories = function(sources, categories, parent) {
                        for (var i = 0; i < sources.length; i += 1) {
                            var category,
                                source = sources[i];

                            if (source.categories) {
                                // make the value of each sub category unique by prepending it's
                                // parent hierarchy
                                category = {value: parent + '|' + source.value, label: source.label};
                                source.value = parent + '|' + source.value;

                                categories.push(category);
                                if (source.children) {
                                    category.children = [];
                                    buildSuggestedCategories(source.children, category.children, category.value);
                                }
                            }
                        }
                    },
                    mapSourceItems = function(source) {
                        return _.map(source.source, function(item) {
                            return {value: item, label: item};
                        });
                    },
                    buidlUrlParemeters = function(source, mapFuncs) {
                        var urlParams = ['value', 'label', 'url', 'category', 'description', 'icon'];

                        urlParams.forEach (
                            function(value) {
                                restSourceMapper(value, source, mapFuncs);
                            }
                        );
                    };

                this._allCategories = [];

                for (var i = 0; i < sourceLength; i += 1) {
                    var category = {},
                        source = sources[i];

                    // backward compatibility for ealier lux category input
                    if (source.categories && Array.isArray(source.categories)) {
                        if (source.categories.length === 1) {
                            legacyCompatibility = true;
                            category = {value: source.categories[0], label: source.categories[0], categories: true};
                            // push cloned object so we don't get the children in the category suggestions.
                            this._allCategories.push(_.clone(category));
                            // add children to category
                            category.children = mapSourceItems(source);
                            // append the category to suggestions
                            sources.push(category);
                        } else {
                            source.categories.forEach(function(cat) {
                                this._allCategories.push({value: cat, label: cat, categories: true});
                            }, this);
                        }
                    } else if (source.categories && typeof source.categories === 'boolean') {
                        // assumes the ojet pattern input for select and combobox.
                        category = {value: source.value, label: source.label};
                        this._allCategories.push(category);
                        if (source.children) {
                            category.children = [];
                            buildSuggestedCategories(source.children, category.children, category.value);
                        }
                    } else if (source.value && source.label) {
                        // otherwise simple list of suggestions using ojet input for select and combobox
                        category = {value: source.value, label: source.label};
                        this._allCategories.push(category);
                    }

                    // REST url
                    if ($.type(source.source) === 'string') {
                        legacyCompatibility = false;
                        // build functions to be executed to get label, value, etc
                        // based on the variations of mapping approaches we accept
                        // (e.g. string name of property or function for returning the right value)
                        mapFuncs = {
                            // default for category is to return the name we already have from API, not look for it in the JSON
                            category: categoryFunctionBuilder(source.categories[0])
                        };
                        buidlUrlParemeters(source, mapFuncs);
                        source.mapFuncs = mapFuncs;
                    }
                }

                if (legacyCompatibility) {
                    // if working on ealier lux source suggestions, need to
                    // transform it to work in this version.
                    sources = _.rest(sources, sourceLength);
                }
            },

            /**
             * Look up combined suggestions from all defined categories and sources
             *
             * @param {request} request should have a .term and optionally a .category property
             * @param {requestCallback} response callback for results. Will have .value and .label properties,
             *                          and sometimes .url, .category, and .description
             */
            getSuggestions: function(request, response) {
                var results = [],
                    callbackTracker,
                    source,
                    funcsToCall = [],
                    self = this,
                    searchString = request.caseSensitive ? request.term : request.term.toLocaleLowerCase(),
                    suggestedSources = this.options.suggestionsSources,
                    // only search sources that contain the selected category. Fall back to all sources if no matches
                    selectedCategory =  self._selectCategory ? self._selectCategory : 'All',
                    // drill down any sub categories looking for the requested
                    // category.
                    searchSources = function filterCategories(suggestions, filteredSuggestions) {
                        var sources = suggestions || suggestedSources,
                            sourcesLength = sources.length,
                            filtered = filteredSuggestions || [];

                        for (var x = 0; x < sourcesLength; x += 1) {
                            // is a parent with categories flag true and is selected parent category
                            if (sources[x].categories &&
                                sources[x].value && sources[x].value === selectedCategory) {
                                // push it in the filter array along with its children
                                filtered.push(sources[x]);
                            } else if (Array.isArray(sources[x].categories) &&
                                      _.contains(sources[x].categories, selectedCategory) &&
                                      sources[x].source) {
                                // legacy sources
                                filtered.push(sources[x]);
                            } else {
                                // otherwise check the chidren for sub categories
                                if (sources[x].children) {
                                    // drill down sub categories looking for a match
                                    // on the selected category
                                    filterCategories(sources[x].children, filtered);
                                }
                            }
                        }
                        return filtered;
                    }(), // invoke function to make filtered catigory items available for matching
                    pushResults = function(value, label, cateogory, isCategory) {
                        var s1 = self.options.searchCaseSensitive ? value : value.toLocaleLowerCase(),
                            foundPos = s1.indexOf(searchString);

                        // NOTE: checking that item is not a category (has categoies attribute) "!isCategory"
                        if ((foundPos === 0 || (foundPos !== -1 && !self.options.searchBeginsWith)) && !isCategory) {
                            results.push({value: value, label: label, category: cateogory});
                        }
                    },
                    searchSuggestions = function (source) {
                        var categoryVal = source.value;

                        for (var j = 0; j < source.children.length; j += 1) {

                            // found children in this category
                            if (source.children[j].categories && source.children[j].children) {
                                searchSuggestions(source.children[j]);
                            }

                            pushResults(source.children[j].value, source.children[j].label, categoryVal, source.children[j].categories);
                        }
                    },
                    successFunc = function(data) {
                        var source;

                        if (data) {
                            source = this;
                            // allow mapping of top-level JSON collection property
                            if (source.responseMap.collection) {
                                data = data[source.responseMap.collection];
                            }

                            $.each(data, function() {
                                var value = source.mapFuncs.value(this),
                                    label = source.mapFuncs.label(this);

                                if (value && label) {
                                    results.push({
                                        value:       value,
                                        label:       label,
                                        url:         source.mapFuncs.url(this),
                                        category:    source.mapFuncs.category(this),
                                        description: source.mapFuncs.description(this),
                                        icon:        source.mapFuncs.icon(this)
                                    });
                                }
                            });
                        }
                    },
                    callBackHandler = function(data) {
                        for (var i = 0; i < data.length; i += 1) {
                            if (!data[i].category) {
                                data[i].category = source.categories[0];
                            }
                            results.push(data[i]);
                        }
                        callbackTracker.resolve();
                    };

                if (!searchSources || searchSources.length === 0) {
                    selectedCategory = null;
                    searchSources = this.options.suggestionsSources;
                }

                // give the request object additional configuration/options for those sources which make use of it
                request = _.extend(request, {
                    category:      selectedCategory,
                    caseSensitive: this.options.searchCaseSensitive,
                    beginsWith:    this.options.searchBeginsWith
                });

                for (var i = 0; i < searchSources.length; i += 1) {
                    source = searchSources[i];

                    // literal array
                    if ($.isArray(source)) {
                        for (var j = 0; j < source.length; j += 1) {
                            if (source[j].value && source[j].lable) {
                                pushResults(source[j].value, source[j].lable, null);
                            }
                        }
                    } else if ($.type(source) === 'string') {
                        results.push(source);
                    } else if (source.categories && source.children) {
                        searchSuggestions(source);
                    } else if (source.value && source.label && !source.categories) {
                        pushResults(source.value, source.lable, null);
                    } else if ($.type(source.source) === 'string') {
                        var url = source.source;

                        if (source.sourceParam) {
                            url = url + source.sourceParam(request);
                        }

                        funcsToCall.push(
                            $.ajax({
                                dataType: 'json',
                                url:      url,
                                context:  source, // the "this" context in the success callback
                                timeout:  this.options.suggestionsTimeoutMilliseconds,
                                success:  successFunc
                            })
                        );

                    } else if ($.isFunction(source.source)) {
                        // when it takes a callback (like the jQueryUI autocomplete standard)
                        callbackTracker = new $.Deferred();

                        funcsToCall.push(callbackTracker);

                        source.source(request, _.bind(callBackHandler, this));
                    }
                }
                if (funcsToCall.length > 0) {
                    // resolve the promises after specified timeout so we don't hold up all
                    // suggestions due to a single misbehaving service note that those created
                    // through jQuery .ajax just use the timeout value
                    setTimeout(function() {
                        _.each(funcsToCall, function(f) {
                            if (f.resolve) {
                                f.resolve();
                            }
                        });
                    }, this.options.suggestionsTimeoutMilliseconds);

                    // wait for all asyc functions to have added their results,
                    $.when.apply(null, funcsToCall).always(function() {
                        // and only then fire the callback
                        response(self._postProcessResults(results));
                    });
                } else {
                    return response(self._postProcessResults(results));
                }
            },

            _sortResults: function(results) {
                if (!this.options.suggestionsSortOnClient || !results || !results.length) {
                    return results;
                }
                var that = this;
                return _.sortBy(results, function(d) {
                    var category = (that.options.suggestionsSeparateByCategory && d.category ? d.category : '');
                    return category + (d.label || '') + (d.value || '');
                });
            },

            _limitResults: function(results) {
                if (!results || !results.length) {
                    return results;
                }
                return results.slice(0, this.options.suggestionsLimit);
            },

            _postProcessResults: function(results) {
                return this._limitResults(this._sortResults(results));
            },

            _changeLabels: function(key, value) {
                var self = this,
                    allText,
                    attrKey = 'title';

                if (self._observablesMap[key]) {
                    if (_.contains(['placeholderText', 'autocompleteFieldLabel', 'suggestionScopeTitle', 'textInputTitle'], key)) {

                        if (key === 'placeholderText') {
                            attrKey = 'placeholder';
                        }

                        self._observablesMap[key].attr(attrKey, value);
                        self.options[key] = value;
                        if (key === 'autocompleteFieldLabel') {
                            self._observablesMap[key].attr('aria-label', value);
                            self._observablesMap[key].text(value);
                        } else if (key === 'suggestionScopeTitle') {
                            self._observablesMap[key][self.suggestionElement]('option', {title: value});
                        } else if (key === 'textInputTitle') {
                            self._observablesMap[key].ojInputText('option', {'title': value});
                        }
                    } else if (key === 'suggestAllCategories') {
                        if (!self.options.multipleSuggestions) {
                            allText = {label: value, value: value};
                            self._allCategories.shift();
                            self._allCategories.unshift(allText);
                            self._observablesMap[key].ojSelect('option', {options: self._allCategories});
                            self._observablesMap[key].ojSelect('reset');
                        }
                    } else {
                        self._observablesMap[key].text(value);
                        self.options[key] = value;
                    }
                    return true;
                }
                return false;
            },

            /**
             * Get/set the selected search category
             *
             * @this {luxsearch}
             * @param {String} [newValue] category which should be marked as selected
             * @returns {String} the currently selected category
             */
            selectedCategory: function(newValue) {
                var self = this;

                if (newValue) {
                    self._selectCategory = newValue;
                }
                return self._selectCategory;
            },

            /**
             * _setOption: evaluates the options and calls the
             *             appropriate function to handle change
             *             event.
             * @param {String} key   key of the changed option
             * @param {String} value new value of the option
             */
            _setOption: function(key, value) {
                var self = this;

                self._changeLabels(key, value);
            },

            /**
             * _setOptions: iterates over all changed options and passes
             *              key values to setOption for evaluation.
             *
             * @param {Object} options array of options
             */
            _setOptions: function(options) {
                var self = this;

                $.each(options, function(key, value) {
                    self._setOption(key, value);
                });
            },

            getInputContainer: function() {
                var self = this;

                return self._inputContainer;
            },

            categoriesElement: function() {
                var self = this;

                return self._categoriesElement;
            },

            /**
             * Get the input control for the search field
             */
            inputElement: function() {
                var self = this;

                return self._inputElement;
            },
            /**
             * @override
             * @private
             */
            _destroy: function() {
                // remove generated elements
                this._containingForm.remove();
            }
        });

    }());
});
