/**
 * Copyright (c) 2013 - 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * luxdymaicform is a jQuery widget that implements functionality to create forms and other HTML elements
 * dynamically through JSON configuration input. Uses dform (https://github.com/daffl/jquery.dform/) as the parse
 * and render the form from a JSON input configuration.
 * @author Joe McClintock
 */

/**
 * @module luxdynamicform
 * @jqbase
 * @requires jquery
 * @requires jqueryui-amd/widget'
 * @requires knockout
 * @requires underscore
 * @requires ojcore
 * @requires ojknockout
 * @requires dform
 * @requires dform.core
 * @requires dform.converters
 * @requires dform.extensions
*/
define(['ojs/ojcore',
    'knockout',
    'jquery',
    'underscore',
    'knockout-mapping',
    'luxdynamicformfields',
    'jqueryui-amd/widget',
    'dform',
    'dform.converters',
    'dform.core',
    'dform.extensions',
    'ojs/ojknockout'

], function(oj, ko, $, _, mapping) {
    'use strict';

    $.widget('GBU.luxdynamicform', {
        version:           '5.1.0',
        defaultElement:    '<div>',
        widgetEventPrefix: 'lux',
        /**
             * Options
             * <dl>
             *     <dt>input</dt>
             *     <dd>
             *         JSON input url form configuration<br />
             *         Type: JSON String<br />
             *         Example:<br />
             *         <pre><code>
             *             input: '/somerelativepath/input.json'
             *             or
             *             input: 'http://server/path/to/input.json'
             *             or
             *             input: '{... a json string ...}'
             *         </code></pre>
             *         <h3>configuration snippet for ojinputtext</h3>
             *         <pre class="prettyprint" style="overflow-x:visible;">
                        "type" : "input",
                        "name" : "username",
                        "id" : "username",
                        "title": "Please enter your email address as your username",
                        "data-bind": {
                            "ojComponent": {
                                "component": "ojInputText",
                                "placeholder": "E.g. user@example.com",
                                "optionChange": "username",
                                "required": true,
                                "validators":
                                [
                                    {
                                        "type": "regExp",
                                        "options": {
                                            "pattern": "(([^<>()[\\]\\.,;:\\s@\"]+(\\.[^<>()[\\]\\.,;:\\s@\"]+)*)|(\".+\"))@((\\[[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\])|(([a-zA-Z\\-0-9]+\\.)+[a-zA-Z]{2,}))",
                                            "hint": "enter e.g. user@example.com",
                                            "messageDetail": "You must enter a valid email address"
                                        }
                                    }
                                ]
                            },
                            "path": "ojs/ojinputtext",
                            "observable": "username"
                        }
             *          </pre>
             *
             *          <p>A callback can be associated the component using the &quot;callback&quot; attribute:</p>
             *          <pre class="prettyprint" style="overflow-x:visible;">

                            "type": "button",
                            "name": "saveButton",
                            "id": "users-save-button",
                            "title": "save",
                            "data-bind": {
                                "ojComponent": {
                                    "component": "ojButton",
                                    "label": "Create New Account"
                                },
                                "path": "ojs/ojbutton",
                                "observable": "save-button",
                                "callback": "users-save-button"
                            }
             *          </pre>
             *          <p>NOTE: the &quot;callback&quot; overrides the optionChange event handler associated the
             *          OJET components.</p>
             *     </dd>
             *     <dt>values</dt>
             *     <dd>
             *         <p>JSON input url for data to populate form field values<br />
             *         Type: JSON String</p>
             *         Example:<br />
             *         <pre><code>
             *             input: '/somerelativepath/path/data.json'
             *             or
             *             input: 'http://server/path/to/data.json'
             *         </code></pre>
             *         <p>Input consists of a JSON input structure of
             *         name value pairs where the name is the "observable"
             *         name defined in the form configuration JSON (the input
             *         parameter above) and the value to be assigned to the
             *         observable and the form field value.</p>
             *         <h3>configuration snippet for data value</h3>
             *         <pre class="prettyprint" style="overflow-x:visible;">
                        {
                            "users-hometown": "Port Orchard",
                            "users-address": "Roosevelt Way NE",
                            "users-city": "Seattle",
                            "users-zip": "98445",
                            "users-state": ["WA"],
                            "users-age": "52",
                            "users-range":
                                {
                                    "max": 700,
                                    "min": 0,
                                    "step": 5,
                                    "value": [200,645]
                                }
                        }
             *          </pre>
             *     </dd>
             *     <dt>manageFormFields</dt>
             *     <dd>
             *         Required form field manager, handles the two-way binding with
             *         knockout observables bound to each form field element <br />
             *         Type: function object<br />
             *         Example:<br />
             *         <code>manageFormFields: 'manageFormFields</code>
             *     </dd>
             *     <dt>onChangeCallback</dt>
             *     <dd>
             *         Exposes manageFormFields manager api default two-way binding.<br />
             *         Type: function<br />
             *         Default: 'Input: '<br />
             *         Example:<br />
             *         <code>onChangeCallback: onChangeCallback</code>
             *     </dd>
             *     <dt>callbacks</dt>
             *     <dd>
             *         Extends manageFormFields manager api, allows for more specific optionChange
             *         callbacks on ojet components and controls <br />
             *         Type: function<br />
             *         Example:<br />
             *         <code>callbacks: callbacks</code>
             *     </dd>
             *     <dt>validators</dt>
             *     <dd>
             *         Custom ojet validation functions.
             *         Type: javascript object<br />
             *         Example:<br />
             *         <code>validators: validators</code>
             *     </dd>
             *  </dl>
             */
        options:           {
            cssClassPrefix:   null,
            dateTimeValue:    null,
            input:            null,
            values:           null,
            onChangeCallback: null,
            callbacks:        {},
            validators:       null,
            manageFormFields: null
        },
        _config:           {},
        _valuesInput:      {},
        _customElements:   [],
        _componentList:    ['ojComponent', 'luxComponent'],
        _dataInput:        ko.observableArray(),
        _mapping:          {'copy': ['max', 'min', 'step', 'value']},
        _subscriptionKeys: ['data-bind', 'onclick-callback'],

        // NOTE: this list is a complete as of 10/01/2015
        _ariaList: ['aria-atomic',
            'aria-busy',
            'aria-controls',
            'aria-describedby',
            'aria-disabled',
            'aria-dropeffect',
            'aria-flowto',
            'aria-grabbed',
            'aria-haspopup',
            'aria-hidden',
            'aria-invalid',
            'aria-label',
            'aria-labelledby',
            'aria-live',
            'aria-owns',
            'aria-relevant',
            'aria-autocomplete',
            'aria-checked',
            'aria-disabled',
            'aria-expanded',
            'aria-haspopup',
            'aria-hidden',
            'aria-invalid',
            'aria-label',
            'aria-level',
            'aria-multiline',
            'aria-multiselectable',
            'aria-orientation',
            'aria-pressed',
            'aria-readonly',
            'aria-required',
            'aria-selected',
            'aria-sort',
            'aria-valuemax',
            'aria-valuemin',
            'aria-valuenow',
            'aria-valuetext'],

        _create: function() {
            this._customElements = [];

            this.id = Date.now();

            try {

                this._containingForm = this.element;

                if (!this.options.manageFormFields) {
                    throw 'missing required manageFormFields';
                }

                if (!this.options.input) {
                    oj.Logger.warn('missing input form configuration on create.');
                } else {
                    this._setup();
                }

            } catch (error) {
                oj.Logger.error(error);
            }
        },

        /**
         * Uses jquery promises and getJSON to request the form
         * configuration and optionally the form field values JSON. If
         * successful, the form is created and form field values are assigned.
         *
         * @param  {Function} callback - Optional callback used for unit tests
         */
        _setup: function(callback) {
            var success = function(data, values) {
                    this._config = data[0];
                    this._values = values || null;
                    // create a list of custom element types and
                    // subscriptions (if any)
                    this._getCustomFieldsFromConfig();

                    if (values) {
                        this._traverseDataValues(values[0]);
                    }

                    this._setUpCustomElements();
                    this._refresh();
                    if (callback && _.isFunction(callback)) {
                        callback({'config': this._config, 'values': this._values});
                    }
                }.bind(this),
                error = function(jqxhr, textStatus, error) {
                    var err = textStatus + ', ' + error;

                    throw 'error loading configuration: ' + err;
                },
                formConfigInput = null,
                valuesInput = null;

            // is the form configuration a JSON string or a URL
            if (this._isJson(this.options.input)) {
                // is string so create a JSON
                formConfigInput = [];
                formConfigInput.push(JSON.parse(this.options.input));
            } else {
                // is a URL so run a getJSON request.
                formConfigInput = $.getJSON(this.options.input);
            }

            // is the form values a JSON string or a URL
            if (this.options.values) {
                // is string so create a JSON
                if (this._isJson(this.options.values)) {
                    valuesInput = [];
                    valuesInput.push(JSON.parse(this.options.values));
                } else {
                    // is a URL so run a getJSON request.
                    valuesInput = $.getJSON(this.options.values);
                }
            }

            // will always get the form configuration input with optional form field values input
            $.when(formConfigInput, valuesInput)
                .done(success)
                .fail(error);
        },

        /**
         * Before dform renders, setup the custom for elements. This is accomplished using
         * dform subscriptions (see dform docs @ https://github.com/daffl/jquery.dform/).
         */
        _setUpCustomElements: function() {
            var addRadioOptions = function BuildRaidoOptions(options, container) {
                    _.each(options.options, function(name, value) {
                        var span = $('<span />', {'class': 'oj-choice-row'});

                        // check to see if radio option is already
                        // rendered.
                        if ($('#' + name + 'op').length > 0) {
                            return;
                        }

                        $('<input />', {
                            'id':    name + 'op',
                            'type':  'radio',
                            'name':  options.name, // name of the radio buttons
                            'value': value
                        }).appendTo(span);

                        $('<label />', {
                            'for':  name + 'op',
                            'text': name
                        }).appendTo(span);

                        span.appendTo(container);
                    });
                },
                addCheckBoxes = function BuildCheckBoxes(options, container) {
                    _.each(options.options, function(name, value) {
                        var span = $('<span />', {'class': 'oj-choice-row'});

                        // check to see if checkbox option is already
                        // rendered.
                        if ($('#' + value + 'op').length > 0) {
                            return;
                        }

                        $('<input />', {
                            'id':    value + 'op',
                            'type':  'checkbox',
                            'value': value
                        }).appendTo(span);

                        $('<label />', {
                            'for':  value + 'op',
                            'text': name
                        }).appendTo(span);

                        span.appendTo(container);
                    });
                },
                applyComponentLevelOptionsFromView = function applyComponentLevelOption(options, viewModel) {
                    _.each(options, function(option, key) {
                        if (_.isObject(option)) {
                            if (option.callback) {
                                options[key] = viewModel[option.callback];
                            } else if (option.observable) {
                                options[key] = ko.isObservable(viewModel[option.observable]) ?
                                    viewModel[option.observable]() : viewModel[option.observable];
                            }
                        }
                    }, this);
                }.bind(this),
                applyOptions = function ApplyWidgetOptions(el, options, custElement) {
                    var formFieldReference = options.observable || options.callback,
                        callbackReference = options.callback,
                        optionChange = options.optionChange,
                        inputObservable = this.options.manageFormFields && this.options.manageFormFields.read(formFieldReference),
                        formatedDate,
                        onClickEvent = function(callbackReference) {
                            el.on('click', function(event) {
                                this.options.callbacks[callbackReference](event);
                            }.bind(this)
                            );
                        }.bind(this),
                        // set options for the component. recursively iterates until the property option
                        // is a not a js object, then uses the "string" as index to the manageFormFields.observablesCallbacks
                        // object to get the component object options.
                        setOptions = function(options) {
                            _.each(options, function(value, name, list) {
                                if (_.isObject(value)) {
                                    setOptions(list[name]);
                                } else {
                                    if (this.options.manageFormFields && this.options.manageFormFields.observablesCallbacks[value]) {
                                        list[name] = this.options.manageFormFields.observablesCallbacks[value];
                                    }
                                }
                            }, this);
                        }.bind(this);

                    setOptions(options);
                    //**********************************************************************
                    // corner cases from OJET
                    // @NOTE if there are any more of these, should separate them into
                    // another js file.
                    // *********************************************************************
                    //
                    if (custElement === 'ojRadioset' && options.options) {
                        addRadioOptions(options, el);
                    } else if (custElement === 'ojCheckboxset' && options.options) {
                        addCheckBoxes(options, el);
                    } else if (custElement === 'ojInputDateTime') {
                        options.value = this.options.dateTimeValue ? this.options.dateTimeValue :
                            oj.IntlConverterUtils.dateToLocalIso(new Date());
                        if (inputObservable && inputObservable()) {
                            formatedDate = new Date(inputObservable());
                            // make sure the date is valid
                            if (isNaN(formatedDate.getTime())) {
                                // nope, display error message.
                                oj.Logger.error('invalid date: ', formFieldReference);
                            } else {
                                // good, use the date
                                options.value = oj.IntlConverterUtils.dateToLocalIso(formatedDate);
                                inputObservable(formatedDate);
                            }
                        }
                    }
                    //********************************************************************
                    // end OJET corner cases.
                    //********************************************************************

                    // If there is one callback on this individual form field
                    if (this.options.callbacks && formFieldReference && this.options.onChangeCallback) {
                        // reference the callback with the optionChange widget option for this individual form field
                        options.optionChange = this.options.onChangeCallback(formFieldReference);

                        // If there's a callback reference on the element, set the on click event to fire the callback
                        if (callbackReference && _.isFunction(this.options.callbacks[callbackReference])) {
                            onClickEvent(callbackReference);
                        }
                    }

                    // override the previously set optionChange function
                    // with a specific optionChange callback that's been
                    // specified in the configuration json.
                    if (this.options.callbacks && optionChange) {
                        options.optionChange = this.options.callbacks[optionChange];
                    }

                    if (this.options.validators && options.validators) {
                        options.validators.forEach(function(validator, index) {
                            if (_.isString(validator)) {
                                options.validators[index] = this.options.validators[validator];
                            }
                        }, this);
                    }
                }.bind(this),
                createCustomElement = function GetCustomElement(el, options, type) {
                    var formFieldReference = options.observable || options.callback,
                        parms = {
                            type:     'GET',
                            url:      null,
                            dataType: 'json',
                            data:     ''
                        },
                        path = options.path ? options.path : null,
                        viewModel = this.options.manageFormFields.getViewModel();

                    if (!path) {
                        oj.Logger.error('Missing path on custom element',
                            type +
                            (formFieldReference ? ' for field ' + formFieldReference + '.' : '.'),
                            'Cannot create custom element.');
                        return;
                    }

                    if (options.observable && this._dataInput[options.observable]) {

                        // get any options that were passed in with the input data json
                        _.each(this._dataInput[options.observable], function(item, key) {
                            options[key] = item();
                        }, this);

                        options.value = this.options.manageFormFields.read(options.observable)();
                    }

                    // If style attribute is specified in the options=, need
                    // to explicitly apply to the DOM element.
                    if (options.style) {
                        el.attr('style', options.style);
                    }

                    if (type === 'ojSlider' && options.slidertype) {
                        options.type = options.slidertype;
                        options.slidertype = null;
                    }

                    // apply viewmodel items to options that
                    // are not form field related, e.g., options that
                    // relate to a grid.
                    _.each(options, function(item) {
                        if (viewModel[item]) {
                            options[item] = viewModel[item];
                        }
                    });

                    // get each custom component / element widget and update the container DOM element
                    // with the "require" widget object
                    require(['ojs/ojcore', 'knockout', 'jquery', 'ojs/ojknockout', path],
                        function applyWidget(oj, ko, $) {

                            applyComponentLevelOptionsFromView(options, viewModel);
                            applyOptions(el, options, type);

                            if (options.options && options.options.url) {
                                parms.url = options.options.url;
                                $.ajax(parms).done(function(data){
                                    // get json data
                                    options.options = data;
                                    // create ojet component with data options
                                    el[type](options);
                                }).fail(function(jqXHR, textStatus, error) {
                                    var err = textStatus + ', ' + error;

                                    oj.Logger.error('error getting select options', err);
                                });
                            } else {
                                el[type](options);
                            }
                        }
                    );
                }.bind(this),
                createCustomDataBind = function(container, options, type) {
                    var vm, reference,
                        bindType = options.value ? 'value' : 'text',
                        bindVar = options.value || options.text,
                        bindVarRef = bindVar,
                        att = document.createAttribute('data-bind'),
                        isComponent = function findComponent() {
                            var componentKey = null;

                            this._componentList.forEach(function(key) {
                                if (options[key]) {
                                    componentKey = key;
                                }
                            }, this);
                            return componentKey;
                        }.bind(this),
                        componentKey = null;

                    // No Manager for form fields so just leave
                    if (!this.options.manageFormFields) {
                        return;
                    }

                    componentKey = isComponent();
                    if (componentKey) {
                        if (!options.path) {
                            oj.Logger.error('missing path to', type, 'in', componentKey);
                            return;
                        }
                        options[componentKey].path = options.path;
                        options[componentKey].observable = options.observable;
                        options[componentKey].callback = options.callback;
                        createCustomElement(container, options[componentKey], options[componentKey].component, type);
                        return;
                    }

                    // get the view model so we can dynamically add references
                    // to bind to the view
                    vm = this.options.manageFormFields.getViewModel();
                    if (options.references) {
                        reference = this.options.manageFormFields.read(options.references);

                        // The json configuration has not described the correct name of the referenced
                        // references so throw a error.
                        if (!reference) {
                            throw 'error binding ' + options.references + ', reference not found!' ;
                        }
                    }

                    // Are we processing elements in a references array?
                    if (bindVar.indexOf('()') > -1) {
                        bindVarRef = bindVar.substring(0, bindVarRef.indexOf('()'));
                        //Is the reference an Array? If not throw error!
                        if (!Array.isArray(reference())) {
                            throw 'error binding ' + options.references + ', is not a observablearray!' ;
                        }
                    } else if (!_.isNull(options.index) && !_.isUndefined(options.index)) {
                        bindVar = bindVar + '()[' + options.index + ']';
                    }

                    // dynamically create references in the view model so the view
                    // can bind to it.
                    if (!vm[bindVarRef]) {
                        vm[bindVarRef] = reference;
                    }

                    att.value = bindType + ': ' + bindVar;

                    container[0].setAttributeNode(att);
                    ko.applyBindings(vm, container[0]);
                }.bind(this),
                createCallback = function(container, options /*, type*/) {
                    var callbackReference = options.name;

                    if (callbackReference &&
                        this.options.manageFormFields &&
                        _.isFunction(this.options.callbacks[callbackReference])) {
                        container.change(function(event) {
                            this.options.callbacks[callbackReference](event);
                        }.bind(this)
                        );

                    }
                }.bind(this);

            // set default dform css class prefix.  if null class names
            // are not generated for each element.
            $.dform.options.prefix = this.options.cssClassPrefix;

            // build subscriptions for aria attributes
            this._ariaList.forEach(function(aria) {
                $.dform.subscribe(aria,
                    function(options) {
                        this.attr(aria, options);
                    }
                );
            });

            // subscription for role attribute
            $.dform.subscribe('role',
                function(options) {
                    this.attr('role', options);
                }
            );

            // subscription for any dform form element that is
            // not a custom widget.
            $.dform.subscribe('onChange',
                function(options, type) {
                    createCallback(this, options, type);
                }
            );

            // subscription for data-bind attribute
            $.dform.subscribe('data-bind',
                function(options, type) {
                    try {
                        createCustomDataBind(this, options, type);
                    } catch (error) {
                        oj.Logger.error(error);
                    }
                }
            );
        },

        _refresh: function() {
            this._containingForm.dform(this._config);
        },

        _destroy: function() {
            // Delete the form json input
            delete this.options.input;
            // Delete the ManageFormFields object so
            // we don't process it again along with
            // the associated view model
            delete this.options.manageFormFields;

            // remove the DOM element from the HTML
            this._containingForm[0].remove();
        },

        /**
         * Builds Array of data-binds.
         */
        _getCustomFieldsFromConfig: function() {
            var walkConfig = function buildCustomFieldList(data) {

                    _.each(data, function(value, key) {

                        if (_.isString(key)) {
                            if (_.contains(this._subscriptionKeys, key)) {
                                this._customElements.push({'type': key, 'options': value});
                            }
                        }
                        if (_.isObject(value)) {
                            walkConfig(value);
                        }
                    }, this);

                }.bind(this),
                optionType;

            walkConfig(this._config);

            if (this.options.manageFormFields) {

                this._customElements.forEach(function createObservable(element) {
                    optionType = element.options.callback || element.options.observable;

                    if (element.options.observable) {

                        _.each(element.options, function(option, key) {
                            if (key !== 'observable' && key !== 'callback' && key !== 'path') {
                                if (option.value) {
                                    this.options.manageFormFields.save(optionType, option.value);
                                } else {
                                    if (!this.options.manageFormFields.read(optionType)) {
                                        this.options.manageFormFields.save(optionType);
                                    } else {
                                        oj.Logger.error('duplicate observable or callback ',
                                            optionType, 'observable and callback name must be unique');
                                    }
                                }
                            }
                        }, this);
                    }
                }, this);
            }
        },

        /**
         * Processes input data separate from the form configuration. Creates a map
         * data input values that are applied to the value attribute of the referenced
         * form field element.
         *
         * @param  JSON inputData - data value input to custom form elements
         */
        _traverseDataValues: function(inputData) {
            var copyOnly = [],
                getInputValues = function(jsonObj, parentKey) {
                    _.each(jsonObj, function(value, key) {
                        // only apply an array or primitive values to
                        // form field observables.
                        if (Array.isArray(value) || !_.isObject(value)) {
                            // if there is a parent
                            if (parentKey) {
                                // if the key is "value", update the value for the parent observable.
                                if (key === 'value') {
                                    if (this.options.manageFormFields) {
                                        this.options.manageFormFields.save(parentKey, value);
                                        copyOnly.push(parentKey);
                                    }
                                }
                            } else {
                                if (this.options.manageFormFields) {
                                    this.options.manageFormFields.save(key, value);
                                    copyOnly.push(key);
                                }
                            }
                        }

                        if (_.isObject(value)) {
                            // recursively get any children attributes of the parent.
                            getInputValues(value, key);
                        }
                    }, this);
                }.bind(this);

            getInputValues(inputData);

            // save ko mapping for any data attributes that are not the initial value.
            // copy only, do not create observables, ignore "value", all ready created
            // an observable with the value of "value".
            this._mapping.copy = this._mapping.copy.concat(copyOnly);
            this._dataInput = mapping.fromJS(inputData, {'ignore': 'value'}, this._mapping);
        },

        _isJson: function(str) {
            try {
                JSON.parse(str);
            } catch (e) {
                return false;
            }
            return true;
        },

        getObservable: function(name) {
            return this.options.manageFormFields.read(name)();
        },

        getValuesData: function() {
            return this._values;
        },

        getFormConfiguration: function() {
            if (Object.getOwnPropertyNames(this._config).length === 0) {
                return null;
            }
            return this._config;
        },

        setFormConfiguration: function(options, callback) {

            try {
                if (!options) {
                    throw 'missing input form configuration';
                }

                this.options.input = options.input || null;
                this.options.values = options.values || null;

                if (!this.options.input) {
                    throw 'missing input form configuration';
                } else {
                    this._setup(callback);
                }
            } catch (error) {
                oj.Logger.error(error);
            }

        },

        getFormFieldManager: function() {
            return this.options.manageFormFields;
        }

    });
});