define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.moment',
    'gridapi.logger',
    'gridapi.grid/configuration',
    'gridapi.grid/utils'
], function ($, _, moment, logger, config, utils) {
    'use strict';

    var formatString = utils.formatString;
    var DefaultScale = function (options) {
        var that = this;
        var defaultOptions = {
            step : 1,
            value : {},
            format : {},
            range: {},
            bucketSize : 40,
            buffer : 5
        };
        this.options = _.merge({}, defaultOptions, options, function (a, b) {
            if (_.isArray(a)) {
                return b;
            }
        });
        var timescaleFns = this.options.timescaleFns;
    // override mediator functions from options
        _.forEach(timescaleFns, function (fnc, fncName) {
            that[fncName] = fnc || that[fncName];
        });

        if (this.options.setRange && !_.isEmpty(this.options.range)) {
            this.setRange(this.options.range);
        }
    };

    DefaultScale.prototype = {
        constructor : DefaultScale,

        getRange : function () {
            return this.options.range;
        },

        setRange : function (range) {
            this.options.range.start = moment(range.start)[config.startof](this.options.value.min).subtract(this.options.buffer, this.options.value.min);
            this.options.range.end = moment(range.end)[config.endof](this.options.value.min).add(this.options.buffer, this.options.value.min);
        },

        getBuckets : function (start, end) {
            start = start || this.options.range.start;
            end = end || this.options.range.end;

            return {
                min : this.getMinBuckets(start.clone(), end.clone()),
                mid : this.getMidBuckets(start.clone(), end.clone()),
                max : this.getMaxBuckets(start.clone(), end.clone())
            };
        },

        isAvailable : function (start, end) {
            if (this.options.isAvailable) {
                return this.options.isAvailable(start, end);
            }
        },

        getMinBuckets : function (start, end) {
            var buckets = [];
            if (this.options.value.min != null) {
                var bucketStart = moment(start)[config.startof](this.options.value.min);
                var bucketEnd = moment(bucketStart)[config.endof](this.options.value.min);
                while (!bucketStart.isAfter(end)) {
                    var bucket = {
                        label : this.getMinLabel(bucketStart, bucketEnd),
                        available : this.isAvailable(bucketStart, bucketEnd),
                        start : moment(bucketStart),
                        end : moment(bucketEnd),
                        width : this.options.bucketSize,
                        units : 1
                    };
                    buckets.push(bucket);
                    bucketStart = bucketEnd.add(1, this.options.value.min)[config.startof](this.options.value.min);
                    bucketEnd = moment(bucketStart)[config.endof](this.options.value.min);
                }
            }
            return buckets;
        },

        getMidBuckets : function (start, end) {
            var buckets = [];
            if (this.options.value.mid != null) {
                var bucketStart = moment(start)[config.startof](this.options.value.min);
                var bucketEnd = moment(bucketStart)[config.endof](this.options.value.mid).add(1, 'ms');

                while (!bucketStart.isAfter(end)) {
                    var units = bucketEnd.diff(bucketStart, this.options.value.min, true);
                    var bucket = {
                        label : this.getMidLabel(bucketStart, bucketEnd),
                        start : moment(bucketStart),
                        end : moment(bucketEnd),
                        units : units,
                        width : Math.floor(units * this.options.bucketSize)
                    };

                    bucketStart = bucketEnd.add(1, this.options.value.min)[config.startof](this.options.value.mid);
                    bucketEnd = moment(bucketStart)[config.endof](this.options.value.mid).add(1, 'ms');
                    buckets.push(bucket);
                }
            }
            return buckets;
        },

        getMaxBuckets : function (start, end) {
            var buckets = [];
            if (this.options.value.max != null) {
                var bucketStart = moment(start)[config.startof](this.options.value.min);
                var bucketEnd = moment(bucketStart)[config.endof](this.options.value.max).add(1, 'ms');

                while (!bucketStart.isAfter(end)) {
                    var units = bucketEnd.diff(bucketStart, this.options.value.min, true);
                    var bucket = {
                        label : this.getMaxLabel(bucketStart, bucketEnd),
                        start : moment(bucketStart),
                        end : moment(bucketEnd),
                        units : units,
                        width : Math.floor(units * this.options.bucketSize)
                    };

                    bucketStart = moment(bucketEnd).add(1, this.options.value.min)[config.startof](this.options.value.max);
                    bucketEnd = moment(bucketStart)[config.endof](this.options.value.max).add(1, 'ms');
                    buckets.push(bucket);
                }
            }
            return buckets;
        },


        getMinLabel : function (bucketStart, bucketEnd) {
            var minLabel = bucketStart.format(this.options.format.min);
            if (this.options.format.minRange) {
                minLabel += ' - ' + bucketEnd.format(this.options.format.min);
            } else if (this.options.format.minPrefix != null) {
                minLabel = formatString(this.options.format.minPrefix, minLabel);
            }
            return minLabel;
        },

        getMidLabel : function (bucketStart, bucketEnd) {
            var midLabel = bucketStart.format(this.options.format.mid);
            if (this.options.format.midRange) {
                midLabel += ' - ' + bucketEnd.format(this.options.format.mid);
            } else if (this.options.format.midPrefix != null) {
                midLabel = formatString(this.options.format.midPrefix, midLabel);
            }
            return midLabel;
        },
        getMaxLabel : function (bucketStart, bucketEnd) {
            var maxLabel = bucketStart.format(this.options.format.max);
            if (this.options.format.maxRange) {
                maxLabel += ' - ' + bucketEnd.format(this.options.format.max);
            } else if (this.options.format.maxPrefix != null) {
                maxLabel = formatString(this.options.format.maxPrefix, maxLabel);
            }
            return maxLabel;
        },

        getOffset : function (date) {
            return date.diff(this.options.range.start, this.options.value.min, true) * this.options.bucketSize;
        },

        getUnits : function (options) {
            var values = {
                start : options.start,
                end : options.end,
                percent : options.percent,
                units : {
                    offset : 0,
                    width : 0
                },
                render : {
                    left : 0,
                    right : 0,
                    width : 0
                }
            };
            if (options.start != null && options.end != null) {
                var units = {
                    offset : options.start.diff(this.options.range.start, this.options.value.min, true),
                    width : options.end.diff(options.start, this.options.value.min, true)
                };
            // Allow invalid date ranges to show 1 pixel line.
                if (units.width < 0) {
                    units.width = 0;
                }

                var left = parseInt(units.offset * this.options.bucketSize, 10),
                    width = parseInt(units.width * this.options.bucketSize, 10) || 1, // show a 1 pixel bar if its too small
                    right = left + width;
                values.units = units;
            // Calculate the percent
                var percent = null;
                if (options.percent != null) {
                    if (moment.isMoment(options.percent)) {
                        percent = options.percent.diff(options.start, this.options.value.min, true) / units.width * 100;
                    } else if (options.percent > 0 && options.percent <= 1) {
                        percent = options.percent * 100;
                    } else {
                        percent = options.percent;
                    }
                }
                values.render = {
                    left : left,
                    right : right,
                    width : width,
                    percent : percent
                };
            }
            return values;
        }
    };

    return DefaultScale;
});
