define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.moment',
    'ganttapi.grid',
    'ganttapi.gantt/ganttView',
    'ganttapi.gantt/enums/barHeight',
    'ganttapi.gantt/enums/recordTypes',
    'ganttapi.gantt/enums/relationshipType',
    'ganttapi.gantt/enums/barStatus',
    'ganttapi.gantt/config/barConfigurator',
    'ganttapi.gantt/config/optionsConfigurator',
    'ganttapi.gantt/locale',
    'gridapi.logger',
    'gridapi.jqueryui'
], function ($, _, moment, grid, GanttView,
             BAR_HEIGHT, RECORD_TYPES, RELATIONSHIP_TYPE, BAR_STATUS,
             barConfigurator, ganttConfigurator, locale, logger) {
    'use strict';

    var defaultGanttAttributes = {
        type : RECORD_TYPES.ACTIVITY,
        status : BAR_STATUS.NONE
    };
    var DUMMY_ROW_ID = 'DUMMY-NEW-RECORD';

    //>>excludeStart("validation", pragmas.production)
    function checkColumnDefinition(columnDefs, dataIndex) {
        if (columnDefs && columnDefs[dataIndex] == null) {
            logger.warn('all bar types configured with the gantt must have a column definition (for label display, formatters, etc)');
        }
    }
    //>>excludeEnd("validation")

    $.widget('pgbu.gantt', {
        // Defaults
        options : {
            debug : false,
            grid : {},
            borders : {
                gantt : 1
            },
            layout : {
                width : '50%',
                height : '100%',
                autoResize : true,
                autoResizeDebounce : 300,
                autoResizeImmediate : false,
                rowHeight : 26,
                timescaleHeight : 45
            },
            locale : locale,
            performance : {
                horizontalScrollThrottle : 30,
                maxRelationshipsPerRow : 20,
                updateRelationshipsDebounce : 300
            },
            scrollbar : {
                extraHeight : 12,
                extraWidth : 12
            },
            bars : {
                rendererFns : null
            },
            relationships : {
                rendererFns : null
            },
            tooltips : {
                constrain : false,
                showNameAndId : false,
                showBarType : true,
                showPercent : false,
                delay : 500
            },
            dataFns : {
                getRelationships : null,
                getRelationship : null,
                getGanttAttributes : null,
                getBarData : null
            },
            dataDate : moment(),
            timescale : {
                range : {
                    start : moment().add(-100, 'days'),
                    end : moment().add(6, 'months')
                }
            },
            config : {
                relationships : {
                    showRelationships : true,
                    showNonDriving : false,
                    showCriticalPath : true
                },
                timescale : {
                    type : 'QUARTER_MONTH',
                    zoom : 100
                },
                chart : {
                    showExpandedSummaryBars : true,
                    showDateLine : false,
                    showProgress : false,
                    showCriticalBars : false,
                    showSightLines : false,
                    showUnavailableTime : false,
                    syncGridViewFocus : false
                },
                layout : {
                    showGrid : true,
                    showGantt : true,
                    ganttPercent : 50
                }
            }
        },

        _create : function () {
            var that = this;
            this._transformGridOptions();

            this.widgetEventPrefix = 'gantt-';
            if (this.options.layout.autoResize) {
                $(window).on('resize' + this.eventNamespace + '-' + that.ganttRowsetId, _.debounce(function () {
                    that.resize();
                }, that.options.layout.autoResizeDebounce, that.options.layout.autoResizeImmediate));
            }
            var ganttRowset = _.findWhere(this.options.grid.layout.rowsets, { ganttRowset : true });
            if (ganttRowset == null) {
                ganttRowset = this.options.grid.layout.rowsets[0];
            }
            this.ganttRowsetId = ganttRowset.id;
            this.view = new GanttView(this);
            this.view.init();

            this.$gridContainer = this.view.$gantt.find('.grid-container');
            this._initGridListeners();
            this.view._syncGridListeners();

            ganttRowset.replacementScroller = this.view.pseudoScrollers.vertical;

            this.grid = this.$gridContainer.grid(this.options.grid).data('pgbu-grid');
            if (this.view.applyLayoutConfig()) {
                this.view.resize();
                this.grid.resize();
            }
            this.element.data('pgbu-grid', this.grid);

            //>>excludeStart("validation", pragmas.production)
            window['gantt' + this.uuid] = this;
            window.gantt = this;
            //>>excludeEnd("validation")

            this.timeouts = {};
        },

        _initGridListeners : function () {
            var that = this;
            this.$gridContainer.on('grid-store-created', function (event, args) {
                that.store = args.store;
            });
            this.$gridContainer.on('grid-before-repopulate', function (event, args) {
                if (!_.get(args, 'args.fromGantt', false)) {
                    that.repopulate(args.recordId, args.columnId, args.args);
                    return false;
                }
                return true;
            });
            this.$gridContainer.on('grid-repopulated', function (event, args) {
                that.view.scrollFocusIntoView({ recordId : _.get(args, 'recordId') });
            });
        },

        updateRowRelationships : function (args) {
            this.view.updateRowRelationships(args);
        },

        updateRelationship : function (args) {
            this.view.updateRelationship(args);
        },

        removeRelationship : function (relationshipId) {
            this.view.removeRelationship(relationshipId);
        },

        showGrid : function (showGrid) {
            this.options.config.layout.showGrid = showGrid;
            if (this.options.config.layout.showGantt === false && showGrid === false) {
                this.view.showGantt(true);
            }
            this.view.showGrid(showGrid);
        },

        showGantt : function (showGantt) {
            this.options.config.layout.showGantt = showGantt;
            if (this.options.config.layout.showGrid === false && showGantt === false) {
                this.view.showGrid(true);
            }
            this.view.showGantt(showGantt);
        },

        getBarLabels : function (args) {
            var labels = [];
            var barLabels = args.barConfig.labels;
            if (barLabels != null && barLabels.length) {
                for (var i = 0; i < barLabels.length; i++) {
                    var labelOptions = barLabels[i];

                    //>>excludeStart("validation", pragmas.production)
                    checkColumnDefinition(this.grid.columnDefinitions, labelOptions.key);
                    //>>excludeEnd("validation")
                    if (this._isGroupRecordId(args.recordId)) {
                        var group = this._getRecord(args.recordId);
                        var groupLabelData;
                        if (group.groupConfig.getBarLabels != null) {
                            groupLabelData = group.groupConfig.getBarLabels({
                                group : group,
                                backingRecord : group.backingRecord,
                                labelOptions : labelOptions
                            });
                            labels.push({
                                data : groupLabelData,
                                labelOptions : labelOptions
                            });
                        }
                    } else {
                        labels.push({
                            data : this._getFormattedLabel({ dataIndex : labelOptions.key, recordId : args.recordId }),
                            labelOptions : labelOptions
                        });
                    }
                }
            }
            return labels;
        },

        _getFormattedLabel : function (args) {
            var label = '';
            var column = this.grid.columnDefinitions[args.dataIndex];
            if (column) {
                if (column.displayType === 'select' || column.displayType === 'picker') {
                    label = column.formatter.format(this.store.getValueFromRecord(this._getRecord(args.recordId), args.dataIndex), column);
                } else {
                    label = column.formatter.format(this.store.getValueFromRecord(this._getRecord(args.recordId), args.dataIndex));
                }
            }
            return label;
        },

        zoomToFit : function () {
            this.view.zoomToFit();
        },

        getTimescaleRange : function (args) {
            var that = this;
            var barDatas = [];
            var min = moment().subtract(6, 'month');
            var max = moment().add(6, 'month');

            this.store.iterateStore(function (record) {
                _.forEach(that.options.config.bars, function (barConfig) {
                    barDatas.push(that.getBarData({
                        recordId : that.grid._getRecordId(record),
                        barConfig : barConfig
                    }));
                });
            });
            var minData = _.filter(barDatas, function (data) {
                return data.start != null;
            });
            if (minData.length) {
                min = _.min(_.pluck(minData, 'start'), function (data) {
                    return data.valueOf();
                });
            }

            var maxData = _.filter(barDatas, function (data) {
                return data.end != null;
            });
            if (maxData.length) {
                max = _.max(_.pluck(maxData, 'end'), function (data) {
                    return data.valueOf();
                });
            }

            return {
                start : min,
                end : max
            };
        },

        getBarData : function (args) {
            var that = this;
            if (args.recordId === DUMMY_ROW_ID) {
                return {};
            }
            var record = this._getRecord(args.recordId);
            var dataSource = record;
            var barType = this.options.barTypes[args.barConfig.type];
            var keys = barType != null ? barType.keys : args.barConfig.keys;
            if (keys == null) {
                return {};
            }
            if (this.store.getBarData != null) {
                args.barType = barType;
                return this.store.getBarData(args);
            }
            if (this._isGroupRecord(record)) {
                var groupConfig = record.groupConfig;
                if (groupConfig.getBarData != null) {
                    return groupConfig.getBarData({
                        barConfig : args.barConfig,
                        barType : barType,
                        keys : keys,
                        group : record,
                        backingRecord : record.backingRecord
                    });
                }
                if (record.rollupValues != null) {
                    dataSource = record.rollupValues;
                } else {
                    return {};
                }
            }
            var data = {};
            if (dataSource != null) {
                data = {
                    start : this.store.getValueFromRecord(dataSource, keys.start),
                    end : this.store.getValueFromRecord(dataSource, keys.end),
                    critical : this.store.getValueFromRecord(dataSource, keys.critical || this.options.grid.recordKeys.critical),
                    percent : this.store.getValueFromRecord(dataSource, keys.percent)
                };
                var rolledup = this.grid._attrs.all(record).rollup;
                if (rolledup) {
                    var recordCache = that.grid.dataCache.getRecordCache(args.recordId);
                    _.forEach(keys, function (dataIndex, key) {
                        var columnDefintion = that.grid.columnDefinitions[dataIndex];
                        if (columnDefintion != null && columnDefintion.rollupRole != null) {
                            data[key] = recordCache.rollupValues[dataIndex];
                        }
                    });
                }
            }

            return data;
        },

        getGanttAttributes : function (args) {
            if (args.recordId === DUMMY_ROW_ID) {
                return {};
            }
            var record = this._getRecord(args.recordId);
            if (this.grid._isGroupRecord(record) && record.groupConfig.getBarData == null) {
                return {
                    status : BAR_STATUS.NONE,
                    type : RECORD_TYPES.SUMMARY
                };
            }
            var callArgs = {
                record : record,
                defaults : _.clone(defaultGanttAttributes),
                enums : {
                    RECORD_TYPES : RECORD_TYPES,
                    BAR_STATUS : BAR_STATUS
                }
            };
            if (this.options.dataFns.getGanttAttributes != null) {
                return this.options.dataFns.getGanttAttributes(callArgs);
            } else if (this.store.getGanttAttributes != null) {
                return this.store.getGanttAttributes(callArgs);
            } else {
                return defaultGanttAttributes;
            }
        },

        getBarTooltipData : function (args) {
            var that = this;
            var tooltipSettings = null;
            var record = this._getRecord(args.recordId);
            var dataSource = record;
            var barType = this.options.barTypes[args.barConfig.type];
            var keys = barType != null ? barType.keys : args.barConfig.keys;
            if (keys == null) {
                return {};
            }
            if (this._isGroupRecord(record)) {
                if (record.rollupValues != null) {
                    dataSource = record.rollupValues;
                } else {
                    return {};
                }
            }
            var data = {
                start : this.store.getValueFromRecord(dataSource, keys.start),
                end : this.store.getValueFromRecord(dataSource, keys.end),
                percent : this.store.getValueFromRecord(dataSource, keys.percent)
            };
            if (this._isGroupRecord(record)) {
                var groupConfig = record.groupConfig;
                if (groupConfig.getBarTooltipData != null) {
                    data = groupConfig.getBarTooltipData({
                        barConfig : args.barConfig,
                        barType : barType,
                        keys : keys,
                        group : record,
                        backingRecord : record.backingRecord
                    });
                }
            }

            var rolledup = this.grid._attrs.all(record).rollup;
            if (rolledup) {
                var recordCache = that.grid.dataCache.getRecordCache(args.recordId);
                _.forEach(keys, function (dataIndex, key) {
                    var columnDefintion = that.grid.columnDefinitions[dataIndex];
                    if (columnDefintion != null && columnDefintion.rollupRole != null) {
                        data[key] = recordCache.rollupValues[dataIndex];
                    }
                });
            }
            if (data.start != null && data.end != null) {
                var tooltipData = {};
                //>>excludeStart("validation", pragmas.production)
                checkColumnDefinition(this.grid.columnDefinitions, keys.start);
                checkColumnDefinition(this.grid.columnDefinitions, keys.end);
                //>>excludeEnd("validation")
                tooltipData[this.options.locale.tooltips.start] = this.grid.columnDefinitions[keys.start].formatter.format(data.start);
                tooltipData[this.options.locale.tooltips.end] = this.grid.columnDefinitions[keys.end].formatter.format(data.end);
                if (this.options.tooltips.showPercent && data.percent != null) {
                    tooltipData[this.options.locale.tooltips.percent] = this.grid.columnDefinitions[keys.percent].formatter.format(data.percent);
                }
                var label;
                if (this._isGroupRecord(record)) {
                    label = record.displayName;
                } else {
                    if (this.options.tooltips.showNameAndId) {
                        label = this.store.getValueFromRecord(record, this.options.tooltips.keys.name) + ' - ' +
                            this.store.getValueFromRecord(record, this.options.tooltips.keys.id);
                    } else {
                        label = this.store.getValueFromRecord(record, this.options.tooltips.keys.label);
                    }
                }
                if (label == null) {
                    label = '';
                }

                tooltipSettings = {
                    label : label,
                    barType : barType != null ? barType.name : args.barConfig.name,
                    data : tooltipData,
                    top : 4 + args.barConfig.render.top + BAR_HEIGHT[args.barConfig.render.size].height,
                    showBarType : this.options.tooltips.showBarType
                };
            }
            return tooltipSettings;
        },

        getBarAttributes : function (args) {
            if (args.recordId === DUMMY_ROW_ID) {
                return { data : {}, labels : {} };
            }
            return {
                data : this.getBarData(args),
                labels : this.getBarLabels(args)
            };
        },

        getWorkWeek : function () {
            if (this.options.workWeek) {
                return this.options.workWeek;
            }
            return null;
        },

        getRelationships : function (args) {
            var relationships = [];
            if (args.recordId !== DUMMY_ROW_ID && !this._isGroupRecordId(args.recordId)) {
                if (this.options.dataFns.getRelationships != null) {
                    relationships = this.options.dataFns.getRelationships.call(this, args) || [];
                } else if (this.store.getRelationships != null) {
                    relationships = this.store.getRelationships(args) || [];
                }
            }

            return _.take(relationships, this.options.performance.maxRelationshipsPerRow);
        },
        getRelationship : function (args) {
            if (this._isGroupRecordId(args.recordId)) {
                return null;
            }
            if (this.options.dataFns.getRelationship != null) {
                return this.options.dataFns.getRelationship.call(this, args);
            } else if (this.store.getRelationship != null) {
                return this.store.getRelationship(args);
            }
            return null;
        },

        _setOption : function (key, value) {
            var newOption = value;
            if (_.isObject(this.options[key]) && !_.isArray(this.options[key])) {
                newOption = {};
                if (key === 'config') {
                    if (value.bars && value.bars.length) {
                        this.options.config.bars = [];
                    }
                    _.merge(newOption, this.options[key], value, function (a, b) {
                        if (_.isArray(a)) {
                            return b;
                        }
                    });
                } else if (moment.isMoment(value)) {
                    newOption = moment(value);
                } else {
                    _.merge(newOption, this.options[key], value, function (a, b) {
                        if (_.isArray(a)) {
                            return b;
                        }
                    });
                }
            }

            $.Widget.prototype._setOption.apply(this, [key, newOption]);

            switch (key) {
            case 'dataDate' :
                this.view.setDataDate();
                break;
            case 'config' :
                this.view.remakeBarRenderers();
                this.view.applyRelationshipOptions();
                this.view.applyChartOptions();
                this.view.changeTimescale(this.options.config.timescale.type, true);
                this.view.applyLayoutConfig();
                break;
            default :
                break;
            }
        },
        getViewConfig : function () {
            return {
                gantt : _.clone(this.options.config, true),
                grid : this.grid.getViewConfig()
            };
        },

        applyViewConfig : function (viewConfig) {
            if (viewConfig != null) {
                if (viewConfig.gantt != null && !_.isEmpty(viewConfig.gantt)) {
                    this._setOption('config', viewConfig.gantt);
                    if (viewConfig.gantt.layout != null) {
                        this._setOption('layout', viewConfig.gantt.layout);
                        if (this.options.config.layout.showGantt) {
                            this.grid._setOption('layout', viewConfig.gantt.layout);
                        }
                    }
                    this.view.ganttItems = {};
                    this.grid.viewManager.queue.repopulate = true;
                }
                if (viewConfig.grid != null && !_.isEmpty(viewConfig.grid)) {
                    this.grid.applyViewConfig(viewConfig.grid);
                } else {
                    this.grid.viewManager.refresh();
                }
            }
        },

        _transformGridOptions : function () {
            this.options.grid.layout.width = '100%';
            this.options.grid.layout.height = '100%';
            this.options.grid.layout.headerHeight = this.options.layout.timescaleHeight;
            this.options.grid.layout.rowHeight = this.options.layout.rowHeight;
            this.options.grid.gantt = {
                enabled : true
            };
            this.options.recordKeys = this.options.grid;// for master detail page
        },

        _getDummyRecord : function () {
            return this.grid._getDummyRecord();
        },

        destroy : function () {
            //>>excludeStart("validation", pragmas.production)
            delete window['gantt' + this.uuid];
            //>>excludeEnd("validation")
            // remove any editors and menus
            _.forEach(this.timeouts, function (timeout) {
                clearTimeout(timeout);
            });
            this.element.children().each(function () {
                $(this).remove();
            });

            this.element.empty();
            this._tearDownTriggers();
            // Does nice things like unbind all namespaced events on the original element
            delete this.store;
            delete this.view;
            $.Widget.prototype.destroy.call(this);
        },

        _tearDownTriggers : function () {
            $(window).off('resize' + this.eventNamespace + '-' + this.ganttRowsetId);
            this.element.off(this.widgetEventPrefix);
            this.grid._tearDownTriggers();
        },

        setStore : function (store, CustomMediator, args) {
            this.store = this.grid.setStore(store, CustomMediator, args, this);
            this.grid.repopulate();
            this.view.updateTimescaleRange();
        },
        resize : function () {
            this.view.resize();
            this.grid.resize();
        },

        getGrid : function () {
            return this.grid;
        },
        _isGroupRecord : function (record) {
            return this.grid._isGroupRecord(record);
        },
        _isGroupRecordId : function (record) {
            return this.grid._isGroupRecordId(record);
        },
        // grid method forwards : this is for generic code in prime that works with a "grid"
        _getRecord : function (recordId) {
            return this.grid._getRecord(recordId);
        },
        _getRecords : function (recordIds) {
            return this.grid._getRecords(recordIds);
        },
        _resetInternalStorage : function () {
            // TODO: handle anything gantt might need
            this.grid._resetInternalStorage();
        },
        _resetGroupStorage : function () {
            this.grid._resetGroupStorage();
        },
        _resetView : function () {
            // TODO: add in gantt support
            this.grid._resetView();
        },
        _recordUpdated : function (record, dataIndex) {
            this.grid._recordUpdated(record, dataIndex);
        },
        validateStore : function () {
            this.grid.validateStore();
        },
        _validateRecord : function (record, hideErrors) {
            this.grid._validateRecord(record, hideErrors);
        },
        _validateRecordCell : function (args) {
            this.grid._validateRecordCell(args);
        },
        _recordsAdded : function (records, select) {
            this.grid._recordsAdded(records, select);
        },
        _recheckStoreUniqueness : function () {
            this.grid._recheckStoreUniqueness();
        },
        _recordsRemoved : function (records) {
            this.grid._recordsRemoved(records);
        },
        _recordMoved : function (record, oldRowset) {
            this.grid._recordMoved(record, oldRowset);
        },
        newRow : function (args) {
            this.grid.newRow(args);
        },
        newChildRow : function (args) {
            this.grid.newChildRow(args);
        },
        deleteCells : function (args) {
            this.grid.deleteCells(args);
        },
        cutCells : function (args) {
            this.grid.cutCells(args);
        },
        copyCells : function (args) {
            this.grid.copyCells(args);
        },
        pasteCells : function (args) {
            this.grid.pasteCells(args);
        },
        _prepareOptions : function (args, methodName) {
            this.grid._prepareOptions(args, methodName);
        },
        deleteRows : function (args) {
            this.grid.deleteRows(args);
        },
        cutRows : function (args) {
            this.grid.cutRows(args);
        },
        copyRows : function (args) {
            this.grid.copyRows(args);
        },
        pasteRows : function (args) {
            this.grid.pasteRows(args);
        },
        moveRowLeft : function (args) {
            this.grid.moveRowLeft(args);
        },
        moveRowRight : function (args) {
            this.grid.moveRowRight(args);
        },
        moveRowUp : function (args) {
            this.grid.moveRowUp(args);
        },
        moveRowDown : function (args) {
            this.grid.moveRowDown(args);
        },
        expandAllRows : function () {
            this.grid.expandAllRows();
        },
        collapseAllRows : function () {
            this.grid.collapseAllRows();
        },
        expandRow : function (recordId, expandChildren) {
            this.grid.expandRow(recordId, expandChildren);
        },
        expandRows : function (args) {
            this.grid.expandRows(args);
        },
        collapseRow : function (recordId, collapseChildren) {
            this.grid.collapseRow(recordId, collapseChildren);
        },
        collapseRows : function (args) {
            this.grid.collapseRows(args);
        },
        _checkExpandCollapseState : function () {
            this.grid._checkExpandCollapseState();
        },
        canEditCell : function (recordId, columnId, recordAttributes) {
            this.grid.canEditCell(recordId, columnId, recordAttributes);
        },
        canEditCells : function (records, columnId) {
            this.grid.canEditCells(records, columnId);
        },
        editCells : function (args) {
            this.grid.editCells(args);
        },
        editRow : function (args) {
            this.grid.editRow(args);
        },
        requestBrowserFocus : function (force) {
            this.grid.requestBrowserFocus(force);
        },
        selectCell : function (recordId, columnId, location, appendToSelection) {
            this.grid.selectCell(recordId, columnId, location, appendToSelection);
        },
        selectHeaderCell : function (columnId) {
            this.grid.selectHeaderCell(columnId);
        },
        selectRows : function (recordIds, append, event, columnId) {
            this.grid.selectRows(recordIds, append, event, columnId);
        },
        deselectRows : function (recordIds, event) {
            this.grid.deselectRows(recordIds, event);
        },
        // Todo: change for gantt as well
        closeMenus : function () {
            this.grid.closeMenus();
        },
        checkRows : function (recordIds, check, all) {
            this.grid.checkRows(recordIds, check, all);
        },
        getSelection : function () {
            return this.grid.getSelection();
        },
        getRowFocus : function () {
            return this.grid.getRowFocus();
        },
        clearSelection : function () {
            this.grid.clearSelection();
        },
        reset : function () {
            this.grid.reset();
        },
        commit : function () {
            this.grid.commit();
        },
        getRecordOrder : function () {
            return this.grid.getRecordOrder();
        },
        sortRows : function (dataIndex, append, order, silent) {
            this.grid.sortRows(dataIndex, append, order, silent);
        },
        groupRows : function (groupConfig, silent) {
            this.grid.groupRows(groupConfig, silent);
        },
        undo : function (change) {
            this.grid.undo(change);
        },
        redo : function (change) {
            this.grid.redo(change);
        },
        repopulate : function (recordId, columnId, args) {
            if (this.store == null) {
                return;
            }
            this.view.visibleRelationships = {};
            this.view.$relationshipContainer.empty();
            this.view.adjustForQuickAddRow(false);
            this.grid.repopulate(recordId, columnId, _.assign({ fromGantt : true }, args));
            this.view.resize();
        },
        closeEditor : function (commit) {
            this.grid.closeEditor(commit);
        },
        searchRows : function (term, args) {
            this.grid.searchRows(term, args);
        },
        clearSearch : function () {
            this.grid.clearSearch();
        },
        toggleColumns : function (columns, toggle) {
            this.grid.toggleColumns(columns, toggle);
        },
        isLoaded : function () {
            return this.grid.isLoaded();
        },
        setLoading : function (loading) {
            // todo: change this to have a part for the grid as well
            this.grid.setLoading(loading);
        },
        // TODO : need disable styling for the gantt
        setDisabled : function (disabled) {
            this.grid.setDisabled(disabled);
        },
        showCellErrors : function (recordId, columnId) {
            this.grid.showCellErrors(recordId, columnId);
        },
        print : function (skipPrint) {
            this.grid.print(skipPrint);
        },
        exportData : function (args) {
            this.grid.exportData(args);
        },
        getPrintContent : function () {
            this.grid.getPrintContent();
        },
        setRowsetMessage : function (message, rowsetId) {
            this.grid.setRowsetMessage(message, rowsetId);
        },
        _markInvalid : function (recordId, dataIndex, validationMessages, skipValidation) {
            this.grid._markInvalid(recordId, dataIndex, validationMessages, skipValidation);
        },
        _clearInvalid : function (recordId, dataIndex, skipValidation) {
            this.grid._clearInvalid(recordId, dataIndex, skipValidation);
        },
        _forceViewRender : function () {
            this.grid._forceViewRender();
        },
        _updateRecordIds : function (recordIdMap) {
            this.grid._updateRecordIds(recordIdMap);
        },
        _silkFindRecord : function (dataIndex, value) {
            return this.grid._silkFindRecord(dataIndex, value);
        },
        checkActionState : function () {
            return this.grid.checkActionState();
        },
        unhideAllErrors : function () {
            this.grid.unhideAllErrors();
        },
        getId : function () {
            return this.id;
        },
        pause : function () {
            this.grid.pause();
        },
        resume : function () {
            this.grid.resume();
        }

    });

    $(window).trigger('ganttapi-loaded');
});
