define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.logger',
    'ganttapi.gantt/enums/barHeight',
    'ganttapi.gantt/enums/relationshipType',
    'ganttapi.gantt/enums/recordTypes',
    'markup!ganttapi.gantt/templates/relationships.htm:.htm-relationship',
    'markup!ganttapi.gantt/templates/relationships.htm:.htm-wrapping-relationship',
    'markup!ganttapi.gantt/templates/relationships.htm:.htm-tooltip'
], function ($, _, logger, BAR_HEIGHT, RELATIONSHIP_TYPE, RECORD_TYPES,
              relationshipTemplate, wrappingTemplate, tooltipTemplate) {
    'use strict';

    var RealtionshipRenderer = function (options) {
        var that = this;
        this.ganttView = options.ganttView;
        this.timescale = options.timescale;
        this.widget = options.widget;
        var rendererFns = this.widget.options.relationships.rendererFns;
        // override mediator functions from options
        _.forEach(rendererFns, function (fnc, fncName) {
            that[fncName] = fnc || that[fncName];
        });
    };

    RealtionshipRenderer.prototype = {
        constructor : RealtionshipRenderer,
        getItemCache : function (options) {
            var itemCache = this.ganttView.ganttItems[options.recordId];
            if (itemCache == null) {
                var record = this.widget._getRecord(options.recordId);
                if (record == null) {
                    return null;
                }
                var barData = this.widget.getBarAttributes({
                    recordId : options.recordId,
                    barConfig : options.barConfig
                }).data;
                var ganttAttributes = this.widget.getGanttAttributes({
                    recordId : options.recordId
                });
                var barCache = this.timescale.getUnits(barData);
                if (ganttAttributes.type === RECORD_TYPES.START_MILESTONE) {
                    barCache.render.right = barCache.render.left + BAR_HEIGHT[options.barConfig.render.size].height;
                    barCache.render.width = BAR_HEIGHT[options.barConfig.render.size].height;
                } else if (ganttAttributes.type === RECORD_TYPES.FINISH_MILESTONE) {
                    barCache.render.left = barCache.render.right;
                    barCache.render.right = barCache.render.left + BAR_HEIGHT[options.barConfig.render.size].height;
                    barCache.render.width = BAR_HEIGHT[options.barConfig.render.size].height;
                }
                if (this.ganttView.ganttItems[options.recordId] == null) {
                    this.ganttView.ganttItems[options.recordId] = {};
                }
                this.ganttView.ganttItems[options.recordId][options.barId] = barCache;
            }
            return this.ganttView.ganttItems[options.recordId][options.barId];
        },

        makeTooltip : function (options) {
            var relationship = this.widget.getRelationship(options);
            var templateOptions = {
                top : options.event.clientY,
                left : options.event.clientX,
                id : this.widget.uuid + 'relationship-tooltip-' + relationship.id,
                locale : this.ganttView.options.locale,
                predecessor : relationship.predecessor,
                successor : relationship.successor,
                type : this.ganttView.options.locale.relationships[RELATIONSHIP_TYPE[relationship.type].label],
                lag : relationship.lag
            };
            return tooltipTemplate(templateOptions);
        },

        remakeRelationship : function (options) {
            var $relationship = this.ganttView.$relationshipContainer.find('.relationship[data-id="' + options.relationship.id + '"]');
            var newRelationship = this.makeRelationship(options);
            if (newRelationship.length) {
                $relationship.replaceWith(newRelationship);
            } else {
                $relationship.remove();
            }
        },

        isRelationshipinBounds : function (relationshipOptions, virtualRows) {
            if (relationshipOptions.predecessorPosition >= virtualRows.displayTop && relationshipOptions.predecessorPosition <= virtualRows.displayBottom) {
                return true;
            }
            if (relationshipOptions.successorPosition >= virtualRows.displayTop && relationshipOptions.successorPosition <= virtualRows.displayBottom) {
                return true;
            }
            return false;
        },

        makeRelationship : function (options) {
            var relationship = options.relationship;
            var relationBuilderOptions = {
                relationship : relationship,
                predecessor : this.getItemCache({ recordId : relationship.predecessor.id, barId : options.barId, barConfig : options.barConfig }),
                successor : this.getItemCache({ recordId : relationship.successor.id, barId : options.barId, barConfig : options.barConfig }),
                successorPosition : this.widget.grid.view.virtualRows[this.ganttView.ganttRowsetId].records.indexOf(relationship.successor.id),
                predecessorPosition : this.widget.grid.view.virtualRows[this.ganttView.ganttRowsetId].records.indexOf(relationship.predecessor.id),
                offset : options.offset,
                barOffset : Math.floor(options.barConfig.render.top + BAR_HEIGHT[options.barConfig.render.size].height / 2 - options.linePadding - options.lineSize / 2),
                linePadding : options.linePadding,
                lineSize : options.lineSize
            };
            if (relationBuilderOptions.predecessor == null ||
                relationBuilderOptions.successor == null ||
                relationBuilderOptions.successorPosition === -1 ||
                relationBuilderOptions.predecessorPosition === -1 || !this.isRelationshipinBounds(relationBuilderOptions, this.widget.grid.view.virtualRows[this.ganttView.ganttRowsetId])) {
                delete this.ganttView.visibleRelationships[relationship.id];
                return '';
            }

            switch (RELATIONSHIP_TYPE[relationship.type]) {
            case RELATIONSHIP_TYPE.FINISH_TO_FINISH:
                return this.makeFinishToFinishRelationship(relationBuilderOptions);
            case RELATIONSHIP_TYPE.START_TO_START :
                return this.makeStartToStartRelationship(relationBuilderOptions);
            case RELATIONSHIP_TYPE.FINISH_TO_START :
                return this.makeFinishToStartRelationship(relationBuilderOptions);
            case RELATIONSHIP_TYPE.START_TO_FINISH :
                return this.makeStartToFinishRelationship(relationBuilderOptions);
            }

            return '';
        },
        // Done
        makeFinishToFinishRelationship : function (options) {
            var startY = Math.min(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;
            var endY = Math.max(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;

            var topActivity, bottomActivity, successorTop;
            if (options.predecessorPosition < options.successorPosition) {
                topActivity = options.predecessor;
                bottomActivity = options.successor;
                successorTop = false;
            } else {
                topActivity = options.successor;
                bottomActivity = options.predecessor;
                successorTop = true;
            }

            // set to the furthest away for finish to finish relationships
            var verticalLineX = Math.max(options.predecessor.render.right, options.successor.render.right) + options.offset;
            var templateOptions = {
                id : this.widget.uuid + 'gantt-relationship-' + options.relationship.id,
                relationshipId : options.relationship.id,
                successorId : options.relationship.successor.id,
                predecessorId : options.relationship.predecessor.id,
                activityTop : startY,
                top : {
                    top : options.barOffset,
                    width : Math.max(topActivity.render.right, verticalLineX) - Math.min(topActivity.render.right, verticalLineX),
                    left : topActivity.render.right,
                    classes : 'relationship-finish',
                    arrow : successorTop
                },
                bottom : {
                    width : Math.max(bottomActivity.render.right, verticalLineX) - Math.min(bottomActivity.render.right, verticalLineX),
                    left : bottomActivity.render.right,
                    top : endY - startY + options.barOffset,
                    classes : 'relationship-finish',
                    arrow : !successorTop
                },
                vertical : {
                    top : options.barOffset + options.linePadding,
                    height : endY - startY + options.lineSize,
                    left : verticalLineX - options.linePadding - options.lineSize,
                    classes : 'relationship-right'
                },
                classes : this.getRelationshipClasses(options).join(' ')
            };
            return relationshipTemplate(templateOptions);
        },
        // Done
        makeStartToStartRelationship : function (options) {
            var startY = Math.min(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;
            var endY = Math.max(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;

            var topActivity, bottomActivity, successorTop;
            if (options.predecessorPosition < options.successorPosition) {
                topActivity = options.predecessor;
                bottomActivity = options.successor;
                successorTop = false;
            } else {
                topActivity = options.successor;
                bottomActivity = options.predecessor;
                successorTop = true;
            }

            // set to the furthest away for finish to finish relationships
            var verticalLineX = Math.min(options.predecessor.render.left, options.successor.render.left) - options.offset;

            var templateOptions = {
                id : this.widget.uuid + 'gantt-relationship-' + options.relationship.id,
                successorId : options.relationship.successor.id,
                predecessorId : options.relationship.predecessor.id,
                relationshipId : options.relationship.id,
                activityTop : startY,
                top : {
                    top : options.barOffset,
                    width : Math.max(topActivity.render.left, verticalLineX) - Math.min(topActivity.render.left, verticalLineX),
                    left : verticalLineX,
                    classes : 'relationship-start',
                    arrow : successorTop
                },
                bottom : {
                    width : Math.max(bottomActivity.render.left, verticalLineX) - Math.min(bottomActivity.render.left, verticalLineX),
                    left : verticalLineX,
                    top : endY - startY + options.barOffset,
                    classes : 'relationship-start',
                    arrow : !successorTop
                },
                vertical : {
                    top : options.barOffset + options.linePadding,
                    height : endY - startY + options.lineSize,
                    left : verticalLineX - options.linePadding,
                    classes : 'relationship-left'
                },
                classes : this.getRelationshipClasses(options).join(' ')
            };
            return relationshipTemplate(templateOptions);
        },

        makeFinishToStartRelationship : function (options) {
            // If the successors start is after the predecessors start draw a simplified line
            if (options.successor.render.left - options.barOffset > options.predecessor.render.right + options.barOffset) {
                return this.makeDirectFinishToStartRelationship(options);
            } else {
                return this.makeWrappingFinishToStartRelationship(options);
            }
        },

        makeDirectFinishToStartRelationship : function (options) {
            var startY = Math.min(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;
            var endY = Math.max(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;


            // set to the furthest away for finish to finish relationships
            var verticalLineX = options.successor.render.left - options.offset - options.linePadding - options.lineSize;

            var topSettings, bottomSettings;
            var predecessorOptions = {
                width : verticalLineX - options.predecessor.render.right + options.linePadding + options.lineSize,
                left : options.predecessor.render.right,
                classes : 'relationship-finish',
                arrow : false
            };
            var successorOptions = {
                width : options.offset,
                left : options.successor.render.left - options.offset,
                classes : 'relationship-start',
                arrow : true

            };
            if (options.predecessorPosition < options.successorPosition) {
                topSettings = predecessorOptions;
                bottomSettings = successorOptions;
            } else {
                topSettings = successorOptions;
                bottomSettings = predecessorOptions;
            }
            topSettings.top = options.barOffset;
            bottomSettings.top = endY - startY + options.barOffset;

            var templateOptions = {
                id : this.widget.uuid + 'gantt-relationship-' + options.relationship.id,
                relationshipId : options.relationship.id,
                successorId : options.relationship.successor.id,
                predecessorId : options.relationship.predecessor.id,
                activityTop : startY,
                top : topSettings,
                bottom : bottomSettings,
                vertical : {
                    top : options.barOffset + options.linePadding,
                    height : endY - startY + options.lineSize,
                    left : verticalLineX,
                    classes : 'relationship-right'
                },
                classes : this.getRelationshipClasses(options).join(' ')
            };
            return relationshipTemplate(templateOptions);
        },

        makeWrappingFinishToStartRelationship : function (options) {
            var startY = Math.min(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;
            var endY = Math.max(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;

            // set to the furthest away for finish to finish relationships
            var wrapLineX = options.successor.render.left - options.offset;
            var predecessorVeritcalX = options.predecessor.render.right + options.offset;

            var topSettings, bottomSettings, topVerticalSettings, bottomVerticalSettings;

            var predecessorOptions = {
                top : options.barOffset,
                width : options.offset,
                left : options.predecessor.render.right,
                classes : 'relationship-finish',
                arrow : false
            };
            var predecessorVerticalOptions = {
                top : options.barOffset + options.linePadding,
                height : endY - startY - options.barOffset,
                left : predecessorVeritcalX - options.linePadding - options.lineSize,
                classes : 'relationship-right'
            };
            var successorOptions = {
                width : options.offset,
                left : wrapLineX,
                classes : 'relationship-start',
                arrow : true,
                top : endY - startY + options.barOffset
            };
            var wrapSettings = {
                left : wrapLineX,
                width : predecessorVeritcalX - wrapLineX,
                top : predecessorVerticalOptions.top + predecessorVerticalOptions.height - options.linePadding - options.lineSize
            };

            var successorVerticalOptions = {
                left : wrapLineX - options.linePadding,
                height : Math.max(wrapSettings.top, successorOptions.top) - Math.min(wrapSettings.top, successorOptions.top) + options.lineSize,
                top : wrapSettings.top + options.linePadding,
                classes : 'relationship-left'
            };

            if (options.predecessorPosition < options.successorPosition) {
                topSettings = predecessorOptions;
                bottomSettings = successorOptions;
                topVerticalSettings = predecessorVerticalOptions;
                bottomVerticalSettings = successorVerticalOptions;
            } else {
                // Need to recalculate for when everything is switched around.
                // Flip the tops of the horizontal bars for the successor and predecessor
                predecessorOptions.top = successorOptions.top;
                successorOptions.top = options.barOffset;
                // Figure out new
                wrapSettings.top = this.ganttView.options.layout.rowHeight - options.lineSize - options.linePadding;
                successorVerticalOptions.top = successorOptions.top + options.linePadding;
                successorVerticalOptions.height = wrapSettings.top - successorVerticalOptions.top + options.lineSize + options.linePadding;

                predecessorVerticalOptions.top = wrapSettings.top + options.linePadding;
                predecessorVerticalOptions.height = predecessorOptions.top - wrapSettings.top + options.linePadding;

                topSettings = successorOptions;
                bottomSettings = predecessorOptions;

                topVerticalSettings = successorVerticalOptions;
                bottomVerticalSettings = predecessorVerticalOptions;
            }

            var templateOptions = {
                id : this.widget.uuid + 'gantt-relationship-' + options.relationship.id,
                relationshipId : options.relationship.id,
                successorId : options.relationship.successor.id,
                predecessorId : options.relationship.predecessor.id,
                activityTop : startY,
                top : topSettings,
                bottom : bottomSettings,
                topVertical : topVerticalSettings,
                bottomVertical : bottomVerticalSettings,
                wrap : wrapSettings,
                classes : this.getRelationshipClasses(options).join(' ')
            };
            return wrappingTemplate(templateOptions);
        },

        makeStartToFinishRelationship : function (options) {
            // If the successors start is after the predecessors start draw a simplified line
            if (options.predecessor.render.left - options.barOffset > options.successor.render.right + options.barOffset) {
                return this.makeDirectStartToFinishRelationship(options);
            } else {
                return this.makeWrappingStartToFinishRelationship(options);
            }
        },
        // Completed
        makeDirectStartToFinishRelationship : function (options) {
            var startY = Math.min(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;
            var endY = Math.max(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;

            // set to the furthest away for finish to finish relationships
            var verticalLineX = options.successor.render.right + options.offset - options.linePadding - options.lineSize;

            var topSettings, bottomSettings;
            var predecessorOptions = {
                width : options.predecessor.render.left - verticalLineX - options.linePadding,
                left : verticalLineX + options.linePadding,
                classes : 'relationship-start',
                arrow : false
            };
            var successorOptions = {
                width : options.offset,
                left : options.successor.render.right,
                classes : 'relationship-finish',
                arrow : true

            };
            if (options.predecessorPosition < options.successorPosition) {
                topSettings = predecessorOptions;
                bottomSettings = successorOptions;
            } else {
                topSettings = successorOptions;
                bottomSettings = predecessorOptions;
            }
            topSettings.top = options.barOffset;
            bottomSettings.top = endY - startY + options.barOffset;

            var templateOptions = {
                id : this.widget.uuid + 'gantt-relationship-' + options.relationship.id,
                relationshipId : options.relationship.id,
                successorId : options.relationship.successor.id,
                predecessorId : options.relationship.predecessor.id,
                activityTop : startY,
                top : topSettings,
                bottom : bottomSettings,
                vertical : {
                    top : options.barOffset + options.linePadding,
                    height : endY - startY + options.lineSize,
                    left : verticalLineX,
                    classes : 'relationship-left'
                },
                classes : this.getRelationshipClasses(options).join(' ')
            };
            return relationshipTemplate(templateOptions);
        },

        makeWrappingStartToFinishRelationship : function (options) {
            var startY = Math.min(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;
            var endY = Math.max(options.successorPosition, options.predecessorPosition) * this.ganttView.options.layout.rowHeight;

            // set to the furthest away for finish to finish relationships
            var wrapLineX = options.successor.render.right + options.offset - options.linePadding - options.lineSize;
            var predecessorVeritcalX = options.predecessor.render.left - options.offset - options.linePadding;

            var topSettings, bottomSettings, topVerticalSettings, bottomVerticalSettings;

            var predecessorOptions = {
                top : options.barOffset,
                width : options.offset,
                left : options.predecessor.render.left - options.offset,
                classes : 'relationship-start',
                arrow : false
            };
            var predecessorVerticalOptions = {
                top : options.barOffset + options.linePadding,
                height : endY - startY - options.barOffset,
                left : predecessorVeritcalX,
                classes : 'relationship-left'
            };
            var successorOptions = {
                width : options.offset,
                left : options.successor.render.right,
                classes : 'relationship-finish',
                arrow : true,
                top : endY - startY + options.barOffset
            };
            var wrapSettings = {
                left : predecessorVeritcalX + options.linePadding,
                width : wrapLineX - predecessorVeritcalX + options.lineSize,
                top : predecessorVerticalOptions.top + predecessorVerticalOptions.height - options.linePadding - options.lineSize
            };

            var successorVerticalOptions = {
                left : wrapLineX,
                height : Math.max(wrapSettings.top, successorOptions.top) - Math.min(wrapSettings.top, successorOptions.top) + options.lineSize,
                top : wrapSettings.top + options.linePadding,
                classes : 'relationship-right'
            };

            if (options.predecessorPosition < options.successorPosition) {
                topSettings = predecessorOptions;
                bottomSettings = successorOptions;
                topVerticalSettings = predecessorVerticalOptions;
                bottomVerticalSettings = successorVerticalOptions;
            } else {
                // Need to recalculate for when everything is switched around.
                // Flip the tops of the horizontal bars for the successor and predecessor
                predecessorOptions.top = successorOptions.top;
                successorOptions.top = options.barOffset;
                // Figure out new
                wrapSettings.top = this.ganttView.options.layout.rowHeight - options.lineSize - options.linePadding;
                successorVerticalOptions.top = successorOptions.top + options.linePadding;
                successorVerticalOptions.height = wrapSettings.top - successorVerticalOptions.top + options.lineSize + options.linePadding;

                predecessorVerticalOptions.top = wrapSettings.top + options.linePadding;
                predecessorVerticalOptions.height = predecessorOptions.top - wrapSettings.top + options.lineSize;

                topSettings = successorOptions;
                bottomSettings = predecessorOptions;

                topVerticalSettings = successorVerticalOptions;
                bottomVerticalSettings = predecessorVerticalOptions;
            }

            var templateOptions = {
                id : this.widget.uuid + 'gantt-relationship-' + options.relationship.id,
                relationshipId : options.relationship.id,
                successorId : options.relationship.successor.id,
                predecessorId : options.relationship.predecessor.id,
                activityTop : startY,
                top : topSettings,
                bottom : bottomSettings,
                topVertical : topVerticalSettings,
                bottomVertical : bottomVerticalSettings,
                wrap : wrapSettings,
                classes : this.getRelationshipClasses(options).join(' ')
            };
            return wrappingTemplate(templateOptions);
        },
        getRelationshipClasses : function (options) {
            var classes = [];
            if (options.relationship.critical) {
                classes.push('relationship-critical');
            }
            if (options.relationship.driving) {
                classes.push('relationship-driving');
            } else {
                classes.push('relationship-non-driving');
            }

            return classes;
        }

    };

    return RealtionshipRenderer;
});
