define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.moment',
    'gridapi.logger',
    'ganttapi.gantt/timescaleHandlers',
    'ganttapi.gantt/defaultTimescale',
    'markup!ganttapi.gantt/templates/timescale.htm:.htm-bucket',
    'markup!ganttapi.gantt/templates/timescale.htm:.htm-range',
    'markup!ganttapi.gantt/templates/timescale.htm:.htm-context-menu',
    'markup!ganttapi.gantt/templates/gantt.htm:.htm-unavailable-overlay'
], function ($, _, moment, logger, timescaleMixins, DefaultTimescale,
             bucketTemplate, rangeTemplate, contextMenuTemplate, unavailableTemplate) {
    'use strict';

    var MIN_TO_SCALE = {
        day : 'WEEK_DAY',
        week : 'MONTH_WEEK',
        month : 'YEAR_MONTH',
        quarter : 'YEAR_QUARTER',
        year : 'YEAR_QUARTER'
    };

    var Timescale = function (options) {
        var that = this;
        this.ganttView = options.ganttView;
        this.options = options.options; // TODO: odd
        this.widget = options.widget;
        this.$timescaleContainer = this.ganttView.$timescaleContainer;
        this.$rangeContainer = this.$timescaleContainer.find('.range-container');

        this.settings = {
            step : 1,
            minValue : 'day',
            midValue : 'week',
            maxValue : 'month',
            range : {
                start : this.options.range.start.clone().day(0).startOf('day'),
                end : this.options.range.end.clone().day(6).startOf('day')
            },
            width : 0
        };
        this.scales = {
            WEEK_DAY : new DefaultTimescale({
                range : this.settings.range,
                granularity : 'DAILY',
                value : {
                    min : 'day',
                    mid : 'week',
                    max : 'month'
                },

                isAvailable : function (start, end) {
                    if (!that.widget.options.config.chart.showUnavailableTime) {
                        return true;
                    }
                    if (that.widget.options.timescale.isAvailable != null) {
                        return that.widget.options.timescale.isAvailable(start, end);
                    }
                    var workweek = that.widget.getWorkWeek();

                    if (workweek && end.diff(start, 'days', true) <= 1) {
                        var dayOfWeek = start.day();
                        return workweek[dayOfWeek] != null ? workweek[dayOfWeek].length !== 0 : false;
                    }
                },

                format : {
                    min : 'ddd D',
                    mid : 'MMMM DD',
                    max : 'MMMM YYYY',
                    midPrefix : this.widget.options.locale.timescale.weekOf
                },
                bucketSize : 50
            }),
            MONTH_WEEK : new DefaultTimescale({
                range : this.settings.range,
                granularity : 'WEEKLY',
                value : {
                    min : 'week',
                    mid : 'month',
                    max : 'year'
                },
                format : {
                    min : 'DD',
                    mid : 'MMMM',
                    max : 'YYYY'
                },
                bucketSize : 80
            }),
            QUARTER_MONTH : new DefaultTimescale({
                range : this.settings.range,
                granularity : 'MONTHLY',
                value : {
                    min : 'month',
                    mid : 'quarter',
                    max : 'year'
                },
                format : {
                    min : 'MMMM',
                    mid : '\\QQ',
                    max : 'YYYY'
                },
                bucketSize : 100,
                buffer : 3
            }),
            YEAR_MONTH : new DefaultTimescale({
                range : this.settings.range,
                granularity : 'MONTHLY',
                value : {
                    min : 'month',
                    mid : 'year'
                },
                format : {
                    min : 'MMMM',
                    mid : 'YYYY'
                },
                bucketSize : 150,
                buffer : 3
            }),
            YEAR_QUARTER : new DefaultTimescale({
                range : this.settings.range,
                granularity : 'QUARTERLY',
                value : {
                    min : 'quarter',
                    mid : 'year'
                },
                format : {
                    min : '\\QQ',
                    mid : 'YYYY'
                },
                bucketSize : 150,
                buffer : 2
            })
        };
    };

    Timescale.prototype = {
        constructor : Timescale,

        createTimeScale : function () {
            this.applyTimescale(this.ganttView.options.config.timescale.type);

            this.initListeners();
        },

        initListeners : function () {
            var that = this;
            this.$timescaleContainer.on('mousedown', that.onMouseDown());
            this.$timescaleContainer.on('contextmenu', that.onContextMenu());
        },

        openContextMenu : function (event) {
            var that = this;
            var $menu = $(contextMenuTemplate({ locale : that.widget.options.locale }));
            $menu.find('.grid-menu-item[data-timescale="' + that.appliedTimescaleId + '"]').addClass('timescale-selected');
            $menu.css('left', event.pageX + 'px');
            $menu.css('top', event.pageY - $(window).scrollTop() + 'px');

            $(document.body).append($menu);
            $menu.find('.set-timescale').on('mousedown', function () {
                var timescale = $(this).data('timescale');
                that.ganttView.changeTimescale(timescale, false, { userAction : true });
            });
            $menu.find('.zoom-timescale').on('mousedown', function () {
                that.ganttView.zoomToFit();
            });
            $(document.body).on('mousedown.close-context', function () {
                $menu.remove();
                $(document.body).off('mousedown.close-context');
            });
        },

        setTimescaleRange : function () {
            var range = {};
            if (this.newMin != null || this.newMax != null) {
                range.start = this.newMin || this.settings.range.start;
                range.end = this.newMax || this.settings.range.end;
                this.newMin = null;
                this.newMax = null;
            } else {
                range = this.widget.getTimescaleRange();
            }

            this.settings.range = range;
            _.forEach(_.values(this.scales), function (timescale) {
                timescale.setRange({ start : range.start, end : range.end });
            });
        },

        zoomToFit : function () {
            var width = this.ganttView.chartWidth;
            var scale = this.whichScale(this.settings.range.start, this.settings.range.end, width, 20);

            var timescale = this.scales[scale.type];
            timescale.options.bucketSize = scale.size;
            timescale.setRange({ start : scale.min, end : scale.max });

            _.forEach(_.values(this.scales), function (timescale) {
                timescale.setRange({ start : scale.min, end : scale.max });
            });

            // assume have range
            // mod the settings
            // applies timescale of best fit (applyTimescale)
            this.ganttView.changeTimescale(scale.type, true, { userAction : true });
        },

        applyTimescale : function (timescale) {
            if (this.scales[timescale] == null) {
                timescale = 'QUARTER_MONTH';
            }
            if (this.scales[timescale] != null) {
                this.emptyTimescale();
                this.appliedTimescale = this.scales[timescale];
                this.appliedTimescaleId = timescale;
                var buckets = this.appliedTimescale.getBuckets();

                this.createRange({
                    buckets : buckets.min,
                    rangeClasses : 'min-range',
                    bucketClasses : ['min-value'],
                    resize : true
                });

                this.createRange({
                    buckets : buckets.mid,
                    rangeClasses : 'mid-range',
                    bucketClasses : ['mid-value']
                });
                this.createRange({
                    buckets : buckets.max,
                    rangeClasses : 'max-range',
                    bucketClasses : ['max-value']
                });

                this.settings.numberOfBuckets = buckets.min.length;
                this.settings.width = this.appliedTimescale.options.bucketSize * this.settings.numberOfBuckets;
                this.$timescaleContainer.css('width', this.settings.width + 'px');

                this.ganttView.$sightlineContainer.css({
                    'background-size' : this.appliedTimescale.options.bucketSize + 'px'
                });
                this.resetUnavailable();
            }
        },

        resetUnavailable : function () {
            var that = this;
            this.ganttView.$unavailableContainer.empty();
            var unavailables = _.reduce(this.$timescaleContainer.find('.min-value.unavailable'), function (memo, unavailable) {
                return memo + unavailableTemplate({
                    width : that.appliedTimescale.options.bucketSize,
                    left : $(unavailable).position().left
                });
            }, '');

            this.ganttView.$unavailableContainer.append(unavailables);
        },

        createRange : function (options) {
            var buckets = options.buckets;
            var $range = $(rangeTemplate({
                classes : options.rangeClasses
            }));
            this.$rangeContainer.prepend($range);

            if (buckets != null && buckets.length > 0) {
                var bucketMarkup = '';
                for (var i = 0; i < buckets.length; i++) {
                    var bucket = buckets[i];
                    var bucketClasses = options.bucketClasses;

                    if (bucket.available === false) {
                        bucketClasses = bucketClasses.concat(['unavailable']);
                    }

                    bucketMarkup += bucketTemplate({
                        resize : options.resize,
                        classes : bucketClasses.join(' '),
                        width : bucket.width,
                        label : bucket.label,
                        units : bucket.units
                    });
                }
                $range.append(bucketMarkup);
            }
        },

        emptyTimescale : function () {
            this.$rangeContainer.empty();
        },

        resizeTimescale : function (newBucketSize, quick) {
            this.appliedTimescale.options.bucketSize = newBucketSize;
            this.settings.width = this.appliedTimescale.options.bucketSize * this.settings.numberOfBuckets;

            this.$rangeContainer.find('.min-value').css({
                'padding-left' : '',
                'padding-right' : '',
                width : newBucketSize + 'px'
            });
            this.$rangeContainer.find('.mid-value,.max-value').each(function () {
                $(this).css({
                    'padding-left' : '',
                    'padding-right' : ''
                });
                $(this).width(parseFloat($(this).data('units')) * newBucketSize);
            });

            this.$timescaleContainer.css('width', this.settings.width + 'px');
            this.ganttView.$sightlineContainer.css({
                'background-size' : newBucketSize + 'px'
            });
            if (!quick) {
                this.resetUnavailable();
            }
        },

        getUnits : function (options) {
            return this.appliedTimescale.getUnits(options);
        },

        getOffset : function (date) {
            return this.appliedTimescale.getOffset(date);
        },

        whichScale : function (min, max, px, padding) {
            var usablePx;

            padding = padding || 0;
            usablePx = px - padding;

            var diffs = _.map(['day', 'week', 'month', 'quarter', 'year'], function (scale) {
                var diff = max.clone().diff(min, scale, true);
                var scaleFactor = usablePx / diff;
                return {
                    type : scale,
                    bucketQty : diff,
                    // Throw all values greater than 20 (min bucket viewable size) by setting to the MAX_SAFE_INTEGER
                    scaleFactor : scaleFactor > 20 ? scaleFactor : diff
                };
            });

            var closest = _.min(diffs, function (pair) {
                return pair.scaleFactor;
            });

            var timePerPx = closest.bucketQty / usablePx;

            return {
                min : min.clone().subtract((padding / 2) * timePerPx, closest.type),
                max : max.clone().add((padding / 2) * timePerPx, closest.type),
                type : MIN_TO_SCALE[closest.type],
                size : (usablePx / closest.bucketQty)
            };
        },

        minBucketSize : function (min, max, px, padding, scale) {
            var usablePx;
            var diff = max.diff(min, scale, true);

            padding = padding || 0;
            usablePx = px - padding;

            return usablePx / diff;
        },

        centerLabels : function (selector, timescaleOffset, timescaleVisibleWidth) {
            var $left, $right, leftPosition, leftWidth, rightPosition, rightWidth;

            var labels = this.$timescaleContainer.find(selector);

            // Wipe out any paddings that were originally on them
            labels.filter('.centered-label').css({
                'padding-left' : '',
                'padding-right' : ''
            }).removeClass('centered-label');

            _.forEach(labels, function (midValue) {
                var bucketLeft, bucketWidth;

                // Once found both left and right - can break out early
                if (!$left || !$right) {
                    bucketLeft = midValue.offsetLeft;
                    bucketWidth = midValue.offsetWidth;
                    if (timescaleOffset + bucketLeft < 0 &&
                        timescaleOffset + bucketLeft + bucketWidth > 0) {
                        $left = $(midValue);
                        leftPosition = bucketLeft;
                        leftWidth = bucketWidth;
                    }

                    if (timescaleOffset + bucketLeft + bucketWidth > timescaleVisibleWidth &&
                        timescaleOffset + bucketLeft < timescaleVisibleWidth) {
                        $right = $(midValue);

                        rightPosition = bucketLeft;
                        rightWidth = bucketWidth;
                    }
                }

                // https://lodash.com/docs#forEach "iterators may exit earlier by explicitly returning false"
                return !($left && $right);
            });

            if ($left &&
                (Math.abs(timescaleOffset) - leftPosition < leftWidth)) {
                $left.addClass('centered-label');
                $left.css({
                    'padding-left' : (Math.abs(timescaleOffset) - leftPosition) + 'px'
                });
            }

            if ($right &&
                (timescaleOffset + rightPosition + rightWidth - timescaleVisibleWidth) < rightWidth) {
                $right.addClass('centered-label');
                $right.css({
                    'padding-right' : (timescaleOffset + rightPosition + rightWidth) - timescaleVisibleWidth + 'px'
                });
            }
        },

        updateBucketLabelPositions : function () {
            var timescaleOffset = this.$timescaleContainer.position().left;
            var timescaleVisibleWidth = this.$timescaleContainer.parents('.timescale-viewport').width();

            this.centerLabels('.min-value', timescaleOffset, timescaleVisibleWidth);
            this.centerLabels('.mid-value', timescaleOffset, timescaleVisibleWidth);
            this.centerLabels('.max-value', timescaleOffset, timescaleVisibleWidth);
        }

    };
    _.assign(Timescale.prototype, timescaleMixins);

    return Timescale;
});
