define([
    'gridapi.lodash',
    'gridapi.logger',
    'gridapi.grid/rollupUtils'
], function (_, logger, rollupUtils) {
    'use strict';

    var NO_ARGS_FLAG = Object.freeze({});

    function RollupBase() {
        //>>excludeStart("validation", pragmas.production)
        logger.assert(this._enabled, 'subclasses must provide _enabled fn');
        logger.assert(this._getNode, 'subclasses must provide _getNode fn');
        logger.assert(this._getChildNodesFromNode, 'subclasses must provide _getChildNodesFromNode');
        logger.assert(this._getRecordsFromNode, 'subclasses must provide _getRecordsFromNode');
        logger.assert(this._getNodeId, 'subclasses must provide _getNodeId');
        logger.assert(this._makeRollupArgs, 'subclasses must provide _makeRollupArgs');
        logger.assert(this.widget && this.widget.view, 'subclasses must provide widget & view');
        //>>excludeEnd("validation")
    }

    RollupBase.prototype = {
        rollupFromLeafs : function (leafIds, dataIndices, rowsetId, args) {
            if (!this._enabled()) {
                return;
            }

            //>>excludeStart("perf", pragmas.compiled)
            window.incrementCallCount('rollupFromLeafs');
            //>>excludeEnd("perf")

            var nodeIds = rollupUtils.rollupOrder(leafIds, this._getNode.bind(this));
            this._processRollupsInOrder(nodeIds, dataIndices, rowsetId, args);
        },

        _processRollupsInOrder : function (groupIds, dataIndices, rowsetId, args) {
            //>>excludeStart("validation", pragmas.production)
            logger.assert(this._enabled(), 'dont call processRollups without checking that rollups are enabled first');
            //>>excludeEnd("validation")

            var commonArgs = this._makeCommonArgs(dataIndices);
            if (commonArgs === NO_ARGS_FLAG) {
                return;
            }

            for (var i = 0, len = groupIds.length; i < len; ++i) {
                var groupId = groupIds[i];
                var group = this._getNode(groupId);
                var childGroups = this._getChildNodesFromNode(group);
                this._processRollups(group, dataIndices, childGroups, rowsetId, commonArgs, args);
            }
        },
        _processRollups : function (node, dataIndices, childNodes, rowsetId, commonArgs, args) {
            if (commonArgs === NO_ARGS_FLAG) {
                return;
            }

            var records = this._getRecordsFromNode(node);
            var forceEmptyRollup = args && args.forceEmptyRollup;

            // Ensure all child groups rolled up and there are some records to roll up
            if (forceEmptyRollup || records.length || childNodes.length) {
                var rollupArgs = this._makeRollupArgs(node, dataIndices, records, childNodes);

                this.rollupStandardColumns(rollupArgs, commonArgs);
                this.rollupTimephasedColumns(rollupArgs, commonArgs);
                this.rollupStackedColumns(rollupArgs, commonArgs);

                if (rowsetId != null) {
                    this.widget.view.updateRow(this._getNodeId(node), rowsetId);
                }
            }
        },
        _makeCommonArgs : function (dataIndices) {
            var args = {
                standardColumns : this.getBasicRollupColumns(dataIndices),
                stackedColumns : this.getStackedColumns(dataIndices),
                stackedSubRows : this.getStackedSubRows(dataIndices),
                timephasedBuckets : this.getTimephasedBuckets(dataIndices),
                timephasedSubRows : this.getTimephasedSubRows(dataIndices)
            };
            var allAreEmpty = _.every(args, function (cols) {
                return _.isEmpty(cols);
            });
            if (allAreEmpty) {
                return NO_ARGS_FLAG;
            }
            return args;
        },

        /*
         *  Columns that rollup to attributes directly on the group. So anything that is not timephased or stacked.
         */
        getBasicRollupColumns : function (dataIndices) {
            var standardColumns;
            if (dataIndices == null) {
                standardColumns = _.values(this.widget.columnDefinitions);
            } else {
                standardColumns = _.pick(this.widget.columnDefinitions, dataIndices);
            }

            return this._sortAndFilterColumns(standardColumns);
        },
        _sortAndFilterColumns : function (columnDefinitions) {
            return _(columnDefinitions)
                .reject(function (colDef) {
                    return colDef.rollupRole == null;
                })
                .sortBy('rollupOrder')
                .value();
        },

        getTimephasedBuckets : function () {
            if (!this.widget.options.timephased.enabled) {
                return;
            }

            return _.filter(this.widget.columnDefinitions, function (colDef) {
                return colDef.timephased === true && colDef.visible;
            });
        },

        getTimephasedSubRows : function (dataIndices) {
            if (!this.widget.options.timephased.enabled) {
                return;
            }

            var that = this;
            var rollupSubRows = [];
            var timephasedColumnGroup = _.find(this.widget.config.groups, function (group) {
                return group.timephased;
            });
            if (dataIndices == null) {
                rollupSubRows = _.filter(this.widget.columnDefinitions, function (definition) {
                    return definition.rollupRole != null && (_.includes(timephasedColumnGroup.stackedCells, definition.dataIndex) ||
                        _.includes(timephasedColumnGroup.rolledUpStackedCells, definition.dataIndex));
                });
            } else {
                _.forEach(dataIndices, function (dataIndex) {
                    var colDef = that.widget.columnDefinitions[dataIndex];
                    if (colDef != null && colDef.rollupRole != null) {
                        if (_.includes(timephasedColumnGroup.stackedCells, colDef.dataIndex) ||
                            _.includes(timephasedColumnGroup.rolledUpStackedCells, colDef.dataIndex)) {
                            rollupSubRows.push(that.widget.columnDefinitions[dataIndex]);
                        }
                    }
                });
            }

            rollupSubRows = _.sortBy(rollupSubRows, 'rollupOrder');
            return rollupSubRows;
        },

        getStackedColumns : function () {
            if (!this.widget.options.stackedCells.enabled || !this.widget.options.stackedCells.rollupEnabled) {
                return;
            }

            var stackedColumns = [];
            var stackedColumnGroup = _.find(this.widget.config.groups, function (columnGroup) {
                return columnGroup.showStackedCells && !columnGroup.timephased && columnGroup.stackedCells.length;
            });
            _.forEach(stackedColumnGroup.columns, function (column) {
                if (column.useStackedContext) {
                    stackedColumns.push(column);
                }
            });

            return stackedColumns;
        },

        getStackedSubRows : function (dataIndices) {
            if (!this.widget.options.stackedCells.enabled || !this.widget.options.stackedCells.rollupEnabled) {
                return;
            }

            var that = this;
            var rollupSubRows = [];
            var stackedColumnGroup = _.find(this.widget.config.groups, function (columnGroup) {
                return columnGroup.showStackedCells && !columnGroup.timephased && columnGroup.stackedCells.length;
            });
            if (dataIndices == null) {
                rollupSubRows = _.filter(this.widget.columnDefinitions, function (definition) {
                    return definition.rollupRole != null && (_.includes(stackedColumnGroup.stackedCells, definition.dataIndex) ||
                        _.includes(stackedColumnGroup.rolledUpStackedCells, definition.dataIndex));
                });
            } else {
                if (!_.isArray(dataIndices)) {
                    dataIndices = [dataIndices];
                }
                _.forEach(dataIndices, function (dataIndex) {
                    var colDef = that.widget.columnDefinitions[dataIndex];
                    if (colDef != null && colDef.rollupRole != null) {
                        if (_.includes(stackedColumnGroup.stackedCells, colDef.dataIndex) ||
                            _.includes(stackedColumnGroup.rolledUpStackedCells, colDef.dataIndex)) {
                            rollupSubRows.push(that.widget.columnDefinitions[dataIndex]);
                        }
                    }
                });
            }

            rollupSubRows = _.sortBy(rollupSubRows, 'rollupOrder');
            return rollupSubRows;
        }
    };

    return RollupBase;
});
