define([
    'gridapi.jquery',
    'gridapi.lodash'
], function ($, _) {
    'use strict';

    /*
     * Data Related Actions: Cut, Copy, Paste, Delete, New    (Is this only for toolbar)
     * Store Related : Sort, Filter, Page, Load, Save, Cancel, Update, Undo, Redo, Load Children,
     * View Related : Resize, Select, ColumnToggle, Column Moved, Row Moved, Group
     *
     * Callable: Find, Find Next, Find Prev
     *
     *
     * Grid Triggered Actions:
     * Filldown, Delete (delete), Paste (paste), Copy (copy),
     *
     */
    var Mediator = function (store, widget) {
        var that = this;
        this.store = store;
        this.tempStore = {};
        this.options = widget.options;
        this.widget = widget;
        var mediatorFns = this.options.mediatorFns;
        // override mediator functions from options
        _.forEach(mediatorFns, function (fnc, fncName) {
            that[fncName] = fnc || that[fncName];
        });

        this.deletedRecords = [];
        this.modified = {};
        this.readOnly = {};
        this.newRecords = [];
        this.invalidFields = {};
        this.dataIcons = {};
        this.recordKeys = this.options.recordKeys;

        this.store = {};
        this.tree = store;
        this.loadChildrenIntoStore(store);
        this.fakeAjaxLoad = {};
    };

    Mediator.prototype = {
        constructor : Mediator,

        loadChildrenIntoStore : function (children) {
            for (var i = 0; i < children.length; i++) {
                if (children[i][this.recordKeys.id] < 0) {
                    this.newRecords[children[i][this.recordKeys.id]] = children[i];
                }
                this.store[children[i][this.recordKeys.id].toString()] = children[i];
                if (children[i].children != null) {
                    this.loadChildrenIntoStore(children[i].children);
                }
            }
        },

        sort : function (args) {
            return false;
        },

        group : function (args) {
            return false;
        },

        filter : function (args) {
            return false;
        },

        getRecord : function (id) {
            var record;
            if (id != null) {
                record = this.store[id];
                if (record == null) {
                    record = this.tempStore[id];
                }
                return record;
            }
        },

        getRecords : function (ids) {
            var records = [];
            var that = this;
            if (ids != null) {
                _.forEach(ids, function (id) {
                    records.push(that.getRecord(id));
                });
            } else {
                records = _.values(this.tree);
            }
            return records;
        },

        newRecord : function (/* type */) {
            var record = {};
            record[this.recordKeys.id] = '' + -(_.uniqueId() + 1);
            return record;
        },

        addRecords : function (records, options) {
            var that = this;
            var stripOut = [];
            var isTemp = options && options.isTemp;
            _.forEach(records, function (record) {
                if (!isTemp) {
                    var recordIndex = _.indexOf(that.deletedRecords, record);
                    if (recordIndex !== -1) {
                        delete that.deletedRecords[recordIndex];
                        stripOut.push(record);
                    }
                    if (record[that.recordKeys.id] < 0) {
                        that.newRecords.push(record);
                    }
                    if (record[that.recordKeys.parentId] != null) {
                        that.store[record[that.recordKeys.parentId]].children.push(record);
                    }
                } else {
                    that.tempStore[record[that.recordKeys.id]] = record;
                }
            });
            if (!isTemp) {
                _.forEach(_.difference(records, stripOut), function (recordToAdd) {
                    that.store[recordToAdd[that.recordKeys.id]] = recordToAdd;
                });
                this.widget._recordsAdded(records);
            }
        },

        copyRecords : function (records) {
            var that = this;
            var copiedRecords = [];

            for (var i = 0; i < records.length; ++i) {
                var newRecord = _.merge(that.newRecord(), records[i]),
                    newRecordId = '' + -(_.uniqueId());
                newRecord[that.recordKeys.id] = newRecordId;
                copiedRecords.push(newRecord);
            }
            return copiedRecords;
        },
        deleteRecords : function (records, options) {
            var isTemp = options && options.isTemp;
            for (var i = 0; i < records.length; i++) {
                var record = records[i];
                var parentId = record[this.recordKeys.parentId];
                if (parentId != null && !isTemp) {
                    this.store[parentId].children.splice(this.store[parentId].children.indexOf(record), 1);
                } else if (isTemp) {
                    delete this.tempStore[record[this.recordKeys.id]];
                }
            }
            if (!isTemp) {
                this.deletedRecords = _.union(this.deletedRecords, records);
                this.newRecords = _.difference(this.newRecords, this.deletedRecords);
                this.widget._recordsRemoved(records);
            }
        },
        attachListeners : function () {
            // Cant really watch an array now can we ....
        },
        placeRecord : function (record) {
            var location = {
                rowsetId : record.__location__,
                table : '$tbody',
                parentId : record.parentId
            };

            if (location.rowsetId == null || location.rowsetId === '') {
                location.rowsetId = this.widget.defaultRowset;
            }
            return location;
        },
        iterateStore : function (functionToRun, thisContext, args) {
            _.forEach(_.values(this.tree), function (record) {
                functionToRun.call(thisContext, record, args);
            });
        },
        updateRecord : function (record, key, data, silent) {
            var originalValue = this.getValueFromRecord(record, key);
            var isTemp = _.has(this.tempStore, record[this.recordKeys.id]);
            this.setRecordValue(record, key, data);

            var newValue = this.getValueFromRecord(record, key);
            if (!isTemp && originalValue !== newValue) {
                var modifiedValues = this.modified[record[this.recordKeys.id]];
                if (modifiedValues == null) {
                    modifiedValues = {};
                    this.modified[record[this.recordKeys.id]] = modifiedValues;
                }
                if (modifiedValues[key] == null) {
                    modifiedValues[key] = originalValue;
                } else if (newValue === modifiedValues[key]) {
                    delete modifiedValues[key];
                }
            }
            if (!silent) {
                this.widget._recordUpdated(record, key, data, true);
            }
        },


        markInvalid : function (recordId, key, messages, append) {
            var invalidFields = this.invalidFields[recordId];
            if (invalidFields == null) {
                invalidFields = {};
                this.invalidFields[recordId] = invalidFields;
            }
            if (append) {
                invalidFields[key] = _.defaults(messages, invalidFields[key]);
            } else if (messages == null || _.keys(messages).length <= 0) {
                delete invalidFields[key];
                if (_.keys(invalidFields).length <= 0) {
                    delete this.invalidFields[recordId];
                }
            } else {
                invalidFields[key] = messages;
            }
        },

        checkRecordUniqueness : function (dataIndex) {
            var that = this;
            var uniqueMatches = this.checkStoreUniqueness(dataIndex, this.getRecords());
            _.forEach(_.pairs(this.invalidFields), function (invalidPair) {
                var invalidMessages = invalidPair[1];
                var invalidRecordId = invalidPair[0];
                if (invalidMessages[dataIndex] != null && invalidMessages[dataIndex].unique != null) {
                    delete invalidMessages[dataIndex].unique;
                    that.markInvalid(invalidRecordId, dataIndex, invalidMessages[dataIndex]);
                    that.widget._updateRecordAndNode(invalidRecordId);
                }
            });
            function markRecord(record) {
                that.markInvalid(record[that.recordKeys.id], dataIndex,
                    {
                        unique : {
                            message : that.options.locale.uniqueError,
                            matches : _.without(matchingRecords, record)
                        }
                    }, true);
                that.widget._updateRecordAndNode(record[that.recordKeys.id]);
            }

            for (var i = 0; i < uniqueMatches.length; i++) {
                var matchingRecords = uniqueMatches[i];
                _.forEach(matchingRecords, markRecord);
            }
        },


        getValueFromRecord : function (record, key) {
            return record[key];
        },

        setRecordValue : function (record, key, data) {
            record[key] = data;
        },

        commit : function () {
            this.modified = {};
            this.readOnly = {};
            this.widget.repopulate();
        },

        // return an array of column keys that will be read only
        checkRecordPermissions : function (/* record */) {
            return [];
        },

        checkRecordIcon : function (/* record */) {
            return {};
        },

        recordMoved : function (record, siblingId, parentId, oldParentId, newRowsetId) {
            var oldRowsetId = null;
            if (oldParentId != null) {
                var oldParent = this.store[oldParentId];
                oldParent.children.splice(oldParent.children.indexOf(record), 1);
            }
            if (parentId != null) {
                var parent = this.store[parentId];
                parent.children.push(record);
            }
            if (newRowsetId != null) {
                oldRowsetId = record.__location__;
                record.__location__ = newRowsetId;
            }
            this.widget._recordMoved(record, oldRowsetId);
        },


        /* Build the record attributes object to help the grid render the rows and columns as a more correct representation of the
         * underlying data. The defaults parameter contains default values, if your mediator becomes outdated you can check
         * the defaults for what the grid expects/can handle.
         * For convenience in overriding behavior breakout methods have been added. They are documented below.
         */
        getRecordAttributes : function (record, defaults) {
            //>>excludeStart("perf", pragmas.compiled)
            window.incrementCallCount('boilerplateTreeMediator#getRecordAttributes');
            //>>excludeEnd("perf")

            var attributes = {};
            _.defaults(attributes, defaults);

            attributes.invalid = this.lookupInvalidFields(record);

            attributes.modified = this.lookupModifiedFields(record);

            attributes.readOnly = this.lookupReadOnlyFields(record);

            var treeDetails = this.lookupRecordTreeDetails(record);
            _.assign(attributes, treeDetails);

            var recordDetails = this.lookupRecordDetails(record);
            _.assign(attributes, recordDetails);

            return attributes;
        },


        /* Should return an object with invalid dataindexes as the keys and an object with errorId as the key and an object with
         * at least a message attribute for the error. This is an object to allow for extra information to be passed along very helpful for when errors are resolved.
         *
         * return {
         *  name : {
         *      format : {
         *          message : 'Name may not contain numbers',
         *      },
         *      unique : {
         *          message : 'Name must be unique',
         *         matching : [records] an array of other matching records
         *     }
         *  },
         *  type : {
         *     required : {
         *         message : 'Must set a type for the project'
         *     }
         *   }
         * }
         */
        lookupInvalidFields : function (record) {
            return this.invalidFields[this.getValueFromRecord(record, this.recordKeys.id)];
        },

        /* Should return an array of modified fields on the record. This will mark the field as changed via css. This should be
         * cached to maintain performance. Values should be set when data is being changed in updateRecord or as a function of
         * your data store.
         *
         * return ['name','description']
         */
        lookupModifiedFields : function (record) {
            return this.modified[this.getValueFromRecord(record, this.recordKeys.id)] || [];
        },

        /* Should return an array of fields that are read only for this record. This is separate from column level where
         * every record on the column is read only. You can use this to check type of object and disable fields. This should be
         * cached to maintain performance. I suggest caching in this method as many it will not change. Example if type is project
         * disable description field. This really only needs to be done one time. If a field that controls readOnlyFields changes,
         * call this method with force set to true to reset.
         *
         *  return ['type','start-date']
         */
        lookupReadOnlyFields : function (record, force) {
            var recordId = this.getValueFromRecord(record, this.recordKeys.id);
            if (force || this.readOnly[recordId] == null) {
                this.readOnly[recordId] = this.checkRecordPermissions(record);
            }
            return this.readOnly[recordId];
        },

        /* This method needs to set the following attributes. They only really need to be set on parent records. They do not
         * need to be set at all for non tree grids.
         *
         * return {
         *  isNode : if it has children / you want the + / icons
         *  hasChildren : determines if a +/- should be there, no children = no icon
         *  expanded : do you want the default to have all the children showing, if not set to false and they will be collapsed
         *  childrenLoaded : set to true unless it is lazyLoaded store. In this case you need a lookup
         *  sortable : Do you want the record to obey column sort. Example Portfolios -> Project -> Activity you may only want activities to obey sort
         *  groupable : Can this record be grouped, in the same concept of sortable often certain types are immune to grouping usually the top most parent
         * }
         */
        lookupRecordTreeDetails : function (record) {
            var attributes = {};
            if (record.children != null && record.children.length) {
                attributes.isNode = true;
                attributes.hasChildren = true;
                attributes.expanded = true;
                if (this.options.fakelazy) {
                    attributes.childrenLoaded = this.fakeAjaxLoad[this.getValueFromRecord(record, this.recordKeys.id)];
                } else {
                    attributes.childrenLoaded = true;
                }
                attributes.sortable = false;
                attributes.groupable = false;
            }

            return attributes;
        },
        /* Should return the icon object for this record. This record should cache the result of checkRecordIcon, if a
         * field that dictates this icon changes you should call this.checkRecordIcon(record)
         * return { icon }
         */
        lookupRecordIcons : function (record, force) {
            var recordId = this.getValueFromRecord(record, this.recordKeys.id);
            if (force || this.dataIcons[recordId] == null) {
                this.dataIcons[recordId] = this.checkRecordIcon(record);
            }
            return this.dataIcons[recordId];
        },

        /* Record details are for attributes that specifically relate to just the record. Expects to set
         * an object with attributes
         *
         * return {
         *  type : if needed for grouping
         *  newRow : if the record is new
         * }
         */
        lookupRecordDetails : function (record) {
            var attributes = {};
            attributes.type = this.getValueFromRecord(record, this.recordKeys.type);
            if (record[this.recordKeys.id] < 0) {
                attributes.newRow = true;
            }
            return attributes;
        },

        // Indicating a row has been dropped into a new rowset
        updateRecordPlacement : function (record, oldRowsetId, newRowsetId) {
            if (oldRowsetId !== newRowsetId) {
                record.__location__ = newRowsetId;
            }
        },

        getChildRecords : function (recordId) {
            var children = this.getRecord(recordId).children;
            if (children == null) {
                children = [];
            }
            return children;
        },

        loadChildren : function (recordId) {
            var that = this;
            var deferred = $.Deferred();
            setTimeout(function () {
                that.widget._childrenLoaded(recordId);
                that.fakeAjaxLoad[recordId] = true;
                deferred.resolve();
            }, 250 * (Math.floor(Math.random() * 8) + 1));
            return deferred;
        },
        updateRecordIds : function (record, idMap) {
            var oldRecordKeys = _.keys(idMap);
            var recordId = record[this.recordKeys.id];
            var parentId = record[this.recordKeys.parentId];
            if (_.includes(oldRecordKeys, recordId)) {
                record[this.recordKeys.id] = idMap[recordId];
                this.store[idMap[recordId]] = record;
                delete this.store[recordId];
                this.children[idMap[recordId]] = this.children[recordId];
                delete this.children[recordId];
            }
            if (_.includes(oldRecordKeys, parentId)) {
                record[this.recordKeys.parentId] = idMap[parentId];
            }
        },

        commitSuccessful : function (serverResponse) {
            var recordToSelect = this.widget.focus.record[this.recordKeys.id];
            if (serverResponse.updatedIdMap != null && _.keys(serverResponse.updatedIdMap).length) {
                this.iterateStore(this.updateRecordIds, this, serverResponse.updatedIdMap);
                if (serverResponse.updatedIdMap[recordToSelect] != null) {
                    recordToSelect = serverResponse.updatedIdMap[recordToSelect];
                }
            }

            this.modified = {};
            this.readOnly = {};
            for (var i = 0; i < this.deletedRecords.length; i++) {
                var deletedRecord = this.deletedRecords[i];
                delete this.store[deletedRecord[this.recordKeys.id]];
            }
            this.newRecords = [];
            this.rootRecordList = _.difference(this.rootRecordList, this.deletedRecords);
            this.deletedRecords = [];
            this.widget.repopulate(recordToSelect, this.widget.focus.columnId);
            this.widget._trigger('is-clean');
        },

        reset : function () {
            var modifiedRecords = _.pairs(this.modified);
            for (var i = 0; i < modifiedRecords.length; i++) {
                var modifiedRecordEntry = modifiedRecords[i];
                var record = this.getRecord(modifiedRecordEntry[0]);
                var modifications = _.pairs(modifiedRecordEntry[1]);
                for (var j = 0; j < modifications.length; j++) {
                    var keyValueModification = modifications[j];
                    record[keyValueModification[0]] = keyValueModification[1];
                }
            }
            this.deletedRecords = [];
            this.replacementMap = {};
            this.newRecords = [];
            this.modified = {};
            this.readOnly = {};
            this.widget.repopulate();
            this.widget._trigger('is-clean');
        },

        hasChanges : function () {
            return _.keys(this.modified).length || this.deletedRecords.length || this.newRecords.length;
        },

        getPostData : function () {
            var postData = null;
            if (this.hasChanges()) {
                var updatedRecords = this.getRecords(_.keys(this.modified));
                postData = {
                    added : this.newRecords,
                    updated : _.difference(updatedRecords, this.newRecords),
                    deleted : _.pluck(this.deletedRecords, this.recordKeys.id),
                    updatedIdMap : this.replacementMap
                };
            }
            return postData;
        }
        // return
    };

    return Mediator;
});
