define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.grid/utils/searchTermRegExpBuilder',
    'gridapi.grid/constants'
], function ($, _, searchTermRegExpBuilder, constants) {
    'use strict';

    var ColumnFilter = function (args) {
        var that = this;
        that.widget = args.widget;
        that.options = args.options;
        that.searchFields = {};
        that.matchingCount = 0;  // number of records that match the filter criteria.
    };

    ColumnFilter.prototype = {
        constructor : ColumnFilter,

        hasFilters : function () {
            return Boolean(this.columnFilter);
        },

        _getMatchingCount : function () {
            return this.matchingCount;
        },

        _clearColumnFilter : function () {
            _.forEach(this.searchFields, function (value) {
                if (value.term) {
                    value.term = '';
                }
            });
            this.resetSearchFields();
            this._columnBasedFilter({ optionUpdate : false });
        },

        _getColumnFilter : function () {
            var columnFilterState = {};
            _.forEach(this.columnFilter, function (val, key) {
                columnFilterState[key] = {
                    'dataIndex' : val.dataIndex,
                    'term' : val.term
                };
            });
            return columnFilterState;
        },

        _setColumnFilter : function (args) {
            this._clearColumnFilter();
            var that = this;
            var searchField;
            _.forEach(args.columnFilter, function (value, key) {
                searchField = that.searchFields[key];
                value.recentlyUpdated = true;
                searchField.term = value.term;
            });
            this._columnBasedFilter({ optionUpdate : false, columnFilter : args.columnFilter });
            this.resetSearchFields();
        },

        resetSearchFields : function () {
            var that = this;
            var gridOptions = that.options;
            var $th;
            var $input;
            if (gridOptions.search.columnFilter && _.findWhere(gridOptions.layout.rowsets, { id : constants.COLUMNFILTER_ROWSET_ID })) {
                _.forEach(that.searchFields, function (searchField) {
                    $th = searchField.$th;
                    $input = $th.find('input');
                    $input.val(searchField.term);
                });
            }
        },

        _columnBasedFilter : function (args) {       // LUX-506
            var that = this;
            var gridWidget = that.widget;
            var gridView = that.widget.view;
            var gridOptions = that.options;

            gridWidget.matchingRecords = { recordOrder : [] }; // clear matchingRecords for Logial AND / OR of filtering

            // don't recalculate everything just for optionUpdate call from gridWidget.options.search.columnFilterMatchAllColumns
            if (!args.optionUpdate) {
                var argColumnFilter = args.columnFilter;

                var columns = _.values(gridView.columnDefinitions);

                that.searchArgs = args;

                if (args != null && args.matchAllTerms != null) {
                    gridOptions.matchAllTerms = args.matchAllTerms;
                }
                if (_.isUndefined(argColumnFilter) || (argColumnFilter && _.keys(argColumnFilter).length === 0)) {
                    that.columnFilter = null;
                    gridWidget.clearSearch({ force : true });
                    that.matchingCount = -1;
                    return false;
                }
                gridWidget.searchCriteria = [];

                // use the cached that.columnFilter
                if (!that.columnFilter) {
                    that.columnFilter = argColumnFilter;
                } else {
                    _.forEach(argColumnFilter, function (value, key) {
                        if (that.columnFilter[key]) {
                            $.extend(that.columnFilter[key], value);
                        } else {
                            that.columnFilter[key] = value;
                        }
                    });
                }

                // if a column is removed from filter criteria(pressing backspace), remove it for further calcualtion.
                that.columnFilter = _.omit(that.columnFilter, function (value, key) {
                    return (argColumnFilter[key] ? false : true);
                });

                _.forEach(that.columnFilter, function (value) {
                    // if not recently updated do not calculate the regEx again
                    if (value.recentlyUpdated) {
                        // recalucate everything for that column (including the regEx for updated term and matchedRecords due to updated term)
                        value.previouslyCalculated = true;
                        value.matchedRecords = []; // will be used in _matchRecordColumnBasedFilter
                        value.term = value.term.trim();
                        var column = _.find(columns, function (col) {
                            return col.dataIndex === value.dataIndex;
                        });
                        column.searchMatcher = searchTermRegExpBuilder(value.term, {
                            matchOnAny : !gridOptions.matchAllTerms,
                            numbersUseStartsWith : false
                        });
                    }
                });
                gridWidget.stringReplacer = new RegExp('htmlReplaceTerm', 'ig'); // dummy
            }

            that._executeFiltering(args);
            that.matchingCount = gridWidget.matchingRecords.recordOrder.length;
        },

        _executeFiltering : function (args) {
            var that = this;
            var gridWidget = that.widget;
            var gridOptions = that.options;

            if (gridWidget.groupingApplied) {
                _.forEach(gridWidget.config.rowsets, function (rowset) {
                    if (rowset.rollupRowset != null) {
                        return;
                    }
                    var nodes = gridWidget.rootNodes['rowset-' + rowset.id] || [];
                    for (var i = 0; i < nodes.length; i++) {
                        gridWidget._matchGroupRecord(gridWidget._getRecord(nodes[i]));
                    }
                });
            } else if (gridOptions.tree && !gridOptions.lazyload) {
                gridWidget.storeRecords = [];
                if (gridOptions.tree && gridOptions.data != null) {
                    gridWidget.storeRecords = gridWidget._getRecordsByData({});
                } else {
                    gridWidget.store.iterateStore(gridWidget._getRecordFromStore, gridWidget, args);
                }

                _.forEach(gridWidget.storeRecords, function (record) {
                    gridWidget._matchTreeRecord(record);
                });
            } else {
                gridWidget.store.iterateStore(that._matchRecordColumnBasedFilter, that, args);
            }

            var oldOption = gridOptions.focusOnSelection;
            gridOptions.focusOnSelection = false;
            if (gridOptions.search.showAll) {
                gridWidget._forceViewRender();
            } else {
                gridWidget.repopulate();
            }
            gridOptions.focusOnSelection = oldOption;
        },

        _matchRecordColumnBasedFilter : function (record) {       // LUX-506
            var that = this;
            var gridWidget = that.widget;
            var gridOptions = that.options;

            var currentRecordId = gridWidget._getRecordId(record);

            var getRecordCellValue = function (record, dataIndex, column) {
                var value = gridWidget.store.getValueFromRecord(record, dataIndex);
                if (_.isFunction(column.getSearchValue)) {
                    return column.getSearchValue(record, dataIndex, column);
                } else if (column.formattedSearch) {
                    return column.formatter.format(value, column);
                }
                return value;
            };
            /* step 1
             * create matchedRecords, it is an array of recordIds for that column,
             * matchedRecords is defined in 'columnBasedFilter' and cleaned for every recentlyUpdated column
             */
            _.forEach(that.columnFilter, function (cFval) {
                var column = that.widget.view.columnDefinitions[cFval.dataIndex];
                if (cFval.recentlyUpdated) {
                    var recordCellVal = getRecordCellValue(record, cFval.dataIndex, column);
                    if (column.searchMatcher.test(recordCellVal)) {
                        cFval.matchedRecords.push(currentRecordId);
                    }
                }
            });

            /*Step two
             * create columnMatches --> array of dataIndexes  (per record)
             * And then most important data-structure 'matchingRecords' which is used for repopulating rows
             */
            var columnMatches = [];
            var columnFilterLength = _.keys(that.columnFilter).length;
            var iteration = 0;
            var returnVal = false;
            _.forEach(that.columnFilter, function (cFval) {
                iteration++;
                if (_.includes(cFval.matchedRecords, currentRecordId)) {
                    columnMatches.push(cFval.dataIndex);
                }
                if (iteration === columnFilterLength) { // at last iteration
                    if (gridOptions.search.columnFilterMatchAllColumns) { // all columns must match
                        if (iteration === columnMatches.length) { // match found in all columns for that 'currentRecordId'
                            gridWidget.matchingRecords[currentRecordId] = columnMatches;
                            gridWidget.matchingRecords.recordOrder.push(currentRecordId);
                            returnVal = true;
                        }
                    } else if (columnMatches.length > 0) { // record is rendered if match found in any column (found in atleast one col)
                        gridWidget.matchingRecords[currentRecordId] = columnMatches;
                        gridWidget.matchingRecords.recordOrder.push(currentRecordId);
                        returnVal = true;
                    }
                }
            });

            return returnVal;
        }


    };
    return ColumnFilter;
});
