define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.textInput',
    'gridapi.grid/utils/keyCode',
    'gridapi.grid/gridMixins',
    'gridapi.grid/constants',
    'gridapi.logger',
    'gridapi.grid/utils'
], function ($, _, textInput, keyCode, mixins, constants, logger, utils) {
    'use strict';
    var inputTextTemplate = utils.basicTemplate(textInput);

    var editor = function () {

    };

    editor.extend = mixins.extendFn;

    _.assign(editor.prototype, {

        init : function (options) {
            var that = this;
            var properties = [];

            _.forEach(options, function (value, name) {
                that[name] = value;
                properties.push(name);
            });

            // use to determine if commit is already in process
            // this is just used to prevent commit being called again from the grid
            this.commitStarted = false;
            this.ignoreSelectText = false;
            this.fromTab = false;
            // cache the property names we have set so we can delete later on
            that.prop = properties;

            this.column = this.modifyColumnSettings(this.column);
        },

        modifyColumnSettings : function (options) {
            return options;
        },

        getEditorId : function () {
            return this.grid.componentIds.editor;
        },

        closeEditorOnMouseDown : function (event) {
            return !$(event.target).parents('.cell-editor').length;
        },

        initialize : function (options) {
            var that = this;

            that.init(options);

            var $editor = that.make();
            $editor.attr('id', this.getEditorId());

            if (options.column.required || options.column.autoNumbered) {
                var message = (options.column.autoNumbered && (options.recordAttributes.newRow || options.fromQuickAddRow)) ? this.grid.options.locale.autoNumbered : this.grid.options.locale.required;
                $editor.find('input').attr('placeholder', '[' + message + ']');
            }

            if (!$editor.is('.cell-editor') && !$editor.is('.modal')) {
                throw new Error('Editor.create() - No "cell-editor" class.');
            }

            // If the grid is in the modal add the in-modal class that should boost the zIndex of the editor
            if (options.grid.options.inModal) {
                $editor.addClass('in-modal');
            }

            this.setPosition($editor);
            this.setSize($editor);

            // append to document as editor is absolute in its positioning
            that.grid.$grid.parent().append($editor);
            var $editorInput = this.getEditorInput();
            this.setAriaMarkup($editor, options);

            if (this.column.maxLength != null && this.column.maxLength > 0) {
                $editorInput.attr('maxlength', this.column.maxLength);
            }
            this.setDisabled($editorInput);
            that.initListeners();

            var $body = $(document.body);
            $body.on('mousedown' + that.grid.widget.eventNamespace + '-editor', function (event) {
                if (that.closeEditorOnMouseDown(event)) {
                    $body.off('mousedown' + that.grid.widget.eventNamespace + '-editor');
                    $editorInput.trigger('blur');
                    that.commit({});
                }
            });

            // Firefox likes to trigger scroll events when nothing has changed make sure something has changed before
            // closing the editors
            var $window = $(window);
            var windowScrollTop = $window.scrollTop();
            var windowScrollLeft = $window.scrollLeft();

            $window.on('scroll' + that.grid.widget.eventNamespace + '-editor', function () {
                if (windowScrollTop !== $window.scrollTop() || windowScrollLeft !== $window.scrollLeft()) {
                    $window.off('scroll' + that.grid.widget.eventNamespace + '-editor');
                    that.commit({});
                }
            });

            // When the window resizes close the editor so the layout does not make it appear incorrect.
            $window.on('resize' + that.grid.widget.eventNamespace + '-editor', function () {
                $window.off('resize' + that.grid.widget.eventNamespace + '-editor');
                that.commit({});
            });


            // Firefox likes to trigger scroll events when nothing has changed make sure something has changed before
            // closing the editors
            this.$scrollParent = that.grid.$grid.scrollParent();
            if (this.$scrollParent.length) {
                var scrollTop = this.$scrollParent.scrollTop();
                var scrollLeft = this.$scrollParent.scrollLeft();

                this.$scrollParent.on('scroll' + that.grid.widget.eventNamespace + '-editor', function () {
                    if (scrollTop !== that.$scrollParent.scrollTop() || scrollLeft !== that.$scrollParent.scrollLeft()) {
                        that.$scrollParent.off('scroll' + that.grid.widget.eventNamespace + '-editor');
                        that.commit({});
                    }
                });
            }


            return $editor;
        },

        setDisabled : function ($editorInput) {
            if (this.disabled) {
                $editorInput.attr('readonly', true);
            }
        },

        setAriaMarkup : function ($editorInput, options) {
            var that = this;
            $editorInput.attr('aria-labelledby', this.ariaLabels);
            if (options.column.describedBy != null && options.column.describedBy.length) {
                var ids = _.map(options.column.describedBy, function (item) {
                    return that.grid.descriptionIds.columnEditors[item] || '';
                });
                var describedBy = $editorInput.attr('arira-describedby');
                if (describedBy != null) {
                    describedBy += ' ' + ids.join(' ');
                } else {
                    describedBy = ids.join(' ');
                }
                $editorInput.attr('aria-describedby', describedBy);
            }
        },

        getEditor : function () {
            return $(document.body).find('#' + this.getEditorId());
        },

        getEditorInput : function () {
            return this.getEditor().find('input');
        },

        initListeners : function () {
            var that = this,
                $editorInput = this.getEditor();

            // stop double click events
            $editorInput.on('dblclick', function (event) {
                event.stopPropagation();
                event.preventDefault();
            });

            // add key events
            $editorInput.on('keypress', function (event) {
                that.onKeyPress(event);
            });

            // disable key events that we are interested in
            $editorInput.on('keyup', function (event) {
                that.onKeyUp(event);
            });


            // disable key events that we are interested in
            $editorInput.on('keydown', function (event) {
                that.onKeyDown(event);
            });

            $editorInput.on('blur', function (event) {
                event.preventDefault();
                event.stopPropagation();
            });
        },

        onKeyPress : function (event) {
            switch (event.keyCode) {
            case keyCode.ESCAPE :
            case keyCode.ENTER :
                event.stopPropagation();
                event.preventDefault();
                break;

            }
        },

        onKeyUp : function (event) {
            var that = this;
            var grid = that.grid;

            switch (event.keyCode) {
            case keyCode.ESCAPE :
                grid.setEscapeMode(false);
                    // escape - cancel editing
                that.close();
                event.stopPropagation();
                event.preventDefault();
                break;
            }
        },

        onTab : function (event) {
            var that = this;
            var grid = this.grid;
            event.stopPropagation();
            event.preventDefault();
            that.getEditorInput().trigger('blur.sync');
            that.commit({});
            grid.widget.editNextCell(!event.shiftKey);
        },

        onKeyDown : function (event) {
            var that = this;
            var grid = this.grid;
            switch (event.keyCode) {
            case keyCode.TAB :
                that.onTab(event);
                break;
                /* falls through */
            case keyCode.ENTER:
                var fromQuickAdd = that.fromQuickAddRow;
                grid.setEscapeMode(false);
                that.commit({ quickAdd: fromQuickAdd && !event.shiftKey });
                if (grid.options.moveDownOnEnter && !fromQuickAdd && !event.shiftKey) {
                    grid.selectionModel.cell.moveDown();
                } else if (fromQuickAdd && grid.widget.options.quickAdd.focusAfterAdd) {
                    grid.widget.selectCell(constants.DUMMY_RECORD_ID, grid.firstVisibleColumn);
                }
                event.stopPropagation();
                event.preventDefault();
                break;
            case keyCode.ESCAPE :
                event.stopPropagation();
                event.preventDefault();
                break;
            }
        },

        make : function () {
            return $(inputTextTemplate({}));
        },

        setPosition : function ($editor) {
            var $cellOffset = this.$cell.offset();
            var $body = $(window);
            $editor.css('top', $cellOffset.top - $body.scrollTop());
            $editor.css('left', $cellOffset.left - $body.scrollLeft());
        },

        setSize : function ($editor) {
            // ensure editor is not larger than the grid
            var suggestedWidth = parseInt(this.$cell.css('width'), 10);
            var gridBoundary = this.grid.$grid.offset().left + this.grid.$grid.width();
            var positionLeft = parseInt($editor.css('left'), 10);
            if (positionLeft + suggestedWidth > gridBoundary) {
                suggestedWidth = gridBoundary - positionLeft;
            }
            $editor.css('width', suggestedWidth);
            $editor.css('height', this.$cell.css('height'));
        },

        getNewData : function () {
            throw new Error('Editor.getNewData() needs to implemented.');
        },

        getCommitStarted : function () {
            return this.commitStarted;
        },

        show : function (startString) {
            if (this.disabled) {
                startString = null;
            }
            this.render(startString);
            var $editor = this.getEditor();

            var currentOwns = (this.grid.$grid.attr('aria-owns') || '').split(' ');
            this.grid.$grid.attr('aria-owns', _.union(currentOwns, [this.getEditorId()]).join(' ').trim());
            $editor.show();
        },

        render : function (startString) {
            if (startString != null) {
                // set typed character into input
                this.getEditorInput().val(startString);
                this.ignoreSelectText = true;
            } else {
                // if the data is not valid then just display it
                this.getEditorInput().val(this.formatValueForInput());
            }
        },

        formatValueForInput : function () {
            if (!this.invalid) {
                return this.column.formatter.format(this.data);
            } else {
                return this.data;
            }
        },

        validateUserInput : function (input) {
            return this.column.formatter.validate(input);
        },

        destroy : function () {
            var $editor = this.getEditor();
            if ($editor.length) {
                $editor.remove();
            }
        },

        drawFocusToInput : function (select) {
            var $input;
            var value;
            var inputElement;

            var $editorInput = this.getEditorInput();
            // draw input cursor into editor
            if ($editorInput.length) {
                $input = $editorInput;
            } else {
                $input = this.getEditor();
            }
            $input.focus();
            value = $input.val();
            select = !this.ignoreSelectText && (select || this.detectSelectText());
            inputElement = $input.get(0);
            if (_.isFunction(inputElement.setSelectionRange)) {
                inputElement.setSelectionRange(select ? 0 : value.length, value.length);
            }
        },

        detectSelectText : function () {
            if (this.column.selectTextTest != null && _.isFunction(this.column.selectTextTest.test)) {
                return this.column.selectTextTest.test(this.data);
            }
            return false;
        },

        selectText : function () {
            var $editorInput = this.getEditorInput();
            $editorInput.select();
        },

        clean : function () {
            var that = this;
            // delete any cached instances of variables that might interfere with the prototype versions
            _.forEach(that.prop, function (name) {
                delete that[name];
            });
            // remove newData property - set in some of the editors
            delete that.newData;

            // reset commit flag
            this.commitStarted = false;
        },

        hasDataChanged : function () {
            return !_.isEqual(this.getNewData(), this.data);
        },

        close : function () {
            if (!this.commitStarted || this.commitFinished) {
                $(document.body).off('mousedown' + this.grid.widget.eventNamespace + '-editor');
                $(window).off('scroll' + this.grid.widget.eventNamespace + '-editor');
                $(window).off('resize' + this.grid.widget.eventNamespace + '-editor');
                if (this.$scrollParent != null) {
                    this.$scrollParent.off('scroll' + this.grid.widget.eventNamespace + '-editor');
                }

                this.destroy();
                // must be called before clean
                var currentOwns = (this.grid.$grid.attr('aria-owns') || '').split(' ');
                var newOwns = _.difference(currentOwns, [this.getEditorId()]).join(' ').trim();

                if (newOwns === '') {
                    this.grid.$grid.removeAttr('aria-owns');
                } else {
                    this.grid.$grid.attr('aria-owns', newOwns);
                }
                this.onCloseFn();
                this.clean();
            }
        },

        commit : function (args) {
            if (!_.isObject(args)) {
                args = {};
            }
            this.commitStarted = true;
            this.getEditorInput().trigger('blur');
            if (!this.disabled) {
                var checkUserInput = this.getNewData();
                if (checkUserInput == null || !_.isObject(checkUserInput) || checkUserInput.valid == null) {
                    logger.warn('Update column config for getNewData to return { valid : true/false, value : value}');
                    checkUserInput = {
                        value : checkUserInput,
                        valid : true
                    };
                }

                var persist = {
                    quickAdd : args.quickAdd,
                    record : this.record,
                    value : checkUserInput.value,
                    valid : checkUserInput.valid,
                    column : this.column,
                    fromTab : this.fromTab
                };
                // persist the new value
                this.onPersistFn(persist);
            }
            this.commitFinished = true;
            // close editor
            this.close();
            return true;
        }
    });

    return editor;
});
