define([
    'gridapi.jquery',
    'gridapi.underscore',
    'gridapi.columns/core/editor',
    'gridapi.columns/core/renderer',
    'gridapi.columns/core/headerRenderer',
    'gridapi.grid/utils'
], function ($, _, editor, renderer, headerRenderer, utils) {
    'use strict';

    var inputTextTemplate = utils.basicTemplate('<div class="cell-editor cell-editor-select"><input type="text"/>' +
        '<a class="btn dropdown-toggle btn-blue-light"><span class="caret"/>' +
        '</a><ul class="hide cell-select-menu"></ul></div>');

    var singleOptionTemplate = utils.basicTemplate('<li class="select-menu-item <%=selected%>" data-id="<%=dataId%>"><%=name%></li>');

    var multiOptionTemplate = utils.basicTemplate('<li class="select-menu-item <%=selected%>" data-id="<%=dataId%>"><input type="checkbox" <%=checked%> class="dropdown-checkbox"><%=name%></li>');


    return {
        defaultOptions : {
            listDelim : ',',
            disableColumnFilter: false
        },
        formatter : {
            format : function (data, column) {
                var text = '';

                if (column.selectMultiple) {
                    // data should be an array of ids
                    _.forEach(data, function (id, index) {
                        text += column.selectValues[id].name;

                        if (index < data.length - 1) {
                            text += column.listDelim;
                        }
                    });
                } else {
                    // data should be one id
                    if (column.selectValues[data] != null) {
                        text = column.selectValues[data].name;
                    }
                }
                return text;
            },

            validate : function (/* data */) {
                return true;
            }
        },

        headerRenderer : headerRenderer.extend({

        }),

        renderer : renderer.extend({
            format : function (renderOptions) {
                var column = renderOptions.column;
                return column.formatter.format(renderOptions.data, column);
            }
        }),

        editor : editor.extend({
            init : function (options) {
                // have to clone as data will be an array for multi-select
                this.newData = _.clone(options.data);
                this.constructor.__super__.init(options);
            },

            updateData : function (id) {
                var that = this;

                // multi-select ->
                // if the id is in the array then remove it
                // if the id is NOT in the array then add it
                // single-select ->
                // replace the id with the new one
                if (that.column.selectMultiple) {
                    var found = false;
                    for (var index = 0; index < that.newData.length; index++) {
                        var thisId = that.newData[index];
                        if (id === thisId) {
                            that.newData.splice(index, 1);
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        that.newData.push(id);
                    }
                } else {
                    that.newData = id;
                }

                // re-render value in input
                that.render();
            },

            getNewData : function () {
                return {
                    value : this.newData,
                    valid : true
                };
            },

            getPossibleSelectValues : function () {
                if (_.isFunction(this.column.selectValues)) {
                    return this.column.selectValues(this.record, this.column);
                } else {
                    return this.column.selectValues;
                }
            },

            make : function () {
                var that = this;

                if (that.column.selectValues == null || that.column.selectValues.length === 0) {
                    throw new Error('DropDownColumn.make() - no values defined.');
                }

                var $markup = $(inputTextTemplate({})),
                    $ul = $markup.find('ul'),
                    $editorInput = $markup.find('input');

                if (that.column.disableTypeahead) {
                    $editorInput.attr('readonly', true);
                    $editorInput.addClass('read-only-editor');
                }

                if (that.column.selectMultiple) {
                    $markup.addClass('cell-editor-select-multi');
                    $ul.addClass('cell-select-multi-menu');
                }
                that.$ul = $ul;
                that.createDropDown($ul, that.sortValues(that.getPossibleSelectValues()));

                // when the drop down button is clicked toggle the drop down menu being visible
                $markup.find('a.dropdown-toggle').on('click', function (event) {
                    $ul.toggleClass('hide');
                    $markup.find('input').get(0).focus();
                    event.stopPropagation();
                    event.preventDefault();
                });

                this.$editor = $markup;

                return $markup;
            },

            createDropDown : function ($ul, values) {
                var that = this;

                // remove any previous drop down content
                $ul.empty();

                // create all the options
                _.forEach(values, function (value) {
                    if (that.column.selectMultiple) {
                        var checked = _.indexOf(that.data, value.id) > -1;
                        $ul.append($(multiOptionTemplate({
                            dataId : value.id,
                            name : value.name,
                            checked : checked ? 'checked="checked"' : '',
                            selected : checked ? 'menu-item-selected' : ''
                        })));
                    } else {
                        var selected = (value.id === that.data);
                        $ul.append($(singleOptionTemplate({
                            dataId : value.id,
                            name : value.name,
                            selected : selected ? 'menu-item-selected' : ''
                        })));
                    }
                });

                var $lis = $ul.find('li');

                // stop any double click events propagating
                $lis.on('dblclick', function (event) {
                    return event.stopPropagation();
                });

                // highlight the value checked
                $lis.find('input[type="checkbox"]').on('change', function (event) {
                    var $input = $(this),
                        $li = $input.closest('li'),
                        id = $li.attr('data-id');

                    if ($input.is(':checked')) {
                        $li.addClass('menu-item-selected');
                    } else {
                        $li.removeClass('menu-item-selected');
                    }
                    that.updateData(id);

                    // draw focus back to input
                    that.drawFocusToInput();

                    // stop other events occurring
                    event.stopPropagation();
                }).on('click', function (event) {
                        // stop div event below being triggered
                    event.stopPropagation();
                });

                $lis.on('mouseover', function (event) {
                    $ul.find('li.item-hovered').addClass('menu-item-hovered-transparent');
                });

                // highlight the value clicked
                $lis.on('click', function (event) {
                    var $span = $(this),
                        $li = $span.closest('li'),
                        id = $li.attr('data-id'),
                        $input = $li.find('input');

                    // if already clicked then removed selected
                    if ($li.is('.menu-item-selected')) {
                        $li.removeClass('menu-item-selected');
                        if ($input.length) {
                            $input.removeAttr('checked');
                        }
                    } else {
                        $input.attr('checked', 'checked');
                        if ($input.length) {
                            $li.addClass('menu-item-selected');
                        }
                    }
                    that.updateData(id);

                    // if in multi select mode then DON'T close drop down on selection
                    if (!that.column.selectMultiple) {
                        that.commit();
                    } else {
                        // draw focus back to input
                        that.drawFocusToInput();
                    }

                    // stop other events occurring
                    event.stopPropagation();
                });
            },

            moveSelection : function ($ul, up) {
                // use current list item selection and move from there, otherwise start from beginning
                var $lis = $ul.children();
                var $li = $lis.filter('.menu-item-hovered');
                if (!$li.length) {
                    $li = $lis.filter('.menu-item-selected');
                }
                if (!$li.length) {
                    $lis.first().addClass('menu-item-hovered');
                } else {
                    $li = $li.eq(0);
                    $li.removeClass('menu-item-hovered');
                    var $nextLi = up ? $li.prev() : $li.next();
                    if (!$nextLi.length) {
                        $nextLi = up ? $lis.last() : $lis.first();
                    }
                    $nextLi.addClass('menu-item-hovered');
                }
            },

            getEditorInput : function () {
                return this.getEditor().find('input[type=text]');
            },

            initListeners : function () {
                var that = this,
                    $ul = that.getEditor().find('.cell-select-menu'),
                    $dropDownButton = that.getEditor().find('a.dropdown-toggle'),
                    $editorInput = that.getEditorInput();

                // add actions to keyboard events
                $editorInput.on('keyup', function (event) {
                    $ul.find('.menu-item-hovered').removeClass('menu-item-hovered-transparent');

                    switch (event.keyCode) {
                    case $.ui.keyCode.ENTER:
                            // select the selected list item
                            // commit selection
                        var $lis = $ul.children();
                        $lis.filter('.menu-item-hovered').trigger('click');
                        break;
                    case $.ui.keyCode.ESCAPE :
                            // cancel editing
                        that.close();
                        break;
                    case $.ui.keyCode.UP:
                            // if the drop down menu is showing then cycle through the list items
                            // start from the selected list item
                            // if no list item is selected then start from the beginning of the list
                        if (!$ul.is('.hide')) {
                            that.moveSelection($ul, true);
                        }
                        break;
                    case $.ui.keyCode.DOWN:
                            // drop the menu into view if not already in view
                            // otherwise move down the items in the list
                            // cycle around once hitting the bottom
                        if ($ul.is('.hide')) {
                            $dropDownButton.trigger('click');
                        } else {
                            that.moveSelection($ul, false);
                        }
                        break;

                    case $.ui.keyCode.SPACE:
                        $ul.find('li.menu-item-hovered').click();
                        if (!that.column.selectMultiple) {
                            that.commit(event.keyCode);
                        }
                        break;
                    case 113 : // F2
                        $dropDownButton.trigger('click');
                        break;
                    case $.ui.keyCode.TAB :
                        var grid = that.grid;
                        if (event.shiftKey) {
                            grid.selectionModel.cell.moveLeft();
                        } else {
                            grid.selectionModel.cell.moveRight();
                        }
                        if (grid.options.editOnClick) {
                            grid.widget.editRow(grid.widget._getRecordId(grid.widget.focus.record),
                                    grid.widget.focus.columnId);
                        }
                        that.commit(false);

                        event.stopPropagation();
                        event.preventDefault();
                        break;
                    default:
                        if (!that.column.disableTypeahead) {
                            that.autoComplete($editorInput.val());
                        }
                    }
                });


                // stop any key events we care about getting any further
                $editorInput.on('keydown', function (event) {
                    switch (event.keyCode) {
                    case $.ui.keyCode.TAB :
                        event.preventDefault();
                        /* falls through */
                    case $.ui.keyCode.ESCAPE:
                    case $.ui.keyCode.ENTER:
                        event.stopPropagation();
                        break;
                    }
                });
            },

            autoComplete : function (text) {
                var that = this,
                    values = [],
                    $ul = that.getEditor().find('.cell-select-menu'),
                    lastValue = text.toLowerCase().split(that.column.listDelim).pop();

                // add any currently selected list items
                if (that.column.selectMultiple) {
                    var $lis = $ul.children().filter('.menu-item-selected');
                    _.forEach(that.column.selectValues, function (value) {
                        _.forEach($lis, function (li) {
                            if (value.id === $(li).attr('data-id')) {
                                values.push(value);
                            }
                        });
                    });
                }

                // find possible values for the entered text
                _.forEach(that.column.selectValues, function (value) {
                    if (value.name.toLowerCase().indexOf(lastValue) === 0) {    // must be the beginning
                        // only add value if it is not already added
                        if (!_.includes(values, value)) {
                            values.push(value);
                        }
                    }
                });

                that.createDropDown($ul, that.sortValues(values));

                // select the first list item
                $ul.children().first().addClass('menu-item-selected');

                // show drop down if not already
                if ($ul.is('.hide')) {
                    $ul.toggleClass('hide');
                }
            },

            sortValues : function (values) {
                if (this.column.sortValues != null) {
                    return this.column.sortValues(values);
                }
                return values;
            },

            placeDropDown : function () {
                var height = this.$ul.height(),
                    position = this.$editor.position(),
                    windowHeight = $(window).height();
                if (position.top + height > windowHeight) {
                    this.$ul.addClass('menu-on-top');
                    this.$ul.css('top', -(height));
                }
            },

            render : function (keyCode, shiftKey) {
                this.placeDropDown();
                if (keyCode == null) {
                    this.getEditorInput().val(this.column.formatter.format(this.newData, this.column));
                } else {
                    this.constructor.__super__.render(keyCode, shiftKey);
                }
            }
        })
    };
});
