define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.grid/utils/keyCode',
    'gridapi.grid/gridMixins',
    'gridapi.logger',
    'gridapi.grid/utils',
    'gridapi.grid/view/menuUtils'
], function ($, _, keyCode, mixins, logger, utils, menuUtils) {
    'use strict';
    var menuTemplate = utils.basicTemplate('<ul  aria-label="<%=menuLabel%>" id="<%=id%>"  role="menu" tabindex="-1" class="grid-menu grid-root-menu"><div class="menu-items"></div></ul>');
    var itemTemplate = utils.basicTemplate('<li role="menuitem" class="grid-menu-item" id="<%=id%>" title="<%=title%>" data-action-index="<%=actionIndex%>" data-trigger="<%=trigger%>"><%=pre%><a href="#" class="item-label"><%=label%></a><%=post%></li>');
    var itemGroupTemplate = utils.basicTemplate('<li role="menuitem" class="grid-menu-group" id="<%=id%>" title="<%=title%>"><%=pre%><a aria-haspopup="true" aria-expanded="false" tabindex="-1" href="#" class="item-label"><%=label%></a><%=post%><ul class="grid-menu grid-sub-menu"></ul></li>');
    var noItemsTemplate = utils.basicTemplate('<div role="menuitem" class="disabled no-items-available" id="<%=id%>" title="<%=label%>"><div class="item-label"><%=label%></div></div>');

    function relativePosition($relativeElement, parentIsBody, $popup) {
        var position;

        if (parentIsBody) {
            position = $relativeElement.offset();
        } else if ($relativeElement.parent().length !== 0) {
            var parent = $relativeElement.parent();
            var additionalPosition = { top : 0, left : 0 };
            position = $relativeElement.position();

            while (parent.offsetParent()[0] !== $popup.offsetParent()[0]) {
                var parentOffset = $(parent.offsetParent()).position();
                additionalPosition.top += parentOffset.top;
                additionalPosition.left += parentOffset.left;
                parent = parent.offsetParent();
            }

            position.top += additionalPosition.top;
            position.left += additionalPosition.left;
        } else {
            position = $relativeElement.position();
        }

        return position;
    }

    var Menu = function (grid) {
        var that = this;
        that.options = grid.options.contextMenu;
        that.grid = grid;
        _.forEach(that.options.menuFns, function (fnc, fncName) {
            that[fncName] = fnc || that[fncName];
        });
        that.isOpen = false;
        that.$menu = that.make();
        that.$focusedItem = $();
        // If the grid is in the modal add the in-modal class that should boost the zIndex of the editor
        if (this.grid.options.inModal) {
            that.$menu.addClass('in-modal');
        }
        this.timeouts = utils.timeoutFactory();
        menuUtils.allowRefocusOn(this, 'close');
    };


    Menu.prototype = {
        constructor : Menu,

        appendMenu : function () {
            this.initListeners();
            // append to document as editor is absolute in its positioning
            this.grid.view.$grid.parent().append(this.$menu);
        },

        initListeners : function () {
            var that = this;

            // disable key events that we are interested in
            this.$menu.on('keyup', function (event) {
                that.onKeyUp(event);
            });


            // disable key events that we are interested in
            this.$menu.on('keydown', function (event) {
                that.onKeyDown(event);
            });

            this.$menu.on('click', function (event) {
                var $target = $(event.target).closest('.grid-menu-item');
                if ($target.length) {
                    var actionIndex = $target.attr('data-action-index');
                    var action = that.options.actions[actionIndex];
                    that.triggerAction(action);
                }
            });

            this.$menu.on('contextmenu', function (event) {
                that.close();
                event.preventDefault();
                event.stopPropagation();
            });


            var $body = $(document.body);
            $body.on('mousedown' + that.grid.eventNamespace + '-menu', function (event) {
                var $target = $(event.target);
                if (!$target.parents('.grid-menu').length && !$target.parents('.row-menu-button').length) {
                    $body.off('mousedown' + that.grid.eventNamespace + '-menu');
                    that.close();
                } else {
                    event.preventDefault();
                }
            });

            // Firefox likes to trigger scroll events when nothing has changed make sure something has changed before
            // closing the editors
            var $window = $(window);
            var windowScrollTop = $window.scrollTop();
            var windowScrollLeft = $window.scrollLeft();
            $window.on('scroll' + that.grid.eventNamespace + '-menu', function () {
                if (windowScrollTop !== $window.scrollTop() || windowScrollLeft !== $window.scrollLeft()) {
                    $window.off('scroll' + that.grid.eventNamespace + '-menu');
                    that.close();
                }
            });

            // When the window resizes close the editor so the layout does not make it appear incorrect.
            $window.on('resize' + that.grid.eventNamespace + '-menu', function () {
                $window.off('resize' + that.grid.eventNamespace + '-menu');
                that.close();
            });


            // Firefox likes to trigger scroll events when nothing has changed make sure something has changed before
            // closing the editors
            this.$scrollParent = that.grid.view.$grid.scrollParent();
            if (this.$scrollParent.length) {
                var scrollTop = this.$scrollParent.scrollTop();
                var scrollLeft = this.$scrollParent.scrollLeft();
                this.$scrollParent.on('scroll' + that.grid.eventNamespace + '-menu', function () {
                    if (scrollTop !== that.$scrollParent.scrollTop() || scrollLeft !== that.$scrollParent.scrollLeft()) {
                        that.$scrollParent.off('scroll' + that.grid.eventNamespace + '-menu');
                        that.close();
                    }
                });
            }
        },
        onClose : function (fn) {
            this.closeFn = fn;
        },
        triggerAction : function (action) {
            if (action.enabled) {
                var args = _.clone(this.getTriggerArgs(action)) || {};
                args.focus = this.context.focus;
                args.records = this.context.records;
                args.focusInSelection = this.context.focusInSelection;


                if ($.isFunction(this.grid[action.trigger])) {
                    this.grid[action.trigger].call(this.grid, args);
                    // trigger action on widget because it exists
                } else {
                    // trigger general action
                    this.grid._trigger(action.trigger, null, args);
                }
                if (!action.leaveOpen) {
                    this.close();
                    if (action.reopenAfter) {
                        var that = this;
                        setTimeout(function () {
                            that.grid.view.$grid.trigger({ type : 'contextmenu' });
                        }, 0);
                    }
                }
            }
        },

        onKeyUp : function (event) {
            switch (event.keyCode) {
            case keyCode.ESCAPE:
                this.close();
                break;
            case keyCode.SPACE:
            case keyCode.ENTER:
                if (this.$focusedItem.length) {
                    var actionIndex = this.$focusedItem.attr('data-action-index');
                    if (actionIndex != null && actionIndex.length) {
                        this.triggerAction(this.options.actions[actionIndex]);
                    }
                } else {
                    this.close();
                }
                break;


            }
        },
        setFocus : function ($item) {
            if ($item.length) {
                if (this.$focusedItem.find($item).length) {
                    this.$focusedItem.addClass('child-focused');
                }
                if ($item.is('.child-focused')) {
                    $item.removeClass('child-focused');
                }
                this.$focusedItem.removeClass('item-focused');

                this.$focusedItem = $item;
                this.$focusedItem.addClass('item-focused');
                this.$menu.attr('aria-activedescendant', $item.attr('id'));
            }
        },

        clearFocus : function () {
            if (this.$focusedItem.length) {
                this.$focusedItem.removeClass('item-focused');
                this.$focusedItem = $();
            }
        },

        onKeyDown : function (event) {
            switch (event.keyCode) {
            case keyCode.DOWN :
                if (this.$focusedItem.length) {
                    var $nextItem = this.$focusedItem.next();
                    if ($nextItem.is('.menu-divider')) {
                        $nextItem = $nextItem.next();
                    }
                    this.setFocus($nextItem);
                } else {
                    this.focusFirstItem();
                }
                event.preventDefault();
                event.stopPropagation();
                break;
            case keyCode.UP :
                if (this.$focusedItem.length) {
                    var $prevItem = this.$focusedItem.prev();
                    if ($prevItem.is('.menu-divider')) {
                        $prevItem = $prevItem.prev();
                    }
                    this.setFocus($prevItem);
                } else {
                    this.focusFirstItem();
                }
                event.preventDefault();
                event.stopPropagation();
                break;
            case keyCode.RIGHT :
                if (this.$focusedItem.length && this.$focusedItem.is('.grid-menu-group')) {
                    this.setFocus(this.$focusedItem.children().children().first('.grid-menu-item,.grid-menu-group'));
                }
                event.preventDefault();
                event.stopPropagation();
                break;
            case keyCode.LEFT :
                if (this.$focusedItem.length) {
                    var $grandParent = this.$focusedItem.parent().parent();
                    if ($grandParent.is('.grid-menu-group')) {
                        this.setFocus($grandParent);
                    }
                }
                event.preventDefault();
                event.stopPropagation();
                break;

            case keyCode.TAB :
                this.close();
                event.preventDefault();
                break;
            }
        },

        make : function () {
            return $(menuTemplate({
                id : this.grid.view.componentIds.menu,
                menuLabel : this.grid.options.locale.cellState.menuLabel
            }));
        },

        setContext : function (record) {
            var selectedRecords = this.grid.getSelection();
            this.context = {
                focus : record,
                records : selectedRecords,
                clipboard : this.grid.clipboard,
                numberOfSelectedRecords : selectedRecords.length,
                focusInSelection : _.includes(selectedRecords, record)
            };
            if (!this.context.focusInSelection && record != null) {
                this.context.records = [record];
            }

            var actionIndexes;
            if (this.grid._getRecordId(record) === 'DUMMY-NEW-RECORD') {
                if (this.options.populate != null) {
                    this.isEmpty = true;
                    return;
                }
                actionIndexes = this.getQuickAddRowActions();
            } else if (this.grid._isGroupRecord(record)) {
                actionIndexes = this.getGroupRowActions();
            } else {
                if (this.options.populate != null) {
                    this.convertPopulateResults(this.options.populate(this.context.focus, this.grid.clipboard));
                    logger.warn('contextMenu.populate is deprecated, Please see documentation to convert to the new standard');
                }
                actionIndexes = this.getActions();
            }


            var $itemContainer = this.$menu.find('.menu-items');
            $itemContainer.empty();
            this.makeMenu(actionIndexes, $itemContainer, {
                actionState : this.grid._preCheckActions()
            });
            if (!$itemContainer.children().length) {
                var $item = $(noItemsTemplate({
                    id : 'menu-item-no-items-available-' + this.grid.uuid,
                    label : this.grid.options.locale.noActionsAvailable
                }));
                $itemContainer.append($item);
                this.$menu.attr('aria-activedescendant', $item.attr('id'));
            }
        },

        convertPopulateResults : function (populateResults) {
            this.options.actions = {};
            this.options.actionOrder = [];
            for (var i = 0; i < populateResults.length; i++) {
                var oldAction = populateResults[i];
                var actionIndex = 'option' + i;
                if (!_.isEmpty(oldAction)) {
                    this.options.actions[actionIndex] = {
                        label : oldAction.label,
                        triggerArgs : oldAction.triggerArgs,
                        trigger : oldAction.action,
                        enabled : !oldAction.disabled
                    };
                    this.options.actionOrder.push(actionIndex);
                } else {
                    this.options.actionOrder.push('');
                }
            }
        },

        focusFirstItem : function () {
            var $item = this.$menu.children().find('.grid-menu-item,.grid-menu-group').first();
            this.setFocus($item);
        },

        makeMenu : function (actionList, $container, args) {
            for (var i = 0; i < actionList.length; i++) {
                var actionIndex = actionList[i];
                if (actionIndex === '') {
                    if (!$container.children().last().is('.menu-divider') && $container.children().length) {
                        $container.append('<div role="separator" class="menu-divider"></div>');
                    }
                } else {
                    $container.append(this.makeActionItem(this.options.actions[actionIndex], actionIndex, args));
                }
            }

            var lastItem = $container.children().last();
            if (lastItem.is('.menu-divider')) {
                lastItem.remove();
            }
        },

        makeActionItem : function (action, actionIndex, args) {
            var $item = $();
            action.actionIndex = actionIndex;

            if (action.cellAction != null) {
                if ((this.openedFromCell && !action.cellAction) || (!this.openedFromCell && action.cellAction)) {
                    return $item;
                }
            }

            if (action.isGroup) {
                $item = $(itemGroupTemplate({
                    id : 'menu-item-' + action.actionIndex + '-' + this.grid.uuid,
                    title : action.title || action.label,
                    actionIndex : action.actionIndex,
                    label : this.getProperLabel(action),
                    pre : this.getPreFragment(action),
                    post : this.getPostFragment(action)
                }));
                var $subMenu = $item.find('.grid-sub-menu');
                this.makeMenu(action.actions, $subMenu);

                // If the subMenu ends up with no children remove it
                if ($subMenu.children().length < 1) {
                    $item = $();
                }
            } else {
                this.setActionEnabled(action, args);
                $item = $(itemTemplate({
                    id : 'menu-item-' + action.actionIndex + '-' + this.grid.uuid,
                    title : action.title || action.label,
                    actionIndex : action.actionIndex,
                    label : this.getProperLabel(action),
                    pre : this.getPreFragment(action),
                    post : this.getPostFragment(action),
                    trigger : action.trigger
                }));
                if (!action.enabled) {
                    if (this.options.hideDisabled) {
                        $item = $();
                    } else {
                        $item.addClass('disabled');
                    }
                }
            }
            return $item;
        },

        getActions : function () {
            return this.options.actionOrder;
        },

        getGroupRowActions : function () {
            return this.options.groupRowActions;
        },

        getQuickAddRowActions : function () {
            return ['quickAddRowNewRow', 'pasteRow'];
        },

        setActionEnabled : function (action, args) {
            var enableAction = true;

            if ($.isFunction(this.grid[action.trigger])) {
                var state = args && args.actionState || this.grid._preCheckActions(this.context.records, this.context.focus);
                if (this.context.focus == null) {
                    state.hasSelection = false;
                    state.selectedRecords = [];
                    state.focusedRecord = null;
                }
                enableAction = this.grid._checkAction(action.trigger, state, action.triggerArgs);
                enableAction = this.overrideGridAction(action, enableAction);
            } else {
                enableAction = this.checkActionEnabled(action);
            }
            action.enabled = enableAction;
        },

        checkActionEnabled : function (action) {
            return action.enable;
        },

        overrideGridAction : function (action, enabled) {
            return enabled;
        },

        getProperLabel : function (action) {
            var actionLabel;
            if (action.pluralLabel != null) {
                if (action.trigger === 'pasteCells' || action.trigger === 'pasteRows') {
                    if (this.context.clipboard.records.length > 1) {
                        actionLabel = this.getPluralLabel(action, this.context.clipboard.records.length);
                    }
                } else if (this.context.numberOfSelectedRecords > 1 && (this.context.focusInSelection || action.trigger === 'newRow')) {
                    actionLabel = this.getPluralLabel(action, this.context.numberOfSelectedRecords);
                }
            }
            if (actionLabel == null) {
                actionLabel = this.getLabel(action);
            }
            return actionLabel;
        },

        getLabel : function (action) {
            if (this.openedFromCell && action.cellMenuLabel != null) {
                return action.cellMenuLabel;
            }
            return action.label;
        },

        getPluralLabel : function (action /*, count */) {
            return action.pluralLabel;
        },

        getTriggerArgs : function (action) {
            var triggerArgs = _.clone(action.triggerArgs);
            if (triggerArgs == null) {
                triggerArgs = {};
            }
            if (action.trigger === 'newRow') {
                if (triggerArgs.skipEdit == null) {
                    triggerArgs.skipEdit = true;
                }
                if (this.grid._isGroupRecord(this.context.focus)) {
                    triggerArgs.isChild = true;
                }
            }
            return triggerArgs;
        },

        getPreFragment : function (action) {
            var iconSpacer = this.options.enableIcons ? '<div class="icon-spacer"></div>' : '';
            return action.preFragment || iconSpacer;
        },
        getPostFragment : function (action) {
            return action.postFragment || '';
        },

        setPosition : function ($relativeElement) {
            var top, left, position,
                offset = $relativeElement.offset(),
                popupHeight = this.$menu.outerHeight(true),
                popupWidth = this.$menu.width();

            // Remove this if it was once positioned above the element
            this.$menu.removeClass('grid-root-menu-up grid-root-menu-down');

            // Determine base relative element coordinates
            position = relativePosition($relativeElement, true, this.$menu);

            // Set up base css top/left coordinates
            top = position.top + $relativeElement.outerHeight();
            left = position.left;

            if ($(document).scrollLeft() !== 0) {
                left -= $(document).scrollLeft();
            }

            if ($relativeElement.is('.row-menu-button')) { // Fix to make it work for larger rows.
                top = position.top + $relativeElement.outerHeight() / 2 + 13;
            }

            // Correct for window dimensions
            if (offset.top + popupHeight + $relativeElement.outerHeight() > window.pageYOffset + $(window).height()) {
                top -= (popupHeight + $relativeElement.outerHeight());
                this.$menu.addClass('grid-root-menu-up');
            } else {
                this.$menu.addClass('grid-root-menu-down');
            }

            if (offset.left + popupWidth > window.pageXOffset + $(window).width()) {
                left -= (popupWidth - $relativeElement.innerWidth());
            }
            //Dont allow context menu to position off the screen.
            if (top < 0) {
                top = 0;
            }

            // Set the coordinates
            this.$menu.css({
                'top' : top - $(window).scrollTop(),
                'left' : left
            });
        },

        show : function (element, offsetX) {
            this.closed = false;
            if (!this.isOpen) {
                this.appendMenu();
            }
            this.setPosition(element, offsetX);
            var currentOwns = (this.grid.view.$grid.attr('aria-owns') || '').split(' ');
            this.grid.view.$grid.attr('aria-owns', _.union(currentOwns, [this.grid.view.componentIds.menu]).join(' ').trim());

            this.requestBrowserFocus();
            this.isOpen = true;
            if (this.openedFromCell) {
                this.grid.view.$grid.addClass('menu-open');
            }
        },

        setOpenedFromCell : function (openedFromCell) {
            this.openedFromCell = openedFromCell;
        },

        destroy : function () {
            if (this.$menu.length) {
                this.$focusedItem = $();
                this.$menu.remove();
            }
        },

        requestBrowserFocus : function () {
            this.allowRefocus();
            this.$menu.get(0).focus();
        },

        close : function () {
            if (!this.closed) {
                $(document.body).off('mousedown' + this.grid.eventNamespace + '-menu');
                $(window).off('scroll' + this.grid.eventNamespace + '-menu');
                $(window).off('resize' + this.grid.eventNamespace + '-menu');
                if (this.$scrollParent != null) {
                    this.$scrollParent.off('scroll' + this.grid.eventNamespace + '-menu');
                }
                this.grid.view.$grid.removeClass('menu-open');
                this.destroy();

                var currentOwns = (this.grid.view.$grid.attr('aria-owns') || '').split(' ');
                var newOwns = _.difference(currentOwns, [this.grid.view.componentIds.menu]).join(' ').trim();

                if (newOwns === '') {
                    this.grid.view.$grid.removeAttr('aria-owns');
                } else {
                    this.grid.view.$grid.attr('aria-owns', newOwns);
                }

                this.isOpen = false;
                // must be called before clean
                if (this.closeFn != null) {
                    this.closeFn();
                    delete this.closeFn;
                }
                this.openedFromCell = false;
                this.grid.requestBrowserFocus(true);
                this.$menu.removeAttr('aria-activedescendant');
                this.closed = true;
                this.timeouts.clear();
            }
        }
    };

    return Menu;
});
