define([
    'gridapi.lodash',
    'gridapi.logger',
    'gridapi.grid/constants'
], function (_, logger, constants) {
    'use strict';

    var ALL_PROPS_TYPE = 'all-props';
    var PARTIAL_PROPS_TYPE = 'partial-props';

    var noopCacheFactory = function () {
        var noopApi = {};

        noopApi.forRecord = function () {
            return {};
        };
        noopApi.apply = function () {
        };
        noopApi.clearRecord = function () {
        };

        return noopApi;
    };
    var realCacheFactory = function () {
        var cacheApi = {};
        var cache = new WeakMap();

        cacheApi.clearRecord = function (record) {
            if (cache.has(record)) {
                return cache.delete(record);
            }
        };

        cacheApi.forRecord = function (record) {
            if (record == null) {
                return {};
            }
            if (cache.has(record)) {
                return cache.get(record);
            }
            var attrs = {};
            cache.set(record, attrs);
            return attrs;
        };
        cacheApi.apply = function (record, attrs) {
            if (record != null) {
                cache.set(record, attrs);
            }
        };

        return cacheApi;
    };

    var NON_SELECTION_ATTRS = 'nonSelection';
    var TREE_ATTRS = 'tree';
    var ALL_ATTRS = 'all';
    var factory = function attrsFactory(options) {
        var api = {};

        var cacheStack = [];
        var cache = noopCacheFactory();
        api.pushCache = function (name) {
            if (!cacheStack.length) {
                cache = realCacheFactory();
            }
            cacheStack.push(name);
        };
        api.popCache = function () {
            cacheStack.pop();
            if (!cacheStack.length) {
                cache = noopCacheFactory();
            }
        };

        api.nonSelection = api.nonSelectionAttrs = function nonSelectionAttrs(record) {
            return api.attrs(record, NON_SELECTION_ATTRS);
        };
        api.tree = api.treeAttrs = function treeAttrs(record) {
            return api.attrs(record, TREE_ATTRS);
        };
        api.all = api.allAttrs = function allAttrs(record) {
            return api.attrs(record, ALL_ATTRS);
        };

        api.attrs = function attrsApi(record, attrGroup) {
            var attrTypeDef = options.attrTypes[attrGroup];

            //>>excludeStart("perf", pragmas.compiled)
            window.incrementCallCount('attrFns#attrs (' + attrGroup + ')');
            //>>excludeEnd("perf")
            //>>excludeStart("validation", pragmas.production)
            logger.assert(attrTypeDef, 'must request one of known attrGroups');
            //>>excludeEnd("validation")

            var fallback = options.fallback;
            if (fallback.mustApply(record)) {
                //>>excludeStart("perf", pragmas.compiled)
                window.incrementCallCount('attrFns#attrs -- fallback to widget#_getRecordAttributes');
                //>>excludeEnd("perf")
                return fallback.strategy(record);
            }

            var knownAttrs = cache.forRecord(record);
            var strategies = attrTypeDef;
            for (var i = 0, len = strategies.length; i < len; ++i) {
                var strat = strategies[i];
                if (strat.test(knownAttrs)) {
                    break;
                }
                if (strat.applicable()) {
                    //>>excludeStart("validation", pragmas.compiled)
                    window.incrementCallCount('attrFns#attrs -- applying strategy ' + strat.name);
                    //>>excludeEnd("validation")
                    knownAttrs = strat.apply(knownAttrs, strat.fn(record), record, cache);
                    break;
                }
            }

            return knownAttrs;
        };

        api.clear = function clearAttrs(record) {
            cache.clearRecord(record);
        };

        //>>excludeStart("testonlyAccess", pragmas.compiled)
        api.getOptions = function getOptions() {
            return options;
        };
        //>>excludeEnd("testonlyAccess")

        return api;
    };

    factory.fromGrid = function fromGrid(grid) {
        var knownTreeAttrs = ['isNode', 'hasChildren', 'childrenLoaded', 'precedence', 'rollup', 'sortable', 'expanded', 'canHaveChildren'];
        var treeAttrStrategy = {
            name : 'tree',
            fn : function getTreeAttrs(record) {
                return grid._getRecordTreeAttributes(record);
            },
            applicable : _.once(function () {
                return !!grid.store.getRecordTreeAttributes;
            }),
            type : PARTIAL_PROPS_TYPE,
            apply : function (knownAttrs, treeAttrs) {
                //>>excludeStart("validation", pragmas.production)
                logger.assert(_.isEmpty(_.omit(treeAttrs, knownTreeAttrs)), 'no props should be returned by getRecordTreeAttributes except for the 6 known keys');
                //>>excludeEnd("validation")
                for (var i = 0; i < knownTreeAttrs.length; i++) {
                    var key = knownTreeAttrs[i];
                    knownAttrs[key] = treeAttrs[key];
                }
                return knownAttrs;
            },
            test : function (knownAttrs) {
                return knownAttrs.hasChildren != null;
            }
        };
        var fullAttrStrategy = {
            name : 'full',
            fn : function fallbackToFullAttrs(record) {
                return grid._getRecordAttributes(record);
            },
            applicable : _.constant(true),
            type : ALL_PROPS_TYPE,
            apply : function (knownAttrs, allAttrs, record, cache) {
                cache.apply(record, allAttrs);
                return allAttrs;
            },
            test : function (knownAttrs) {
                return knownAttrs.selected != null;
            }
        };
        var EXCLUDE_SELECTION = { includeSelection : false };
        var nonSelectionAttrStrategy = _.defaults({
            name : 'nonSelection',
            fn : function getNonSelectionAttrs(record) {
                return grid._getRecordAttributes(record, EXCLUDE_SELECTION);
            },
            test : function (knownAttrs) {
                return knownAttrs.editable != null;
            }
        }, fullAttrStrategy);

        var options = {
            fallback : {
                strategy : fullAttrStrategy.fn,
                mustApply : function shouldFallback(record) {
                    if (grid._isGroupRecord(record)) {
                        return true;
                    }
                    var recordId = grid._getRecordId(record);
                    if (recordId === constants.DUMMY_RECORD_ID) {
                        return true;
                    }
                    var proxyRecord = grid._getProxyRecord();
                    if (proxyRecord != null) {
                        return recordId === grid._getRecordId(proxyRecord);
                    }
                    return false;
                }
            },
            attrTypes : {
                nonSelection : [
                    nonSelectionAttrStrategy,
                    fullAttrStrategy
                ],
                tree : [
                    treeAttrStrategy,
                    fullAttrStrategy
                ],
                all : [fullAttrStrategy]
            }
        };
        return factory(options);
    };

    factory.decorateCache = function (privateProp, cacheName, fn) {
        return function cacheMethodWrapper() {
            this[privateProp].pushCache(cacheName);
            // assuming here that the fn will never throw an error...gonna be a bad time if so, granted
            var result = fn.apply(this, arguments);
            this[privateProp].popCache();
            return result;
        };
    };

    return factory;
});
