define([
    'gridapi.jszip',
    'gridapi.moment',
    'gridapi.lodash',
    'gridapi.grid/constants',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-worksheets-sheet-header',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-worksheets-sheet-footer',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-rels-rels',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-docProps-app',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-docProps-core',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-calcChain',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-workbook',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-styles',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-theme-theme1',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-xl-rels-workbook',
    'markup!gridapi.grid/templates/exportToXlsx.htm:.htm-xlsx-content-types'
], function (JSZip, moment, _, constants, exportWorksheetHeader, exportWorksheetFooter, exportXlsxRelsRels,
             exportXlsxDocPropsApp, exportXlsxDocPropsCore, exportXlsxXlCalcChain, exportXlsxXlWorkbook,
             exportXlsxXlStyles, exportXlsxXlThemeTheme, exportXlsxXlRelsWorkbook, exportXlsxContentTypes) {
    'use strict';

    var CELL_NAME_REGEX = /([A-Z]+)(\d+)/;

    function getDownloadContentFormat(fileType) {
        var format = {};
        switch (fileType) {
        case 'csv':
            format.mimeType = 'data:text/csv';
            format.fileExtention = '.csv';
            break;
        default:
            format.mimeType = 'application/vnd.ms-office';
            format.fileExtention = '.xlsx';
        }
        return format;
    }

    function downloadContent(content, fileName, fileType) {
        var format = getDownloadContentFormat(fileType);
        var blob = new window.Blob([content], { type: format.mimeType });
        var url = window.URL.createObjectURL(blob);
        var link = window.document.createElement('a');
        var ev;

        link.href = url;
        link.download = fileName + format.fileExtention;
        link.textContent = 'save';

        if (window.navigator.msSaveOrOpenBlob) {
            window.navigator.msSaveOrOpenBlob(blob, link.download);
        } else if (window.document.createEvent) {
            ev = window.document.createEvent('MouseEvents');
            ev.initMouseEvent('click', true, true, window,
                0, 0, 0, 0, 0, false, false, false, false, 0, null);
            link.dispatchEvent(ev);
        } else {
            link.click();
        }
    }

    function getTopLeftCell(frozenRows, frozenColumns) {
        var columns = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K'];
        var topLeftCol = frozenColumns <= 10 ? columns[frozenColumns] : columns[10];
        var topLeftRow = frozenRows + 1;
        return topLeftCol + topLeftRow;
    }

    function getFrozenRows(gridView) {
        var frozenRows = 1;
        var seenNonFixed = false;
        var fixedRowset = _.find(gridView.config.rowsets, function (rowset) {
            if (!(rowset.id === 'headers' || rowset.id === constants.QUICK_ADD_ROWSET_ID)) {
                if (rowset.type !== 'fixed') {
                    seenNonFixed = true;
                }
                return !seenNonFixed;
            }
            return false;
        });
        if (fixedRowset != null) {
            frozenRows += gridView.virtualRows[fixedRowset.id].records.length;
        }
        return frozenRows;
    }

    function getFrozenColumns(gridView) {
        var frozenColumns = 0;
        var fixedGroup = _.find(gridView.config.groups, function (group) {
            return !group.internalDisplayOnly && group.type === 'fixed';
        });
        if (fixedGroup != null) {
            frozenColumns = fixedGroup.columns.length;
        }
        return frozenColumns <= 10 ? frozenColumns : 10; // limit to 10
    }

    /**
     * Get the Excel name for a cell from its row and column in a spreadsheet.
     * For example:
     *  getExcelCellName(1, 1) => 'A1'
     *  getExcelCellName(5, 10) => 'J5'
     *  getExcelCellName(1, 70) => 'BR1'
     *
     * @method getExcelCellName
     * @private
     * @param rowNum Cell's row in the spreadsheet
     * @param columnNum Cell's column in the spreadsheet
     * @returns {string} Excel name of the cell at (row, column)
     */
    function getExcelCellName(rowNum, columnNum) {
        var ALPHABET_LENGTH = 26;
        var columnName = '';

        while (columnNum > 0) {
            var mod = (columnNum - 1) % ALPHABET_LENGTH;
            columnName = String.fromCharCode(65 + mod) + columnName;
            columnNum = parseInt((columnNum - mod) / ALPHABET_LENGTH);
        }

        return columnName + rowNum;
    }

    function getColorFill(color) {
        return '<fill><patternFill patternType="solid"><fgColor rgb="' + color.replace('#', '') + '"/><bgColor indexed="64"/></patternFill></fill>';
    }

    function getFillId(color, exportData) {
        if (exportData.fillIds[color]) {
            return exportData.fillIds[color];
        } else {
            var id = exportData._nextFillId;
            exportData.fills.push(getColorFill(color));
            exportData.fillIds[color] = id;
            exportData._nextFillId++;
            return id;
        }
    }

    function getXfId(xf, exportData) {
        if (_.isNumber(exportData.xfIds[xf])) {
            return exportData.xfIds[xf];
        } else {
            var id = exportData._nextXfId;
            exportData.xfs.push(xf);
            exportData.xfIds[xf] = id;
            exportData._nextXfId++;
            return id;
        }
    }

    function buildXf(cellData, exportData) {
        var xf = '<xf borderId="0" ';
        if (cellData.applyNumberFormat && cellData.percent) {
            xf += 'numFmtId="9" applyNumberFormat="1" ';
        }
        if (cellData.color) {
            xf += 'fillId="' + getFillId(cellData.color, exportData) + '" applyFill="1" ';
        }
        if (cellData.link) {
            xf += 'fontId="1" xfId="1" ';
        }

        xf += 'applyAlignment="1"><alignment vertical="center" ';

        if (cellData.outlineLevel) {
            xf += 'indent="' + cellData.outlineLevel + '" ';
        }
        if (cellData.wrapText) {
            xf += 'wrapText="1" ';
        }

        xf += '/></xf>';

        return xf;
    }

    function addMergedCell(name, exportData) {
        var cellColRow = name.match(CELL_NAME_REGEX); // "AA100" => ["AA100", "AA", "100"]
        var column = cellColRow[1];
        var cellAbove = column + (parseInt(cellColRow[2]) - 1);

        if (!exportData.mergedCellRanges[column]) {
            exportData.mergedCellRanges[column] = [];
        }

        var lastMerge = _.last(exportData.mergedCellRanges[column]);

        if (!lastMerge) {
            exportData.mergedCellRanges[column].push({
                from: cellAbove,
                to: name
            });
        } else if (lastMerge.to === cellAbove) {
            lastMerge.to = name;
        } else {
            exportData.mergedCellRanges[column].push({
                from: cellAbove,
                to: name
            });
        }
    }

    function getRow(rowData, exportData) {
        var row = '<row ' + rowData.attrs + '>';
        _.forEach(rowData.cells, function (cellData) {
            row += '<c r="' + cellData.name + '" ';

            var xf = buildXf(cellData, exportData);
            var xfId = getXfId(xf, exportData);
            if (xfId) {
                row += 's="' + xfId + '" ';
            }

            if (cellData.mergedCell) {
                addMergedCell(cellData.name, exportData);
            }

            if (!cellData.value) {
                row += '><v></v></c>';
            } else if (cellData.inline) {
                // the added blank character is *not* a space, but rather ALT+255
                // this is to ensure inline values are considered strings by Excel
                // allowing numbers and costs to be exported from text columns
                row += 't="inlineStr"><is><t>' + cellData.value + ' ' + '</t></is></c>';
            } else if (cellData.link) {
                row += 't="str"><f>HYPERLINK("' + cellData.url + '", "' + cellData.value + '")</f><v>' + cellData.value + '</v></c>';
                exportData.calcChains.push('<c l="1" i="1" r="' + cellData.name + '"/>');
            } else {
                row += '><v>' + cellData.value + '</v></c>';
            }
        });
        return row + '</row>';
    }

    function getMergedCells(exportData) {
        var merges = [];
        _.forEach(Object.keys(exportData.mergedCellRanges), function (column) {
            _.forEach(exportData.mergedCellRanges[column], function (range) {
                merges.push('<mergeCell ref="' + range.from + ':' + range.to + '"/>');
            });
        });
        return merges;
    }

    function exportToXlsx(rowsAndHeaders, gridView) {
        var zip = new JSZip();
        var rels = zip.folder('_rels');
        var docProps = zip.folder('docProps');
        var xl = zip.folder('xl');
        var xlRels = xl.folder('_rels');
        var xlTheme = xl.folder('theme');
        var xlWorksheets = xl.folder('worksheets');

        var frozenRows = getFrozenRows(gridView);
        var frozenColumns = getFrozenColumns(gridView);
        var topLeftCell = getTopLeftCell(frozenRows, frozenColumns);

        var exportData = {
            xfs: [],
            xfIds: {},
            fills: [],
            fillIds: {},
            calcChains: [],
            mergedCellRanges: {},
            // Excel boilerplate styles at indices 0 and 1, custom styles start at index 2
            _nextFillId: 2,
            _nextXfId: 0
        };

        var headers = getRow(rowsAndHeaders.headers, exportData);
        var rows = '';
        _.forEach(rowsAndHeaders.rows, function (rowData) {
            rows += getRow(rowData, exportData);
        });
        rows = headers + rows;

        var mergedCells = getMergedCells(exportData);

        zip.file('[Content_Types].xml', exportXlsxContentTypes());
        rels.file('.rels', exportXlsxRelsRels());
        docProps.file('app.xml', exportXlsxDocPropsApp());
        docProps.file('core.xml', exportXlsxDocPropsCore({
            created: moment().format(),
            creator: gridView.options.export.user
        }));
        xl.file('styles.xml', exportXlsxXlStyles({
            colorFills: exportData.fills.join(''),
            xfs: exportData.xfs.join('')
        }));
        xl.file('workbook.xml', exportXlsxXlWorkbook({
            worksheetName: gridView.options.export.worksheetName
        }));
        xlRels.file('workbook.xml.rels', exportXlsxXlRelsWorkbook({
            calcChain : exportData.calcChains.length
        }));
        xlTheme.file('theme1.xml', exportXlsxXlThemeTheme());
        xlWorksheets.file('sheet1.xml', exportWorksheetHeader({
            frozenRows: frozenRows,
            frozenColumns: frozenColumns,
            topLeftCell: topLeftCell
        }) + rows + exportWorksheetFooter({
            numMergedCells: mergedCells.length,
            mergedCells: mergedCells.join('')
        }));

        if (exportData.calcChains.length) {
            xl.file('calcChains.xml', exportXlsxXlCalcChain({
                calcChains: exportData.calcChains.join('')
            }));
        }

        zip.generateAsync({
            type: 'blob',
            compression: 'DEFLATE',
            compressionOptions: {
                level: 6
            }
        }).then(function (content) {
            downloadContent(content, gridView.options.export.fileName);
        });
    }

    function escapeCSVField(value) {
        return ((value === null) || (value === undefined)) ? '' : value.replace(/"/g, "\"\"");
    }

    function generateCSVContent(exportData) {
        var csvData = '';
        //headers
        _.forEach(exportData.headers.cells, function (cell) {
            csvData = csvData + '"' + escapeCSVField(cell.value.trim()) + '",';
        });
        csvData += '\n';

        // data
        _.forEach(exportData.rows, function (row) {
            _.forEach(row.cells, function (cell) {
                csvData = csvData + '"' + escapeCSVField(cell.value) + '",';
            });
            csvData += '\n';
        });

        return csvData;
    }

    function exportToCSV(exportData, fileName) {
        var csvData = generateCSVContent(exportData);
        downloadContent(csvData, fileName, 'csv');
    }

    return {
        toExcel : exportToXlsx,
        toCSV : exportToCSV,
        getExcelCellName : getExcelCellName,
        _getTopLeftCell : getTopLeftCell
    };
});
