define([
    'gridapi.moment',
    'gridapi.i18n/localizedStrings',
    'gridapi.grid/constants'
], function (moment, locale, constants) {
    'use strict';
    /**
     * @typedef grid~options#actionConfig
     * @property {String} trigger The method to trigger on the grid when the cell toolbar button is pressed.
     * @property {Boolean} gridAction Whether or not this is a grid action or a customized action outside of the grid.
     * @property {String} iconClass The icon to be displayed with this action.
     * @property {Boolean} enabled Used to enable/disable the action.
     */
    /**
     * @typedef grid~options#cellToolbarConfig
     *
     * @property {Boolean} [enabled=false] Used to enable/disable the cellToolbar
     * @property {Array<String>} [actionOrder=['newRow', 'newChildRow', 'deleteRows']] Used to specify the order of actions in the cell toolbar.
     * @property {Object<String, grid~options#actionConfig>} [actions] Used to define actions that are available to the cell toolbar.
     * @property {grid~options#actionConfig} [actions.newRow] Creates a new row in the grid. In hierarchical grids, it is created at the current level.
     * @property {String} [actions.newRow.trigger='newRow']
     * @property {Boolean} [actions.newRow.gridAction=true]
     * @property {String} [actions.newRow.iconClass='pgbu-icon-add-button']
     * @property {Boolean} [actions.newRow.enabled=true]
     * @property {grid~options#actionConfig} [actions.newChildRow] Creates a new child row in the grid.  Only applicable to hierarchical grids.
     * @property {String} [actions.newChildRow.trigger='newChildRow']
     * @property {Boolean} [actions.newChildRow.gridAction=true]
     * @property {String} [actions.newChildRow.iconClass='pgbu-icon-forced-in']
     * @property {Boolean} [actions.newChildRow.enabled=true]
     * @property {grid~options#actionConfig} [actions.deleteRows] Deletes the currently selected rows from the grid.
     * @property {String} [actions.deleteRows.trigger='deleteRows']
     * @property {Boolean} [actions.deleteRows.gridAction=true]
     * @property {String} [actions.deleteRows.iconClass='pgbu-icon-delete-button']
     * @property {Boolean} [actions.deleteRows.enabled=true]
     *
     * @example
     * //This is not currently used in Prime, please check with core team if you are using this as it may have outdated styles.
     * cellToolbar : {
                        enabled : false,
                        actionOrder : ['newRow', 'newChildRow', 'deleteRows'],
                        actions : {
                            newRow : {
                                trigger : 'newRow',
                                gridAction : true,
                                iconClass : 'pgbu-icon-add-button',
                                enabled : true
                            },
                            newChildRow : {
                                trigger : 'newChildRow',
                                gridAction : true,
                                iconClass : 'pgbu-icon-forced-in',
                                enabled : true
                            },
                            deleteRows : {
                                trigger : 'deleteRows',
                                gridAction : true,
                                iconClass : 'pgbu-icon-delete-button',
                                enabled : true
                            }
                        }
                    }
     */
    /**
     * @typedef grid~options#contextMenuConfig
     *
     * @property {Boolean} [enabled=false] Whether or not the context menu is displayed.
     * @property {Boolean} [enableIcons=false] Whether or not the context menu can display icons. Do not change this for Prime.
     * @property {Boolean} [enableHoverRowMenuButton=true] Whether or not the context menu will be available on hover.
     * @property {Boolean} [enableRowMenuButton=false] Whether or not the context menu will be available as the 'gear' menu.
     * @property {Number} [rowMenuButtonWidth=27] Setting for resize calculations, do not change unless styles are modified.
     * @property {Number} [offsetTop=0] Number of pixels to offset the top of the rendered context menu.
     * @property {Boolean} [hideDisabled=true] Whether or not to hide disabled context menu options. Prime should leave this as true.
     * @property {Boolean} [allowDisjointSelection=false] When opening context menus should the grid select each row.
     * @property {Array} [actionOrder] Defines the order that the actions will appear in the context menu list.  Actions should be in a comma separated list.  To
     *     add a separator between actions, use an empty character ''.
     * <br/><br/>Example:<br/>
     * // continuous action list<br/>
     * actionOrder: ['newRow', 'newChildRow', 'deleteRows']<br/><br/>
     * // separated action list - '' will add a horizontal rule to action list.<br/>
     * actionOrder: ['newRow', 'newChildRow', 'deleteRows', '', 'copyCells']
     * @property {Object} [actions] Action definitions
     * @property {String} [actions.action.label='Add'] Label for the action to appear in the menu
     * @property {String} [actions.action.trigger='newRow'] Trigger for the action to have, either a grid method call or custom grid-triggerValue event will be
     *     fired
     * @property {Object} [actions.action.triggerArgs=''] Arguments to pass into the grid method call.
     * @property {Boolean} [actions.action.leaveOpen='false'] Should the context menu stay open when you trigger the action. Turning this on would allow for
     *     multiple clicking of the action.
     * @property {Object} [menuFns] Overrides for the contextMenu methods, defined as { methodToOverride : function, method2ToOverride : function }
     * @property {Function} [checkActionEnabled] Method that can be used to control when an action is enabled/disabled in the context menu list of actions.
     *     Typically used to enable/disable custom actions.
     * <br/><br/>Example:<br/>
     * menuFns : {<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;checkActionEnabled : function (action) {<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;if (action.trigger === 'customAction') {<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;return this.grid.store.canPerformCustomAction();<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;} else {<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;return action.enable;<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;}<br/>
     * &nbsp;&nbsp;&nbsp;&nbsp;}<br/>
     * }
     * @property {Array} [groupRowActions] Actions that can be called on a group record.  Should be a comma separated list of action names.
     * @example
     *  //For more information see the following blog :  https://goo.gl/3Ljcdx
     * contextMenu : {
                        enabled : false,
                        enableIcons : false,
                        enableHoverRowMenuButton : true,
                        enableRowMenuButton : false,
                        rowMenuButtonWidth : 27,
                        offsetTop : 0,
                        hideDisabled : true,
                        allowDisjointSelection : false,
                        actions : {
                            quickAddRowNewRow : {
                                label : 'Add',
                                trigger : 'newRow'
                            }
                        },
                        groupRowActions : []
                }
     */
    /**
     * @typedef grid~options#timephasedConfig
     *
     * @property {Boolean} [enabled=false] Whether or not the grid will support timephase column groups
     * @property {Boolean} [supportsTimephased=false] Timephased grids will need to set this to true in order to export view data and toggle timephased settings
     * @property {Array<String>} [allowedTimescales=['WEEK_DAY', 'MONTH_WEEK', 'QUARTER_MONTH', 'YEAR_MONTH', 'YEAR_QUARTER', 'REPORTING_PERIOD']]
     * The timescale options that are available for to a timephase grid.<br/><br/>
     * This option should not be changed.  If changes are needed, please file a core request.
     * @property {Object} [range] Used to constrain the bounds of the timephased section of the grid.
     * @property {moment} [range.start=moment('1960-01-01')]
     * @property {moment} [range.end=moment('2060-01-01')]
     * @property {Object} [view] The dates that are currently visible in the grid.
     * @property {moment} [view.start=null] Left most visible date in the grid.
     * @property {String} [scale='MONTH_WEEK'] The current timescale of the timephase grid.
     * @property {String} [bucketIdFormat='YYYYMMDDHH'] The format of bucket ids.<br/><br/>This option should not be changed.  If changes are needed, please file a
     *     core request.
     * @property {Number} [bucketCount=15] Number of buckets the grid will render at a time.
     * @property {Number} [shiftCount=10] Number of buckets the grid will shift to when the left or right chevrons are clicked.
     * @property {Number} [bucketWidth=100] Width, in pixels, of each bucket column, this is overriden by user changes in the view.
     * @property {Boolean} [usePeriodLabels=false] Whether or not to use reporting cycle labels.
     * @property {Function} [canEditTimephasedBucket=<br/>function (args) {<br/>&thinsp;&thinsp;return true;<br/>}] Function to disable editing of buckets, usually
     *     by dates.
     * @example
     * //For more information see the blog at http://goo.gl/m7GNTJ
     * timephased : {
                    enabled : false,
                    allowedTimescales : ['WEEK_DAY', 'MONTH_WEEK', 'QUARTER_MONTH', 'YEAR_MONTH', 'YEAR_QUARTER', 'REPORTING_PERIOD'],
                    range : {
                        start : moment('1960-01-01'),
                        end : moment('2060-01-01')
                    },
                    view : {
                        start : null
                    },
                    scale : 'MONTH_WEEK',
                    bucketIdFormat : 'YYYYMMDDHH',
                    bucketCount : 15,
                    shiftCount : 10,
                    bucketWidth : 100,
                    usePeriodLabels : false,
                    canEditTimephasedBucket : function (args) {
                        return true;
                    }
                }
     */
    /**
     * @interface grid~options
     * @type {Object}
     */
    return {
        /**
         * @member grid~options.keybindings
         * @type {Object}
         * @property {Array} [disable] The key bindings to disable, given by the code. Defaults to `[]`
         */
        keybindings : {
            disable : []
        },
        /**
         * Whether or not to select the column editors text when it opens.
         *
         * @member grid~options.selectTextOnEdit
         * @type {Boolean}
         * @default true
         */
        selectTextOnEdit : true,
        /**
         * Whether or not to match all the terms during a filter.
         *
         * @member grid~options.matchAllTerms
         * @type {Boolean}
         * @default true
         */
        matchAllTerms : true,
        /**
         * Whether or not the grid will continue to wrap from row to row when tabbing.
         *
         * @member grid~options.wrapSelection
         * @type {Boolean}
         * @default true
         */
        wrapSelection : true,
        /**
         * Whether or not to move selection down when the ENTER key is pressed
         *
         * @member grid~options.moveDownOnEnter
         * @type {Boolean}
         * @default true
         */
        moveDownOnEnter : true,
        /**
         * Whether or not to have a row default to selected when created.
         *
         * @member grid~options.selectRowByDefault
         * @type {Boolean}
         * @default true
         */
        selectRowByDefault : true,
        /**
         * Whether or not to have a cell default to checked when selected.
         *
         * @member grid~options.checkRowByDefault
         * @type {Boolean}
         * @default false
         */
        checkRowByDefault : false,
        /**
         * Whether or not to override the grids internal validation.
         *
         * @member grid~options.overrideValidation
         * @type {Boolean}
         * @default false
         */
        overrideValidation : false,
        /**
         * Whether or not to enable the trigger for when a cell is changed.
         *
         * @member grid~options.enableCellChangeTrigger
         * @type {Boolean}
         * @default false
         */
        enableCellChangeTrigger : false,
        /**
         * Whether or not the cell errors are hidden until they open and close the editor.
         *
         * @member grid~options.hideErrorsUntilEdit
         * @type {Boolean}
         * @default false
         */
        hideErrorsUntilEdit : false,
        addColumnOnShowCellError : true,
        filldown : false,
        /**
         * @member grid~options.methodDefaults
         * @type {Object}
         * @property {Object} [newRow]
         * @property {Boolean} [newRow.addAfter=true] Whether or not to add the new row after the currently selected row.
         * @property {Boolean} [newRow.batchAddAtFocus=false] Whether or to batch the triggers to prevent multiple validation, or repopulate calls.
         * @property {Boolean} [newRow.hideErrors=true] Whether or not to hide the errors on a new row.
         * @property {Object} [_recordsAdded]
         * @property {Boolean} [_recordsAdded.select=true] Whether or not to select the newly added record automatically.
         * @property {Object} [pasteCells]
         * @property {Boolean} [pasteCells.strictSelection=false] Use to enforce strict cell selection for pasting records
         */
        methodDefaults : {
            newRow : {
                addAfter : true,
                batchAddAtFocus : false,
                hideErrors : true
            },
            '_recordsAdded' : {
                select : true
            },
            pasteCells : {
                strictSelection : false
            },
            pasteRows : {
                autoCopy : true
            }
        },
        /**
         * Whether or not the grid will populate changes from one editor to the other selected cells in the column.
         *
         * @member grid~options.multiCellEdit
         * @type {Boolean}
         * @default true
         */
        multiCellEdit : true,
        /**
         * @member grid~options.shortCutLevel
         * @type {Object}
         * @property {String} [cutCopyPaste='cell']
         * @property {String} [delete='cell']
         */
        shortCutLevel : {
            cutCopyPaste : 'cell',
            delete : 'cell'
        },
        /**
         * @member grid~options.gantt
         * @type {Object}
         * @property {Boolean} [enabled=false] Whether or not the gantt component is enabled for this grid.
         */
        gantt : {
            enabled : false
        },
        /**
         * @member grid~options.stackedCells
         * @type {Object}
         * @property {Boolean} [enabled=false] Whether or not the grid will support stacked cells.
         * @property {Boolean} [rollupEnabled=false] Whether or not to enable rollup on stacked cells
         * @property {Number} [rowHeight=20] Height of a stacked cell row, in pixels.
         */
        stackedCells : {
            enabled : false,
            rollupEnabled : false,
            rowHeight : 20
        },
        /**
         * @member grid~options.cellToolbar
         * @type {grid~options#cellToolbarConfig}
         */
        cellToolbar : {
            enabled : false,
            actionOrder : ['newRow', 'newChildRow', 'deleteRows'],
            actions : {
                newRow : {
                    trigger : 'newRow',
                    gridAction : true,
                    iconClass : 'pgbu-icon-add-button',
                    enabled : true
                },
                newChildRow : {
                    trigger : 'newChildRow',
                    gridAction : true,
                    iconClass : 'pgbu-icon-forced-in',
                    enabled : true
                },
                deleteRows : {
                    trigger : 'deleteRows',
                    gridAction : true,
                    iconClass : 'pgbu-icon-delete-button',
                    enabled : true
                }
            }
        },
        /**
         * Array to hold indexes of columns that use icons.
         *
         * @member grid~options.iconColumns
         * @type {Array}
         * @default []
         */
        iconColumns : [],
        /**
         * @member grid~options.quickAdd
         * @type {Object}
         * @property {Boolean} [enabled=false] Turn on the quick add row
         * @property {Boolean} [focusAfterAdd=true] Whether or not to focus the record after adding
         * @property {String} [addAfterRowset=headers] Rowset to add the quick add row after
         * @property {Function} [getQuickAddType='record'] Function that returns a string to fill the Add ___ template
         * @property {Boolean} [addRecordAtTop=false]
         */
        quickAdd : {
            enabled : false,
            focusAfterAdd : true,
            addAfterRowset : constants.HEADERS_ROWSET_ID,
            getQuickAddType : function (args) {
                return args.options.locale.quickAddHintType;
            },
            addRecordAtTop : false,
            hintColumn : null
        },
        /**
         * Whether or not the grid will print using multiple lines
         *
         * @member grid~options.printMultiline
         * @type {Boolean}
         * @default true
         */
        printMultiline : true,
        /**
         * Whether or not the grid is in a modal dialog.
         *
         * @member grid~options.inModal
         * @type {Boolean}
         * @default false
         */
        inModal : false,
        /**
         * Whether or not the grid will be placed into edit mode when the user clicks on the grid.
         *
         * @member grid~options.editOnClick
         * @type {Boolean}
         * @default false
         */
        editOnClick : false,
        /**
         * Whether or not the grid will be placed into edit mode when the user tabs to the grid.
         *
         * @member grid~options.editOnTab
         * @type {Boolean}
         * @default true
         */
        editOnTab : true,
        /**
         * Whether or not the grid will create a new record when the user tabs to the grid.
         *
         * @member grid~options.newRecordOnTab
         * @type {Boolean}
         * @default true
         */
        newRecordOnTab : true,
        /**
         * @member grid~options.classes
         * @type {Object}
         * @property {Array<String>} [grid=['']]
         * @property {String} [disabled='grid-disabled']
         * @property {String} [loading='grid-loading']
         * @property {String} [cellSelected='cell-selected']
         * @property {String} [cellSelectedTop='cell-selected-top']
         * @property {String} [cellSelectedBottom='cell-selected-bottom']
         * @property {String} [cellFocused='cell-focused']
         * @property {String} [rowSelected='row-selected']
         * @property {String} [rowSelectedTop='row-selected-top']
         * @property {String} [rowSelectedBottom='row-selected-bottom']
         * @property {String} [rowFocused='row-focused']
         * @property {String} [rowHovered='row-hovered']
         */
        classes : {
            grid : [''],
            disabled : 'grid-disabled',
            loading : 'grid-loading',
            cellFullSelection : 'cell-full-selection',
            cellSelected : 'cell-selected ',
            cellSelectedTop : 'cell-selected-top ',
            cellSelectedBottom : 'cell-selected-bottom ',
            cellFocused : 'cell-focused ',
            rowSelected : 'row-selected ',
            rowSelectedTop : 'row-selected-top ',
            rowSelectedBottom : 'row-selected-bottom ',
            rowFocused : 'row-focused ',
            rowHovered : 'row-hovered '
        },
        /**
         * Whether or not to show vertical lines within the grid.
         * @member grid~options.showVerticalLines
         * @type {Boolean}
         * @default false
         */
        showVerticalLines : true,
        /**
         * @member grid~options.markup
         * @type {Object}
         * @property {String} [rowDragger] Defaults to `<i class="pgbu-icon-small pgbu-icon-drag-drop-dots "></i>`
         * @property {String} [lockedCell] Defaults to `<i class="pgbu-icon-small pgbu-icon-locked"></i>`
         * @property {String} [collapsedRow] Defaults to `<i class="pgbu-icon-small pgbu-icon-tree-collapsed"></i>`
         * @property {String} [expandedRow] Defaults to `<i class="pgbu-icon-small pgbu-icon-tree-expanded"></i>`
         * @property {String} [cellLoading] Defaults to `<i class="pgbu-icon-small pgbu-icon-loading"></i>`
         */
        markup : {
            rowDragger : '<i class="pgbu-icon-small pgbu-icon-drag-drop-dots "></i>',
            lockedCell : '<i class="pgbu-icon-small pgbu-icon-locked"></i>',
            collapsedRow : '<i class="pgbu-icon-small pgbu-icon-tree-collapsed"></i>',
            expandedRow : '<i class="pgbu-icon-small pgbu-icon-tree-expanded"></i>',
            cellLoading : '<i class="pgbu-icon-small pgbu-icon-loading"></i>'
        },
        headerOverlay : '',
        /**
         * Whether or not to display a loading spinner when the grid is initially loading data.
         *
         * @member grid~options.showLoading
         * @type {Boolean}
         * @default true
         */
        showLoading : true,
        /**
         * Whether or not the grid is a tree structure.
         *
         * @member grid~options.tree
         * @type {Boolean}
         * @default false
         */
        tree : false,
        hierarchy : {
            byType : true
        },
        rowExpansion : {
            enabled : false,
            heightInRows : 5,
            placement : 'rowset'
        },
        /**
         * The maximum amount of levels the tree is allowed to have.
         *
         * @member grid~options.treeMaxLevel
         * @type {Number}
         * @default null
         */
        treeMaxLevel : null,
        /**
         * Whether or not during a sort the grid should also sort root nodes.
         *
         * @member grid~options.sortRootNodes
         * @type {Boolean}
         * @default true
         */
        sortRootNodes : true,
        /**
         * Whether or not the focus should follow the selection.
         *
         * @member grid~options.focusOnSelection
         * @type {Boolean}
         * @default false
         */
        focusOnSelection : false,
        /**
         * Whether or not the grid should always call mediator.recordUpdated on editing a cell even if no change in value is detected.
         *
         * @member grid~options.alwaysResetValue
         * @type {Boolean}
         * @default false
         */
        alwaysResetValue : false,
        /**
         * @member grid~options.contextMenu
         * @type {grid~options#contextMenuConfig}
         */
        contextMenu : {
            enabled : false,
            enableIcons : false,
            enableHoverRowMenuButton : true,
            enableRowMenuButton : false,
            rowMenuButtonWidth : 27,
            offsetTop : 0,
            hideDisabled : true,
            allowDisjointSelection : false,
            actions : {
                quickAddRowNewRow : {
                    label : 'Add',
                    trigger : 'newRow'
                }
            },
            groupRowActions : []
        },
        /**
         * Options pertaining to the grid's right click column menu.
         *
         * Note: to be able to hide columns through this menu, both the grid `menus.canHideColumns` option must be true, as well as the columnGroup's `canHideColumns` option.
         *
         * @member grid~options.columnMenu
         * @type {Object}
         * @property {Boolean} [enabled=false]
         * @property {Object} [columnToggleMenu={}]
         */
        columnMenu : {
            enabled : false,
            columnToggleMenu : {}
        },
        /**
         * Options pertaining to the grid's columnGroup menu.
         *
         * Column group menus will only be shown if some functionality can be accessed through the menu - notably, that means that if `menus.canHideColumns` is false, the column group menu will not be shown.
         *
         * @member grid~options.columnGroupMenu
         * @type {Object}
         * @property {Boolean} [enabled=false]
         */
        columnGroupMenu : {
            enabled : false
        },
        /**
         * Options pertaining to grid menus - especially the columnGroup and column menus.
         *
         * @member grid~options.menus
         * @type {Object}
         * @property {Boolean} [canHideColumns=true]
         * @property {Boolean} [canHideRequiredColumns=true]
         */
        menus : {
            canHideColumns : true,
            canHideRequiredColumns : false
        },
        /**
         * @member grid~options.recordIcons
         * @type {Object}
         * @property {String} [default=null]
         */
        recordIcons : {
            'default' : null
        },
        /**
         * @member grid~options.editable
         * @type {Boolean}
         * @default false
         */
        editable : false,
        debug : false,
        /**
         * Option for cut/copy/paste undo
         *
         * @member grid~options.undo
         * @type {Boolean}
         * @default false
         */
        undo : false,
        /**
         * @member grid~options.lazyload
         * @type {Boolean}
         * @default false
         */
        lazyload : false,
        /**
         * @member grid~options.search
         * @type {Object}
         * @property {Boolean} [showAll=false] if set to true, results are not filtered; instead, the grid will simply be redrawn with matches highlighted
         * @property {Boolean} [showParents=false]
         * @property {Boolean} [numericSearchEnabled=true]
         * @property {Boolean} [columnFilter=false]
         * @property {Boolean} [columnFilterMatchAllColumns=true]
         */
        search : {
            showAll : false,
            showParents : false,
            forceLazyload : false,
            numericSearchEnabled : true,
            columnFilter : false,  // LUX-506
            columnFilterMatchAllColumns : true  // LUX-506
        },
        /**
         * Shows the advanced column filters at the top of the grid directly underneath the header row.
         * Please refer to <a href="https://confluence.oraclecorp.com/confluence/display/gbulux/Lux+Grid-Advanced+Filters">confluence page</a> for more details.
         *
         * @member grid~options.advancedFilters
         * @type {Object}
         * @property {Boolean} [enabled=false] toggle for enabling the advanced filter rowset
         * @property {Boolean} [client=false] toggle for enabling clientside filtering
         */
        advancedFilters : {
            enabled : false,
            client : false
        },
        /**
         * @member grid~options.sortClickOrder
         * @type {Array<String>}
         * @default ['asc', 'dsc', 'none']
         */
        sortClickOrder : ['asc', 'dsc', 'none'],
        /**
         * @member grid~options.sortBy
         * @type {Array}
         * @default []
         */
        sortBy : [],
        /**
         * @member grid~options.sorters
         * @type {Object}
         * @default {}
         */
        sorters : {},
        /**
         * @member grid~options.groupBy
         * @type {Array}
         * @default []
         */
        groupBy : [],
        /**
         * @member grid~options.performance
         * @type {Object}
         * @property {Number} [bulkThreshold=2]
         * @property {Number} [minimumSearchCharacters=1]
         * @property {Number} [selectDebounceInterval=300]
         * @property {Object} [selectDebounceOptions={}]
         * @property {Number} [verticalScrollThrottle=20]
         * @property {Number} [verticalScrollTimeout=100]
         * @property {Number} [horizontalScrollThrottle=30]
         * @property {Number} [actionStateInterval=500]
         * @property {Object} [actionStateDebounceOptions={}]
         */
        performance : {
            bulkThreshold : 2,
            minimumSearchCharacters : 1,
            selectDebounceInterval : 300,
            selectDebounceOptions : {},
            verticalScrollThrottle : 20,
            verticalScrollTimeout : 100,
            horizontalScrollThrottle : 30,
            actionStateInterval : 500,
            actionStateDebounceOptions : {}
        },
        /**
         * @member grid~options.grouping
         * @type {Object}
         * @property {Boolean} [enableRollup=true]
         * @property {Boolean} [enableSequence=false]
         * @property {Boolean} [showColumnTitle=false]
         * @property {Boolean} [showEmptyBands=false]
         * @property {Number} [maxEmptyBands=10000]
         * @property {Number} [maxRollupGroups=30000]
         * @property {String} [nameFormat=both]
         * @property {Boolean} [boldBands=true]
         * @property {Boolean} [collapseByDefault=false]
         * @property {Array<String>} [bandClasses=['group-background-0', 'group-background-1', 'group-background-2', 'group-background-3', 'group-background-4', 'group-background-5']]
         */
        grouping : {
            enableRollup : true,
            enableSequence : false,
            showColumnTitle : false,
            showEmptyBands : false,
            maxEmptyBands : 10000,
            maxRollupGroups : 30000,
            nameFormat : 'both',
            boldBands : true,
            collapseByDefault : false,
            bandClasses : [
                'group-background-0',
                'group-background-1',
                'group-background-2',
                'group-background-3',
                'group-background-4',
                'group-background-5'
            ]
        },
        /**
         * @member grid~options.contextMenu
         * @type {grid~options#timephasedConfig}
         */
        timephased : {
            enabled : false,
            supportsTimephased : false,
            allowedTimescales : ['WEEK_DAY', 'MONTH_WEEK', 'QUARTER_MONTH', 'YEAR_MONTH', 'YEAR_QUARTER', 'REPORTING_PERIOD'],
            range : {
                start : moment('1960-01-01'),
                end : moment('2060-01-01')
            },
            view : {
                start : null
            },
            scale : 'MONTH_WEEK',
            bucketIdFormat : 'YYYYMMDDHH',
            bucketCount : 14,
            shiftCount : 10,
            bucketWidth : 100,
            usePeriodLabels : false,
            hideZeroValues : true,
            canEditTimephasedBucket : function (args) {
                return true;
            },
            tooltipFormat : 'LL'
        },
        /**
         * @member grid~options.theme
         * @type {Object}
         * @property {String} [style='MODERN_NOTCHES']
         * @property {String} [colors='GREY']
         * @property {Boolean} [readOnlyStyling=false]
         */
        theme : {
            style : 'NONE',
            colors : 'GREY',
            readOnlyStyling : false
        },
        /**
         * @member grid~options.layout
         * @type {Object}
         * @property {String} [width='500px']
         * @property {String} [height='500px']
         * @property {String} [widthShim='0px']
         * @property {String} [heightShim='0px']
         * @property {Array} rowsets
         * @property {Array<grid#columnGroup>} columnGroups
         * @property {Boolean} [autoResize=true]
         * @property {Number} [autoResizeDebounce=300]
         * @property {Boolean} [autoResizeImmediate=false]
         * @property {Number} [rowHeight=26]
         * @property {Boolean} [removeHeaderRow=false]
         * @property {Number} [headerHeight=27]
         * @property {Array<String>} [lockedColumns]
         */
        layout : {
            width : '500px',
            height : '500px',
            widthShim : '0px',
            heightShim : '0px',
            rowsets : [],
            columnGroups : [],
            autoResize : true,
            autoResizeDebounce : 300,
            autoResizeImmediate : false,
            rowHeight : 26,
            removeHeaderRow : false,
            headerHeight : 27,
            lockedColumns : []
        },
        statusBar : {
            enabled : false,
            height : 25,
            recordCountEnabled : true,
            readOnlyStylingEnabled : false,
            toggleItems : []
        },
        export : {
            enabled : false,
            fileName : locale.export.fileName,
            worksheetName : locale.export.worksheetName,
            user : locale.export.user
        },
        /**
         * @member grid~options.scrollbar
         * @type {Object}
         * @property {Number} [extraHeight=12]
         * @property {Number} [extraWidth=12]
         */
        scrollbar : {
            extraHeight : 12,
            extraWidth : 12
        },
        /**
         * @member grid~options.borders
         * @type {Object}
         * @property {Number} [viewportLeftRight=2]
         * @property {Number} [size=1]
         * @property {Number} [grid=1]
         * @property {Number} [row=1]
         */
        borders : {
            viewportLeftRight : 2, // Compensate for both sides of the viewport
            size : 1,
            grid : 1,
            row : 1
        },
        /**
         * @member grid~options.sizing
         * @type {Object}
         * @property {Number} [minColumnWidth=40]
         * @property {Number} [minGroupWidth=100]
         */
        sizing : {
            minColumnWidth : 40,
            minGroupWidth : 100
        },
        /**
         * @member grid~options.recordKeys
         * @type {Object}
         * @property {String} [id='id']
         * @property {String} [siblingId='siblingId']
         * @property {String} [parentId='parentId']
         * @property {String} [type='type']
         * @property {String} [bucketId='bucketId']
         * @property {String} [sequence=null]
         */
        recordKeys : {
            id : 'id',
            siblingId : 'siblingId',
            parentId : 'parentId',
            type : 'type',
            bucketId : 'bucketId',
            sequence : null
        },
        /**
         * @member grid~options.rows
         * @type {Object}
         * @property {Object} [rendererFns={}]
         * @property {Boolean} [draggable=false]
         * @property {Function} [classes]
         * @property {Function} [attributes]
         */
        rows : {
            rendererFns : {},
            draggable : false,
            classes : function (/* record, groupId, rowsetId */) {
                // Space separated list of classes
                // Example 'my-row-type icon-class awesome-hightlight-class'
                return '';
            },
            attributes : function (/* record, groupId, rowsetId */) {
                // Add attributes
                return '';
            }
        },
        /**
         * @member grid~options.shortcuts
         * @type {Object}
         * @property {Object} [insert]
         * @property {Function} [insert.primary] returns newRow args, can also be Object
         * @property {Function} [insert.secondary] returns newRow args, can also be Object
         */
        shortcuts : {
            insert : {
                primary : function () {
                    return {};
                },
                secondary : function () {
                    return {};
                }
            }
        },
        /**
         * @member grid~options.mediatorFns
         * @type {Object}
         * @default {}
         */
        mediatorFns : {},
        /**
         * @member grid~options.selectionMode
         * @type {String}
         * @default default
         */
        selectionMode : 'default',
        /**
         * @member grid~options.checkOnSelect
         * @type {Boolean}
         * @default true
         */
        checkOnSelect : true,
        /**
         * @member grid~options.multiSelect
         * @type {Boolean}
         * @default true
         */
        multiSelect : true,
        /**
         * @member grid~options.disableMultiRowsetSelection
         * @type {Boolean}
         * @default true
         */
        disableMultiRowsetSelection : true,
        /**
         * @member grid~options.singleColumn
         * @type {Boolean}
         * @default false
         */
        singleColumn : false,
        /**
         * @member grid~options.locale
         * @type {Object}
         */
        locale : locale,
        /**
         * @member grid~options.columns
         * @type {Object}
         * @property {Object<String, grid#columnDefinition>} columns={}
         */
        columns : {},
        defaultColumnOptions : {},
        /**
         * @member grid~options.dataRollup
         * @type {Object}
         * @property [enabled=false] Note rollups are only applicable when you have tree-structured data and are un-grouped
         */
        dataRollup : {
            enabled : false
        },
        /**
         * @member grid~options.renamedViewColumns
         * @type {Object}
         * @property {Object<String, String>} renamedViewColumns - with columnGroups-based columns lists, use this option to indicate that viewdata-derived
         * column names have been renamed to a new string; e.g. {'oldCodeDataIndex': 'newCodeDataIndex'} would map the old name to the new one before apply.
         * PRIMCR-2174 has more context into the problem.
         * @default null
         */
        renamedViewColumns : null,
        /**
         * Options pertaining to the grid's touch interactions.
         *
         * @member grid~options.touchInteractions
         * @type {Object}
         * @property {Boolean} [enabled=true] Enable various touch interactions on grid.
         */
        touchInteractions : {
            enabled : false
        }
    };
});
