define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.moment',
    'gridapi.logger',
    'gridapi.grid/constants',
    'gridapi.grid/groupers',
    'gridapi.grid/RollupBase',
    'gridapi.grid/rollupUtils'
], function ($, _, moment, logger, constants, groupers, RollupBase, rollupUtils) {
    'use strict';

    var FORCE_EMPTY_ROLLUP_ARGS = { forceEmptyRollup : true };
    var rollups = rollupUtils.rollups;
    var normalizeDataIndices = rollupUtils.normalizeDataIndices;

    function mapRecords(recordIds) {
        var result = [];
        var widget = this.widget;
        for (var i = 0, len = recordIds.length; i < len; ++i) {
            var record = widget._getRecord(recordIds[i]);
            if (record) {
                result.push(record);
            }
        }
        return result;
    }

    var RecordGrouper = function (args) {
        var that = this;
        this.widget = args.widget;

        // override mediator functions from optionsmn
        _.forEach(this.widget.options.grouping.recordGrouperFns, function (fnc, fncName) {
            that[fncName] = fnc || that[fncName];
        });
        this.resetGroupStorage();

        RollupBase.call(this);
    };

    RecordGrouper.prototype = _.create(RollupBase.prototype, {
        constructor : RecordGrouper,

        setStore : function (store) {
            this.store = store;
        },

        resetGroupStorage : function (skipNodeReset) {
            this.recordGroups = {};
            this.widget.rootNodes = {};
            this.widget.siblingIds = {};
            this.rootNodes = this.widget.rootNodes;
            this.recordToGroupMapping = {};
            this.groupValues = [];
            this.groupValueIndices = [];
            this.groupDataValues = [];
            this.hierarchyMapping = {};
            if (!skipNodeReset) {
                this.resetVirtualRowNodes();
            }
            this.endChildCache();
        },

        resetVirtualRowNodes : function () {
            _.forEach(this.widget.view.virtualRows, function (virtualRowSet) {
                virtualRowSet.nodes = {};
            });
        },

        groupRecords : function (groupConfig /*, silent */) {
            var that = this;

            this.resetGroupStorage(true);
            this.updateGroupConfig(groupConfig);
            this.startChildCache();

            if (that.widget.config.groupBy != null && that.widget.config.groupBy.length && _.every(that.widget.config.groupBy, function (grouper) {
                return !grouper.fakeGroup;
            })) {
                this.resetVirtualRowNodes();
                var recordsToGroup = this.getRecordsToGroup();
                for (var i = 0, len = recordsToGroup.length; i < len; ++i) {
                    that.groupRecord(recordsToGroup[i]);
                }
                that.orderGroups();
                var totalGroups = _.size(that.recordGroups);
                if (totalGroups < this.widget.options.grouping.maxRollupGroups) {
                    var toRollup = [];
                    _.forEach(that.recordGroups, function initiateChildRollups(group) {
                        if (!group.childGroups.length && !group.hierarchyGroups.length) {
                            toRollup.push(group.id);
                        }
                    });
                    this.rollupFromLeafs(toRollup);
                } else {
                    this.enableRollup = false;
                    this.widget._trigger('performance-optimization', null, {
                        event: 'groupRollupsDisabled',
                        groupsToRollup : totalGroups,
                        message : 'Grid disabled rollups for groupBands to prevent potential issues.',
                        limit : this.widget.options.grouping.maxRollupGroups
                    });
                }
                that.widget.groupingApplied = true;
            } else {
                that.widget.groupingApplied = false;
            }

            var focusedRecordId = null;
            var focusRecord = that.widget.focus.record;
            if (focusRecord != null && !that.widget._isGroupRecord(focusRecord)) {
                focusedRecordId = that.widget._getRecordId(focusRecord);
            }
            if (that.widget._hasFilters()) {
                that.widget.searchRows(that.widget.searchTerm, that.widget.searchArgs);
            } else {
                that.widget.repopulate(focusedRecordId, that.widget.focus.columnId);
            }
            that.widget.view.resetGridRole(); // reset the aria role

            this.endChildCache();
        },

        updateGroupConfig : function (groupConfig) {
            var that = this;
            var groupsToRemove = [];
            _.forEach(groupConfig, function (group, index) {
                group.level = index;
                if (group.sortOrder == null) {
                    group.sortOrder = 'asc';
                }
                if (group.getGroupValue == null) {
                    group.getGroupValue = _.identity;
                }
                if (group.sortOrder === 'hierarchy') {
                    group.hierarchy = true;
                }
                that.groupValues[index] = [];
                that.groupValueIndices[index] = {};
                that.groupDataValues[index] = {};
                var column = that.widget.view.columnDefinitions[group.dataIndex];
                if (column != null) {
                    group.columnDefinition = column;
                    var keys, dataValues,
                        columnGroupingFns = column.groupingFns;
                    if (column.dataType === 'date' || column.dataType === 'datetime') {
                        group.getGroupValue = groupers.date;
                    }
                    if (columnGroupingFns != null) {
                        _.forEach(columnGroupingFns, function (fnc, fncName) {
                            group[fncName] = fnc || group[fncName];
                        });
                    }

                    if (column.selectValues != null && !_.isFunction(column.selectValues)) {
                        keys = _.keys(column.selectValues);
                        that.groupValues[index] = keys;
                        dataValues = {};
                        _.forEach(keys, function (key, i) {
                            dataValues[key] = key;
                            that.groupValueIndices[index][key] = i;
                        });
                        that.groupDataValues[index] = dataValues;
                    } else if (column.groupValues != null && column.groupValues.length) {
                        keys = column.groupValues;
                        that.groupValues[index] = keys;
                        dataValues = {};
                        _.forEach(keys, function (key, i) {
                            dataValues[key] = key;
                            that.groupValueIndices[index][key] = i;
                        });
                        that.groupDataValues[index] = dataValues;
                    }
                    if (group.hierarchy && that.widget.isLoaded()) {
                        if (group.recordType != null) {
                            that.hierarchyMapping[group.dataIndex] = that.buildDataHierarchyMapping(group);
                        } else {
                            that.hierarchyMapping[group.dataIndex] = that.buildHierarchyMapping(group);
                        }

                        that.groupValues[index] = _.uniq(_.pluck(that.hierarchyMapping[group.dataIndex].groups, 'value'));
                        _.forEach(that.groupValues[index], function (key, i) {
                            that.groupValueIndices[index][key] = i;
                        });
                    }
                } else {
                    groupsToRemove.push(group.dataIndex);
                    logger.warn(group.dataIndex + ' Group removed due to no existing column on the grid.');
                }
            });
            that.widget.config.groupBy = _.filter(groupConfig, function (group) {
                return !_.includes(groupsToRemove, group.dataIndex);
            });
        },

        buildHierarchyMapping : function (grouper) {
            var mapping = {
                rootItems : [],
                rootGroups : [],
                groups : {}
            };
            mapping.groups[constants.NO_GROUP] = {
                item : null,
                id : null,
                parentId : null,
                level : 1,
                children : [],
                squashedGroup : false,
                value : constants.NO_GROUP
            };
            if (!this.widget.isLoaded() || !_.isFunction(this.widget.store.getGroupingStructure)) {
                return mapping;
            }
            var groupUpdates = this.widget.store.getGroupingStructure({
                dataIndex : grouper.dataIndex, config : grouper,
                defaults : {
                    recordKeys : {
                        parentId : 'parentId',
                        id : 'id'
                    },
                    structure : []
                },
                getGroupKey : function (value) {
                    return value;
                }
            });

            var structure = groupUpdates.structure;
            var recordKeys = groupUpdates.recordKeys;
            var maxLevel = grouper.toLevel;
            if (grouper.toLevel === 'all') {
                maxLevel = 999999;
            }

            if (_.isFunction(grouper.getRootNodes)) {
                mapping.rootItems = grouper.getRootNodes(structure);
            } else {
                mapping.rootItems = _.filter(structure, recordKeys.parentId, null);
            }
            function addGroups(node, parent) {
                var groupId = node[recordKeys.id];
                var level = parent != null ? parent.level + 1 : 1;
                var children = _.filter(structure, recordKeys.parentId, node[recordKeys.id]);
                var squashedGroup = false;
                var groupChildren = _.pluck(children, recordKeys.id);
                var value = groupId;

                if (level > maxLevel) {
                    parent = mapping.groups[parent[recordKeys.id]];
                    value = parent.value;
                    level = maxLevel;
                    parent.children.push(node[recordKeys.id]);
                    squashedGroup = true;
                    groupChildren = [];
                }
                var parentId;
                if (parent != null) {
                    if (parent.squashedGroup) {
                        parentId = parent.parentId;
                        value = parentId;
                    } else {
                        parentId = parent.id;
                    }
                }
                var group = {
                    item : node,
                    id : groupId,
                    parentId : parentId,
                    level : level,
                    children : groupChildren,
                    squashedGroup : squashedGroup,
                    value : value

                };

                mapping.groups[node[recordKeys.id]] = group;
                _.forEach(children, function (child) {
                    addGroups(child, group);
                });
                return group;
            }

            _.forEach(mapping.rootItems, function (rootNode) {
                mapping.rootGroups.push(addGroups(rootNode, null));
            });

            return mapping;
        },

        buildDataHierarchyMapping : function (grouper) {
            var that = this;
            var typeDefinition = this.widget.options.data.types[grouper.recordType];
            var typeMapping = typeDefinition.mapping[grouper.recordType];
            var mapping = {
                rootItems : this.widget._getRecordsByData({ types : [grouper.recordType] }),
                rootGroups : [],
                groups : {}
            };
            var childCache = {};
            if (grouper.singlePassMapping && that.store.getAllRecordsByType != null) {
                var items = that.store.getAllRecordsByType({
                    type : grouper.recordType
                });

                for (var i = 0; i < items.length; i++) {
                    var item = items[i];
                    var parentId = that.store.getValueFromRecord(item, typeMapping.foreignKey);
                    if (childCache[parentId] != null) {
                        childCache[parentId].push(item);
                    } else {
                        childCache[parentId] = [item];
                    }
                }
            }

            var maxLevel = grouper.toLevel;
            if (grouper.toLevel === 'all') {
                maxLevel = 999999;
            }
            if (grouper.getHierarchyKey == null) {
                grouper.getHierarchyKey = function (record) {
                    if (record != null) {
                        return that.store.getValueFromRecord(record, typeMapping.localKey);
                    }
                };
            }
            // For two way joins add a way to get the indirect assignment to this group.
            if (grouper.getDataValue == null && !_.intersection(typeDefinition.childTypes, this.widget.options.data.baseTypes).length && typeDefinition.chainedGroup) {
                // Join by from top of hierarchy down to lowest, reverse to do lookups from source data of record
                var joinTypes = _.map(typeDefinition.joinedBy, function (value) {
                    return { type : value, definition : that.widget.options.data.types[value] };
                });
                joinTypes = joinTypes;
                var typeOrder = joinTypes.concat([{ type : grouper.recordType, definition : typeDefinition }]);

                grouper.getDataValue = function (args) {
                    if (args.record != null) {
                        var value = null;
                        var lookupObject = args.record;
                        var lookupObjectType = that.widget._getTypeFromId(that.widget._getRecordId(args.record));
                        // Find values if you can until you get to the grouper you are looking for. Ex baseType-> n-1 -> n
                        _.forEach(typeOrder, function (joinType) {
                            if (lookupObject != null) {
                                value = _.get(lookupObject, joinType.definition.mapping[lookupObjectType].foreignKey);
                                if (value != null) {
                                    lookupObject = that.store.getRecordByType(joinType.type, value);
                                    lookupObjectType = joinType.type;
                                } else {
                                    lookupObject = null;
                                }
                            }
                        });
                        return lookupObject;
                    }
                };
                grouper.filterEmptyBands = function (args) {
                    return _.filter(args.rootHierarchyNodes, function (rootHierarchyNode) {
                        var lookupObject = rootHierarchyNode.item;
                        var lookupType = args.childType;
                        var joinMapping = typeDefinition.mapping[lookupType];

                        return _.get(lookupObject, joinMapping.foreignKey) === _.get(args.backingRecord, joinMapping.localKey) &&
                            _.every(joinMapping.constraints, function (constraint) {
                                return _.get(lookupObject, constraint.key) === constraint.value;
                            });
                    });
                };
            }
            function addGroups(args) {
                var groupId = that.store.getValueFromRecord(args.record, typeMapping.localKey);
                var children;
                if (grouper.singlePassMapping) {
                    children = childCache[groupId];
                } else {
                    children = that.widget._getRecordsByData({
                        types : [grouper.recordType],
                        parentId : groupId,
                        parent : args.record,
                        parentType : grouper.recordType
                    });
                }

                var squashedGroup = false;
                var groupChildren = _.pluck(children, typeMapping.localKey);
                var value = groupId;
                var parent = args.parentGroup;
                var level = args.parentGroup != null ? args.parentGroup.level + 1 : 1;
                if (level > maxLevel) {
                    parent = mapping.groups[args.parentGroup.id];
                    value = parent.value;
                    level = maxLevel;
                    parent.children.push(groupId);
                    squashedGroup = true;
                    groupChildren = [];
                }
                var parentId;
                if (parent != null) {
                    if (parent.squashedGroup) {
                        parentId = parent.parentId;
                        value = parentId;
                    } else {
                        parentId = parent.id;
                    }
                }
                var group = {
                    item : args.record,
                    id : groupId,
                    parentId : parentId,
                    level : level,
                    children : groupChildren,
                    squashedGroup : squashedGroup,
                    value : value

                };

                mapping.groups[groupId] = group;
                _.forEach(children, function (child) {
                    addGroups({
                        record : child,
                        parentGroup : group
                    });
                });
                return group;
            }

            _.forEach(mapping.rootItems, function (rootNode) {
                mapping.rootGroups.push(addGroups({ record : rootNode, parentId : null }));
            });
            mapping.groups[constants.NO_GROUP] = {
                item : null,
                id : null,
                parentId : null,
                level : 1,
                children : [],
                squashedGroup : false,
                value : constants.NO_GROUP
            };
            return mapping;
        },

        getRecordsToGroup : function () {
            var recordsToGroup = [];
            if (_.isFunction(this.store.getRecordsToGroup)) {
                recordsToGroup = this.store.getRecordsToGroup();
            } else if (this.widget.options.data != null) {
                recordsToGroup = this.widget._getRecordsByData({
                    types : this.widget.options.data.baseTypes
                });
            } else {
                this.store.iterateStore(this.widget._getRecordFromStore, this.widget);
                recordsToGroup = this.widget.storeRecords;
            }
            return recordsToGroup;
        },

        getGroupQualifiers : function (location) {
            var qualifiers = {};
            qualifiers.group = location.rowsetId;
            qualifiers.root = 'rowset-' + location.rowsetId;
            return qualifiers;
        },

        getGroupValue : function (grouper, dataValue, record) {
            var groupValue = dataValue;
            if (grouper.hierarchy) {
                if (grouper.getHierarchyKey != null) {
                    groupValue = grouper.getHierarchyKey(dataValue, {
                        dataValue : dataValue,
                        record : record,
                        grouper : grouper
                    });
                }
                var hierarchyGroup = this.hierarchyMapping[grouper.dataIndex].groups[groupValue];
                if (groupValue == null || hierarchyGroup == null) {
                    groupValue = constants.NO_GROUP;
                }
                if (hierarchyGroup != null) {
                    groupValue = hierarchyGroup.value;
                }
            } else {
                groupValue = grouper.getGroupValue(dataValue, {
                    record : record,
                    groupValue : groupValue,
                    groupConfig : grouper
                });
            }

            return groupValue;
        },

        getDataValue : function (grouper, record) {
            var dataValue = this.store.getValueFromRecord(record, grouper.dataIndex);
            if (grouper.getDataValue != null) {
                dataValue = grouper.getDataValue({
                    dataValue : dataValue,
                    store : this.store,
                    record : record,
                    dataIndex : grouper.dataIndex,
                    grouper : grouper
                });
            }
            return dataValue;
        },

        groupRecord : function (record) {
            var recordId = this.widget._getRecordId(record);
            var that = this,
                groupId = 'group-',
                groupers = that.widget.config.groupBy,
                newGroups = [];
            var qualifiers = this.getGroupQualifiers(that.widget._getRecordPlacement(record, true));
            var groupQualifier = qualifiers.group;
            var rootQualifier = qualifiers.root;

            groupId += groupQualifier + '-';
            if (this.rootNodes[rootQualifier] == null) {
                this.rootNodes[rootQualifier] = [];
            }

            var parentGroupId = null;
            for (var i = 0; i < groupers.length; ++i) {
                var grouper = groupers[i];
                var dataValue = this.getDataValue(grouper, record);

                var groupValue = this.getGroupValue(grouper, dataValue, record);
                if (groupValue == null) {
                    groupValue = null;
                    dataValue = null;
                }

                var groupKeyPart = that.groupValueIndices[i][groupValue];
                if (groupKeyPart == null) {
                    that.groupValues[i].push(groupValue);
                    that.groupDataValues[i][groupValue] = dataValue;
                    groupKeyPart = that.groupValues[i].length - 1;
                    that.groupValueIndices[i][groupValue] = groupKeyPart;
                }
                var groupIdStart = groupId + (i > 0 ? ':' : '');
                groupId += (i > 0 ? ':' : '') + groupKeyPart;
                var recordGroup = that.recordGroups[groupId];
                if (recordGroup == null) {
                    recordGroup = this.buildGroup({
                        grouperIndex : i,
                        hierarchyMapping : this.hierarchyMapping[grouper.dataIndex],
                        groupId : groupId,
                        baseId : groupIdStart,
                        groupValue : groupValue,
                        dataValue : dataValue,
                        parentGroupId : parentGroupId,
                        grouper : grouper,
                        rootQualifier : rootQualifier,
                        newGroups : newGroups
                    });
                }
                // Add record to the last grouping level
                if (i === groupers.length - 1 && recordGroup.recordLocations[recordId] == null) {
                    var insertIndex = recordGroup.records.length;
                    var siblingId = this.widget.siblingIds && this.widget.siblingIds[recordId];
                    var siblingIndex = recordGroup.recordLocations[siblingId];
                    //If siblingId is explicitly null, the user is adding above the topmost grouped record
                    //If siblingId is defined, yet the index is not, the user is adding below on the group band
                    //If siblingId and siblingIndex are both defined, the record is inserted into the group
                    if (siblingId === null || (siblingId != null && siblingIndex == null)) {
                        insertIndex = 0;
                    } else if (siblingId != null && siblingIndex != null) {
                        insertIndex = siblingIndex + 1;
                    }
                    recordGroup.records.push(recordId);
                    recordGroup.orderedRecords.splice(insertIndex, 0, recordId);
                    this.recordToGroupMapping[recordId] = recordGroup.id;
                    this.indexRecords(recordGroup, insertIndex);
                }

                parentGroupId = recordGroup.id;
            }
            return newGroups;
        },

        indexRecords : function (group, startingIndex) {
            var i = startingIndex;
            var records = group.orderedRecords;
            var recordLocations = group.recordLocations;
            var len = records.length;
            for (i; i < len; ++i) {
                recordLocations[records[i]] = i;
            }
        },

        buildGroup : function (args) {
            var that = this;
            var groupId = args.groupId,
                grouper = args.grouper,
                parentGroupId = args.parentGroupId,
                level = args.grouper.level,
                addToHierarchy = false,
                backingRecord = null;
            var hierarchyGroup, parentGroup, hierarchyDataValue;
            var recordGroup = this.recordGroups[groupId];
            if (recordGroup == null) {
                if (grouper.hierarchy) {
                    hierarchyGroup = args.hierarchyMapping.groups[args.groupValue];
                    if (hierarchyGroup.parentId != null) {
                        var parent = args.hierarchyMapping.groups[hierarchyGroup.parentId];
                        var originalParentGroupId = parentGroupId;
                        parentGroupId = args.baseId + this.groupValues[args.grouperIndex].indexOf(parent.id);
                        parentGroup = this.buildGroup({
                            hierarchyMapping : args.hierarchyMapping,
                            groupValue : parent.id,
                            dataValue : _.isFunction(grouper.getHierarchyDataValue) ? grouper.getHierarchyDataValue(parent.item) : parent.item,
                            grouper : grouper,
                            level : parent.level,
                            groupId : parentGroupId,
                            rootQualifier : args.rootQualifier,
                            baseId : args.baseId,
                            grouperIndex : args.grouperIndex,
                            parentGroupId : originalParentGroupId,
                            newGroups : args.newGroups
                        });

                        addToHierarchy = true;
                        if (hierarchyGroup.squashedGroup) {
                            return parentGroup;
                        }
                    } else if (hierarchyGroup.item) {
                        hierarchyDataValue = _.isFunction(grouper.getHierarchyDataValue) ? grouper.getHierarchyDataValue(hierarchyGroup.item) : hierarchyGroup.item;
                    }

                    backingRecord = hierarchyGroup.item;
                    level = level + hierarchyGroup.level - 1;
                    recordGroup = this.recordGroups[groupId];
                    if (recordGroup != null) {
                        return recordGroup; // Was generated by the parent so return
                    }
                }
                recordGroup = {
                    parentId : parentGroupId,
                    groupValue : args.groupValue,
                    dataIndex : grouper.dataIndex,
                    dataValue : hierarchyDataValue ? hierarchyDataValue : args.dataValue,
                    records : [],
                    recordLocations : {},
                    childGroups : [],
                    hierarchyGroups : [],
                    level : level,
                    noGroupValue : grouper.noGroupValue || this.widget.options.locale.noGroupValue,
                    formatValue : grouper.formatValue,
                    setGroupValueOnRecord : grouper.setGroupValueOnRecord,
                    ignoreGroupValueOnNewRecord : grouper.ignoreGroupValueOnNewRecord,
                    id : groupId,
                    isGroup : true,
                    rootQualifier : args.rootQualifier,
                    rollupValues : {},
                    timephasedRollupValues : {},
                    stackedRollupValues : {},
                    groupConfig : grouper,
                    locationIndex : -1,
                    backingRecord : backingRecord,
                    orderedRecords : [],
                    location : ''
                };
                this.recordGroups[groupId] = recordGroup;
                args.newGroups.push(recordGroup);
                // Add group to parent or to the root groups
                if (parentGroupId != null) {
                    if (addToHierarchy) {
                        this.recordGroups[parentGroupId].hierarchyGroups.push(groupId);
                        recordGroup.location = 'hierarchyGroups';
                        recordGroup.locationIndex = this.recordGroups[parentGroupId].hierarchyGroups.length - 1;
                    } else {
                        this.recordGroups[parentGroupId].childGroups.push(groupId);
                        recordGroup.location = 'childGroups';
                        recordGroup.locationIndex = this.recordGroups[parentGroupId].childGroups.length - 1;
                    }
                } else {
                    this.rootNodes[args.rootQualifier].push(groupId);
                    recordGroup.locationIndex = this.rootNodes[args.rootQualifier].length - 1;
                }
                if (grouper.hierarchy && this.widget.options.grouping.showEmptyBands) {
                    _.forEach(hierarchyGroup.children, function (childGroupId) {
                        var childGroup = args.hierarchyMapping.groups[childGroupId];
                        that.buildGroup({
                            hierarchyMapping : args.hierarchyMapping,
                            groupValue : childGroup.id,
                            dataValue : _.isFunction(grouper.getHierarchyDataValue) ? grouper.getHierarchyDataValue(childGroup.item) : childGroup.item,
                            grouper : grouper,
                            level : childGroup.level,
                            groupId : args.baseId + that.groupValues[args.grouperIndex].indexOf(childGroup.id),
                            rootQualifier : args.rootQualifier,
                            baseId : args.baseId,
                            grouperIndex : args.grouperIndex,
                            parentGroupId : groupId,
                            noGroupValue : grouper.noGroupValue || that.widget.options.locale.noGroupValue,
                            newGroups : args.newGroups
                        });
                    });
                }
            }

            return recordGroup;
        },

        getAllChildRecordsFromGroup : function (group) {
            if (this._shouldCacheRecordsFromAllGroups) {
                if (!this._childCache) {
                    this._childCache = this._buildChildRecordCache();
                }
                if (!this._childCache[group.id]) {
                    this._childCache[group.id] = mapRecords.call(this, group.records);
                }
                return this._childCache[group.id];
            }
            return this._getAllChildRecordsFromGroup(group);
        },
        startChildCache : function () {
            this._shouldCacheRecordsFromAllGroups = true;
        },
        endChildCache : function () {
            this._shouldCacheRecordsFromAllGroups = false;
            this._childCache = null;
        },

        _buildChildRecordCache : function () {
            var that = this;
            var childCache = {};
            _.forEach(that.recordToGroupMapping, function buildCacheForRecord(groupedId, recordId) {
                var groups = that._identifyGroups(groupedId);
                _.forEach(groups, function (groupId) {
                    var group = that.recordGroups[groupId];
                    if (!group) {
                        return;
                    }

                    var allChildRecords;
                    if (!childCache[groupId]) {
                        allChildRecords = childCache[groupId] = mapRecords.call(that, group.records);
                    } else {
                        allChildRecords = childCache[groupId];
                    }

                    if (group.childGroups.length || group.hierarchyGroups.length) {
                        allChildRecords.push(that.widget._getRecord(recordId));
                    }
                });
            });
            return childCache;
        },
        _identifyGroups : function (groupedId) {
            var ids = groupedId.split(':');
            var accum = ids[0];
            var result = [accum];
            for (var i = 1, len = ids.length; i < len; ++i) {
                var id = ids[i];
                if (id) {
                    accum += ':' + ids[i];
                    result.push(accum);
                }
            }
            return result;
        },

        /*
         * Gets all children nodes, recursively, of the given group.
         *
         * Say you have groups nested two levels deep; their ids might look like:
         *
         * group-proj1
         * group-proj1:5
         *
         * So, a recordToGroupMapping with a node under group-proj1 looks like:
         *
         * { a1 : group-proj1 }
         *
         * And a record under group-proj1:5 would have an entry like:
         *
         * { a2 : group-proj1:5 }
         *
         * And based on startsWith('group-proj1'), you'd get both nodes.
         * startsWith('group-proj1:5') would only get a2.
         */
        _getAllChildRecordsFromGroup : function (group) {
            /* eslint-disable no-restricted-syntax, no-continue, guard-for-in */
            var that = this;
            var recordId;
            var expectedGroupId = group.id;
            var allChildRecords = mapRecords.call(this, group.records);

            // Grab all the child records for averages,medians and range
            if (group.childGroups.length || group.hierarchyGroups.length) {
                // Abuses the fact that child groupIds will have the parent id as their start, group-rowset-a-0:0 child: group-rowset-a-0:0:1
                // Which allows for very quick access to all child records.
                var mapping = that.recordToGroupMapping;
                for (recordId in mapping) {
                    if (mapping.hasOwnProperty(recordId)) {
                        var groupId = mapping[recordId];
                        if (_.startsWith(groupId, expectedGroupId)) {
                            allChildRecords.push(that.widget._getRecord(recordId));
                        }
                    }
                }
            }

            return allChildRecords;
            /* eslint-enable no-restricted-syntax, no-continue, guard-for-in */
        },

        getRecordsFromGroup : function (group) {
            return mapRecords.call(this, group.records);
        },

        _enabled : function () {
            return this.widget.options.grouping.enableRollup;
        },
        _getRecordsFromNode : function (group) {
            return this.getRecordsFromGroup(group);
        },
        _getNode : function getGroup(groupId) {
            return this.recordGroups[groupId];
        },
        _getChildNodesFromNode : function (group) {
            return group.childGroups.concat(group.hierarchyGroups);
        },
        _getNodeId : function (group) {
            return group.id;
        },
        _makeRollupArgs : function (group, dataIndices, records, childGroups) {
            return {
                group : group,
                dataIndices : normalizeDataIndices(dataIndices),
                records : records,
                childGroups : childGroups,
                allChildRecords : this.getAllChildRecordsFromGroup(group)
            };
        },

        getHierarchyChildrenFromGroup : function (group) {
            var children = [];
            for (var i = 0; i < group.hierarchyGroups.length; i++) {
                var childGroup = this.recordGroups[group.hierarchyGroups[i]];
                children.push(childGroup.backingRecord);
                children = children.concat(this.getHierarchyChildrenFromGroup(childGroup));
            }
            return children;
        },

        rollupGroupValues : function (groupId, dataIndices, rowsetId) {
            logger.warn('rollupGroupValues is deprecated - use rollupFromLeafs(<Array<Id>, dataIndices, rowsetId) instead');
            this.rollupFromLeafs([groupId], dataIndices, rowsetId);
        },

        rollupStackedColumns : function (args, commonArgs) {
            if (!this.widget.options.stackedCells.enabled || !this.widget.options.stackedCells.rollupEnabled) {
                return;
            }

            var stackedColumns = commonArgs.stackedColumns;
            var stackedSubRows = commonArgs.stackedSubRows,
                records = args.records,
                childGroups = args.childGroups,
                allChildRecords = args.allChildRecords,
                group = args.group,
                that = this;
            _.forEach(stackedSubRows, function (stackedSubRow) {
                _.forEach(stackedColumns, function (stackedColumn) {
                    var valueKey = stackedColumn.dataIndex + '-' + stackedSubRow.dataIndex;
                    var rollupValue = null;
                    var values = _.map(records, function (record) {
                        return that.widget._getStackedValueFromRecord(record, stackedColumn.dataIndex, stackedSubRow.dataIndex);
                    });

                    if (childGroups.length) {
                        // Average median and range need to look at all records to properly calculate
                        if (stackedSubRow.rollupRole === 'average' || stackedSubRow.rollupRole === 'median' || stackedSubRow.rollupRole === 'range') {
                            values = _.map(allChildRecords, function (record) {
                                return that.widget._getStackedValueFromRecord(record, stackedColumn.dataIndex, stackedSubRow.dataIndex);
                            });
                        } else {
                            _.forEach(childGroups, function (childGroupId) {
                                values.push(that.recordGroups[childGroupId].stackedRollupValues[valueKey]);
                            });
                        }
                    }

                    values = _.filter(values, function (value) {
                        return value != null;
                    });

                    if (values.length || stackedSubRow.rollupRole === 'store' || stackedSubRow.rollupRole === 'function') {
                        if (stackedSubRow.rollupRole === 'store') {
                            if (that.store.rollupValues != null) {
                                rollupValue = that.store.rollupValues(values, stackedSubRow.dataIndex, {
                                    records : records,
                                    childGroups : childGroups,
                                    values : values,
                                    allChildRecords : allChildRecords,
                                    group : group,
                                    rollupValues : group.rollupValues
                                });
                            }
                        } else if (stackedSubRow.rollupRole === 'function') {
                            if (stackedSubRow.rollupFn != null) {
                                rollupValue = stackedSubRow.rollupFn({
                                    values : values,
                                    records : records,
                                    childGroups : childGroups,
                                    allChildRecords : allChildRecords,
                                    group : group,
                                    rollupValues : group.rollupValues
                                });
                            }
                        } else {
                            rollupValue = rollups[stackedSubRow.rollupRole](values);
                        }
                        group.stackedRollupValues[valueKey] = rollupValue;
                    } else if (group.stackedRollupValues && group.stackedRollupValues[valueKey]) {
                        delete group.stackedRollupValues[valueKey];
                    }
                });
            });
        },

        rollupStandardColumns : function (args, commonArgs) {
            var that = this;
            var standardColumns = commonArgs.standardColumns,
                records = args.records,
                childGroups = args.childGroups,
                allChildRecords = args.allChildRecords,
                group = args.group;
            _.forEach(standardColumns, function (column) {
                var rollupValue = null;
                var values = _.map(records, function (record) {
                    return that.store.getValueFromRecord(record, column.dataIndex);
                });

                if (childGroups.length) {
                    // Average median and range need to look at all records to properly calculate
                    if (column.rollupRole === 'average' || column.rollupRole === 'median' || column.rollupRole === 'range') {
                        values = _.map(allChildRecords, function (record) {
                            return that.store.getValueFromRecord(record, column.dataIndex);
                        });
                    } else {
                        _.forEach(childGroups, function (childGroupId) {
                            values.push(that.recordGroups[childGroupId].rollupValues[column.dataIndex]);
                        });
                    }
                }

                if (column.rollupRole !== 'minDateNull' && column.rollupRole !== 'maxDateNull') {
                    values = _.filter(values, function (value) {
                        return value != null;
                    });
                }

                if (values.length || column.rollupRole === 'store' || column.rollupRole === 'function') {
                    if (column.rollupRole === 'store') {
                        if (that.store.rollupValues != null) {
                            rollupValue = that.store.rollupValues(values, column.dataIndex, {
                                records : records,
                                childGroups : childGroups,
                                values : values,
                                allChildRecords : allChildRecords,
                                group : group,
                                rollupValues : group.rollupValues
                            });
                        }
                    } else if (column.rollupRole === 'function') {
                        if (column.rollupFn != null) {
                            rollupValue = column.rollupFn({
                                values : values,
                                records : records,
                                childGroups : childGroups,
                                allChildRecords : allChildRecords,
                                group : group,
                                rollupValues : group.rollupValues
                            });
                        }
                    } else {
                        rollupValue = rollups[column.rollupRole](values);
                    }
                    group.rollupValues[column.dataIndex] = rollupValue;
                } else if (group.rollupValues && group.rollupValues[column.dataIndex]) {
                    delete group.rollupValues[column.dataIndex];
                }
            });
        },

        rollupTimephasedColumns : function (args, commonArgs) {
            if (!this.widget.options.timephased.enabled) {
                return;
            }
            var timephasedBuckets = commonArgs.timephasedBuckets;
            var timephasedSubRows = commonArgs.timephasedSubRows,
                records = args.records,
                childGroups = args.childGroups,
                allChildRecords = args.allChildRecords,
                group = args.group,
                that = this;
            _.forEach(timephasedSubRows, function (timephasedSubRow) {
                _.forEach(timephasedBuckets, function (timephasedBucket) {
                    var rollupValue = null;
                    var bucketKey = timephasedBucket.granularity + '-' + timephasedBucket.bucketId;
                    var valueKey = bucketKey + '.' + timephasedSubRow.dataIndex;
                    var values = _.map(records, function (record) {
                        return that.widget._getTimephasedValueFromRecord({
                            recordId : that.widget._getRecordId(record),
                            record : record,
                            dataIndex : timephasedSubRow.dataIndex,
                            bucketId : timephasedBucket.bucketId,
                            granularity : timephasedBucket.granularity
                        });
                    });

                    if (childGroups.length) {
                        // Average median and range need to look at all records to properly calculate
                        if (timephasedSubRow.rollupRole === 'average' || timephasedSubRow.rollupRole === 'median' || timephasedSubRow.rollupRole === 'range') {
                            values = _.map(allChildRecords, function (record) {
                                return that.widget._getTimephasedValueFromRecord({
                                    recordId : that.widget._getRecordId(record),
                                    record : record,
                                    dataIndex : timephasedSubRow.dataIndex,
                                    bucketId : timephasedBucket.bucketId,
                                    granularity : timephasedBucket.granularity
                                });
                            });
                        } else {
                            _.forEach(childGroups, function (childGroupId) {
                                values.push(_.get(that.recordGroups[childGroupId].timephasedRollupValues, valueKey));
                            });
                        }
                    }

                    values = _.filter(values, function (value) {
                        return value != null;
                    });

                    if (values.length || timephasedSubRow.rollupRole === 'store' || timephasedSubRow.rollupRole === 'function') {
                        if (timephasedSubRow.rollupRole === 'store') {
                            if (that.store.rollupValues != null) {
                                rollupValue = that.store.rollupValues(values, timephasedSubRow.dataIndex, {
                                    records : records,
                                    childGroups : childGroups,
                                    values : values,
                                    allChildRecords : allChildRecords,
                                    group : group,
                                    rollupValues : group.rollupValues,
                                    timephasedRollupValues : group.timephasedRollupValues[bucketKey],
                                    granularity : timephasedBucket.granularity,
                                    bucketId : timephasedBucket.bucketId,
                                    dataIndex : timephasedSubRow.dataIndex
                                });
                            }
                        } else if (timephasedSubRow.rollupRole === 'function') {
                            if (timephasedSubRow.rollupFn != null) {
                                rollupValue = timephasedSubRow.rollupFn({
                                    values : values,
                                    records : records,
                                    childGroups : childGroups,
                                    allChildRecords : allChildRecords,
                                    group : group,
                                    rollupValues : group.rollupValues,
                                    granularity : timephasedBucket.granularity,
                                    bucketId : timephasedBucket.bucketId,
                                    dataIndex : timephasedSubRow.dataIndex,
                                    timephasedRollupValues : group.timephasedRollupValues[bucketKey]
                                });
                            }
                        } else {
                            rollupValue = rollups[timephasedSubRow.rollupRole](values);
                        }
                        _.set(group.timephasedRollupValues, valueKey, rollupValue);
                    } else if (_.get(group.timephasedRollupValues, valueKey)) {
                        delete group.timephasedRollupValues[bucketKey][timephasedSubRow.dataIndex];
                    }
                });
            });
        },

        shouldCreateNewGroup : function (indices) {
            var noIndices = _.compact(indices).length === 0;
            var hasGroupableIndex = noIndices || _.intersection(_.pluck(this.widget.config.groupBy, 'dataIndex'), indices).length > 0;
            return noIndices || hasGroupableIndex;
        },

        rerollGroupValues : function (args) {
            var that = this;
            var oldGroup = this.recordGroups[this.recordToGroupMapping[args.recordId]];
            var newGroups;
            var toRollup = [];
            var rollupAll = false;

            if (this.shouldCreateNewGroup(args.dataIndex)) {
                newGroups = this.groupRecord(args.record);
                if (newGroups.length) {
                    _.forEach(newGroups, function (newGroup) {
                        toRollup.push(newGroup.id);
                    });
                }
                var newGroup = this.recordGroups[this.recordToGroupMapping[args.recordId]];
                if (newGroup != null && oldGroup != null) {
                    if (newGroup !== oldGroup) {
                        that.moveRecordGroup(args.record, oldGroup, newGroups);
                        // If the group has changed recalculate all dataIndexes
                        rollupAll = true;
                        toRollup.push(oldGroup.id);
                    }
                    toRollup.push(newGroup.id);
                }
            } else if (oldGroup != null) {
                toRollup.push(oldGroup.id);
            }

            if (toRollup.length) {
                this.rollupFromLeafs(toRollup, rollupAll ? null : args.dataIndex, args.rowsetId);
            }
        },

        forceAllRollups : function (args) {
            if (this.widget.options.grouping.enableRollup) {
                var dataIndices = args && args.dataIndices;
                var toRollup = [];
                _.forEach(this.recordGroups, function (group) {
                    if (_.isEmpty(group.hierarchyGroups) && _.isEmpty(group.childGroups)) {
                        toRollup.push(group.id);
                    }
                });
                if (toRollup.length) {
                    this.rollupFromLeafs(toRollup, dataIndices);
                    this.widget._forceViewRender();
                }
            }
        },

        recordsAdded : function (all) {
            var that = this;
            var toRollup = [];
            var rowset;
            _.forEach(all, function (args) {
                var newGroups = that.groupRecord(args.record);
                if (newGroups.length) {
                    _.forEach(newGroups, function (newGroup) {
                        toRollup.push(newGroup.id);
                    });
                }
                rowset = rowset || that.widget._getRecordPlacement(args.record).rowsetId;
                var group = that.recordGroups[that.recordToGroupMapping[args.recordId]];
                if (group != null) {
                    toRollup.push(group.id);
                }
                if (group != null) {
                    that.addRecordGroupsToView(newGroups);
                }
            });
            if (toRollup.length) {
                that.rollupFromLeafs(toRollup, null, rowset);
            }
        },

        addRecordGroupsToView : function (groupsToAdd) {
            // Make new groups if needed
            for (var i = 0; i < groupsToAdd.length; i++) {
                var group = groupsToAdd[i];
                var groupLocation = this.getGroupPlacement(group);
                this.widget.view.updateNode(group.id, groupLocation, this.widget._attrs.treeAttrs(group));
                this.widget.view.addRecord({ recordId : group.id, location : groupLocation });
            }
        },

        recordsRemoved : function (recordIds) {
            var that = this;
            var visitedGroupIds = {};
            var visitedGroups = [];
            var allPrunedGroups = [];
            var parentsToRollup = [];
            _.forEach(recordIds, function handleRecordDeletion(recordId) {
                var group = that.recordGroups[that.recordToGroupMapping[recordId]];
                delete that.recordToGroupMapping[recordId];
                if (group == null) { // already removed
                    logger.warn('Trying to remove a record that isnt here, you might be duplicating calls');
                    return;
                }

                if (!visitedGroupIds[group.id]) {
                    visitedGroups.push(group);
                    visitedGroupIds[group.id] = true;
                }
                if (group.parentId && !visitedGroupIds[group.parentId]) {
                    parentsToRollup.push(group.parentId);
                }

                var loc = group.records.indexOf(recordId);
                if (loc >= 0) {
                    group.records.splice(loc, 1);
                }
                delete group.recordLocations[recordId];
                var groupsToDelete = [];
                if (!that.widget.options.grouping.showEmptyBands) {
                    groupsToDelete = that.pruneEmptyGroups(group, groupsToDelete);
                    allPrunedGroups = allPrunedGroups.concat(groupsToDelete);
                }
            });

            visitedGroups = _.reject(visitedGroups, function rejectEmptyGroups(visitedGroup) {
                return _.includes(allPrunedGroups, visitedGroup.id);
            });
            var groupIds = [];
            _.forEach(visitedGroups, function reindexGroups(group) {
                that.indexRecords(group, 0);
                groupIds.push(group.id);
            });

            // we want to rollup all groups that still exist, and parent groups of those that were deleted - dupes to rollupFromLeafs are OK, but we do want to avoid passing groups that no longer exist
            var toRollup = groupIds.concat(_.difference(parentsToRollup, allPrunedGroups));
            this.rollupFromLeafs(toRollup, null, null, FORCE_EMPTY_ROLLUP_ARGS);

            for (var i = 0; i < allPrunedGroups.length; i++) {
                delete this.recordGroups[allPrunedGroups[i]];
            }
            this.widget.view.deleteRecords(_.uniq(allPrunedGroups), false);
        },

        recordMoved : function (args) {
            var oldGroup = this.recordGroups[this.recordToGroupMapping[args.recordId]];
            var newGroups = this.groupRecord(args.record);
            var toRollup = [];
            if (newGroups.length) {
                _.forEach(newGroups, function (newGroup) {
                    toRollup.push(newGroup.id);
                });
            }
            var newGroup = this.recordGroups[this.recordToGroupMapping[args.recordId]];
            if (newGroup !== oldGroup) {
                this.moveRecordGroup(args.record, oldGroup, newGroups);
                toRollup.push(newGroup.id);
                toRollup.push(oldGroup.id);
            }
            this.rollupFromLeafs(toRollup, null, args.location.rowsetId);
        },

        removeRecordFromGroup : function (recordId, group) {
            // Remove record from old group
            if (group.recordLocations[recordId] != null) {
                group.orderedRecords.splice(group.recordLocations[recordId], 1);
                group.records.splice(group.records.indexOf(recordId), 1);
                var i = group.recordLocations[recordId];
                delete group.recordLocations[recordId];
                this.indexRecords(group, i);
            }
            var groupsToDelete = [];
            if (!this.widget.options.grouping.showEmptyBands) {
                groupsToDelete = this.pruneEmptyGroups(group, groupsToDelete);
            }
            this.widget.view.deleteRecords(groupsToDelete, true);
            return groupsToDelete;
        },

        moveRecordGroup : function (record, oldGroup, newGroups) {
            var recordId = this.widget._getRecordId(record),
                location = this.widget._getRecordPlacement(record);

            // Remove record from old group
            var groupsToDelete = this.removeRecordFromGroup(recordId, oldGroup);

            this.widget.view.deleteRecords([recordId], true);

            for (var i = 0; i < groupsToDelete.length; i++) {
                delete this.recordGroups[groupsToDelete[i]];
            }

            this.addRecordGroupsToView(newGroups);
            // Add record back in
            this.widget.view.updateNode(recordId, location, this.widget._attrs.treeAttrs(record));
            // Only add into teh view if the parent is not expanded
            var shouldAdd = !this.widget.view.ensureExpanded({
                nodeId : location.parentId,
                rowsetId : location.rowsetId
            });
            if (shouldAdd) {
                this.widget.view.addRecord({ recordId : recordId, location : location });
            }

            this.widget.view.resize();

            if (this.widget.focus.recordId === recordId) {
                this.widget.selectCell(null, null, 'middle');
            }
        },

        getGroupChildIds : function (groupId, args) {
            var childIds = [];

            var group = this.recordGroups[groupId];
            if (group.childGroups.length > 0) {
                childIds = _.clone(group.childGroups);
            } else {
                childIds = this.widget._orderChildIds({
                    recordId : groupId,
                    childRecords : this.getRecordsFromGroup(group),
                    forSearch : args.forSearch
                });
                group.orderedRecords = childIds;
            }
            var orderedIds = _.clone(childIds);

            if (group.hierarchyGroups.length) {
                if (group.orderedHierarchyGroups == null) {
                    var sortedHierarchy = _.sortBy(group.hierarchyGroups, function (groupId) {
                        return group.groupConfig.groupingOrder.indexOf(parseInt(groupId.slice(Math.max(groupId.lastIndexOf(':'), groupId.lastIndexOf('-')) + 1), 10));
                    });
                    group.orderedHierarchyGroups = sortedHierarchy;
                }
                orderedIds = orderedIds.concat(group.orderedHierarchyGroups);
            }
            //Keep recordLocations in sync after updating sort order
            for (var i = 0; i < childIds.length; i++) {
                group.recordLocations[childIds[i]] = i;
            }
            return orderedIds;
        },

        getGroupPlacement : function (record) {
            var siblingId, parentId;
            var rowsetId = record.id.split('-')[1];
            if (record.parentId != null && this.recordGroups[record.parentId] != null) {
                parentId = record.parentId;
                var parentGroup = this.recordGroups[record.parentId];
                siblingId = null;

                if (record.location === 'childGroups') {
                    if (record.locationIndex === 0) {
                        siblingId = _.last(parentGroup.records);
                    } else {
                        siblingId = parentGroup.childGroups[record.locationIndex - 1];
                    }
                } else if (record.location === 'hierarchyGroups') {
                    if (record.locationIndex === 0) {
                        siblingId = _.last(parentGroup.records) || _.last(parentGroup.childGroups);
                    } else {
                        siblingId = parentGroup.hierarchyGroups[record.locationIndex - 1];
                    }
                }
            } else if (record.locationIndex === 0) {
                siblingId = null;
            } else {
                    // eslint-disable-next-line lodash/path-style
                siblingId = _.get(this.rootNodes, [record.rootQualifier, record.locationIndex - 1], null);
            }
            return {
                rowsetId : rowsetId,
                parentId : parentId,
                siblingId : siblingId
            };
        },

        getRecordPlacement : function (record) {
            var location, siblingId, indexOfChild;
            var recordId = this.widget._getRecordId(record);
            var groupId = this.recordToGroupMapping[recordId];
            if (groupId != null) {
                var group = this.recordGroups[this.recordToGroupMapping[recordId]];
                indexOfChild = group.recordLocations[recordId];
                siblingId = null;

                if (indexOfChild > 0) {
                    siblingId = group.orderedRecords[indexOfChild - 1];
                }

                location = {
                    parentId : group.id,
                    rowsetId : group.id.split('-')[1],
                    siblingId : siblingId
                };
            } else {
                location = {
                    rowsetId : this.widget.defaultRowset
                };
            }
            return location;
        },

        /*
         * Removes empty groups if they have no records or children. Recursively calls to clean the parents as well.
         */
        pruneEmptyGroups : function (group, emptyGroups) {
            var reindexGroups;
            if (group.records.length === 0 && group.childGroups.length === 0 && group.hierarchyGroups.length === 0) {
                emptyGroups.push(group.id);
                if (group.parentId != null && this.recordGroups[group.parentId] != null) {
                    var parentGroup = this.recordGroups[group.parentId];
                    // Remove child group but also remove it from the pruned empty groups if it exists
                    var subGroup = this.recordGroups[group.id];
                    var location = _.get(subGroup, 'location');
                    var subGroupIndex = _.get(subGroup, 'locationIndex', -1);
                    if (location === 'childGroups') {
                        parentGroup.childGroups.splice(subGroupIndex, 1);
                    }
                    if (location === 'hierarchyGroups') {
                        parentGroup.hierarchyGroups.splice(subGroupIndex, 1);
                    }
                    reindexGroups = parentGroup[location];
                    this.pruneEmptyGroups(this.recordGroups[group.parentId], emptyGroups);
                } else {
                    this.rootNodes[group.rootQualifier].splice(this.rootNodes[group.rootQualifier].indexOf(group.id), 1);
                    reindexGroups = this.rootNodes[group.rootQualifier];
                }
                _.forEach(reindexGroups, function (groupId, idx) {
                    var group = this.recordGroups[groupId];
                    if (group) {
                        group.locationIndex = idx;
                    }
                }.bind(this));
            }
            return emptyGroups;
        },
        // This is the biggest mistake. Thanks Brian.
        orderGroups : function () {
            var groupers = this.widget.config.groupBy;
            var that = this,
                i;
            var previousGroups = 1;
            for (i = 0; i < groupers.length; i++) {
                var grouper = groupers[i];
                var groupValues = this.groupValues[i];
                var groupDataValues = this.groupDataValues[i];
                var sortValueArgs = {
                    grouper : grouper,
                    groupingOptions : that.widget.options.grouping
                };
                var sortedValues = grouper.sortValues !=
                null ? grouper.sortValues(_.clone(groupValues), _.clone(groupDataValues), sortValueArgs) : _.clone(groupValues).sort();
                if (grouper.sortOrder === 'dsc') {
                    sortedValues = sortedValues.reverse();
                }
                var sortMap = [];
                var indexes = {};
                for (var x = 0; x < groupValues.length; x++) {
                    indexes[groupValues[x]] = x;
                }
                for (var j = 0; j < sortedValues.length; j++) {
                    var index = indexes[sortedValues[j]];
                    if (index < 0) {
                        groupValues.push(sortedValues[j]);
                        index = groupValues.length - 1;
                    }
                    sortMap.push(index);
                }
                grouper.groupingOrder = sortMap;
                if (groupValues.length > this.widget.options.grouping.maxEmptyBands || ((groupValues.length * previousGroups) > this.widget.options.grouping.maxEmptyBands)) {
                    grouper.ignoreEmptyBands = true;
                    this.widget._trigger('performance-optimization', null, {
                        event: 'emptyBandsDisabled',
                        grouper: grouper,
                        dataIndex : grouper.dataIndex,
                        totalBandCount : groupValues.length * previousGroups,
                        message : 'Grid disabled empty bands for this grouper to prevent potential issues.',
                        limit : this.widget.options.grouping.maxEmptyBands
                    });
                }
                previousGroups *= groupValues.length;
            }

            var firstGroup = _.first(groupers);
            var groupsToAdd = [];

            //If store is empty there will be no rootNodes, in the case of showEmptyBands inject empty array to force
            //empty bands to show
            if (_.isEmpty(that.rootNodes) && that.widget.options.grouping.showEmptyBands) {
                that.rootNodes['rowset-' + that.widget.defaultRowset] = [];
            }

            _.forEach(that.rootNodes, function (rowsetRootList, rootQualifier) {
                var rowsetId = rootQualifier.substring(rootQualifier.lastIndexOf('-'));
                var baseId = 'group' + rowsetId + '-';

                var properlyOrderedRootGroups = [];

                for (var i = 0; i < firstGroup.groupingOrder.length; i++) {
                    properlyOrderedRootGroups.push(baseId + firstGroup.groupingOrder[i]);
                }

                if (that.widget.options.grouping.showEmptyBands && !firstGroup.ignoreEmptyBands) {
                    if (!firstGroup.hierarchy) {
                        // Find what groups are not there.
                        var missing = _.difference(properlyOrderedRootGroups, rowsetRootList);
                        _.forEach(missing, function (emptyBand) {
                            var groupValue = that.groupValues[firstGroup.level]
                                [parseInt(emptyBand.substring(emptyBand.lastIndexOf('-') + 1), 10)];

                            var bandValue = that.groupDataValues[firstGroup.level][groupValue];

                            if (groupValue == null) {
                                groupValue = null;
                                bandValue = null;
                            }
                            var newGroup = {
                                id : emptyBand,
                                groupValue : groupValue,
                                dataIndex : firstGroup.dataIndex,
                                formatValue : firstGroup.formatValue,
                                setGroupValueOnRecord : firstGroup.setGroupValueOnRecord,
                                ignoreGroupValueOnNewRecord : firstGroup.ignoreGroupValueOnNewRecord,
                                rootQualifier : rootQualifier,
                                noGroupValue : grouper.noGroupValue || that.widget.options.locale.noGroupValue,
                                level : firstGroup.level,
                                isGroup : true,
                                records : [],
                                groupConfig : firstGroup,
                                childGroups : [],
                                hierarchyGroups : [],
                                rollupValues : {},
                                timephasedRollupValues : {},
                                stackedRollupValues : {},
                                dataValue : bandValue,
                                orderedRecords : [],
                                recordLocations : {}
                            };
                            groupsToAdd.push(newGroup);
                        });
                        that.rootNodes[rootQualifier] = properlyOrderedRootGroups;
                    } else if (that.hierarchyMapping[firstGroup.dataIndex] != null) {
                        var rootHierarchyNodes = that.hierarchyMapping[firstGroup.dataIndex].rootGroups;
                        _.forEach(rootHierarchyNodes, function (rootNode) {
                            var groupKeyPart = that.groupValueIndices[0][rootNode.id];

                            var groupValue = rootNode.id;
                            var dataValue = rootNode.item;
                            if (groupKeyPart < 0) {
                                that.groupValues[0].push(groupValue);
                                that.groupValueIndices[0][groupValue] = that.groupValues[0].length - 1;
                                that.groupDataValues[0][groupValue] = dataValue;
                            }

                            that.buildGroup({
                                hierarchyMapping : that.hierarchyMapping[firstGroup.dataIndex],
                                groupValue : groupValue,
                                dataValue : _.isFunction(firstGroup.getHierarchyDataValue) ? firstGroup.getHierarchyDataValue(dataValue) : dataValue,
                                grouper : firstGroup,
                                level : firstGroup.level,
                                groupId : baseId + that.groupValueIndices[0][groupValue],
                                rootQualifier : rootQualifier,
                                baseId : baseId,
                                grouperIndex : 0,
                                noGroupValue : firstGroup.noGroupValue || that.widget.options.locale.noGroupValue,
                                newGroups : []
                            });
                        });
                        that.rootNodes[rootQualifier] = _.intersection(properlyOrderedRootGroups, rowsetRootList);
                    }
                } else {
                    that.rootNodes[rootQualifier] = _.intersection(properlyOrderedRootGroups, rowsetRootList);
                }
                _.forEach(that.rootNodes[rootQualifier], function (groupId, idx) {
                    var group = that.recordGroups[groupId];
                    if (group) {
                        group.locationIndex = idx;
                    }
                });
            });
            if (groupers.length > 1) {
                _.forEach(_.values(that.recordGroups), function (group) {
                    if (group.groupConfig.level < groupers.length - 1) {
                        var childGrouper = groupers[group.groupConfig.level + 1];
                        // Generate all possible child groups in proper order
                        var properlyOrderedChildGroups = [];
                        for (var i = 0; i < childGrouper.groupingOrder.length; i++) {
                            properlyOrderedChildGroups.push(group.id + ':' + childGrouper.groupingOrder[i]);
                        }
                        if (that.widget.options.grouping.showEmptyBands && !childGrouper.ignoreEmptyBands) {
                            if (!childGrouper.hierarchy) {
                                // Find what groups are not there.
                                var missing = _.difference(properlyOrderedChildGroups, group.childGroups);
                                _.forEach(missing, function (emptyBand) {
                                    var groupValue = that.groupValues[childGrouper.level]
                                        [parseInt(emptyBand.substring(emptyBand.lastIndexOf(':') + 1), 10)];

                                    var bandValue = that.groupDataValues[childGrouper.level][groupValue];

                                    if (groupValue == null) {
                                        groupValue = null;
                                        bandValue = null;
                                    }
                                    var newGroup = {
                                        id : emptyBand,
                                        parentId : group.id,
                                        groupValue : groupValue,
                                        dataIndex : childGrouper.dataIndex,
                                        formatValue : childGrouper.formatValue,
                                        setGroupValueOnRecord : childGrouper.setGroupValueOnRecord,
                                        ignoreGroupValueOnNewRecord : childGrouper.ignoreGroupValueOnNewRecord,
                                        noGroupValue : grouper.noGroupValue || that.widget.options.locale.noGroupValue,
                                        level : childGrouper.level,
                                        isGroup : true,
                                        records : [],
                                        groupConfig : childGrouper,
                                        childGroups : [],
                                        hierarchyGroups : [],
                                        rollupValues : {},
                                        timephasedRollupValues : {},
                                        dataValue : bandValue,
                                        orderedRecords : [],
                                        recordLocations : {}
                                    };
                                    groupsToAdd.push(newGroup);
                                });
                                group.childGroups = properlyOrderedChildGroups;
                            } else if (that.hierarchyMapping[childGrouper.dataIndex] != null) {
                                var rootHierarchyNodes = that.hierarchyMapping[childGrouper.dataIndex].rootGroups;
                                if (_.isFunction(group.groupConfig.filterEmptyBands)) {
                                    rootHierarchyNodes = group.groupConfig.filterEmptyBands({
                                        rootHierarchyNodes : rootHierarchyNodes,
                                        backingRecord : group.backingRecord,
                                        childType : childGrouper.recordType
                                    });
                                }
                                _.forEach(rootHierarchyNodes, function (rootNode) {
                                    var groupKeyPart = that.groupValueIndices[childGrouper.level][rootNode.id];
                                    var groupValue = rootNode.id;
                                    var dataValue = rootNode.item;
                                    if (groupKeyPart < 0) {
                                        that.groupValues[childGrouper.level].push(groupValue);
                                        that.groupValueIndices[childGrouper.level][groupValue] = that.groupValues[childGrouper.level].length - 1;
                                        that.groupDataValues[childGrouper.level][groupValue] = dataValue;
                                    }
                                    var baseId = group.id + ':';

                                    that.buildGroup({
                                        hierarchyMapping : that.hierarchyMapping[childGrouper.dataIndex],
                                        groupValue : groupValue,

                                        dataValue : _.isFunction(childGrouper.getHierarchyDataValue) ? childGrouper.getHierarchyDataValue(dataValue) : dataValue,
                                        grouper : childGrouper,
                                        level : childGrouper.level,
                                        parentGroupId : group.id,
                                        groupId : baseId + that.groupValues[childGrouper.level].indexOf(groupValue),
                                        baseId : baseId,
                                        grouperIndex : childGrouper.level,
                                        noGroupValue : childGrouper.noGroupValue || that.widget.options.locale.noGroupValue,
                                        newGroups : []
                                    });
                                });
                                group.childGroups = _.intersection(properlyOrderedChildGroups, group.childGroups);
                            }
                        } else {
                            // use intersection to apply the correct order to the child groups, it will use the order of the first array passed in
                            group.childGroups = _.intersection(properlyOrderedChildGroups, group.childGroups);
                        }
                    }
                    _.forEach(group.childGroups, function (groupId, idx) {
                        var group = that.recordGroups[groupId];
                        if (group) {
                            group.locationIndex = idx;
                        }
                    });
                });
            }

            // Add empty bands if need be
            if (groupsToAdd.length) {
                _.forEach(groupsToAdd, function (groupToAdd) {
                    that.recordGroups[groupToAdd.id] = groupToAdd;
                });
            }
        },
        setGroupBandValues : function (newRecord, focus) {
            focus = focus || this.widget.focus.record;
            if (this.widget.groupingApplied && focus != null) {
                var group;
                var setValues = [];
                if (this.widget._isGroupRecord(focus)) {
                    group = focus;
                } else {
                    group = this.recordGroups[this.recordToGroupMapping[this.widget._getRecordId(focus)]];
                }
                if (group != null) { // When adding from a dummy new row
                    if (!group.ignoreGroupValueOnNewRecord) {
                        if (group.setGroupValueOnRecord != null) {
                            group.setGroupValueOnRecord({
                                record : newRecord,
                                dataIndex : group.dataIndex,
                                value : group.dataValue,
                                mediator : this.store
                            });
                        } else if (this.store.setGroupValueOnRecord != null) {
                            this.store.setGroupValueOnRecord({
                                record : newRecord,
                                dataIndex : group.dataIndex,
                                value : group.dataValue
                            });
                        } else {
                            this.store.setRecordValue(newRecord, group.dataIndex, group.dataValue);
                        }
                        setValues.push(group.dataIndex);
                    }
                    while (group.parentId != null) {
                        group = this.recordGroups[group.parentId];
                        if (!_.includes(setValues, group.dataIndex) && !group.ignoreGroupValueOnNewRecord) {
                            if (group.setGroupValueOnRecord != null) {
                                group.setGroupValueOnRecord({
                                    record : newRecord,
                                    dataIndex : group.dataIndex,
                                    value : group.dataValue,
                                    mediator : this.store
                                });
                            } else if (this.store.setGroupValueOnRecord != null) {
                                this.store.setGroupValueOnRecord({
                                    record : newRecord,
                                    dataIndex : group.dataIndex,
                                    value : group.dataValue
                                });
                            } else {
                                this.store.setRecordValue(newRecord, group.dataIndex, group.dataValue);
                            }
                            setValues.push(group.dataIndex);
                        }
                    }
                }
            }
        },

        updateRecordIds : function (recordIdMap) {
            /* eslint-disable no-restricted-syntax, no-continue, guard-for-in */
            var that = this;

            // Update the group objects to reference new id
            var recordGroups = that.recordGroups;
            for (var prop in recordGroups) {
                if (!recordGroups.hasOwnProperty(prop)) {
                    continue;
                }
                var records = recordGroups[prop].records;
                _.forEach(records, function updateRecordsInner(oldId, index) {
                    var newId = recordIdMap[oldId];
                    if (newId != null && newId !== oldId) {
                        records[index] = newId;
                    }
                });
            }

            for (var oldId in recordIdMap) {
                var newId = recordIdMap[oldId];
                if (newId === oldId) {
                    continue;
                }
                if (!recordIdMap.hasOwnProperty(oldId)) {
                    continue;
                }

                // Add new ids mapping for its group, delet the old one
                var recordsGroup = that.recordToGroupMapping[oldId];
                if (recordsGroup != null) {
                    that.recordToGroupMapping[newId] = recordsGroup;
                    delete that.recordToGroupMapping[oldId];
                }
            }
            /* eslint-enable no-restricted-syntax, no-continue, guard-for-in */
        }
    });

    return RecordGrouper;
});
