define([
    'gridapi.lodash',
    'gridapi.moment',
    'gridapi.logger'
], function (_, moment, logger) {
    'use strict';

    var NUMERIC_FLAG = { numeric : true };

    var standardCompare = function (value1, value2, asc) {
        var a, b;
        if (asc) {
            a = value1;
            b = value2;
        } else {
            a = value2;
            b = value1;
        }
        if (a == null && b == null) return 0;
        if (a == null) return -1;
        if (b == null) return 1;
        return a < b ? -1 : a > b ? 1 : 0;
    };

    var sorters = {
        float : function (valueA, valueB, asc) {
            var a = parseFloat(valueA),
                b = parseFloat(valueB);
            if (isNaN(a)) {
                a = null;
            }
            if (isNaN(b)) {
                b = null;
            }
            return standardCompare(a, b, asc);
        },
        boolean : function (valueA, valueB, asc) {
            var a = valueA === 'true' || valueA === true || valueA === 'Y' || valueA === 'yes' || valueA === 1,
                b = valueB === 'true' || valueB === true || valueB === 'Y' || valueB === 'yes' || valueB === 1;
            return standardCompare(a, b, asc);
        },
        integer : function (valueA, valueB, asc) {
            var a = parseInt(valueA, 10),
                b = parseInt(valueB, 10);
            if (isNaN(a)) {
                a = null;
            }
            if (isNaN(b)) {
                b = null;
            }
            return standardCompare(a, b, asc);
        },
        string : function (valueA, valueB, asc, options) {
            //>>excludeStart("validation", pragmas.production)
            if ((valueA != null && !_.isString(valueA)) ||
                (valueB != null && !_.isString(valueB))) {
                logger.error('grid/sorters | an attempt was made to sort, with the "string" sorter, one or more values that were not strings.  (did you specify the dataType property for your columns?  note that the default dataType for columns is string)');
            }
            //>>excludeEnd("validation")
            var a = valueA != null && _.isString(valueA) ? valueA.toLowerCase() : '',
                b = valueB != null && _.isString(valueB) ? valueB.toLowerCase() : '',
                match = 0;
            if (options && options.numeric) {
                if (asc) {
                    match = a.localeCompare(b, undefined, NUMERIC_FLAG);
                } else {
                    match = b.localeCompare(a, undefined, NUMERIC_FLAG);
                }
            } else if (asc) {
                match = a.localeCompare(b);
            } else {
                match = b.localeCompare(a);
            }

            return match;
        },
        date : function (valueA, valueB, asc) {
            return standardCompare(valueA, valueB, asc);
        },
        datetime : function (valueA, valueB, asc) {
            return standardCompare(valueA, valueB, asc);
        },
        check : function (valueA, valueB, asc) {
            var a = valueA === 'true' || valueA === true || valueA === 'Y' || valueA === 'yes' || valueA === 1,
                b = valueB === 'true' || valueB === true || valueB === 'Y' || valueB === 'yes' || valueB === 1;
            return standardCompare(a, b, asc);
        },
        'enum' : function (valueA, valueB, asc) {
            return standardCompare(valueA, valueB, asc);
        },
        picker : function (valueA, valueB, asc) {
            var a = valueA != null ? valueA.length : null,
                b = valueB != null ? valueB.length : null;
            return standardCompare(a, b, asc);
        }
    };

    return function (requestedSorter) {
        return sorters[requestedSorter]; // TODO can we give the user a default sorter?  Or throw an error?
    };
});
