define([
    'gridapi.jquery',
    'gridapi.lodash',
    'gridapi.grid/utils/keyCode',
    'gridapi.grid/gridMixins',
    'gridapi.logger',
    'gridapi.grid/touchInteractions/utils',
    'gridapi.grid/touchInteractions/gestures'
], function ($, _, keyCode, mixins, logger, utils, Gestures) {
    'use strict';

    var DragHandles = function (args, loggerInstance) {
        var that = this;
        that.eventHandlers = args.eventHandlers;
        that.options = args.options;
        that.widget = args.widget;
        that.view = args.gridView;
        that._isTouch = utils.isTouchDevice();
        that.logger = loggerInstance;
    };

    var dHOffset = {
        upper: 12,
        lower: 11
    };

    DragHandles.prototype = {
        constructor: DragHandles,
        upper: '<div style="position: absolute; z-index: 100;" class="lux-datagrid-touch-area upper-drag" ><div class="lux-datagrid-touch-selection-affordance-top" role="button" aria-label="Top selection handle"></div></div>',
        lower: '<div style="position: absolute; z-index: 100;" class="lux-datagrid-touch-area lower-drag" ><div class="lux-datagrid-touch-selection-affordance-bottom" role="button" aria-label="Bottom selection handle"></div></div>',

        isItDragHandle: function ($ele) {
            return ($ele.is('.lux-datagrid-touch-area') || $ele.parents().is('.lux-datagrid-touch-area'));
        },

        findOverlay: function () {
            var that = this;
            var overlay = that.view.$rowsets[that.view.selectionModel.focus.rowsetId].find('.rowset-overlay');
            return overlay;
        },

        createDragHandles: function () {
            var that = this;
            var overlay = that.findOverlay();

            that.$upper = $(that.upper);
            overlay.append(that.$upper);

            that.$lower = $(that.lower);
            overlay.append(that.$lower);
        },

        repositionDragHandles: function () {
            var that = this;
            var gridOffsetLeft = that.view.$grid.offset().left;
            var gridWidth = that.view.$grid.width();
            var $topTr = that.view.$grid.find('.row-selected-top');
            var $bottomTr = that.view.$grid.find('.row-selected-bottom');

            if ($topTr.length) {
                var topTr_offsetTop = $topTr.offset().top;
                that.$upper.offset({ top: (topTr_offsetTop - dHOffset.upper), left: (gridOffsetLeft + gridWidth / 2) });
            }
            if ($bottomTr.length) {
                var rowHeight = that.view.ROW_HEIGHT;
                var bottomTr_offsetTop = $bottomTr.offset().top;
                that.$lower.offset({ top: (bottomTr_offsetTop + rowHeight - dHOffset.lower), left: (gridOffsetLeft + gridWidth / 2) });
            }
            that.showDragHandles();
        },

        hideDragHandles: function (event) {
            var that = this;
            if (that.$lower) {
                that.$upper.css({ 'opacity': 0 }); // note hidden elemnts x y position can not be calcualted.
                that.$lower.css({ 'opacity': 0 });
            }
        },

        showDragHandles: function () {
            var that = this;
            if (that.$lower) {
                that.$upper.css({ 'opacity': 1 });
                that.$lower.css({ 'opacity': 1 });
            }
        },

        isLowerHalf: function (event) {
            var that = this;
            var rowHeight = that.view.ROW_HEIGHT;
            var currentY = event.originalEvent.touches[0].clientY + $(window).scrollTop();
            var $tr = that.getTrUnderTouch(event);
            var trOffsetTop = $tr.offset().top;
            return (currentY >= (trOffsetTop + (rowHeight / 2))); // returns boolean
        },

        exist: function () {
            var that = this;
            return (that.$lower && that.$lower.length >= 1) ? true : false;
        },

        getTrUnderTouch : function (event) {
            var x = event.originalEvent.touches[0].pageX;
            var y = event.originalEvent.touches[0].pageY;
            var $elem = $(document.elementFromPoint(x, y - $(window).scrollTop())); // y : A vertical position within the current viewport.
            var $td = $elem.closest('td');
            var $tr = $td.parent();
            return $tr;
        },

        /*
         * The move function fills the gaps when dragging fast. Otherwise you would see holes in row-selection.
         * it feels all rows 'from' --> to'
         */
        moveFn: function ($tr, recordId, upperOrLower) {
            var that = this;
            var fromId, toId, fromIndex, toIndex, fromRowset, toRowset;
            var focus = that.view.selectionModel.focus;

            if (upperOrLower === 'upper') {
                fromId = recordId;
                toId = (focus.pivotBottomId === null || focus.pivotBottomId === undefined) ? focus.recordId : focus.pivotBottomId;
            } else {
                fromId = (focus.pivotTopId === null || focus.pivotTopId === undefined) ? focus.recordId : focus.pivotTopId;
                toId = recordId;
            }

            fromRowset = focus.rowsetId;
            toRowset = $tr.closest('div.viewport').data().viewport.rowset;
            fromIndex = that.view.getRecordIndex(fromId, fromRowset);
            toIndex = that.view.getRecordIndex(toId, toRowset);

            if (fromRowset === toRowset && fromIndex <= toIndex) { // fromIndex<=toIndex don't cross upper and lower drag handles.
                that.rowsToSelect = that.view.virtualRows[focus.rowsetId].records.slice(fromIndex, toIndex + 1);
                that.widget.selectRows(that.rowsToSelect, false, null, that.view.selectionModel.focus.columnId);
                // setter (current focus do not change)
                focus.pivotIndex = fromIndex;
                focus.pivotId = that.view.virtualRows[focus.rowsetId].records[focus.pivotIndex];
                focus.pivotTopId = _.first(that.rowsToSelect);
                focus.pivotBottomId = _.last(that.rowsToSelect);
            }
            that.repositionDragHandles();
        },

        trackTouchMove: function () {
            var that = this;

            var gestureLower = new Gestures(that.$lower, that.logger);
            gestureLower.register('touchMoveAll', function (event) {
                event.preventDefault();
                event.stopPropagation();
            });
            gestureLower.register('touchMoveThrottle', function (event, direction) {
                var $tr = that.getTrUnderTouch(event);
                var recordId = that.view.fixRecordIdType($tr.attr('data-id'));
                if (recordId !== undefined && that.isLowerHalf(event)) {
                    that.moveFn($tr, recordId, 'lower', direction);
                }
            });

            var gestureUpper = new Gestures(that.$upper, that.logger);
            gestureUpper.register('touchMoveAll', function (event) {
                event.preventDefault();
                event.stopPropagation();
            });
            gestureUpper.register('touchMoveThrottle', function (event, direction) {
                var $tr = that.getTrUnderTouch(event);
                var recordId = that.view.fixRecordIdType($tr.attr('data-id'));
                if (recordId !== undefined && !that.isLowerHalf(event)) {
                    that.moveFn($tr, recordId, 'upper', direction);
                }
            });
        }
    };

    return DragHandles;
});
