/**
 * Copyright (c) 2013 - 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implements luxsearch and luxsearchresults widgets for client-side interaction
 * with multiple REST services providing search suggestions and search results
 * @author Joe McClintock
 */

/**
* The luxsearch widget
* @module luxsearch
* @jqbase
* @requires ojet
* @requires jquery
* @requires knockout
* @requires underscore
* @requires jqueryui
* @requires luxautocomplete
*/
define(['ojs/ojcore', 'knockout', 'jquery', 'underscore', './luxsearchmodel', 'luxautocompleteinput', 'jqueryui-amd/widget',
    'ojs/ojbutton', 'ojs/ojinputtext', 'ojs/ojmenu', 'ojs/ojselectcombobox'],

function(oj, ko, $, _, ojSearchModel) {
    'use strict';
    (function() {
        $.widget('GBU.luxsearch', {
            version:           '5.1.0',
            defaultElement:    '<div>',
            widgetEventPrefix: 'lux',
            /**
             * Options
             * <dl>
             *     <dt>cssClass</dt>
             *     <dd>
             *         CSS class applied to the main element. Default: 'searchForm'<br />
             *         (does not typically need to be overridden)
             *     </dd>
             *     <dt>textInputId</dt>
             *     <dd>
             *         Id value applied to the search text input element. (does not typically need to be overridden)<br />
             *         Type: String<br />
             *         Default: 'luxsearch_input'<br />
             *         Example:<br />
             *         <code>textInputId: 'luxsearch_input'</code>
             *     </dd>
             *     <dt>textInputLabelId</dt>
             *     <dd>
             *         Id value applied to the search text input label element. (does not typically need to be overridden)<br />
             *         Type: String<br />
             *         Default: 'luxsearch_input_label'<br />
             *         Example:<br />
             *         <code>textInputLabelId: 'luxsearch_input_label'</code>
             *     </dd>
             *     <dt>searchCategoriesId</dt>
             *     <dd>
             *         Id value applied to the drop down category list. (does not typically need to be overridden)<br />
             *         Type: String<br />
             *         Default: 'luxsearch_combobox'<br />
             *         Example:<br />
             *         <code>searchCategoriesId: 'luxsearch_combobox'</code>
             *     </dd>
             *     <dt>placeholderText</dt>
             *     <dd>
             *         Optional placeholder text inside the search text box.<br />
             *         Type: String<br />
             *         Default: 'Enter your search term here'<br />
             *         Example:<br />
             *         <code>placeholderText: 'Enter your search term here'</code>
             *     </dd>
             *     <dt>searchFieldLabel</dt>
             *     <dd>
             *         Text for search label.<br />
             *         Type: String<br />
             *         Default: 'Search'<br />
             *         Example:<br />
             *         <code>searchFieldLabel: 'Search'</code>
             *     </dd>
             *     <dt>textInputTitle</dt>
             *     <dd>
             *         Text for search text input element title.<br />
             *         Type: String<br />
             *         Default: 'Search Term'<br />
             *         Example:<br />
             *         <code>textInputTitle: 'Search Term'</code>
             *     </dd>
             *     <dt>suggestionScopeTitle</dt>
             *     <dd>
             *         Text for category selection drop down title.<br />
             *         Type: String<br />
             *         Default: 'Select search scope'<br />
             *         Example:<br />
             *         <code>suggestionScopeTitle: 'Select search scope'</code>
             *     </dd>
             *     <dt>includeSearchButton</dt>
             *     <dd>
             *         Should the search button be visible?<br />
             *         Default: true<br />
             *         When set to false, enter key can be used to invoke search.
             *     </dd>
             *     <dt>searchCaseSensitive</dt>
             *     <dd>
             *         Should searches and suggestions be case-sensitive?<br />
             *         Default: false<br />
             *         Applies only when literal array is provided as data source. Server-side or custom function determines this in the other cases
             *     </dd>
             *     <dt>searchBeginsWith</dt>
             *     <dd>
             *         Should searches and suggestions match beginning of value only?<br />
             *         Default: true<br />
             *         Applies only when literal array is provided as data source. Server-side or custom function determines this in the other cases
             *     </dd>
             *     <dt>suggestionsEnabled</dt>
             *     <dd>
             *         Use suggestions from drop down category list to narrow the auto-complete<br />
             *         Default: true<br />
             *         NOTE: only works if categories/sources supplied for suggestions source.
             *     </dd>
             *     <dt>suggestionsLimit</dt>
             *     <dd>
             *         Maximum number of suggestions shown.<br />
             *         Default: 15<br />
             *         Note: limit is across all sources, not per suggestion source
             *     </dd>
             *     <dt>suggestionsSeparateByCategory</dt>
             *     <dd>
             *         List suggestions separately by category?<br />
             *         Default: true<br />
             *         When set to false, suggestions are combined according to sort logic
             *     </dd>
             *     <dt>suggestionsSortOnClient</dt>
             *     <dd>
             *         Alphabetically sort the suggestions on the client after they come back from server?<br />
             *         Default: false<br />
             *         (assume server knows more about what's most relevant/important)
             *     </dd>
             *     <dt>suggestionsEnableDirectLink</dt>
             *     <dd>
             *         Can the user navigate directly to a suggested item?<br />
             *         Default: true<br />
             *         Direct navigation depends on having suggestion source(s) that populate the url property of the suggestion
             *         When set to false, selecting the suggestion simply places its value in the search box
             *     </dd>
             *     <dt>searchResultsId</dt>
             *     <dd>
             *         Id of the search results wrapper element.  <br />
             *         Default: null<br />
             *         Type: String</br>
             *         Create the aria-controls attribute with this value. Used by screen readers to name
             *         parts of the page controlled by the search button.  If there is no search button
             *         there will be no aria-control event of it specified in the options.
             *     </dd>
             *     <dt>Inline</dt>
             *     <dd>
             *         Lines up the form fields vertically &quote;true&quote; or horizontally &quote;false&quote;.
             *         Default is true
             *     </dd>
             *     <dt>categoryOptionChange</dt>
             *     <dd>
             *         Callback function to run when the search option field changes.<br />
             *         Can be used to effect a behavior or update the
             *         presentation based on the search option selected<br />
             *         Example: <br />
        <pre class="prettyprint" style="overflow-x:visible;">
                    categoryOptionChange: function(event, data)
                            console.log('search option: ' + data.value);
                    }
        </pre>
             *     </dd>
             *     <dt>inputChangeCallback</dt>
             *     <dd>
             *         Callback function to run when the search term field changes.<br />
             *         Can be used to effect a behavior or update the
             *         presentation after the search term is entered<br />
             *         Example: <br />
        <pre class="prettyprint" style="overflow-x:visible;">
                    categoryOptionChange: function(event, data)
                            console.log('search term value: ' + data.value);
                    }
        </pre>
             *     </dd>
             *     <dt>searchButtonClick</dt>
             *     <dd>
             *         Callback function to run when the search icon button is clicked.<br />
             *         Can be used to effect a behavior such as fire
             *         as search message that the search results will use to run a search.<br />
             *         Example: <br />
        <pre class="prettyprint" style="overflow-x:visible;">
                    searchButtonClick: function(event, data)
                            console.log('search term value: ' + event.currentTarget.id + ' search term: ' + data.get('searchTerm'));
                    }
        </pre>
             *     </dd>
             *     <dt>searchButtonId</dt>
             *     <dd>
             *         Unique ID of the search icon button element.
             *     </dd>
             *     <dt>searchModel</dt>
             *     <dd>
             *         Optional search model containing search related data, search option term, search term, preferencesLinks.
             *     </dd>
             *     <dt>suggestionsSources</dt>
             *     <dd>
             *         Array of suggestion sources and which categories they supply.
             *         Categories drop down is populated by the total unique list.
             *         Different types of sources can be mixed together in the array
             *         Each source can have the following properties:<br />
    <pre class="prettyprint" style="overflow-x:visible;">
    format 1:
    value                   : property name holding the value
    label                   : property name holding the display label
    categories              : boolean, indicates node is a category.
    children                : array, contains value, label attributes and can have category and children attributes

    format 2:
    categories (required)   : string array
    source (required)       : array, function, or REST url (string). See code examples below.
    sourceParam             : function to invoke to generate a string to append to end of the REST url based on the search values/category
    responseMap             : used to map JSON responses to search suggestions when the REST service does not return items with value and
                                     lable properties
       collection (optional): top-level property of the JSON response that holds the collection of suggestions
       value                : string property name holding the value. Or, function to execute on the JSON item to generate it.
       label                : string property name holding the display label. Or, function to execute on the JSON item to generate it.
       url (optional)       : string property name holding the url for navigating to this searchs suggestion. Or, function to execute
                                on the JSON item to generate the value.
       icon (optional)      : string property name holding the url of the icon for this suggestion. Or, function to execute on the
                                JSON item to generate the value.
       category (optional)  : string property name holding the category of this suggestion. Or, function to execute on the JSON
                                 item to generate the value.
    </pre>
             *
             *     </dd>
             *     <dt>suggestionsTimeoutMilliseconds</dt>
             *     <dd>
             *         Number of milliseconds to wait for each source to return suggestions. Default: 1000 (1 second)
             *         If there are multiple suggestion sources defined, the suggestions for some might be returned even if we timed out on others
             *     </dd>
             *     <dt>preferencesLinks</dt>
             *     <dd>
             *         Array of search preferences links to show in the preferences button dropdown.
             *         Default: null, preferences button not shown
             *         Each item the must must be a javascript object:
             *           [
             *                   {text: 'Advanced Search', callback: advanceSearchLink },
             *                   {text: 'Preferences Link', callback: preferencesLink},
             *                   {text: 'Profile Link', callback: profileLink}
             *           ],
             *           the callback references a javascript function defined in the view model.
             *
             *           view.advanceSearchLink = function() {
             *               var advanceLink = 'developers/advanced.html';
             *
             *               alert("Advanced link: " + advanceLink);
             *           };
             *
             *     </dd>
             *     <dt>multipleSuggestions</dt>
             *     <dd>
             *         Allows for selection of multiple suggestion.<br />
             *         Default: false
             *     </dd>
             * </dl>
             * @example
             * <caption>Arrays suggestion source, format 1</caption>
                DeepSearchCategories:[{
                    value: "Birds",
                    label: "Birds",
                    categories: true,
                        children: [{
                            value: "Eagle",
                            label: "Eagle"
                        }, {
                            value: "Heron",
                            label: "Heron"
                        }, {
                            value: "Robin",
                            label: "Robin"
                        }, {
                            value: "Mockingbird",
                            label: "Mockingbird"
                        }, {
                            value: "Bluejay",
                            label: "Bluejay"
                        }, {
                            value: "Hawk",
                            label: "Hawk"
                        }]
                    }, {
                    value: "Fish",
                    label: "Fish",
                    categories: true,
                    children: [{
                        value: "Bass",
                        label: "Bass"
                    }, {
                        value: "Pike",
                        label: "Pike"
                    }, {
                        value: "Trout",
                        label: "Trout"
                    }, {
                        value: "Salmon",
                        label: "Salmon"
                    }, {
                        value: "Bluegill",
                        label: "Bluegill"
                    }, {
                        value: "Halibut",
                        label: "Halibut"
                    }, ]
                }],

             * @example
             * <caption>Arrays suggestion source, form 2</caption>
             * [ { categories: ['Birds'], source: ['Eagle', 'Heron', 'Robin', 'Mockingbird', 'Bluejay', 'Hawk'] },
             * { categories: ['Fish'],  source: ['Bass', 'Pike', 'Trout', 'Salmon', 'Bluegill', 'Halibut' ] } ]
             *
             * @example
             * <caption>Example using REST source</caption>
             * [ {
             *   categories: ['Sponsors'],
             *   source: 'webresources/entities.sponsor',
             *   sourceParam: function(req) { return '?beginsWith=' + req.term; },
             * } ]
             *
             * @example
             * <caption>using REST source with complex response mapping</caption>
             * [ {
             *   categories: ['Cities'],
             *   source: 'http://ws.geonames.org/searchJSON',
             *   sourceParam: function(req) { return '?featureClass=P&style=full&maxRows=10&name_startsWith=' + req.term; },
             *   responseMap: {
             *      collection: 'geonames',
             *      value: 'name',
             *      label: function(item) { return item.name + (item.adminName1 ? ", " + item.adminName1 : "") + ", " + item.countryName; },
             *      url: function(item) { return 'https://www.google.com/search?q=' + item.name; },
             *      icon: function(item) { return 'styles/img/confirmation.png?' + item.name; }
             *      category: function(item) { return item.countryName; }
             *   }
             * } ]
             *
             * @example
             * <caption>Example using custom function</caption>
             *  [ {
             *    categories: ['My Category'],
             *    source: function(request, response) {
             *        var arr = [
             *            {value: 'One', label: 'One' },
             *            {value: 'Two', label: 'Two' },
             *            {value: 'Demo', label: 'Demo Custom' }
             *        ];
             *        response($.grep(arr, function (s) { return s.value.indexOf(request.term) === 0; }));
             *    }
             *  } ]
             */
            options:           {
                cssClass:                       'searchForm',
                textInputId:                    'luxsearch_input',
                textInputLabelId:               'luxsearch_input_label',
                textInputTitle:                 'Search Term',
                searchCategoriesId:             'luxsearch_categories',
                inline:                         true,
                includeSearchButton:            true,
                searchBeginsWith:               true,
                searchFieldLabel:               null,
                placeholderText:                null,
                suggestionScopeTitle:           'Select search scope',
                searchAdvancedLink:             'Advanced',
                startSearchButton:              'Start Search',
                suggestAllCategories:           'All',
                searchCaseSensitive:            false,
                suggestionsEnabled:             true,
                suggestionsEnableDirectLink:    true,
                suggestionsLimit:               15,
                suggestionsSeparateByCategory:  true,
                suggestionsSortOnClient:        false,
                suggestionsSources:             [],
                suggestionsTimeoutMilliseconds: 1500,
                multipleSuggestions:            false,
                categoryOptionChange:           null,
                inputChangeCallback:            null,
                searchButtonClick:              null,
                searchButtonId:                 null,
                searchResultsId:                null,
                preferencesLinks:               [],
                searchModel:                    {}
            },
            _allCategories:  [],
            _observablesMap: {},

            /**
             * @override
             * @private
             */
            _create: function () {
                this._setup();

                // FIXME should require a sounding element of some type?
                this._searchForm = $(document).find('.' + this.options.cssClass);
                if (this._searchForm.length === 0) {
                    this._searchForm = $('<div/>', {'class': this.options.cssClass}).appendTo(this.element);
                }
                this._searchForm.addClass('oj-form oj-md-odd-cols-2 oj-md-labels-inline lux-search');
                this._refresh();
            },

            /**
             * @override
             * @private
             */
            _setup: function() {
                // extend model for search criteria
                var SearchModel = ojSearchModel;
                // create the search model to hold search criteria.
                // if the search model is bound to the view model, the options will contain a search model, otherwise
                // the search model will be empty and we create our own search model
                this._searchModel = _.isEmpty(this.options.searchModel) ? new SearchModel() : this.options.searchModel;
            },

            /**
             * @override
             * @private
             */
            _refresh: function() {
                var self = this,
                    searchText = this.options.searchFieldLabel,
                    prefsList,
                    prefsText = 'Preferences',
                    prefLinks = this.options.preferencesLinks,
                    inputContainer = $(document.createElement('div')).addClass('oj-flex-item')
                        .attr('aria-labelledby', self.options.textInputLabelId)
                        .attr('role', 'group'),
                    labelEl = $('<label/>', {'for': self.options.textInputId, text: this.options.searchFieldLabel})
                        .attr('id', self.options.textInputLabelId);

                this._searchForm.empty();

                _.extend(self.options, {
                    // overrides to the autocomplete widget
                    autocompleteFieldLabel: self.options.searchFieldLabel,
                    searchCategoriesId:     self.options.searchCategoriesId,
                    // append to the autocomplete widget
                    resultsModel:           this._searchModel, // pass the model object
                    formContainer:          self._searchForm, // pass form container
                    inputContainer:         inputContainer, // pass the input wrapper
                    labelEl:                labelEl // pass the label element
                });
                // autocomplete controls the input text for the search
                // term so always need to invoke the widget.
                self._searchForm.luxautocompleteinput(self.options);

                if (this.options.includeSearchButton) {
                    this._buttonElement = $('<button/>');
                    if (this.options.searchButtonId) {
                        this._buttonElement.attr('id', this.options.searchButtonId);
                    }
                    if (this.options.searchResultsId) {
                        this._buttonElement.attr('aria-controls', this.options.searchResultsId);
                    }
                    inputContainer.append(this._buttonElement);
                    this._buttonElement.ojButton({
                        display: 'icons',
                        icons:   {start: 'oj-fwk-icon oj-fwk-icon-magnifier'},
                        label:   self.options.startSearchButton,
                        title:   searchText
                    });

                    this._observablesMap.startSearchButton = this._buttonElement;

                    this._buttonElement.on('click',
                        function(event) {
                            self.search(event);
                        }
                    );

                    // respond to keyboard interaction
                    $('#' + self.options.textInputId)[0].addEventListener('keydown', function(event) {
                        // but only when the "enter" or "return" keys are pressed.
                        if (event.keyCode === 13) {
                            self._searchModel.set({'term': $('#' + self.options.textInputId)[0].value});
                            self.search(event);
                        }
                    });
                }

                if (prefLinks.length > 0) {
                    // if not horizontal break so the search fields are stacked.
                    if (this.options.inline) {
                        $('<br/>').appendTo(inputContainer);
                    }

                    if (prefLinks.length > 1) {
                        this._prefsElement = $('<button/>', {title: prefsText, text: prefsText, 'class': 'oj-button-half-chrome'});
                        inputContainer.append(this._prefsElement);

                        prefsList = $('<ul/>').addClass('advanced-search-menu').css('display', 'none');
                        this._observablesMap.preferencesLinks = [];
                        prefLinks.forEach(function(item) {
                            var li = $('<li/>'),
                                prefLink = self._createPrefLink(li, item);

                            this._observablesMap.preferencesLinks.push(prefLink[0].childNodes[0]);
                            prefsList.append(prefLink);
                        }, self);
                        inputContainer.append(prefsList);

                        this._prefsElement.ojButton({
                            label:    self.options.searchAdvancedLink,
                            chroming: 'half',
                            menu:     '.advanced-search-menu'
                        });

                        this._observablesMap.searchAdvancedLink = this._prefsElement;

                        prefsList.ojMenu({});

                        prefsList.on({'ojselect': function(event, ui) {
                            self._searchModel.set({'prefsLink': ui.item.children('a').text()});
                            ui.item.children('a')[0].click();
                        }
                        });
                    } else {
                        inputContainer.append(this._createPrefLink($('<span />'), prefLinks[0]));
                    }
                } else {
                    this._prefsElement = null;
                }

                // respond to "enter" or "return" key press in the current "form" object
                this._searchForm.keypress(function (event) {
                    self._checkForEnterKey(event);
                });
            },

            /**
             * @override
             * @private
             */
            _createPrefLink: function(el, item) {
                var itemText = '',
                    advancedLinks = '';

                // **** legacy code that does not enable keyboard interaction
                if (item.jquery && item.length && item[0].tagName === 'A') {
                    item.addClass('oj-form-control lux-form-link').
                        attr('tabindex', 0);
                } else if (item.href && item.text) {
                    return el.append($('<a />',
                        {href: item.href, text: item.text, title: item.title, tabindex: '0', target: item.target}).
                        addClass('oj-form-control lux-form-link')
                    );
                // ****
                } else if (item.text && item.callback) {
                    // a callback is the only way that keyboard interaction works (pressing either the space bar or enter/return)
                    if (ko.isObservable(item.text)) {
                        itemText = item.text();
                    } else {
                        itemText = item.text;
                    }
                    advancedLinks = $('<a />', {href: '', text: itemText, title: itemText, tabindex: '0', 'data-bind': 'click:' + item.callback + ', clickBubble: false'});
                    if (item.id) {
                        advancedLinks.attr({id: item.id});
                    }
                    return el.append(advancedLinks);

                }
                return el.append(item);
            },

            /**
             * @override
             * @private
             */
            _preProcessCategories: function() {
                var mapFuncs,
                    sources = this.options.suggestionsSources,
                    sourceLength = sources.length,
                    legacyCompatibility = false,
                    categoryFunctionBuilder = function(categoryName) {
                        return function() {
                            return categoryName;
                        };
                    },
                    restSourceMapper = function(val, source, mapFuncs) {
                        if (!mapFuncs[val]) {
                            // default is to use myjsonItem.value, myjsonItem.label, myjsonItem.url,
                            mapFuncs[val] = function(item) {
                                return item[val];
                            };
                        }
                        // but if they provided a map for this property,
                        if (source.responseMap && source.responseMap[val]) {
                            // assume it's a function
                            mapFuncs[val] = source.responseMap[val];
                            // but if it's a string, build a function based on that string key
                            if ($.type(source.responseMap[val]) === 'string') {
                                mapFuncs[val] = function(key) {
                                    return function(item) {
                                        return item[key];
                                    };
                                }(source.responseMap[val]);
                            }
                        }
                    },
                    buildSuggestedCategories = function(sources, categories, parent) {
                        for (var i = 0; i < sources.length; i += 1) {
                            var category,
                                source = sources[i];

                            if (source.categories) {
                                // make the value of each sub category unique by prepending it's
                                // parent hierarchy
                                category = {value: parent + '|' + source.value, label: source.label};
                                source.value = parent + '|' + source.value;

                                categories.push(category);
                                if (source.children) {
                                    category.children = [];
                                    buildSuggestedCategories(source.children, category.children, category.value);
                                }
                            }
                        }
                    },
                    mapSourceItems = function(source) {
                        return _.map(source.source, function(item) {
                            return {value: item, label: item};
                        });
                    },
                    buidlUrlParemeters = function(source, mapFuncs) {
                        var urlParams = ['value', 'label', 'url', 'category', 'description', 'icon'];

                        urlParams.forEach (
                            function(value) {
                                restSourceMapper(value, source, mapFuncs);
                            }
                        );
                    };

                this._allCategories = [];

                for (var i = 0; i < sourceLength; i += 1) {
                    var category = {},
                        source = sources[i];

                    // backward compatibility for ealier lux category input
                    if (source.categories && source.categories.length) {
                        legacyCompatibility = true;
                        category = {value: source.categories[0], label: source.categories[0], categories: true};
                        // push cloned object so we don't get the children in the category suggestions.
                        this._allCategories.push(_.clone(category));
                        // add children to category
                        category.children = mapSourceItems(source);
                        // append the category to suggestions
                        sources.push(category);
                    } else if (source.categories) {
                        // assumes the ojet pattern input for select and combo box.
                        category = {value: source.value, label: source.label};
                        this._allCategories.push(category);
                        if (source.children) {
                            category.children = [];
                            buildSuggestedCategories(source.children, category.children, category.value);
                        }
                    } else if (source.value && source.label) {
                        // otherwise simple list of suggestions using ojet input for select and combo box
                        category = {value: source.value, label: source.label};
                        this._allCategories.push(category);
                    }

                    // REST url
                    if ($.type(source.source) === 'string') {
                        legacyCompatibility = false;
                        // build functions to be executed to get label, value, etc
                        // based on the variations of mapping approaches we accept
                        // (e.g. string name of property or function for returning the right value)
                        mapFuncs = {
                            // default for category is to return the name we already have from API, not look for it in the JSON
                            category: categoryFunctionBuilder(source.categories[0])
                        };
                        buidlUrlParemeters(source, mapFuncs);
                        source.mapFuncs = mapFuncs;
                    }
                }

                if (legacyCompatibility) {
                    // if working on ealier lux source suggestions, need to
                    // transform it to work in this version.
                    sources = _.rest(sources, sourceLength);
                }
            },

            /**
             * @override
             * @private
             */
            _destroy: function() {
                // remove generated elements
                this._searchForm.remove();
            },

            /**
             * @override
             * @private
             */
            _changeLabels: function(key, value) {
                var self = this,
                    option = {};

                if (_.contains([
                    'searchFieldLabel',
                    'placeholderText',
                    'textInputTitle',
                    'suggestionScopeTitle',
                    'suggestAllCategories'], key)) {
                    if (key === 'searchFieldLabel') {
                        key = 'autocompleteFieldLabel';
                    }
                    // set the new option value
                    option[key] = value;
                    // update the option in the autocomplete widget
                    self._searchForm.luxautocompleteinput('option', option);
                    return true;
                }

                if (self._observablesMap[key]) {
                    if (key === 'searchAdvancedLink' && this.options.preferencesLinks.length > 1) {
                        self._observablesMap[key].ojButton('option', {label: value, menu: '.advanced-search-menu'});
                        self.options[key] = value;
                        self._observablesMap[key][0].setAttribute('title', value);
                    } else if (key === 'preferencesLinks') {
                        this._observablesMap.preferencesLinks.forEach(function(link, index) {
                            link.innerText = value[index].text();
                            link.setAttribute('title', value[index].text());
                        });
                    } else if (key === 'startSearchButton') {
                        self._observablesMap[key].ojButton('option', {label: value});
                        self.options[key] = value;
                    } else {
                        self._observablesMap[key].text(value);
                        self.options[key] = value;
                    }
                    return true;
                }

                return false;
            },

            /**
             * _setOption: evaluates the options and calls the
             *             appropriate function to handle change
             *             event.
             * @param {String} key   key of the changed option
             * @param {String} value new value of the option
             */
            _setOption: function(key, value) {
                var self = this;

                self._changeLabels(key, value);
            },

            /**
             * _setOptions: iterates over all changed options and passes
             *              key values to setOption for evaluation.
             *
             * @param {Object} options array of options
             */
            _setOptions: function(options) {
                var self = this;

                $.each(options, function(key, value) {
                    self._setOption(key, value);
                });

            },

            _checkForEnterKey: function(e) {
                var self = this;

                if (e.which === $.ui.keyCode.ENTER) {
                    self.search(e);
                }
            },

            _sortResults: function(results) {
                if (!this.options.suggestionsSortOnClient || !results || !results.length) {
                    return results;
                }
                var that = this;
                return _.sortBy(results, function(d) {
                    var category = (that.options.suggestionsSeparateByCategory && d.category ? d.category : '');
                    return category + (d.label || '') + (d.value || '');
                });
            },

            _limitResults: function(results) {
                if (!results || !results.length) {
                    return results;
                }
                return results.slice(0, this.options.suggestionsLimit);
            },

            _postProcessResults: function(results) {
                return this._limitResults(this._sortResults(results));
            },

            ////////////////////////// INTENDED TO BE PUBLIC //////////////////////////

            /**
             * Sends responses back to the view model
             *
             */
            search: function(e) {
                var self = this;

                // stop other default browser events (e.g. form submit)
                if (e && e.preventDefault) {
                    e.preventDefault();
                }

                if (this.options.searchButtonClick) {
                    this.options.searchButtonClick(e, self._searchModel);
                }
            },

            /**
             * Get the div node of the categories selector
             */
            categoriesElement: function() {
                var self = this;

                return self._searchForm.luxautocompleteinput('categoriesElement');
            },

            /**
             * Get/set the selected search category
             *
             * @this {luxsearch}
             * @param {String} [newValue] category which should be marked as selected
             * @returns {String} the currently selected category
             */
            selectedCategory: function(newValue) {
                if (newValue) {
                    return this._searchModel.set({'searchCategory': newValue});
                } else {
                    return this._searchModel.get('searchCategory') ? this._searchModel.get('searchCategory') : 'ALL';
                }
            },

            /**
             * Get the input control for the search field
             */
            inputElement: function() {
                var self = this;

                return self._searchForm.luxautocompleteinput('inputElement');
            },

            /**
             * Get the search button. May be null if the search button option is turned off.
             */
            buttonElement: function() {
                return this._buttonElement ? this._buttonElement : null;
            },

            getSuggestions: function(term, callBack) {
                var self = this;

                self._searchForm.luxautocompleteinput('getSuggestions', term, callBack);
            },

            /**
             * Get the preferences button. May be null if no preferences links were passed to options
             */
            preferencesElement: function() {
                var self = this;

                return self._prefsElement;
            }
        });

    }());
});
