define([
    'ojs/ojcore',
    'knockout',
    'jquery',
    'lodash',
    'luxUtils',
    'luxhierarchy',
    'luxtree'],
function(oj, ko, $, _, luxUtils, LuxHierarchy){
    'use strict';

    var CONSTANTS = {},
        /**
         * Generalized function to load children for a JSON object array
         * in a luxHierarchy.
         * @param {object} children luxhierarchy node
         * @param {object} parent luxhierarchy node
         * @returns {void}
         */
        createHierarchy = function(children, parent) {

            children.forEach(function(node) {
                var id = node.attr.id,
                    nodeObj = {action: id, actionLabel: node.title, data: {}},
                    newNode;

                if (node.attr) {
                    nodeObj.data = node.attr;
                }

                newNode = parent.addChild().parseJSON(JSON.stringify(nodeObj));

                if (node.children && node.children.length) {
                    createHierarchy(node.children, newNode);
                }

            });
        },

        /**
         * Generates a unique node id (unique in the context of the tree)
         * This is a modified copy of the nodeid generator from luxtree. luxtree
         * has a bug when added a node to an existing tree, the count of
         * nodes does not reflect the node to be added therefore the nodeid geneterated
         * for the new node is a duplicated of the sibling node prior to it.
         * This version of generateNodeId finds the maximum current id and increments
         * that value by 1 to create a id for the new node.
         *
         * @param {object} parent luxhierarchy node
         * @param {string|null} separator
         * @returns {*}
         */
        generateNodeId = function(parent, separator) {
            var increment = 0, maxId = 0;

            // find current max id
            parent.children.forEach(function(child) {
                var id = +child.id.substr(child.id.lastIndexOf('/') + 1);

                if (id > maxId) {
                    maxId = id;
                }

            });

            increment = maxId + 1;

            separator = separator || '/';
            if (parent) {
                return [parent.id, increment].join(separator);
            } else {
                return '0';
            }

        };

    /**
     * Creates a new DataTree object with input in the form
     * of a array of JSON objects.
     * @param {array} data - array of JSON objects
     * @param {object} CONSTS - constants from the shuttle
     * @param {string} prefix - 'selected-' indicates this is the selected list.
     * @constructor
     * @example [{'title': 'Forms', 'attr':{'id': 'forms', 'children': [{'title': 'Child1', 'attr':{'id': 'child1'}}, {}}}]
     * <br />
     */
    function DataTree(data, CONSTS, prefix) {
        var self = this,
            jsonObj = {'action': 'home', 'actionLabel': 'Home'};

        self.data = data;
        self.prefix = prefix;
        // update the constants from the shuttle;
        CONSTANTS = CONSTS;

        // create the hierarchy root node of 'home'
        self.treeHierarchy = new LuxHierarchy().parseJSON(JSON.stringify(jsonObj));

        self.treeHierarchy.findRoot().data = {'prefix': prefix};

        self.loadData();
    }

    /**
     * Builds the lux hierarchy object from the array of JSON objects passed
     * in through the constructor.
     */
    DataTree.prototype.loadData = function() {
        var self = this;

        createHierarchy(self.data, self.treeHierarchy);
    };
    /**
     *  Adds a node from the luxhierarchy source list.
     *
     * @param {string} sourceNodeId - luxhierarchy action id
     * @param {object} sourceTree - source luxhierarchy, contains the sourceNode.
     * @param {string} prefix - either 'selected-' or empty string
     * @param {string} position - relative position, before or after, the targetNode
     * @param {string|null} targetNodeId - id of the target node in this luxhierarchy
     * @returns {string|*}
     */
    DataTree.prototype.addNode = function(sourceNodeId, sourceTree, prefix, position, targetNodeId) {
        var self = this,
            sourceNode = sourceTree.findNodeByAction(sourceNodeId),
            currentNode,
            parents = sourceNode.listParents(),
            location = self.treeHierarchy.findRoot();

        parents.forEach(function(node) {
            var id, parentNode;

            id = node.action !== 'home' ? self._getActionId(node.action) : node.action;
            parentNode = self.treeHierarchy.findNodeByAction(id);
            // Does the parent node already exist?
            if (!parentNode) {
                // no parent node so need to create one
                parentNode = self._createHierarchyNodeFromSource(node, location);
                oj.Logger.info('addNode, add parent: ', parentNode);
                // move the node relative to the target node
                self.moveNode(parentNode.action, targetNodeId, position);
            }
            // save to this parent node so subsequent nodes are added as children
            location = parentNode;
        });

        // Does the node already exist?
        currentNode = self.treeHierarchy.findNodeByAction(self._getActionId(sourceNode.action));
        if (!currentNode) {
            // Nope so need to create node.
            currentNode = self._createHierarchyNodeFromSource(sourceNode, location);
            // move the node relative to the target node
            oj.Logger.info('addNode, add child: ', currentNode);
            self.moveNode(currentNode.action, targetNodeId, position);
        }

        // If there are any children associated with this node
        // move them as children too.
        if (sourceNode.children && sourceNode.children.length) {
            self.addChildren(sourceNode.children, currentNode);
        }

        return currentNode;
    };

    /**
     * Iterates over the child lux hierarchy nodes array from a source
     * lux hierarchy.
     * @param {array} children - array of child lux hierarchy node
     * @param {object} parent - parent lux hierarchy node
     */
    DataTree.prototype.addChildren = function(children, parent) {
        var self = this;

        // iterate over the children in reverse order so each child is
        // added in the order of the source hierarchy.
        children.reverse().forEach(function(child) {
            var currentNode;

            currentNode = self.treeHierarchy.findNodeByAction(self._getActionId(child.action));
            if (!currentNode) {
                currentNode = self._createHierarchyNodeFromSource(child, parent);
                self.moveNode(currentNode.action, null, null);
            }

            if (child.children && child.children.length) {
                self.addChildren(child.children, currentNode);
            }
        });
    };

    /**
     * Removes a lux hierarchy node from the tree.  If there are
     * any children, they are deleted too (Lux Hierarchy docs).
     * @param {string} nodeId - id of node to delete
     * @return {void}
     */
    DataTree.prototype.removeNode = function(nodeId) {
        var self = this,
            node = self.treeHierarchy.findNodeByAction(nodeId),
            parents = node.listParents();

        node.remove();
        parents.forEach(function(parent) {
            if (!parent.children || !parent.children.length) {
                self.removeNode(parent.action);
            }
        });
    };

    /**
     * Creates a lux hierarchy node added to a parent node.
     * @param {object} sourceHierarchyNode - lux hierarchy to clone to a new node under parent node
     * @param {object} parent - lux hierarchy node that will be parent of the new node
     * @return {*}
     * @private
     */
    DataTree.prototype._createHierarchyNodeFromSource = function(sourceHierarchyNode, parent) {
        var self = this,
            jsonObj = {},
            newNode,
            cloneData = function(source) {
                return {'action': self._getActionId(source.action), actionLabel: source.actionLabel, 'data': $.extend({}, source.data)};
            };

        jsonObj = cloneData(sourceHierarchyNode);
        jsonObj.data.id = jsonObj.action;
        jsonObj.data.isselected = (!!self.prefix.length).toString();

        newNode = parent.addChild().parseJSON(JSON.stringify(jsonObj));
        // the default node id generator creates duplicate ids so
        // so override with custom generator.
        newNode.id = generateNodeId(newNode.parent);
        return newNode;
    };

    /**
     * Returns a Lux Hierarchy node based on a node id (action)
     * @param {string} sourceId - node id (action)
     * @return {*}
     * @private
     */
    DataTree.prototype._getActionId = function(sourceId) {
        var prefix = sourceId.indexOf(CONSTANTS.SELECTED_ID_PREFIX),
            id;

        if (prefix > -1) {
            // available list is the target so remove the selected prefix
            id = sourceId.replace(CONSTANTS.SELECTED_ID_PREFIX, CONSTANTS.AVAILABLE_ID_PREFIX);
        } else {
            // selected list is the target so add the selected prefix
            id = CONSTANTS.SELECTED_ID_PREFIX + sourceId;
        }

        return id;
    };

    /**
     * Based on the current node updates the isselected and status partialLabel in the node title
     * @param {string} currentNodeId - id of node to update
     * @param {object} sourceTree - the opposite tree to this tree
     * @param {string} partialLabel - localized value for status partialLabel in the node title (actionLabel).
     * @return {void}
     */
    DataTree.prototype.updateParentStatus = function(currentNodeId, sourceTree, partialLabel) {
        var self = this,
            id = self.prefix.length ? self.prefix + currentNodeId : currentNodeId.replace(CONSTANTS.SELECTED_ID_PREFIX, CONSTANTS.AVAILABLE_ID_PREFIX),
            currentNode = self.treeHierarchy.findNodeByAction(id),
            parents = currentNode.listParents();

        // iterate over patents of the currentNode and update each
        // with the appropriate isselected and title values.
        parents.forEach(function(parent) {
            var oppositeParentId = self.prefix.length ?
                    parent.action.replace(CONSTANTS.SELECTED_ID_PREFIX, CONSTANTS.AVAILABLE_ID_PREFIX) :
                    (CONSTANTS.SELECTED_ID_PREFIX + parent.action),
                oppositeParentNode;

            if (parent.action !== 'home') {
                // does the opposite list have same parent?
                oppositeParentNode = sourceTree.findNodeByAction(oppositeParentId);
                if (oppositeParentNode) {
                    // yes so update isselected and title
                    parent.data.isselected = 'partial';
                    parent.actionLabel = parent.actionLabel.search(partialLabel) > -1 ? parent.actionLabel : (parent.actionLabel + ' ' + partialLabel);
                    oppositeParentNode.data.isselected = 'partial';
                    if (oppositeParentNode.actionLabel.indexOf(partialLabel) < 0) {
                        oppositeParentNode.actionLabel += ' ' + partialLabel;
                    }
                } else {
                    // no so update the title and isselected according to the list.
                    parent.data.isselected = (!!self.prefix.length).toString();
                    parent.actionLabel = parent.actionLabel.replace(partialLabel, '');
                }
                // update the parent DOM title per the parent actionLabel just derived above.
                $('#' + luxUtils.encodeId(parent.action)).find('.oj-tree-title').first().text(parent.actionLabel);
            }
        });

        currentNode.data.isselected = (!!self.prefix.length).toString();
        if (currentNode.children && currentNode.children.length) {
            self.updateChildrenSelectedStatus(currentNode, partialLabel);
        }

        // is the current node in the opposite list?
        if (!sourceTree.findNodeByAction(currentNodeId)) {
            // Nope, update the title and isselected according to the list.
            currentNode.actionLabel = currentNode.actionLabel.replace(partialLabel, '');
            currentNode.data.isselected = (!!self.prefix.length).toString();
        }
    };

    /**
     * Support function that updates the isselected boolean and the status
     * of children of a particular parent.
     * @param {object} currentNode - current parent node
     * @param {object} partialLabel - localized label
     * @return {*}
     */
    DataTree.prototype.updateChildrenSelectedStatus = function(currentNode, partialLabel) {
        var self = this;

        currentNode.children.forEach(function(child) {
            child.data.isselected = (!!self.prefix.length).toString();
            if (child.children && child.children.length) {
                child.actionLabel = child.actionLabel.replace(partialLabel, '');
                self.updateChildrenSelectedStatus(child);
            }
        });
    };

    /**
     * Move a node up or down at the same level in the selected hierarchy
     * @param {string} nodeId - id of node to move
     * @param {string} siblingNodeId - id of sibling node
     * @param {string} position - where to move relative to the sibling, before or after
     * @return {void}
     */
    DataTree.prototype.moveNode = function(nodeId, siblingNodeId, position) {
        var self = this,
            action =  nodeId,
            current = self.treeHierarchy.findNodeByAction(action),
            parent = current.listParents().pop(),
            moveItem,
            currentIndex,
            siblingIndex;

        // get the current index of item to move
        parent.children.some(function(node, index) {
            if (node.action === current.action) {
                currentIndex = index;
                return true;
            }
        });
        // Remove the item from the tree
        moveItem = parent.children.splice(currentIndex, 1)[0];

        // Check the case where we want to move the item always to the top (DnD from one panel to other panel).
        if (!siblingNodeId) {
            siblingIndex = 0;
        } else {
            // get the index of sibling to move to
            parent.children.some(function(node, index) {
                if (node.action === siblingNodeId) {
                    siblingIndex = index;
                    return true;
                }
            });

            // We ONLY need to handle the case where we move the item AFTER a specified
            // node by changing to move the item BEFORE the next node instead, since we are
            // always going the use splice() to insert the item at the specified
            // index. In this case, we just need to adjust the siblingIndex to the
            // next node index.
            if (position === CONSTANTS.MOVE_LOCATION_AFTER) {
                siblingIndex += 1;
            }
        }

        // Insert the move item where it should be.
        parent.children.splice(siblingIndex, 0, moveItem);
    };

    /**
     * Returns a JSON object array representing the LuxHierarchy tree.
     * @returns {Array}
     */
    DataTree.prototype.toJSON = function() {
        var self = this,
            root = self.treeHierarchy.findRoot(),
            arrayContainer = [],
            createJSON = function(children, container) {

                children.forEach(function(child) {
                    var childNode = {title: child.actionLabel, attr: {id: child.action}};

                    if (child.data) {
                        childNode.attr = child.data;
                    }

                    container.push(childNode);

                    if (child.children && child.children.length) {
                        childNode.children = [];
                        createJSON(child.children, childNode.children);
                    }
                });
            };

        createJSON(root.children, arrayContainer);

        return arrayContainer;
    };

    DataTree.prototype.changePartialLabelForLocale = function(id, partialLabel) {
        var self = this,
            node = self.treeHierarchy.findNodeByAction(id);

        if (!node) {
            return;
        }

        node.actionLabel = node.actionLabel.substring(0, node.actionLabel.search(/\(/) - 1) + ' ' + partialLabel;
    };

    DataTree.prototype.findNodeById = function(id) {
        var self = this;

        return self.treeHierarchy.findNodeByAction(id);
    };

    DataTree.prototype.getDataTree = function() {
        var self = this;

        if (arguments.length && arguments[0] !== 'undefined') {
            return self.treeHierarchy.findNodeByAction(arguments[0]);
        }

        return this.treeHierarchy;
    };

    /**
     * Returns ojLogger info display of the luxhierarchy object
     * @param {string} message - optional prefix message.
     * @return {void}
     */
    DataTree.prototype.toConsole = function(message) {
        var self = this,
            msg = message || '';

        oj.Logger.info(msg, 'DataTree:', self.treeHierarchy);
    };

    return DataTree;
});