/**
 * Copyright (c) 2013 - 2015 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * The luxshuttle component provides you with functionality to manage 2 lists / trees of
 * options.  On the left side are the available options and the on the right are the
 * selected options. The options can be move from the available
 * list to the selected list and back again.  Items can be moved done individually or
 * by selectively highlighting items and pressing the move > button.  Options call also
 * be move "on-mass" with >> move all button. Items in the selected options can be removed in
 * as similar fashion using the individual remove < button or the remove all << button.
 * The option exists to allow the end user to organize the selected list items my moving
 * individual or blocks items up or down the list with the "move up", "move down",
 * "move to top", "move bottom"; buttons.
 *
 * By default an item move to the opposite list place at the top of the list.  You can override this
 * behavior my highlighting the a item in the opposite list and the item is them place just below the
 * highlighted item.
 *
 * To 2 lists are actually ojet data tree (ojtree)s.  As such you must follow the ojet api to create the
 * lists as found here: http://jet.us.oracle.com/5.1.0/jsdocs/oj.ojTree.html. With that the
 * shuttle can be internalized with the available options list only, both options list, or the selected options
 * list only.  There is also the ability to create both list from 1 to the "availableValues" option if you
 * indicate the "isselected": "true" attribute on selected items in the list.
 *
 * To save the selected option list you must provide and callback function the takes the selected options
 * list as input.  The selected options list is passed as a JSON string.  Its up to you to then do something
 * with it.
 *
 * @author Joe McClintock
 /**
 * The luxshuttle widget
* @module oj-lux-shuttle
* @jqbase
* @requires ojs/ojcore
* @requires jquery
* @requires knockout
* @requires underscore
 */
define(['ojs/ojcore',
    'knockout',
    'jquery',
    'lodash',
    'luxshuttleDataTree',
    'luxUtils',
    'ojs/ojtree',
    'ojs/ojbutton',
    'ojs/ojinputtext',
    'ojs/ojdialog',
    'ojs/ojjsontreedatasource'],
function(oj, ko, $, _, DataTree, luxUtils){
    'use strict';
    // private constants
    var SHUTTLE = {
        AVAILABLE_LIST_ID:            'availablevalues',
        SELECTED_LIST_ID:             'selectedvalues',
        AVAILABLE_LIST_CSS_CLASS:     'lux-shuttle-available-values',
        SELECTED_LIST_CSS_CLASS:      'lux-shuttle-selected-values',
        SELECTED_ID_PREFIX:           'selected-',
        AVAILABLE_ID_PREFIX:          '',
        MOVE_LOCATION_AFTER:          'after',
        MOVE_LOCATION_BEFORE:         'before',
        PARTIAL_TEXT:                 '(partial)',
        PARTIAL_STATUS:               'partial',
        MOVE_UP:                      'up',
        MOVE_DOWN:                    'down',
        MOVE_TOP:                     'top',
        MOVE_BOTTOM:                  'bottom',
        AVAILABLE_CONTAINER_CLASS:    'lux-shuttle-available-items',
        SELECTED_CONTAINER_CLASS:     'lux-shuttle-selected-items',
        DIALOG_MESSAGE:               'The selected items have been moved but one or more of the times are hidden by the filter on {listName} list.',
        DIALOG_TITLE:                 'Filtered Moved Items',
        NO_ITEMS_MATCH_MSG:           'No items match the current filter',
        FILTER_AVAILABLE_PLACEHOLDER: 'Type to filter',
        FILTER_SELECTED_PLACEHOLDER:  'Type to filter',
        AVAILABLE_DESCRIPTION_ID:     'availablevaluedescription',
        SELECTED_DESCRIPTION_ID:      'selecetedvaluedescription'
    };
    var DND = {
        DROP_EFFECT_NONE:     'none',
        DROP_EFFECT_MOVE:     'move',
        DROP_POSITION_FIRST:  'first',
        DROP_POSITION_AFTER:  'after',
        DROP_POSITION_BEFORE: 'before',
        DROP_POSITION_INSIDE: 'inside'
    };

    (function() {
        $.widget('GBU.luxshuttle', {
            version:           '5.1.0',
            defaultElement:    '<div>',
            widgetEventPrefix: 'lux',
            /**
            * Options
            * <dl>
            *     <dt>selected-values</dt>
            *     <dd>Uses ojTree to create the Tree View of Selected values.
            *        ojTree contents can be specified in JSON format or a
            *        a function the returns a JSON format.<br />
            *        Type: String, Function, or HTML</br>
            *        Default: null</br>
            *        Example:
             *        as widget<br />
             *        <code>selectedValues: selectedList</code><br />
             *        as CCA<br />
             *       <code>selected-values="{{selectedList}}"</code>
            *        <br/>
            *        <p>a function that returns a JSON formatted structure</p>
            *        <pre class="prettyprint" style="overflow-x:visible;">
            *            self.getAvailableValues = function (node, fn) {
            *                var data = [{
            *                        "title": "Pin",
            *                        "attr": { "id": "pin" }
            *                    }, {
            *                        "title": "Slider",
            *                        "attr": { "id": "slider" }
            *                    }, {
            *                        "title": "Buttons",
            *                        "attr": { "id": "buttons" },
            *                        "children": [{
            *                            "title": "Push Button",
            *                            "attr": { "id": "pushButton" }
            *                         }, {
            *                             "title": "Toggle Button",
            *                             "attr": { "id": "toggleButton" }
            *                         }, {
            *                             "title": "Menu Button",
            *                             "attr": { "id": "menuButtonButtons" }
            *                         }, {
            *                             "title": "Iconic Button",
            *                             "attr": { "id": "iconicButton" }
            *                         }, {
            *                             "title": "Button Chroming",
            *                             "attr": { "id": "buttonChroming" }
            *                         }, {
            *                             "title": "Toolbar",
            *                             "attr": { "id": "toolbar" }
            *                         }]
            *                   },
            *                .
            *                .
            *                }
            *            ];
            *            fn(data) ;
            *        };
            *       </pre>
            *       <p>or a JSON formatted structure</p>
            *      <pre class="prettyprint" style="overflow-x:visible;">
            *       self.selected = [
            *           { "title": "Scroll Position",
            *             "attr": {"id": "scrollPosition"}
            *           },
            *           { "title": "Autocomplete",
            *             "attr": {"id": "autocomplete"}
            *           },
            *           { "title": "Forms",
            *             "attr": {"id": "forms"},
            *             "children": [ { "title": "Label",
            *                             "attr": {"id": "label"},
            *                             "children": [ { "title": "Side Label",
            *                                             "attr": {"id": "sideLabel"}
            *                                           }
            *                                         ]
            *                           },
            *                           { "title": "Combobox",
            *                             "attr": {"id": "combobox"}
            *                           },
            *                           { "title": "Select",
            *                             "attr": {"id": "select"}
            *                           },
            *                           { "title": "Checkboxset",
            *                             "attr": {"id": "checkboxset"}
            *                           },
            *            .
            *            .
            *            .
            *            }]
            *       </pre>
            *       </br>
            *        <p>selected-values: variable containing the selected json structure or function</p>
            *     </dd>
            *     <dt>available-values</dt>
            *     <dd>Uses ojTree to create the Tree View of Available values.
            *        ojTree contents can be specified in JSON format or
            *        a function the returns a JSON format.<br />
            *        Type: String or Function</br>
            *        Default: null</br>
            *        Example:
            *        as widget<br />
             *        <code>availableValues: availableList</code><br />
             *        as CCA<br />
             *       <code>available-values="{{availableList}}"</code>
            *        <p>a function that returns a JSON formatted structure</p>
            *        <pre class="prettyprint" style="overflow-x:visible;">
            *        self.getSelectedValues = function (node, fn) {
            *             var data = [{
            *                "title": "Scroll Position",
            *                "attr": {
            *                    "id": "scrollPosition",
            *                    "description": "This is a scroll position control",
            *                    "isselected": "true"
            *                }
            *                }, {
            *                    "title": "Autocomplete",
            *                    "attr": {
            *                        "id": "autocomplete",
            *                        "description": "This is a autocomplete control",
            *                        "isselected": "true"
            *                        }
            *                }, {
            *                    "title": "Pin",
            *                    "attr": { "id": "pin", "description": "This is a pin widget" }
            *                }, {
            *                    "title": "Slider",
            *                    "attr": { "id": "slider", "description": "This is a slider control" }
            *                },
            *                .
            *                .
            *                }
            *            ];
            *            fn(data) ;
            *        };
            *        </pre>
            *        <p>or a JSON formatted structure</p>
            *        <pre class="prettyprint" style="overflow-x:visible;">
            *        available = [
            *           { "title": "Pin",
            *             "attr": {"id": "pin"}
            *           },
            *           { "title": "Slider",
            *             "attr": {"id": "slider"}
            *           },
            *           { "title": "Buttons",
            *             "attr": {"id": "buttons"},
            *                      "children": [ { "title": "Push Button",
            *                             "attr": {"id": "pushButton"}
            *                           },
            *                           { "title": "Toggle Button",
            *                             "attr": {"id": "toggleButton"}
            *                           },
            *                           { "title": "Menu Button",
            *                             "attr": {"id": "menuButtonButtons"}
            *                           },
            *                           { "title": "Iconic Button",
            *                             "attr": {"id": "iconicButton"}
            *                           },
            *                           { "title": "Button Chroming",
            *                             "attr": {"id": "buttonChroming"}
            *                           },
            *                           { "title": "Toolbar",
            *                             "attr": {"id": "toolbar"}
            *                           }
            *                         ]
            *            },
            *            .
            *            .
            *            .
            *            }];
            *        </pre>
            *        <p>available-values: variable containing the available json structure or function </p>
            *     </dd>
            *     <dt>available-values-id</dt>
             *     <dd>
             *         The DOM element ID of the Available list.<br />
             *         Type: String <br />
             *         Default: 'availablevalues'<br />
             *         Example: <br />
             *         as widget<br />
             *         <code>availableValuesId: 'availableValues'</code><br />
             *         as CCA<br />
             *         <code>available-values-id="[['availableValues']]"</code>
             *     </dd>
             *
             *     <dt>selected-list-id</dt>
             *     <dd>
             *         The DOM element ID of the Selected list.<br />
             *         Type: String <br />
             *         Default: 'selectedvalues'<br />
             *         Example: <br />
             *         as widget<br />
             *         <code>selectedListId: 'selectedvalues'</code><br />
             *         as CCA<br />
             *         <code>selected-list-id="[['selectedvalues']]"</code>
             *     </dd>
             *
             *     <dt>available-values-heading</dt>
            *     <dd>
            *        Text to appear above the available values list</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *       as widget<br />
            *        <code>availableValuesHeading: 'Available Values List'</code><br />
             *       as CCA<br />
             *       <code>available-values-heading="[['Available Values List']]"</code>
            *     </dd>
            *     <dt>available-value-description</dt>
            *     <dd>
            *        Initial text to appear below the available values list.
            *        The value of the highlighted available option is displayed here. If boolean false
            *        is supplied as the value, no description will appear.</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>availableValueDescription: 'A long description of the list'</code><br />
             *         as CCA<br />
             *       <code>available-value-description="[['A long description of the list']]"</code>
            *     </dd>
            *     <dt>selected-values-heading</dt>
            *     <dd>
            *        Initial text to appear above the selected values list.
            *        The value of the highlighted selected option is displayed here.</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>selectedValuesHeading: 'Selected Values List'</code><br />
             *        as CCA<br />
             *       <code>selected-values-heading="[['Selected Values List']]"</code>
            *     </dd>
            *     <dt>selected-value-description</dt>
            *     <dd>
            *        Text to appear below the selected values list.
            *        The value of the highlighted selected option is displayed here. If boolean false
            *        is supplied as the value, no description will appear.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>selectedValueDescription: 'A long description of the list'</code><br />
             *        as CCA<br />
             *       <code>selected-value-description="[['A long description of the list']]"</code>
            *     </dd>
            *     <dt>available-description-heading</dt>
            *     <dd>
            *        Title that appears under the available list box</br>
            *        Type: String</br>
            *        Default: null, textarea description does not appear</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>availableDescriptionHeading: 'Description'</code><br />
             *        as CCA<br />
             *        <code>available-description-heading="[['Description']]"</code>
            *     </dd>
            *     <dt>selected-description-heading</dt>
            *     <dd>
            *        Title that appears under the selected list box</br>
            *        Type: String</br>
            *        Default:  null, textarea description does not appear</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>selectedDescriptionHeading: 'Description'</code><br />
             *        as CCA<br />
             *        <code>selected-description-heading="[['Description']]"</code>
            *     </dd>
            *     <dt>description-area-rows</dt>
            *     <dd>
            *        Number of row in the textarea description</br>
            *        Type: number</br>
            *        Default: 3</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>descriptionAreaRows: 3</code><br />
             *        as CCA<br />
             *        <code>description-area-rows="[[3]]"</code>
            *     </dd>
            *     <dt>move-button-label</dt>
            *     <dd>
            *        "title" attribute of move button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveButtonLabel: 'Move'</code><br />
             *        as CCA<br />
             *        <code>move-button-label="[['Move']]"</code>
            *     </dd>
            *     <dt>move-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;move&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of move_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id move_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveAccessibility: 'Move the selected items in the Available Values list to the Selected Values list'</code><br />
             *        as CCA<br />
             *        <code>move-accessibility="[['Move the selected items in the Available Values list to the Selected Values list']]"</code>
            *     </dd>
            *     <dt>move-all-button-label</dt>
            *     <dd>
            *        "title" attribute of move all button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveAllButtonLabel: 'Move All'</code><br />
             *        as CCA<br />
             *        <code>move-all-button-label="[['Move All']]"</code>
            *     </dd>
            *     <dt>move-all-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;move all&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of move_all_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id move_all_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveAllAccessibility: 'Move all items to the Selected Values list'</code><br />
             *        as CCA<br />
             *        <code>move-all-accessibility="[['Move all items to the Selected Values list']]"</code>
            *     </dd>
            *     <dt>remove-button-label</dt>
            *     <dd>
            *        "title" attribute of remove button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>removeButtonLabel: 'Remove'</code><br />
             *        as CCA<br />
             *        <code>remove-button-label="[['Remove']]"</code>
            *     </dd>
            *     <dt>remove-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;remove&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of remove_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id remove_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>removeAccessibility: 'Move the selected items in the Selected Values list to the Available Values list'</code><br />
             *        as CCA<br />
             *        <code>remove-accessibility="[['Move the selected items in the Selected Values list to the Available Values list']]"</code>
            *     </dd>
            *     <dt>remove-all-button-label</dt>
            *     <dd>
            *        "title" attribute of move all button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>removeAllButtonLabel: 'Remove All'</code>
             *        as CCA<br />
             *        <code>remove-all-button-label="[['Remove All']]"</code>
            *     </dd>
            *     <dt>remove-all-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;remove all&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of remove_all_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id remove_all_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>removeAllAccessibility: 'Move all items to the Available Values list'</code><br />
             *        as CCA<br />
             *        <code>remove-all-accessibility="[['Move all items to the Available Values list']]"</code>
            *     </dd>
            *     <dt>move-to-top-button-label</dt>
            *     <dd>
            *        "title" attribute of move top button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveToTopButtonLabel: 'Move to top'</code><br />
             *        as CCA<br />
             *        <code>move-to-top-button-label="[['Move to top']]"</code>
            *     </dd>
            *     <dt>move-to-top-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;move to top&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of move_to_top_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id move_to_top_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveToTopAccessibility: 'Move each selected item in the Selected Values list to the top of that list, or to be the first child of its parent item'</code>
            *         <br />as CCA<br />
             *        <code>move-to-top-accessibility="[['Move each selected item in the Selected Values list to the top of that list, or to be the first child of its parent item']]"</code>
            *     </dd>
            *     <dt>move-up-button-label</dt>
            *     <dd>
            *        "title" attribute of move up button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveUpButtonLabel: 'Move up'</code>
             *        <br />as CCA<br />
             *        <code>move-up-button-label="[['Move up'</code>
            *     </dd>
            *     <dt>move-up-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;move up&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of move_up_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id move_up_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveUpAccessibility: 'Move each selected item in the Selected Values list one position earlier in the list, or earlier in the order within its parent item'</code>
            *         <br />as CCA<br />
             *        <code>move-up-accessibility="[['Move each selected item in the Selected Values list one position earlier in the list, or earlier in the order within its parent item']]"</code>
            *     </dd>
            *     <dt>move-to-bottom-button-label</dt>
            *     <dd>
            *        "title" attribute of move bottom button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveToBottomButtonLabel: 'Move to bottom'</code>
             *        <br />as CCA<br />
             *        <code>move-to-bottom-button-label="[['Move to bottom']]"</code>
            *     </dd>
            *     <dt>move-to-bottom-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;move to bottom&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of move_to_bottom_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id move_to_bottom_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *         as widget<br />
            *        <code>moveToBottomAccessibility: 'Move each selected item in the Selected Values list to the bottom of that list, or to be the last child of its parent item'</code>
             *        <br />as CCA<br />
             *        <code>move-to-bottom-accessibility="[['Move each selected item in the Selected Values list to the bottom of that list, or to be the last child of its parent item']]"</code>
            *     </dd>
            *     <dt>move-down-button-label</dt>
            *     <dd>
            *        "title" attribute of move down button</br>
            *        Type: String</br>
            *        Default: Empty string</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveDownButtonLabel: 'Move down'</code>
             *         <br />as CCA<br />
             *        <code>move-down-button-label="[['Move down']]"</code>
            *     </dd>
            *     <dt>move-down-accessibility</dt>
            *     <dd>
            *        Creates alternate &quot;move down&quot; button information for sight impaired. With this option a DIV element is created with the class oj-helper-hidden-accessible so the DIV element is hidden and a id of move_down_accessibility.  On the corresponding move button element the attribute aria-describedby is created referencing the DIV with id move_down_accessibility.</br>
            *        Type: String</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>moveDownAccessibility: 'Move each selected item in the Selected Values list one position later in the list, or later in the order within its parent item'</code>
             *       <br /> as CCA<br />
             *        <code>move-down-accessibility="[['Move each selected item in the Selected Values list one position later in the list, or later in the order within its parent item']]"</code>
            *     </dd>
            *     <dt>allow-reorder</dt>
            *     <dd>
            *        specifies whether or not the end user can reorder the selected values</br>
            *        Type: boolean</br>
            *        Default: true</br>
            *        Example:</br>
             *          as widget<br />
            *        <code>allowReorder: false</code>
             *        <br />as CCA<br />
             *       <code>allow-reorder="[[false]]"</code>
            *     </dd>
            *     <dt>allow-dn-d</dt>
            *     <dd>
            *        specifies whether or not to enable the DnD (Drag and Drop) feature</br>
            *        Type: boolean</br>
            *        Default: true</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>: allowDnD: false</code>
             *         <br />as CCA<br />
             *      <code>: allow-dn-d="[[false]]"</code>
            *     </dd>
            *     <dt>on-save-callback</dt>
            *     <dd>
            *        specifies the callback to execute with the selected values when
            *        ever selected values change.  REQUIRED to do anything interesting.</br>
            *        Type: function</br>
            *        Default: null</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>onSaveCallback: function(data) { this.selectedValues(data); }</code>
             *         <br />as CCA<br />
             *        <code>on-save-callback=[[function(data) { this.selectedValues(data); }]]</code>
            *     </dd>
            *     <dt>partial-label</dt>
            *     <dd>
            *        specifies the locale specific label next to the available and selected parent items
            *        that are partially selected</br>
            *        Type: String</br>
            *        Default: '(partial)'</br>
            *        Example:</br>
             *        as widget<br />
            *        <code>partialLabel: '(partial)'</code>
             *        <br />as CCA<br />
             *       <code>partial-label="[['(partial)']]"</code>
            *     </dd>
             *     <dt>filter-lists</dt>
             *     <dd>
             *       This option adds filtering to capability both the available and selected lists.  Input fields
             *       are displayed above each list. A character or string of characters entered in a input field
             *       does a case insensitive filter on the corresponding list.<br />
             *       Type: boolean<br />
             *       Default: false<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>filterLists: true</code>
             *        <br />as CCA<br />
             *        <code>filter-lists=[[true]]</code>
             *     </dd>
             *     <dt>filter-available-id</dt>
             *     <dd>
             *       With the filterLists option, sets the placeholder id available filter text input element.<br />
             *       Type: String<br />
             *       Default: 'availableFilterId'<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>filterAvailableId: 'AvailableFilterId'</code>
             *       <br />as CCA<br />
             *       <code>filter-available-id="[['AvailableFilterId']]"</code>
             *     </dd>
             *     <dt>filter-selected-id</dt>
             *     <dd>
             *       With the filterLists option, sets the placeholder id selected filter text input element.<br />
             *       Type: String<br />
             *       Default: 'selectedFilterId'<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>filterSelectedId: 'SelectedFilterId'</code>
             *       <br />as CCA<br />
             *       <code>filter-selected-id="[['SelectedFilterId']]"</code>
             *     </dd>
             *     <dt>filter-available-placeholder</dt>
             *     <dd>
             *       With the filterLists option, sets the placeholder value of the available list input field.<br />
             *       Type: String<br />
             *       Default: 'Type to filter'<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>filterAvailablePlaceholder: 'Type to filter'</code>
             *       <br />as CCA<br />
             *       <code>filter-available-placeholder="[['Type to filter']]"</code>
             *     </dd>
             *     <dt>filter-selected-placeholder</dt>
             *     <dd>
             *       With the filterLists option, sets the placeholder value of the selected list input field.<br />
             *       Type: String<br />
             *       Default: 'Type to filter'<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>filterSelectedPlaceholder: 'Type to filter'</code>
             *       <br />as CCA<br />
             *       <code>filter-selected-placeholder="[['Type to filter']]"</code>
             *     </dd>
             *     <dt>filter-warning-dialog-timer</dt>
             *     <dd>
             *       With the filterLists option, sets the time the popup dialog warning displays when a item is moved to a
             *       filtered list which filters out that item.<br />
             *       Type: number<br />
             *       Default: null<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>filterWarningDialogTimer: 5000</code>
             *       <br />as CCA<br />
             *       <code>filter-warning-dialog-timer: "[[5000]]"</code>
             *     </dd>
             *     <dt>expand-parent-on-move</dt>
             *     <dd>
             *       With this option the parent list item is expanded in the target list when a child item is moved.<br />
             *       Type: boolean<br />
             *       Default: false<br />
             *       Example:<br />
             *       as widget<br />
             *       <code>expandParentOnMove: true</code>
             *       <br />as CCA<br />
             *       <code>expand-parent-on-move="[[true]]"</code>
             *     </dd>
            * </dl>
             * <div>
             *     <h2>Filtered List Messages and Locales</h2>
             *     <p>There are 2 resource buddle items to include for localization that should not be part the the
             *     view model mainTranslatableTextIDs data structure: dialogMessage and noItemsMatchMsg.  Add these two items
             *     in the resources/nls/menu.js and resources/nls/[locale]/menu.js files</p>
             *     Example: <br />
             *     as widget<br />
             *     'dialogMessage': 'The selected items have been moved but one or more of the times are hidden by the filter on {listName} list.',<br/>
             *     'noItemsMatchMsg': 'No items match the current filter.'
             *      <br />as CCA<br />
             *     'dialog-message'="[['The selected items have been moved but one or more of the times are hidden by the filter on {listName} list.']]"<br/>
             *     'no-items-match-msg'="[['No items match the current filter.']]"
             *     <p>NOTE: Include the '{listName}' to display the list name in the dialog message</p>
             * </div>
            */

            options: {
                availableListId:             SHUTTLE.AVAILABLE_LIST_ID,
                selectedListId:              SHUTTLE.SELECTED_LIST_ID,
                selectedValues:              null,
                availableValues:             null,
                availableValuesHeading:      '',
                availableValueDescription:   true,
                selectedValueDescription:    true,
                availableValueDescriptionId: SHUTTLE.AVAILABLE_DESCRIPTION_ID,
                selectedValueDescriptionId:  SHUTTLE.SELECTED_DESCRIPTION_ID,
                selectedValuesHeading:       '',
                availableDescriptionHeading: null,
                selectedDescriptionHeading:  null,
                descriptionAreaRows:         3,
                moveButtonLabel:             '',
                filterLists:                 false,
                filterAvailableId:           'availableFilterId',
                filterSelectedId:            'selectedFilterId',
                filterAvailablePlaceholder:  SHUTTLE.FILTER_AVAILABLE_PLACEHOLDER,
                filterSelectedPlaceholder:   SHUTTLE.FILTER_SELECTED_PLACEHOLDER,
                filterWarningDialogTimer:    null,
                moveAllButtonLabel:          '',
                removeButtonLabel:           '',
                removeAllButtonLabel:        '',
                moveToTopButtonLabel:        '',
                moveUpButtonLabel:           '',
                moveDownButtonLabel:         '',
                moveToBottomButtonLabel:     '',
                moveAccessibility:           null,
                moveAllAccessibility:        null,
                removeAccessibility:         null,
                removeAllAccessibility:      null,
                moveToTopAccessibility:      null,
                moveUpAccessibility:         null,
                moveDownAccessibility:       null,
                moveToBottomAccessibility:   null,
                itemCountLabel:              '',
                allowReorder:                true,
                allowDnD:                    true,
                onSaveCallback:              null,
                partialLabel:                SHUTTLE.PARTIAL_TEXT,
                ariaLabelledby:              null,
                showcounts:                  true,
                expandParentOnMove:          false
            },

            _observablesMap:    {},
            _ignoredAttributes: ['role', 'class', 'aria-selected'],
            _buttonRefreshList: [
                {'label': 'moveButtonLabel', 'button': 'moveButton'},
                {'label': 'moveAllButtonLabel', 'button': 'moveAllButton'},
                {'label': 'removeButtonLabel', 'button': 'removeButton'},
                {'label': 'removeAllButtonLabel', 'button': 'removeAllButton'},
                {'label': 'moveToTopButtonLabel', 'button': 'moveToTopButton'},
                {'label': 'moveUpButtonLabel', 'button': 'moveUpButton'},
                {'label': 'moveDownButtonLabel', 'button': 'moveDownButton'},
                {'label': 'moveToBottomButtonLabel', 'button': 'moveToBottomButton'}
            ],
            _accessibilityText: [
                {'name': 'moveAccessibility', 'id': 'move_accessibility'},
                {'name': 'moveAllAccessibility', 'id': 'move_all_accessibility'},
                {'name': 'removeAccessibility', 'id': 'remove_accessibility'},
                {'name': 'removeAllAccessibility', 'id': 'remove_all_accessibility'},
                {'name': 'moveToTopAccessibility', 'id': 'move_to_top_accessibility'},
                {'name': 'moveDownAccessibility', 'id': 'move_down_accessibility'},
                {'name': 'moveUpAccessibility', 'id': 'move_up_accessibility'},
                {'name': 'moveToBottomAccessibility', 'id': 'move_to_bottom_accessibility'}
            ],
            _accessibilityDivs: {},
            _selectedIds:       [],
            _availableIds:      [],
            _hasAvailableDescr: false,
            _hasSelectedDescr:  false,

            _filterPlaceHolderMap: {
                filterAvailablePlaceholder: 'availableFilter',
                filterSelectedPlaceholder:  'selectedFilter'
            },

            _create: function() {
                var self = this;

                self._dndContext = {};
                self._setup();
                self._refresh();
            },

            _setup: function() {
                var self = this;

                self._observablesMap.partialLabel = self.options.partialLabel;
            },

            _refresh: function() {
                var self = this;

                self._mainContainer = $('<div />', {'class': 'lux-shuttle' + self._styleContainerForMobile() + 'oj-flex', 'role': 'group'}).appendTo(self.element);

                self._accessiblityContainer = $('<div />', {'style': 'display: none;'});
                self._mainContainer.append(self._accessiblityContainer);

                self._createAccessibilityDivs(self._accessiblityContainer);

                if (self.options.ariaLabelledby) {
                    self._mainContainer.attr('aria-labelledby', self.options.ariaLabelledby);
                }

                self._leftContainer = $('<div />',
                    {'class': SHUTTLE.AVAILABLE_CONTAINER_CLASS + ' oj-flex-item'}).appendTo(self._mainContainer);

                self._observablesMap.availableValuesHeading = $('<label />',
                    {'id':    self.options.availableListId + 'label',
                        'class': 'lux-shuttle-header oj-label',
                        'text':  self.options.availableValuesHeading}).appendTo(self._leftContainer);

                if (self.options.filterLists) {
                    self._availableFilterText = ko.observable();
                    self._observablesMap.availableFilter = $('<input />', {id: self.options.filterAvailableId}).appendTo(self._leftContainer);
                    ko.applyBindingsToNode(self._observablesMap.availableFilter, {event: {keyup: self._handleAvailableFilterKeyUp}}, self);
                    self._observablesMap.availableFilter.ojInputText({'placeholder': self.options.filterAvailablePlaceholder});
                }

                self._availableOptionsContainer = $('<div />', {
                    class:      SHUTTLE.AVAILABLE_LIST_CSS_CLASS,
                    'id':       self.options.availableListId,
                    'tabindex': '-1'
                }).appendTo(self._leftContainer);
                self._leftTree = self._availableOptionsContainer.ojTree({
                    data:          {data: self.options.availableValues || self._emptyAvailableOptions},
                    selectionMode: 'multiple',
                    icons:         false,
                    optionChange:  _.bind(self._selectionChange, self)
                });

                self._observablesMap.itemCountLabels = {
                    'available': {'el': null},
                    'selected':  {'el': null}
                };

                if (self.options.showcounts) {
                    self._leftTree.find('ul').attr('aria-labelledby', self.options.availableListId + 'label');
                    self._leftTree.find('ul').attr('aria-describedby', self.options.availableListId + '-counts');
                    self.availableCountLabel = $('<span />', {text: self.options.itemCountLabel, 'class': 'lux-shuttle-count-label'});
                    self._observablesMap.itemCountLabels.available.el = self.availableCountLabel;
                    self.availableCount = $('<span />', {text: 0});
                    $('<div />', {'id': self.options.availableListId + '-counts', 'class': 'lux-shuttle-counts'}).append(self.availableCount).append(self.availableCountLabel).appendTo(self._leftContainer);
                }

                // if available description heading is specified, create the
                // heading and the textarea.
                if (self.options.availableDescriptionHeading) {
                    self._observablesMap.availableDescriptionHeading = $('<label />',
                        {'for':  self.options.availableValueDescriptionId,
                            'text': self.options.availableDescriptionHeading}).appendTo(self._leftContainer);
                }
                if (self.options.availableValueDescription) {
                    self._leftAvailableValue = $('<textarea />', {
                        'id':        self.options.availableValueDescriptionId,
                        'rows':      self.options.descriptionAreaRows,
                        'aria-live': 'polite'
                    }).appendTo(self._leftContainer);

                    self._leftAvailableValue.ojTextArea({
                        value:    '',
                        readOnly: true
                    });
                    self._hasAvailableDescr = true;
                }

                self._buttonsContainer = $('<div />', {'class': 'lux-shuttle-actions oj-flex-item'}).appendTo(self._mainContainer);
                self._moveButtonContainer = $('<div />', {'class': 'lux-shuttle-move'}).appendTo(self._buttonsContainer);

                // move available to selected
                self._moveButton = $('<button />').appendTo(self._moveButtonContainer);
                if (self.options.moveAccessibility) {
                    self._moveButton.attr('aria-describedby', self._accessibilityDivs.moveAccessibility);
                }
                self._moveButton.ojButton({
                    display:  'icons',
                    icons:    {start: 'lux-move-icon oj-icon'},
                    label:    self.options.moveButtonLabel,
                    disabled: true
                });
                this._moveButton.on('click', _.bind(self._moveButtonClick, self));
                self._observablesMap.moveButtonLabel = self._moveButton.find('.oj-button-text');

                // move all from available to selected
                self._moveAllButton = $('<button />').appendTo(self._moveButtonContainer);
                if (self.options.moveAllAccessibility) {
                    self._moveAllButton.attr('aria-describedby', self._accessibilityDivs.moveAllAccessibility);
                }
                self._moveAllButton.ojButton({
                    display: 'icons',
                    icons:   {start: 'lux-move-all-icon oj-icon'},
                    label:   self.options.moveAllButtonLabel
                });
                this._moveAllButton.on('click', _.bind(self._moveAllButtonClick, self));
                self._observablesMap.moveAllButtonLabel = self._moveAllButton.find('.oj-button-text');

                // move selected to available
                self._removeButtonContainer = $('<div /> ', {'class': 'lux-shuttle-remove'}).appendTo(self._buttonsContainer);
                self._removeButton = $('<button />').appendTo(self._removeButtonContainer);
                if (self.options.removeAccessibility) {
                    self._removeButton.attr('aria-describedby', self._accessibilityDivs.removeAccessibility);
                }
                self._removeButton.ojButton({
                    display:  'icons',
                    icons:    {start: 'lux-remove-icon oj-icon'},
                    label:    self.options.removeButtonLabel,
                    disabled: true
                });
                this._removeButton.on('click', _.bind(self._removeButtonClick, self));
                self._observablesMap.removeButtonLabel = self._removeButton.find('.oj-button-text');

                // move all selected to available
                self._removeAllButton = $('<button />').appendTo(self._removeButtonContainer);
                if (self.options.removeAllAccessibility) {
                    self._removeAllButton.attr('aria-describedby', self._accessibilityDivs.removeAllAccessibility);
                }
                self._removeAllButton.ojButton({
                    display: 'icons',
                    icons:   {start: 'lux-remove-all-icon oj-icon'},
                    label:   self.options.removeAllButtonLabel
                });
                this._removeAllButton.on('click', _.bind(self._removeAllButtonClick, self));
                self._observablesMap.removeAllButtonLabel = self._removeAllButton.find('.oj-button-text');

                self._selectedContainer = $('<div />',
                    {'class': 'lux-shuttle-selected oj-flex-item oj-flex'}).appendTo(self._mainContainer);

                self._rightContainer = $('<div />',
                    {'class': SHUTTLE.SELECTED_CONTAINER_CLASS + ' oj-flex-item'}).appendTo(self._selectedContainer);

                self._observablesMap.selectedValuesHeading = $('<label />',
                    {'id':    self.options.selectedListId + 'label',
                        'class': 'lux-shuttle-header oj-label',
                        'text':  self.options.selectedValuesHeading}).appendTo(self._rightContainer);

                if (self.options.filterLists) {
                    self._selectedListFilterText = ko.observable();
                    self._observablesMap.selectedFilter = $('<input />', {id: self.options.filterSelectedId}).appendTo(self._rightContainer);
                    ko.applyBindingsToNode(self._observablesMap.selectedFilter, {event: {keyup: self._handleSelectedFilterKeyUp}}, self);
                    self._observablesMap.selectedFilter.ojInputText({'placeholder': self.options.filterSelectedPlaceholder});
                }

                self._selectedOptionsContainer = $('<div />', {
                    class:      SHUTTLE.SELECTED_LIST_CSS_CLASS,
                    'id':       self.options.selectedListId,
                    'tabindex': '-1'
                }).appendTo(self._rightContainer);
                self._rightTree = self._selectedOptionsContainer.ojTree({
                    data:          {data: self.options.selectedValues || self._getSelectedValues()},
                    selectionMode: 'multiple',
                    icons:         false,
                    optionChange:  _.bind(self._selectionChange, self)
                });

                if (self.options.showcounts) {
                    self._rightTree.find('ul').attr('aria-labelledby', self.options.selectedListId + 'label');
                    self._rightTree.find('ul').attr('aria-describedby', self.options.selectedListId + '-counts');
                    self.selectedCountLabel = $('<span />', {text: self.options.itemCountLabel, 'class': 'lux-shuttle-count-label'});
                    self._observablesMap.itemCountLabels.selected.el = self.selectedCountLabel;
                    self.selectedCount = $('<span />', {text: 0});
                    $('<div />', {'id': self.options.selectedListId + '-counts', 'class': 'lux-shuttle-counts'}).append(self.selectedCount).append(self.selectedCountLabel).appendTo(self._rightContainer);
                }

                // if selected description heading is specified, create the
                // heading and the textarea.
                if (self.options.selectedDescriptionHeading) {
                    self._observablesMap.selectedDescriptionHeading = $('<label />',
                        {'for':  self.options.selectedValueDescriptionId,
                            'text': self.options.selectedDescriptionHeading}).appendTo(self._rightContainer);
                }
                if (self.options.selectedValueDescription) {
                    self._rightSelectedValue = $('<textarea />',
                        {'id':        self.options.selectedValueDescriptionId,
                            'rows':      self.options.descriptionAreaRows,
                            'aria-live': 'polite'}).appendTo(self._rightContainer);

                    self._rightSelectedValue.ojTextArea({
                        value:    '',
                        readOnly: true
                    });
                    self._hasSelectedDescr = true;
                }

                if (self.options.allowReorder) {
                    self._reorderButtonsContainer = $('<div />', {'class': 'lux-shuttle-reorder oj-flex-item'}).appendTo(self._selectedContainer);
                    self._moveUpButtonContainer = $('<div />', {'class': 'lux-shuttle-up'}).appendTo(self._reorderButtonsContainer);

                    // move to top button
                    self._moveToTopButton = $('<button />').appendTo(self._moveUpButtonContainer);
                    if (self.options.moveToTopAccessibility) {
                        self._moveToTopButton.attr('aria-describedby', self._accessibilityDivs.moveToTopAccessibility);
                    }
                    self._moveToTopButton.ojButton({
                        display: 'icons',
                        icons:   {start: 'lux-to-top-icon oj-icon'},
                        label:   self.options.moveToTopButtonLabel
                    });
                    this._moveToTopButton.on('click', _.bind(self._moveToTopButtonClick, self));
                    self._observablesMap.moveToTopButtonLabel = self._moveToTopButton.find('.oj-button-text');

                    // move up one button
                    self._moveUpButton = $('<button />').appendTo(self._moveUpButtonContainer);
                    self._moveUpButton.ojButton({
                        display: 'icons',
                        icons:   {start: 'lux-up-icon oj-icon'},
                        label:   self.options.moveUpButtonLabel
                    });
                    this._moveUpButton.on('click', _.bind(self._moveUpButtonClick, self));
                    if (self.options.moveUpAccessibility) {
                        self._moveUpButton.attr('aria-describedby', self._accessibilityDivs.moveUpAccessibility);
                    }
                    self._observablesMap.moveUpButtonLabel = self._moveUpButton.find('.oj-button-text');

                    // move down one button
                    self._moveDownButtonContainer = $('<div />', {'class': 'lux-shuttle-down'}).appendTo(self._reorderButtonsContainer);
                    self._moveDownButton = $('<button />').appendTo(self._moveDownButtonContainer);
                    if (self.options.moveDownAccessibility) {
                        self._moveDownButton.attr('aria-describedby', self._accessibilityDivs.moveDownAccessibility);
                    }
                    self._moveDownButton.ojButton({
                        display: 'icons',
                        icons:   {start: 'lux-down-icon oj-icon'},
                        label:   self.options.moveDownButtonLabel
                    });
                    this._moveDownButton.on('click', _.bind(self._moveDownButtonClick, self));
                    self._observablesMap.moveDownButtonLabel = self._moveDownButton.find('.oj-button-text');

                    // move to bottom button
                    self._moveToBottomButton = $('<button />').appendTo(self._moveDownButtonContainer);
                    if (self.options.moveToBottomAccessibility) {
                        self._moveToBottomButton.attr('aria-describedby', self._accessibilityDivs.moveToBottomAccessibility);
                    }
                    self._moveToBottomButton.ojButton({
                        display: 'icons',
                        icons:   {start: 'lux-to-bottom-icon oj-icon'},
                        label:   self.options.moveToBottomButtonLabel
                    });
                    this._moveToBottomButton.on('click', _.bind(self._moveToBottomButtonClick, self));
                    self._observablesMap.moveToBottomButtonLabel = self._moveToBottomButton.find('.oj-button-text');
                }

                // Check if we need to enable DnD feature
                if (self.options.allowDnD) {
                    self._dndOption();
                }

                if (self.options.selectedValues) {
                    self._setSelectedValuesAttributes();
                }
                self._saveSelections();

                self._setItemCountDisplay();

                if (self.options.filterLists) {
                    self._createFilterStructures();
                }
            },

            _styleContainerForMobile: function() {
                return ' ' + luxUtils.getTouchEnabledStyleClassName() + ' ';
            },

            _createAccessibilityDivs: function(container) {
                var self = this;

                self._accessibilityText.forEach(function(item) {
                    var div;

                    if (self.options[item.name] && !_.isEmpty(self.options[item.name])) {
                        div = self['_' + item.name] = $('<div >', {
                            id:   item.id,
                            text: self.options[item.name]
                        });
                    }

                    container.append(div);
                    self._accessibilityDivs[item.name] = item.id;
                }, self);

            },

            _createFilterStructures: function() {
                var self = this;

                self._mainContainer.addClass('lux-shuttle-has-filters');

                self._filterMoveWarningDialog = $('<div />', {'style': 'display:none;', 'id': 'filterWarningDialog', 'title': SHUTTLE.DIALOG_TITLE});
                self._filterMoveWarningTextDiv = $('<div />', {'class': 'oj-dialog-body', 'id': 'filterWarningText'});
                self._filterMoveWarningDialog.append(self._filterMoveWarningTextDiv);
                self._mainContainer.append(self._filterMoveWarningDialog);
                self._filterMoveWarningDialog.ojDialog({
                    'initialVisibility': 'hide',
                    'modality':          'modeless'
                });

                self._initializeFilteredLists();
            },

            _initializeFilteredLists: function() {
                var self = this;

                self._shadowLists = {};
                self._filteredList = {};
                self._filteredListItems = {};
                self._filteredListItems[self._availableOptionsContainer[0].id] = {};
                self._filteredListItems[self._selectedOptionsContainer[0].id] = {};
                self._dataTree = {};
                self._setShadowList();
            },

            /*
             * _dndOption: Private function to enable/disable DnD feature based on the value
             * of options.allowDnD variable.
             */
            _dndOption: function _dndOption() {
                var self = this,
                    // Setting up drag and drop options that are common for the shuttle trees.
                    dragOption = {
                        node: {
                            dataTypes: ['application/ojtreenodes+json'],
                            dragStart: _.bind(self._dragStart, self),
                            dragEnd:   _.bind(self._dragEnd, self)
                        }
                    },
                    dropOption = {
                        node: {
                            dataTypes: ['application/ojtreenodes+json'],
                            dragOver:  _.bind(self._dragOver, self),
                            dragLeave: _.bind(self._dragLeave, self),
                            drop:      _.bind(self._drop, self)
                        }
                    };

                if (self.options.allowDnD) {
                    self._leftTree.ojTree('option', {
                        dnd: {
                            drag: dragOption,
                            drop: dropOption
                        },
                        before: _.bind(self._dndBeforeHandler, self)
                    });
                    self._rightTree.ojTree('option', {
                        dnd: {
                            reorder: (self.options.allowReorder) ? 'enable' : 'disable',
                            drag:    dragOption,
                            drop:    dropOption
                        },
                        before: _.bind(self._dndBeforeHandler, self)
                    });
                } else { // This is to handle the case we implement a demo 'button' to toggle the DnD feature.
                    self._leftTree.ojTree('option', {dnd: undefined, before: undefined});
                    self._rightTree.ojTree('option', {dnd: undefined, before: undefined});
                }
            },

            /*
             * This function is called when the dnd drag starts, i.e. initial mouse drag.
             *
             * @param event: the jQuery wrapped native event object.
             *
             * @return boolean: should return true to indicate that the drag is permissible or false otherwise.
             */
            _dragStart: function(event, ui) {
                var self = this,
                    selections, checkPathIds, item, sourcePathIds, path,
                    sourceTree = (self._rightTree[0] === event.originalEvent.currentTarget) ? self._rightTree : self._leftTree,
                    targetTree = (sourceTree === self._leftTree) ? self._rightTree : self._leftTree,
                    fromSameParent = true,
                    sourceItems = [];

                // Single or multiple-selected item(s) drag
                if (sourceTree.ojTree('isSelected', event.target)) {
                    selections = sourceTree.ojTree('option', 'selection');
                    for (var i = 0; i < selections.length; i++) {
                        item = selections[i];
                        sourcePathIds = self._getPathIds(sourceTree, item);
                        sourcePathIds.pop();
                        if (!checkPathIds) {
                            checkPathIds = sourcePathIds;
                        }
                        if (fromSameParent && (checkPathIds.length === 0 && sourcePathIds.length > 0) ||
                             !_.isEqual(checkPathIds, _.intersection(checkPathIds, sourcePathIds))) {
                            fromSameParent = false;
                        }
                        path = sourceTree.ojTree('getPath', item);
                        sourceItems.push(item);
                    }
                // Single un-selected item drag
                } else {
                    sourceItems.push(event.target);
                }

                // Save source information for use later
                self._dndContext.sourceTree = sourceTree;
                self._dndContext.sourceItems = sourceItems;

                // Flag to indicate if source items are from the same parent
                self._dndContext.sourceItemsFromSameParent = fromSameParent;

                // Clear any target selection on drag start
                targetTree.ojTree('deselectAll');

                return true;
            },

            /*
             * This function is called when the dnd drag cursor is over a drop target.
             * (Note that there will be multiple calls to this function as the mouse moves over the node).
             *
             * @param event: the jQuery wrapped native event object.
             * @param ui: tree parameters
             *
             * @return boolean: return false to indicate that the dragged data can be accepted at this position, or true otherwise.
             */
            _dragOver: function(event, ui) {
                var self = this,
                    dndSourceParent, dndTargetParent, sourceParentPathIds, targetParentPathIds,
                    isSourceLeafNode, isTargetLeafNode, targetRootIds, checkPathIds,
                    isValidLeafNode = false,
                    isValidTreeNode = false,
                    dropRejected = true,
                    // Information about the selected target tree (drop zone)
                    dndTargetTree = (self._rightTree[0] === event.originalEvent.currentTarget) ? self._rightTree : self._leftTree,
                    dndTargetItem = ui.item,
                    // Information about the source tree
                    dndSourceTree = self._dndContext.sourceTree,
                    dndSourceItems = self._dndContext.sourceItems,
                    // Is this a reorder drag?
                    reorder = (dndSourceTree === dndTargetTree),
                    // Use the first source item so we can check if source is accepted to drop at this target position or not
                    dndSourceItem = dndSourceItems[0],
                    // Get the path to the target item (path also includes the last parent node)
                    // Use these to test if source/target nodes are from the same partial tree node
                    sourcePathIds = self._getPathIds(dndSourceTree, dndSourceItem),
                    targetPathIds = self._getPathIds(dndTargetTree, dndTargetItem),
                    pathMinCount = Math.min(sourcePathIds.length, targetPathIds.length),
                    // Get path up to the last parent of either source/target path
                    checkIdx = (pathMinCount >= 2) ? pathMinCount - 2 : pathMinCount - 1;

                // Nothing we can do if there is no drag source.
                if (_.isEmpty(dndSourceItems)) {
                    return (dropRejected);
                }

                // We decide to go with the container drag-and-drop implementation,
                // so this is where we highlight the target container if drop is
                // acceptable, and let the shuttle code handle the move position.
                if (!reorder) {
                    // Collapse all tree nodes
                    dndTargetTree.ojTree('collapseAll');

                    // Highlight target container to indicate dnd operation
                    self._dndContainerBorderHighlight(dndTargetTree, true);
                    return (false);
                }

                // NOTE: The following code is for the true drag-and-drop implementation,
                // where user can decide on a drop position. To use true dnd, need to
                // comment the above container dnd code block.
                // Also, part of the code below handles the dnd reordering.

                // Special handling if source items are not from the same parent
                if (!reorder && !self._dndContext.sourceItemsFromSameParent) {

                    // Collapse all tree nodes
                    dndTargetTree.ojTree('collapseAll');

                    // Highlight target container to indicate multi-parent drag
                    self._dndContainerBorderHighlight(dndTargetTree, true);

                    // we only allow drop positions other than 'inside'
                    if (ui.position !== DND.DROP_POSITION_INSIDE) {
                        return (false);
                    }

                    return (dropRejected);
                }

                // handle the drop position
                if (ui.position === DND.DROP_POSITION_FIRST || ui.position === DND.DROP_POSITION_BEFORE ||
                    ui.position === DND.DROP_POSITION_AFTER || (!reorder && ui.position === DND.DROP_POSITION_INSIDE)) {

                    // This will tell if source item is a nested node (i.e. has a parent node)
                    dndSourceParent = dndSourceTree.ojTree('getParent', dndSourceItem);

                    // This will tell if target item is a nested node (i.e. has a parent node)
                    dndTargetParent = dndTargetTree.ojTree('getParent', dndTargetItem);

                    sourceParentPathIds = self._getPathIds(dndSourceTree, dndSourceParent);
                    targetParentPathIds = self._getPathIds(dndTargetTree, dndTargetParent);

                    // Use these to test if source/target item is a leaf node (has no children)
                    isSourceLeafNode = dndSourceTree.ojTree('isLeaf', dndSourceItem);
                    isTargetLeafNode = dndTargetTree.ojTree('isLeaf', dndTargetItem);
                    targetRootIds = self._getRootIds(dndTargetTree);

                    // check if source item is accepted to drop at this position.
                    // If target is a top-level leaf node ...
                    if (_.isNull(dndTargetParent)) {
                        // Accept position drop-position!=inside
                        isValidLeafNode = (ui.position !== DND.DROP_POSITION_INSIDE &&
                                            // and both source and target are leaf nodes and
                                            isTargetLeafNode &&
                                            // source is a leaf node and is not a target partial yet
                                            ((isSourceLeafNode || !_.contains(targetRootIds, dndSourceItem.id)) && _.isNull(dndSourceParent)));

                        // Check further if source is a tree node
                        if (!isValidLeafNode) {
                            // Accept position if source and target are from the same partial tree node
                            isValidTreeNode = (ui.position === DND.DROP_POSITION_INSIDE && targetPathIds.length > 0 &&
                                                 (sourcePathIds[0] === targetPathIds[0]));
                            if (!isValidTreeNode) {
                                // Accept the case where source item is a tree node that is not a target partial yet
                                isValidTreeNode = (ui.position !== DND.DROP_POSITION_INSIDE && !_.contains(targetRootIds, sourcePathIds[0]));
                            }
                        }

                        dropRejected = !(isValidLeafNode || isValidTreeNode);

                        // if drop rejected but this is a reorder instead, we check some more
                        if (dropRejected && reorder) {
                            // accept if source is from the top-level
                            // or accept only the move to first/after the target, and both source/target is from the same parent
                            dropRejected = !((_.isNull(dndSourceParent) &&
                                              (isTargetLeafNode || ((ui.position === DND.DROP_POSITION_BEFORE ||
                                                                     ui.position === DND.DROP_POSITION_AFTER) &&
                                                                    targetParentPathIds.length === 0))) ||
                                              (targetPathIds.length > 0 &&
                                               (ui.position !== DND.DROP_POSITION_BEFORE && _.isEqual(sourceParentPathIds, targetPathIds))));
                        }
                    } else { // target is a tree node

                        // To accept a drop, source item should also come from a tree node ...
                        if (!_.isNull(dndSourceParent)) {

                            if (isSourceLeafNode && isTargetLeafNode) { // if both source and target are leaf node
                                // accept if the two paths are the same
                                dropRejected = !(ui.position !== DND.DROP_POSITION_INSIDE && _.isEqual(sourceParentPathIds, targetParentPathIds));
                            } else if (isSourceLeafNode) { // if source is a leaf node and target is a tree node
                                // accept if source parent path (i.e. excluding the source leaf node) is the same as the target path
                                dropRejected = !(ui.position !== DND.DROP_POSITION_BEFORE && sourceParentPathIds.length === targetPathIds.length && _.isEqual(sourceParentPathIds, targetPathIds));
                            } else if (dropRejected && isTargetLeafNode) { // if target is a leaf node
                                // accept if source path is the same as target parent path
                                dropRejected = !(ui.position !== DND.DROP_POSITION_FIRST && sourcePathIds.length === targetParentPathIds.length && _.isEqual(sourcePathIds, targetParentPathIds));
                            } else { // both source and target are tree nodes
                                // accept if source path is the same as target path
                                dropRejected = !(ui.position !== DND.DROP_POSITION_FIRST && ui.position !== DND.DROP_POSITION_BEFORE && sourcePathIds.length === targetPathIds.length && _.isEqual(sourcePathIds, targetPathIds));
                            }

                            // if drop still rejected but this is a reorder instead, we check some more
                            if (dropRejected && reorder) {
                                // accept if source path is the same as the target path.
                                // If position is 'first', that is, first position after the last parent node, we use the
                                // path that includes the last parent node.
                                checkPathIds = (ui.position === DND.DROP_POSITION_FIRST) ? targetPathIds : targetParentPathIds;
                                dropRejected = !_.isEqual(sourceParentPathIds, checkPathIds);
                            }
                        }

                    }
                }

                // Set dragOverPos to a valid position if drop is accepted, undefined otherwise.
                self._dndContext.dragOverPos = !dropRejected ? ui.position : undefined;

                return dropRejected;
            },

            /*
             * This function is called when dnd drop is valid on a target.
             * Drop event occurs only for a valid drop, and before dragEnd event.
             *
             * @param event: the jQuery wrapped native event object.
             * @param ui: tree parameters
             *
             * @return boolean: return false to indicate that the dragged data is accepted, or true otherwise.
             */
            _drop: function(event, ui) {
                var sourceData,
                    self = this,
                    selectedListId = self._selectedOptionsContainer[0].id,
                    dndTargetTree = (self._rightTree[0] === event.originalEvent.currentTarget) ? self._rightTree : self._leftTree,
                    dt = event.originalEvent.dataTransfer;

                // If reorder is true, the drop callback should NOT perform any tree restructuring
                // since this is automatically performed as part of the reorder.
                if (!ui.reorder) {

                    // We know we have tree node data because we used dataTypes in both the
                    // drag source tree and this drop target tree.
                    sourceData = dt.getData('application/ojtreenodes+json');
                    sourceData = JSON.parse(sourceData);
                    // Looks like if we follow the OJET cookbook example, there is an issue with the dragEnd callback
                    // not triggered if we call _dndMove here (which calls ojTree('remove')), so we save the data and call
                    // _dndMove from dragEnd callback instead.
                    self._dndContext.sourceData = sourceData;
                    self._dndContext.targetRef = ui.reference ? ui.reference[0].id : null; // reference is null if target tree is empty
                    self._dndContext.targetPos = ui.position;
                }

                //*************** With list filters option ***********************
                // Support DND reorder updates to the shadow selected list.
                if (self.options.filterLists && ui.reorder) {
                    sourceData = dt.getData('application/ojtreenodes+json');
                    sourceData = JSON.parse(sourceData);
                    sourceData.forEach(function(source) {
                        self._dataTree[selectedListId].moveNode(source.attr.id, ui.reference[0].id, ui.position);

                        //************** update the shadow lists **********************************
                        // update the new node configuration to the target shadow list.
                        self._shadowLists[selectedListId] = self._dataTree[selectedListId].toJSON();
                    });
                }
                //****************************************************************

                // Reset target container highlight
                self._dndContainerBorderHighlight(dndTargetTree);

                if (self._dndContext.dragOverPos) {
                    // We save the DnD changes asynchronously after the DOM changes settled.
                    ko.tasks.schedule(function() {
                        self._saveSelections();
                    });
                }

                // accept the drop if dragOverPos is defined (from dragOver)
                return (!self._dndContext.dragOverPos);
            },

            /*
             * This function is called when the dnd drag operation moves out of the last entered node.
             *
             * @param event: the jQuery wrapped native dragend event object.
             * @param ui: tree parameters
             */
            _dragLeave: function(event, ui) {
                var self = this,
                    dndTargetTree = (self._rightTree[0] === event.originalEvent.currentTarget) ? self._rightTree : self._leftTree;
                // Reset target container highlight
                self._dndContainerBorderHighlight(dndTargetTree);
            },

            /*
             * This function is called when the dnd drag ends, i.e. mouse releases.
             * This event always occurs, even for invalid drop.
             *
             * @param event: the jQuery wrapped native dragend event object.
             * @param ui: tree parameters
             */
            _dragEnd: function(event, ui) {
                var self = this;

                // If reorder is false the dragEnd callback should remove any nodes from the
                // tree that were moved (that is, dropEffect is "move").
                if (!ui.reorder && event.originalEvent.dataTransfer.dropEffect === DND.DROP_EFFECT_MOVE && self._dndContext.sourceData) {
                    // NOTE: that here we use the container drop implementation here, so we pass null as targetValue,
                    // and position as 'after' to use the shuttle move logic.
                    // To use true dnd drop implementation, call the _dndMove and pass in self._dndContext.targetRef,
                    // and self._dndContext.targetPos as targetValue, position respectively.
                    self._dndMove(self._dndContext.sourceData, null, SHUTTLE.MOVE_LOCATION_AFTER);
                    self._setItemCountDisplay();
                }

                // clean up
                self._dndContext = {};
            },

            /*
             * Function to move the selected source node(s) to the intended target node at the specified position.
             *
             * @param  sourceData: source node data for the target drop
             * @param  targetValue: target node ID
             */
            _dndMove: function(sourceData, targetValue, position) {
                var self = this,
                    sourceList = (self._dndContext.sourceTree === self._leftTree) ? self._availableOptionsContainer : self._selectedOptionsContainer,
                    targetList = (sourceList === self._selectedOptionsContainer) ? self._availableOptionsContainer : self._selectedOptionsContainer,
                    isFromSelectedToAvailable = (sourceList === self._selectedOptionsContainer),
                    shuttleIdPrefix = isFromSelectedToAvailable ? SHUTTLE.AVAILABLE_ID_PREFIX : SHUTTLE.SELECTED_ID_PREFIX;

                window.setTimeout(function() {
                    // setTimeout so the asynchronous drag and drop event finishes before the move logic executes.
                    // This allows all the parent anc child moved and status. title changes to work as it would with
                    // actions initiated with the various move buttons.
                    sourceData.forEach(function(data) {
                        self._moveItems(sourceList, targetList, shuttleIdPrefix, position, targetValue, data.attr.id, isFromSelectedToAvailable, true);
                    }, self);
                });

                // delay updating the filtered list until the dnd has had time
                // to finish.
                _.delay(function(selfP) {
                    selfP._updateFilteredLists(targetList[0].id, sourceList[0].id);
                }, 100, self);

                self._removeButton.ojButton('option', 'disabled', true);
                self._moveButton.ojButton('option', 'disabled', true);
                // important to clear source and target selected items
                // so not to pollute the next move or remove
                self._availableValue = null;
                self._selectedValue = null;
                targetList.ojTree('deselectAll');
                if (self._hasAvailableDescr) {
                    self._leftAvailableValue.ojTextArea('option', {'value': ''});
                }
            },

            /*
             * This function is to set/clear container border highlight.
             *
             * @param container: container to set border highlight
             * @param entering: true if entering container div, false (or omit) otherwise.
             */
            _dndContainerBorderHighlight: function(container, entering) {
                var self = this;

                // Reset target container highlight
                if (container) {
                    // Add class to highlight container's border and hide ojTree drop marker UI upon entering,
                    // and remove class on exit.
                    if (entering) {
                        $('body').addClass('lux-oj-tree-hide-drop-markers');
                        container.addClass('lux-oj-tree-drag-over');
                    } else {
                        $('body').removeClass('lux-oj-tree-hide-drop-markers');
                        container.removeClass('lux-oj-tree-drag-over');
                    }
                }
            },

            /*
             * Intercept handler callback before an event triggered.
             * The following events can be vetoed during before event processing by returning false from the before event handler
             * (omitting a return value or returning true permits the event processing to continue):
             * collapse, expand, hover, select, remove, rename.
             */
            _dndBeforeHandler: function(event, ui) {
                var self = this;

                // We are interested in the expand event
                if (ui.func === 'expand') {
                    // and since we implement drop on target container, we disable the tree node expansion for
                    // shuttle DnD and reordering, and only allow tree node expansion for manual tree node expansion.
                    return (self._dndContext.sourceTree === undefined);
                }
            },

            /**
             * Gets list of node IDs in the tree path from the parent node down to the specified node.
             * @param {object} tree - ojTree where to get the path from
             * @param {object} node - tree node to look up the path
             * @return {*}
             */
            _getPathIds: function(tree, node) {
                var idList = [];
                var pathIds = (!tree || !node) ? idList : tree.ojTree('getPath', node, true);
                _.each(pathIds, function(id) {
                    idList.push(id.substring(id.indexOf('-') + 1));
                });
                return (idList);
            },

            /*
             * return list of node IDs of the first-level tree nodes from root.
             *
             * @param  tree: ojTree where to get the path from
             */
            _getRootIds: function(tree) {
                var idList = [];
                var rootNodes = tree.ojTree('getChildren');
                _.each(rootNodes, function(node) {
                    idList.push(node.id.substring(node.id.indexOf('-') + 1));
                });
                return (idList);
            },

            /**
             * This function is called to pull selected options from the available options list only if just the
             * available options list is input into the widget.
             * @return {array} - array of selected options objects.
             * @private
             */
            _getSelectedValues: function() {
                var self = this,
                    available = self._availableOptionsContainer.ojTree('getRoot'), // get root
                    selectedItems = [],
                    getSelected = function(available, treeContainer, isSelected) {
                        // iterate over the collection of tree node objects
                        _.each(available, function(item) {
                            var selectedItem = {},
                                childNodes,
                                nodeItem,
                                nodeSelected;

                            // Ignore invalid item during ABCS initialization
                            if (!item.id || item.id.length === 0) {
                                return;
                            }

                            // Initialize the nodes, only after we checked that item id is valid
                            nodeItem = $('#' + luxUtils.encodeId(item.id));
                            nodeSelected = isSelected || nodeItem.attr('isselected');

                            // create a new selected object from the available object
                            selectedItem.title = nodeItem.find('.oj-tree-title').first().text();
                            selectedItem.attr = self._copyAttributes(nodeItem);
                            selectedItem.attr.id = (selectedItem.attr.id.search(SHUTTLE.SELECTED_ID_PREFIX) !== -1) ?
                                item.id :
                                SHUTTLE.SELECTED_ID_PREFIX + selectedItem.attr.id;

                            // if there are children iterate over them
                            if (self._hasChildren(nodeItem)) {
                                selectedItem.children = [];
                                // get child nodes
                                childNodes = self._availableOptionsContainer.ojTree('getChildren', nodeItem);
                                // recursive call against children
                                getSelected(childNodes, selectedItem.children, nodeSelected);
                            }

                            // if the node is flagged "isSelected" or the node has children that are "isSelected"
                            // push the node to the selected treeContainer
                            if (nodeSelected || (Array.isArray(selectedItem.children) && selectedItem.children.length > 0)) {
                                treeContainer.push(selectedItem);
                                // if there are no children or all the child nodes have been selected, delete the node from
                                // the available options list
                                if (!Array.isArray(selectedItem.children) || selectedItem.children.length === childNodes.length) {
                                    // if there are no children or the available children for this node is zero
                                    // remove the node from the available list.
                                    if (!Array.isArray(selectedItem.children) || self._deepCountAvailableChildren(nodeItem) === 0) {
                                        self._availableOptionsContainer.ojTree('remove', nodeItem);
                                    } else {
                                        // but if there are available children show selected node as partially selected and do the
                                        // same with the matching available node in the available options list
                                        selectedItem.title += ' ' + self.options.partialLabel;
                                        nodeItem.find('.oj-tree-title').first().text(selectedItem.title);
                                        selectedItem.attr.isselected = SHUTTLE.PARTIAL_STATUS;
                                    }
                                    // but if not all child nodes are selected, show selected node as partially selected and do the
                                    // same with the matching available node in the available options list
                                } else if (Array.isArray(selectedItem.children) || selectedItem.children.length !== childNodes.length) {
                                    selectedItem.title += ' ' + self.options.partialLabel;
                                    nodeItem.find('.oj-tree-title').first().text(selectedItem.title);
                                    selectedItem.attr.isselected = SHUTTLE.PARTIAL_STATUS;
                                }
                            }

                        }, self);
                    };

                if (available[0].id.length === 0) { // no root so get top level nodes
                    available = self._availableOptionsContainer.ojTree('getChildren', -1); // get top level nodes
                }

                getSelected(available, selectedItems);

                return selectedItems;
            },

            /**
             * Does a deep count of children under the selected
             * dom node
             * @param  {object} nodeItem  - selected DOM node
             * @param  {number} count of available children
             * @return {number} - count of available nodes.
             * count of available children
             */
            _deepCountAvailableChildren: function _deepCountAvailableChidren(nodeItem, count) {
                var self = this,
                    childNodes = self._availableOptionsContainer.ojTree('getChildren', nodeItem),
                    childCount = count || 0,
                    selectedItemAttr = nodeItem[0] ? self._copyAttributes(nodeItem) : null,
                    isSelected = selectedItemAttr ? selectedItemAttr.isselected : null;

                if (!isSelected) {
                    childCount += 1;
                }

                _.each(childNodes, function(node) {
                    if (self._hasChildren(node)) {
                        self._deepCountAvailableChildren(node, childCount);
                    }
                });

                return childCount;
            },

            /**
             * When the filter options is declared, sets up the shadow available and selected lists
             * @private
             * @return {*}
             */
            _setShadowList: function() {
                var self = this,
                    selectedId = self._selectedOptionsContainer[0].id,
                    availableId = self._availableOptionsContainer[0].id;

                if (!self.options.filterLists) {
                    return;
                }

                self._shadowLists[availableId] = self._createShadowList(self._availableOptionsContainer);
                self._shadowLists[selectedId] = self._createShadowList(self._selectedOptionsContainer);

                if (!self._dataTree[availableId]) {
                    self._dataTree[availableId] = new DataTree(self._shadowLists[availableId], SHUTTLE, SHUTTLE.AVAILABLE_ID_PREFIX);
                    self._dataTree[availableId].toConsole('********** Available list');
                    oj.Logger.info('*********** back to JSON', self._dataTree[availableId].toJSON());
                }
                if (!self._dataTree[selectedId]) {
                    self._dataTree[selectedId] = new DataTree(self._shadowLists[selectedId], SHUTTLE, SHUTTLE.SELECTED_ID_PREFIX);
                    self._dataTree[selectedId].toConsole('********** Selected list');
                    oj.Logger.info('*********** back to JSON', self._dataTree[selectedId].toJSON());
                }
            },

            /**
             * When the filter options is declared, creates a shadow list object in the form of a JSON structure.
             * @param {object} container - DOM element attached to ojTree
             * @return {Array}
             * @private
             */
            _createShadowList: function(container) {
                var self = this,
                    topLevelNodes = container.ojTree('getChildren', -1),
                    output = [],
                    getChildren = function(parent, children) {

                        _.each(children, function(item) {
                            var nodeItem = $('#' + luxUtils.encodeId(item.id)),
                                selectedItem = {},
                                childNodes = self._selectedOptionsContainer.ojTree('getChildren', nodeItem);

                            selectedItem.title = nodeItem.find('.oj-tree-title').first().text();
                            selectedItem.attr = self._copyAttributes(nodeItem);

                            if (childNodes) {
                                selectedItem.children = [];
                                getChildren(selectedItem.children, childNodes);
                            }

                            parent.push(selectedItem);

                        }, self);
                    };

                getChildren(output, topLevelNodes);

                return output;

            },

            /**
             * If 2 separated lists, available and selected options, are input to the widget, this
             * function is called to create the appropriate labeling for each parent node.
             * @return {*}
             */
            _setSelectedValuesAttributes: function() {
                var self = this,
                    selectedOptions = self._selectedOptionsContainer,
                    selectedItemList = selectedOptions.ojTree('getRoot'), // get root
                    getSelected = function(selected) {

                        _.each(selected, function(item) {
                            var childNodes,
                                availableNode,
                                title,
                                hasAvailable = false,
                                nodeItem = $('#' + luxUtils.encodeId(self.options.selectedListId) + ' #' + luxUtils.encodeId(item.id));

                            if (self._hasChildren(nodeItem)) {
                                childNodes = selectedOptions.ojTree('getChildren', nodeItem);
                                getSelected(childNodes);
                                availableNode = $('#' + luxUtils.encodeId(self.options.availableListId) + ' #' + luxUtils.encodeId(item.id));
                                title = availableNode.find('.oj-tree-title').first();
                                title.text(title.text() + ' ' + self.options.partialLabel);
                                // is this node is in the available list?
                                if (title.length) {
                                    // only if there is a title in the available list
                                    hasAvailable = true;
                                }

                                title = nodeItem.find('.oj-tree-title').first();
                                // if this node is in the available list
                                if (hasAvailable) {
                                    // append the title with the "partial" label.
                                    title.text(title.text() + ' ' + self.options.partialLabel);
                                }

                                nodeItem.attr('id', SHUTTLE.SELECTED_ID_PREFIX + item.id);
                                nodeItem.attr('isselected', hasAvailable ? 'true' : 'partial');
                                if (item.description) {
                                    nodeItem.attr('description', item.description);
                                }
                                availableNode.attr = self._copyAttributes(nodeItem);

                            } else {
                                nodeItem.attr('id', SHUTTLE.SELECTED_ID_PREFIX + item.id);
                                nodeItem.attr('isselected', 'true');
                                if (item.description) {
                                    nodeItem.attr('description', item.description);
                                }
                            }

                        }, this);

                    };

                if (selectedItemList[0].id.length === 0) { // no root so get top level nodes
                    selectedItemList = selectedOptions.ojTree('getChildren', -1); // get top level nodes
                }

                getSelected(selectedItemList);
            },

            /**
             * Function to get and set the available and selected list counts. Also
             * manages the the moveAll and removeAll buttons 'disabled' option.  If
             * a list is empty its associated "move" ALL button is disabled.
             *
             * @returns {*}
             * @private
             */
            _setItemCountDisplay: function() {
                var self = this,
                    availabeCount,
                    selectedCount,
                    // get the item count label that can be stored in a knockout
                    // observable or it can be string.
                    itemCountLabel = {
                        single: ko.isObservable(self.options.itemCountLabel.single) ?
                            self.options.itemCountLabel.single() :
                            self.options.itemCountLabel.single,
                        plural: ko.isObservable(self.options.itemCountLabel.plural) ?
                            self.options.itemCountLabel.plural() :
                            self.options.itemCountLabel.plural
                    };

                availabeCount = self._getNodeCount(self._availableOptionsContainer);
                selectedCount = self._getNodeCount(self._selectedOptionsContainer);

                self._moveAllButton.ojButton('option', 'disabled', false);
                self._removeAllButton.ojButton('option', 'disabled', false);

                if (availabeCount === 0) {
                    // list is empty so disable the moveAll button
                    self._moveAllButton.ojButton('option', 'disabled', true);
                }

                if (selectedCount === 0) {
                    // list is empty so disable the removeAll button
                    self._removeAllButton.ojButton('option', 'disabled', true);
                }

                // do we want to display counts?
                if (! self.options.showcounts) {
                    // if not, leave
                    return;
                }

                if (self.options.itemCountLabel) {
                    if (availabeCount === 1) {
                        self._observablesMap.itemCountLabels.available.el.text(itemCountLabel.single);
                    } else {
                        self._observablesMap.itemCountLabels.available.el.text(itemCountLabel.plural);

                    }
                    if (selectedCount === 1) {
                        self._observablesMap.itemCountLabels.selected.el.text(itemCountLabel.single);
                    } else {
                        self._observablesMap.itemCountLabels.selected.el.text(itemCountLabel.plural);
                    }
                }

                self.availableCount.text(availabeCount);
                self.selectedCount.text(selectedCount);
            },

            _getNodeCount: function getCount(container) {
                var self = this,
                    sourceTree = container.ojTree('getChildren', -1),
                    count = 0,
                    countcountChildren = function countChidren(children) {
                        var childArray = _.map(children, function(item) {
                                return item;
                            }, this),
                            childNodes;

                        childArray.forEach(function(item) {
                            var nodeItem = $('#' + luxUtils.encodeId(item.id));

                            count += 1;
                            if (self._hasChildren(item)) {
                                childNodes = container.ojTree('getChildren', nodeItem);
                                countcountChildren(childNodes);
                            }
                        });

                    },
                    listArray = _.map(sourceTree, function(item) {
                        return item;
                    }, this);

                listArray.forEach(function(item) {
                    var childNodes;

                    count += 1;
                    if (self._hasChildren(item) && item.id) {
                        childNodes = container.ojTree('getChildren', $('#' + luxUtils.encodeId(item.id)));
                        countcountChildren(childNodes);
                    }
                });

                return count;
            },

            // Convert a jQuery list of html element nodes to string containing node id's.
            _arrayToStr: function(arr) {
                var s = '',
                    d = '',
                    self = this;

                if (!_.isArray(arr)) {
                    return;
                }

                $.each(arr, function(i /*, val*/) {
                    var id = $(arr[i]).attr('id'),
                        descr = $(arr[i]).attr('description');

                    if (i) {
                        s += ', ';
                        d += ', ';
                    }
                    s += id;
                    d += descr;

                }.bind(self));

                return {'id': s, 'description': d};
            },

            _hasChildren: function(item) {
                var self = this;

                if (self._availableOptionsContainer.ojTree('isLeaf', item)) {
                    return false;
                }
                return true;
            },

            /**
             * Function to manage on optionChange against the 2 options lists.
             *
             * @param {object} event - object contains the id of the values list
             * @param {object} data - object contains the selected data
             * @return {*}
             */
            _selectionChange: function _selectionChange(event, data) {
                var self = this,
                    title,
                    selected = self._arrayToStr(data.value),
                    ids = selected ? (selected.id !== 'undefined' ? selected.id.split(', ') : null) : null,
                    nodeItem,
                    selectedId,
                    theTextArea,
                    actionButton,
                    isSelectedDown = function(selectedIds, optionValues, currentId) {
                        var lastSelected,
                            currentNode;

                        //***************************************************************************************************
                        // Find out if the the last selected item in the range of selected items is the current
                        // item selected. If it is, then the selection is from the current selection is up the list, returning
                        // false otherwise the selected range is down the list, returning true.
                        //***************************************************************************************************

                        // get the next sibling node from the last item in the selected range
                        lastSelected = optionValues.ojTree('getNextSibling', $('#' + luxUtils.encodeId(selectedIds[selectedIds.length - 1])));
                        // if is a next sibling
                        if (lastSelected) {
                            currentNode = $('#' + luxUtils.encodeId(currentId[currentId.length - 1]));
                            // and the current selected node is also the next sibling in the range of selected items
                            // then the selection has been made from the current selected node up the list of items.
                            if (currentNode[0] === lastSelected[0]) {
                                return false;
                            }
                        }
                        return true;
                    },
                    setSelectedId = function() {
                        var selectedDown = true;

                        // get the diff of current selected ids and the
                        // stored ids.  will return 1 to n element array
                        // of the newly selected ids.
                        if (event.target.id === self.options.availableListId) {
                            selectedId = ids ? _.difference(ids, self._availableIds) : null;
                            theTextArea = self._leftAvailableValue;
                            actionButton = self._moveButton;
                            // selected a range if items?
                            if (selectedId && selectedId.length > 1) {
                                selectedDown = isSelectedDown(selectedId, self._availableOptionsContainer, self._availableIds);
                            }
                        } else if (event.target.id === self.options.selectedListId) {
                            selectedId = ids ? _.difference(ids, self._selectedIds) : null;
                            theTextArea = self._rightSelectedValue;
                            actionButton = self._removeButton;
                            // selected a range if items?
                            if (selectedId && selectedId.length > 1) {
                                selectedDown = isSelectedDown(selectedId, self._selectedOptionsContainer, self._selectedIds);
                            }
                        }

                        if (selectedId) {
                            if (selectedId.length === 0) {
                                // there is no selected id because the item was already in the previous range of
                                // selected items. in this case apply the id to the selected item ids.
                                selectedId = ids;
                            }
                            selectedId = selectedDown ? selectedId[selectedId.length - 1] : selectedId[0];
                        }
                    },
                    uiTreeSelectionScan = function(targetTreeListId, moveButton) {
                        var target = $('#' + luxUtils.encodeId(targetTreeListId)),
                            items = _.map(target.ojTree('getChildren', -1)),
                            nIntervalId,
                            areSelectedItems = function areSelectedItems(itemsList, button, context) {
                                var someSelected = [];
                                var filterList = function(children, filterArray) {

                                    children.forEach(function (item) {

                                        // if item is selected, push to the array
                                        if (target.ojTree('isSelected', $('#' + luxUtils.encodeId(item.id)))) {
                                            filterArray.push(item.id);
                                        }

                                        // if there are children check for children that are selected
                                        if (!_.isUndefined(this) && (this._hasChildren && this._hasChildren(item))) {
                                            // recursive call
                                            filterList(_.map(target.ojTree('getChildren', item)), filterArray);
                                        }

                                    }, context);
                                };

                                // fill the someSelected array with selected items
                                filterList(itemsList, someSelected);

                                // if there are no selected items then disable the button and
                                // stop the interval
                                if (someSelected.length === 0) {
                                    button.ojButton('option', 'disabled', true);
                                    clearInterval(nIntervalId);
                                }
                            };

                        // set the interval timer that runs the selected items check
                        nIntervalId = setInterval(areSelectedItems, 300, items, moveButton, self);
                    };

                // if data.option is not 'selection', that is selection
                // of a ojTree node or the variable ids is null
                // don't do anything quit the function.
                if (data.option !== 'selection' || !ids) {
                    return;
                }

                setSelectedId();
                if (theTextArea) {
                    // Then empty the textarea value.
                    theTextArea.ojTextArea('option', {'value': ''});
                    theTextArea.ojTextArea('reset');
                }

                if (selectedId) {
                    nodeItem = $('#' + luxUtils.encodeId(selectedId));
                    title = nodeItem[0].attributes.description ?
                        nodeItem[0].attributes.description.value :
                        nodeItem.find('.oj-tree-title').first().text();

                    actionButton.ojButton('option', 'disabled', false);
                    if (_.isUndefined(event.which)) {
                        uiTreeSelectionScan(event.target.id, actionButton);
                    }

                    if (theTextArea) {
                        theTextArea.ojTextArea('option', {'value': title});
                    }
                } else {
                    // node in list has been unselected so make the
                    // disable the button.
                    actionButton.ojButton('option', 'disabled', true);
                }
                if (event.target.id === self.options.availableListId) {
                    self._availableValue = ids && ids.join(', ');
                    self._availableIds = ids;
                } else if (event.target.id === self.options.selectedListId) {
                    self._selectedValue = ids && ids.join(', ');
                    self._selectedIds = ids;
                }

                self._setItemCountDisplay();
            },

            _moveButtonClick: function(/*event*/) {
                var self = this,
                    sourceList = self._availableOptionsContainer,
                    targetList = self._selectedOptionsContainer,
                    targetValue = self._selectedValue,
                    sourceValue = self._availableValue.split(', ');

                sourceValue.forEach(function(sourceVal) {
                    targetValue = self._verifyTarget(sourceList, targetList, targetValue);
                    self._moveItems(sourceList, targetList, SHUTTLE.SELECTED_ID_PREFIX, SHUTTLE.MOVE_LOCATION_AFTER, targetValue, sourceVal, false);
                }, this);

                self._updateFilteredLists(targetList[0].id, sourceList[0].id);

                self._removeButton.ojButton('option', 'disabled', true);
                self._moveButton.ojButton('option', 'disabled', true);
                // important to clear source and target selected items
                // so not to pollute the next move or remove
                self._availableValue = null;
                self._selectedValue = null;
                targetList.ojTree('deselectAll');
                // set focus on selected list
                document.getElementById(self.options.selectedListId).firstChild.focus();
                if (self._hasAvailableDescr) {
                    self._leftAvailableValue.ojTextArea('option', {'value': ''});
                }
            },

            _removeButtonClick: function(/*event*/) {
                var self = this,
                    sourceList = self._selectedOptionsContainer,
                    targetList = self._availableOptionsContainer,
                    targetValue = self._availableValue,
                    sourceValue = self._selectedValue.split(', ');

                sourceValue.forEach(function(sourceVal) {
                    targetValue = self._verifyTarget(sourceList, targetList, targetValue);
                    self._moveItems(sourceList, targetList, SHUTTLE.AVAILABLE_ID_PREFIX, SHUTTLE.MOVE_LOCATION_AFTER, targetValue, sourceVal, true);
                }, this);

                self._updateFilteredLists(targetList[0].id, sourceList[0].id);

                self._removeButton.ojButton('option', 'disabled', true);
                self._moveButton.ojButton('option', 'disabled', true);
                // important to clear source and target selected items
                // so not to pollute the next move or remove
                self._availableValue = null;
                self._selectedValue = null;
                targetList.ojTree('deselectAll');
                // set focus on available list
                document.getElementById(self.options.availableListId).firstChild.focus();
                if (self._hasSelectedDescr) {
                    self._rightSelectedValue.ojTextArea('option', {'value': ''});
                }
            },

            /**
             * If the filtered list option is on, and after moving items between lists, runs
             * the _handleFilteredList function to filter any nodes that should not be in
             * the filtered list.
             * @param {string} targetListId - DOM id of the filtered list
             * @param {string} sourceListId - DOM id of the source filtered list
             * @private
             */
            _updateFilteredLists: function(targetListId, sourceListId) {
                var self = this;

                if (self.options.filterLists) {
                    if (self._filteredListItems[targetListId] && self._filteredListItems[targetListId].items) {
                        // if the list is filtered update the changes
                        if (self._filteredListItems[targetListId] && self._filteredListItems[targetListId].text) {
                            // check for empty filter
                            if (self._filteredListItems[targetListId].text.peek().length === 0) {
                                if (self._filteredListItems[targetListId].items) { // if items, remove them
                                    self._filteredListItems[targetListId].items = [];
                                }
                                // check the source list filter is empty
                                self._checkForEmptySourceList(sourceListId, self._filteredListItems[sourceListId].text);
                                return; // empty filter so do not check filters on this target;
                            }
                            // pass attributes of the filteredText js object.
                            self._handleFilteredList(targetListId,
                                self._filteredListItems[targetListId].text,
                                self._filteredListItems[targetListId].contanterClass);
                            self._filterMoveWarning(targetListId);
                        }
                        self._filteredListItems[targetListId].items = [];
                    }
                    // check the source list filter is empty
                    self._checkForEmptySourceList(sourceListId, self._filteredListItems[sourceListId].text);
                }
            },

            _verifyTarget: function(sourceList, targetList, targetValue) {
                var targetParent, sourceParent;

                if (targetValue) {

                    targetParent = targetList.ojTree('getParent') ? targetList.ojTree('getParent')[0].id : null;
                    sourceParent = sourceList.ojTree('getParent') ? sourceList.ojTree('getParent')[0].id : null;

                    if (targetParent) {
                        targetParent = targetParent.substring(targetParent.indexOf('-') + 1);
                    }
                    if (sourceParent) {
                        sourceParent = sourceParent.substring(sourceParent.indexOf('-') + 1);
                    }

                    if ((!targetParent) && (!sourceParent)) {
                        return targetValue;
                    }
                    if (targetParent && (!sourceParent)) {
                        return null;
                    }
                    if (sourceParent && (sourceParent !== targetParent)) {
                        return null;
                    }
                }

                return targetValue;
            },

            _setParentTitle: function(id) {
                var self = this,
                    availableList = $('#' + luxUtils.encodeId(id)),
                    selectedList = $('#' + SHUTTLE.SELECTED_ID_PREFIX + luxUtils.encodeId(id)),
                    availableChildren = self._availableOptionsContainer.ojTree('getChildren', availableList),
                    selectedChildren = self._selectedOptionsContainer.ojTree('getChildren', selectedList),
                    newTitle = '',
                    updateParentTitle = function() {
                        if (selectedChildren && availableChildren) {

                            // set parent to "partial"
                            availableList.attr('isselected', SHUTTLE.PARTIAL_STATUS);

                            newTitle = availableList.find('.oj-tree-title').first().text();
                            if (newTitle.search(self.options.partialLabel) === -1) {
                                availableList.find('.oj-tree-title').first().text(newTitle + ' ' + self.options.partialLabel);
                            }

                            // set parent to "partial"
                            selectedList.attr('isselected', SHUTTLE.PARTIAL_STATUS);

                            newTitle = selectedList.find('.oj-tree-title').first().text();
                            if (newTitle.search(self.options.partialLabel) === -1) {
                                selectedList.find('.oj-tree-title').first().text(newTitle + ' ' + self.options.partialLabel);
                            }

                        } else if (selectedChildren) {
                            selectedList.attr('isselected', 'true');
                            newTitle = selectedList.find('.oj-tree-title').first().text().replace(self.options.partialLabel, '');
                            selectedList.find('.oj-tree-title').first().text(newTitle);
                        } else if (availableChildren) {
                            newTitle = availableList.find('.oj-tree-title').first().text().replace(self.options.partialLabel, '');
                            availableList.find('.oj-tree-title').first().text(newTitle);
                        }
                    };

                // update parents titles.
                updateParentTitle();
            },

            _moveItems: function(sourceList, targetList, targetIdPrefix, targetPosition, targetValue, sourceValue, remove, isDnd) {
                var self = this,
                    position = targetPosition,
                    path = sourceList.ojTree('getPath', $('#' + luxUtils.encodeId(sourceValue)), true),
                    selectList = targetList.ojTree('getChildren', -1), // get top level nodes
                    // get target and source list DOM ids. Used to identify the associated DataTree object.
                    targetListId = targetList[0].id,
                    sourceListId = sourceList[0].id,
                    currentLuxHierarchyNode,
                    location = targetValue ? $('#' + luxUtils.encodeId(targetValue)) : null,
                    moveChildren = function(selected, children) {
                        var childArray = _.map(children, function(item) {
                            return item;
                        }, this);

                        childArray.reverse().forEach(function(item) {
                            var nodeItem = $('#' + luxUtils.encodeId(item.id)),
                                childNodes,
                                selectedItem = {},
                                targetId = targetIdPrefix + (remove ? item.id.replace(SHUTTLE.SELECTED_ID_PREFIX, '') : item.id);

                            selectedItem.title = nodeItem.find('.oj-tree-title').first().text();
                            selectedItem.attr = self._copyAttributes(nodeItem);
                            selectedItem.attr.id = targetId;
                            if (!remove) {
                                selectedItem.attr.isselected = 'true';
                            }

                            if (!document.getElementById(targetId)) {
                                targetList.ojTree('create', selected, 'inside', selectedItem);
                            }

                            // if this node has children, it's parent have moved, so just
                            // move over all the children
                            if (self._hasChildren(item)) {
                                childNodes = sourceList.ojTree('getChildren', nodeItem);
                                moveChildren($('#' + luxUtils.encodeId(selectedItem.attr.id)), childNodes);
                                // get the updated list of children
                                childNodes = sourceList.ojTree('getChildren', nodeItem);
                            }

                            // if no children remove the node from the value list
                            if (!childNodes) {
                                sourceList.ojTree('remove', nodeItem);
                            }

                            oj.Logger.info('***** moved', item.id);
                            // This is a parent node, so we reset the (partial).
                            // This is to fix the case where a partial nested parent label (partial)
                            // not getting updated when moving its parent node.
                            self._setParentTitle(targetId.replace(SHUTTLE.SELECTED_ID_PREFIX, ''));
                        }, self);
                    };

                /********** (1) GET THE TARGET LOCATION ************************/
                // get the location where to move to in the selected values list.
                // defaults to before the first item in the list. if a item is
                // selected in the selected values list, the item selected in
                // the available value list is created "before" that item.
                if (!location && selectList) {
                    location = _.first(_.toArray(selectList));
                    location = $('#' + luxUtils.encodeId(location.id));
                    position = SHUTTLE.MOVE_LOCATION_BEFORE;
                }
                /***************************************************************/

                /********** (2) MOVE SELECTED FROM SOURCE TO TARGET ************/
                // iterate over list of parents of the selected available item
                // and move the parents to the selected values list (if its not
                // already there).
                path.forEach(function(item) {
                    var selected = $('#' + luxUtils.encodeId(item)),
                        selectedItem = {},
                        targetId = targetIdPrefix + (remove ? item.replace(SHUTTLE.SELECTED_ID_PREFIX, '') : item);

                    // create the tree node
                    selectedItem.title = selected.find('.oj-tree-title').first().text();
                    selectedItem.attr = self._copyAttributes(selected);
                    selectedItem.attr.id = targetId;
                    if (!remove) {
                        (selectedItem.attr.isselected = 'true');
                    }

                    /******* IF TARGET OJTREE IS EMPTY *************************/
                    // if location is null the target is an empty ojTree
                    if (_.isNull(location)) {
                        // so need to re-populate the ojTree.
                        // wrap the json in 1 element array
                        selectedItem = [selectedItem];
                        // populate ojtree
                        targetList.ojTree('option', {data: selectedItem});
                        // get the location where the next item is to be added.
                        location = $('#' + luxUtils.encodeId(targetId));
                        position = SHUTTLE.MOVE_LOCATION_AFTER;
                    }
                    /***********************************************************/

                    // if not already created in the selected options list,
                    // create it.
                    if (!document.getElementById(targetId)) {
                        targetList.ojTree('create', location, position, selectedItem);
                    }

                    if (self.options.expandParentOnMove &&
                        targetList.ojTree('getParent', $('#' + luxUtils.encodeId(targetId)))) {
                        targetList.ojTree('expand', targetList.ojTree('getParent', $('#' + luxUtils.encodeId(targetId))));
                    }

                    // change the position so subsequent nodes are added as
                    // children
                    position = 'inside';
                    // to this node
                    location = $('#' + luxUtils.encodeId(targetId));

                }, self);
                /***************************************************************/

                /************ (3) MOVE CHILDREN FROM SOURCE TO TARGET ***********/
                // Flip the path array so the items are moved to target list in
                // the same order as they where removed from the source list.
                // Then iterate over path checking to see if children need to be moved
                path.reverse().forEach(function(item, index, path) {
                    var selected = $('#' + luxUtils.encodeId(item)),
                        moveTheChildren = true,
                        children = sourceList.ojTree('getChildren', selected),
                        targetId = targetIdPrefix + (remove ? item.replace(SHUTTLE.SELECTED_ID_PREFIX, '') : item);

                    if (children) {
                        // iterate over children, if one child has already been
                        // move to the target list, the rest of the children will
                        // not be moved.
                        _.each(children, function(child) {
                            if (_.contains(path, child.id)) {
                                moveTheChildren = false;
                            }
                        }, this);

                        if (moveTheChildren) {
                            moveChildren($('#' + luxUtils.encodeId(targetId)), children);
                            // get the updated list of children
                            children = sourceList.ojTree('getChildren', selected);
                        }
                    }
                    // If filter option is on.
                    if (self.options.filterLists) {
                        // If a filter has been applied to the source list, look for any
                        // children that should be moved but have been filtered out
                        if (self._filteredListItems[sourceListId].text &&
                            self._filteredListItems[sourceListId].text.peek().length) {
                            self._updateTargetFromFilteredSource(targetIdPrefix, targetId, targetList, item, sourceListId, path);
                        }
                    }
                }, self);
                /***************************************************************/

                /********** (4) REMOVE SOURCE ITEM *****************************/
                // now check to see source item need to be remove, has no children
                path.forEach(function(item) {
                    var selected, children,
                        removeParentFromSourceAndSetTitle = function() {

                            selected = $('#' + luxUtils.encodeId(item));
                            children = sourceList.ojTree('getChildren', selected);

                            if (!children) {
                                // TODO: This could be JET issue. This is a hack for now to deselect
                                // the selection to prevent ojtree to replay the selection when applying
                                // a new data to the tree (see _handleFilteredList method below, LUX-1591).
                                sourceList.ojTree('deselect', selected);
                                // when filtering the list don't remove the item, just yet.
                                if (!self.options.filterLists) {
                                    sourceList.ojTree('remove', selected);
                                }
                            }

                            // if no more children in the source list, remove the parent.
                            self._setParentTitle(item.replace(SHUTTLE.SELECTED_ID_PREFIX, ''));
                        };

                    removeParentFromSourceAndSetTitle();
                }, self);

                /***************************************************************/
                // now determine the displayed list counts
                self._setItemCountDisplay();

                /*********** (5) IF FILTERED LISTS ******************************/
                // If the filtered list option is set update the shadow list to
                // reflect changes.
                if (self.options.filterLists) {
                    // add selected node from source shadow list to target shadow list.
                    currentLuxHierarchyNode = self._dataTree[targetListId].addNode(sourceValue, self._dataTree[sourceListId].getDataTree(), targetIdPrefix, targetPosition, targetValue);

                    // remove node from source shadow list.
                    self._dataTree[sourceListId].removeNode(sourceValue);

                    /********** (5.1) REMOVE SOURCE ITEM *****************************/
                    // Now check to see source item needs to be remove, is not in the source dataTree.
                    // AND, Update the parent title
                    path.forEach(function(item) {
                        var node = self._dataTree[sourceListId].findNodeById(item),
                            filterText, remove,
                            selected = $('#' + luxUtils.encodeId(item));

                        if (self._filteredListItems[sourceListId].text &&
                            self._filteredListItems[sourceListId].text.peek().length) {
                            filterText = self._filteredListItems[sourceListId].text();
                            remove = !self._isFiltered(selected.id, selected.text(), filterText);
                        }

                        if (_.isNull(node) || remove) { // not in source dataTree
                            selected = $('#' + luxUtils.encodeId(item));
                            sourceList.ojTree('remove', selected); // remove from the view
                        }

                        self._setParentTitle(item.replace(SHUTTLE.SELECTED_ID_PREFIX, ''));
                    }, self);

                    /***************************************************************/
                    // now determine the displayed list counts
                    self._setItemCountDisplay();

                    // update parent status
                    self._dataTree[targetListId].updateParentStatus(sourceValue, self._dataTree[sourceListId].getDataTree(), self.options.partialLabel);

                    //************** update the shadow lists ***************************************//
                    // update the new node configuration to the target shadow list.
                    self._shadowLists[targetListId] = self._dataTree[targetListId].toJSON();
                    // update the new node configuration to the source shadow list.
                    self._shadowLists[sourceListId] = self._dataTree[sourceListId].toJSON();
                    //******************************************************************************//

                    // add current node to list of nodes added to the targeted list.
                    if (self._filteredListItems[targetListId]) {
                        if (self._filteredListItems[targetListId].items) {
                            self._filteredListItems[targetListId].items.push(currentLuxHierarchyNode);
                        }

                        oj.Logger.info('***** filtering new node:', currentLuxHierarchyNode.actionLabel);
                    }
                }

                this._saveSelections();
            },

            /**
             * ONLY used with the filter option. Copies the child nodes from the shadow source luxhierarchy list
             * to the target list in the DOM when the source list in the DOM has hidden filtered children.
             *
             * @param {string} targetIdPrefix - the target node id prefix
             * @param {string} targetId - the node target DOM id
             * @param {object} targetList - the target list in the DOM
             * @param {string} sourceId - the node source DOM id
             * @param {string} sourceListId - the id of the source list in the DOM
             * @param {string} path - hierarchy of parents
             * @return {void}
             * @private
             */
            _updateTargetFromFilteredSource: function(targetIdPrefix, targetId, targetList, sourceId, sourceListId, path) {
                var self = this,
                    moveTheChildren = true,
                    node = self._dataTree[sourceListId].findNodeById(sourceId),
                    children = node.children,
                    moveChildren = function(selected, children) {
                        children.forEach(function(item) {
                            var selectedItem = {},
                                targetId = targetIdPrefix + item.action.replace(SHUTTLE.SELECTED_ID_PREFIX, '');

                            selectedItem.title = item.actionLabel;
                            selectedItem.attr = item.data;
                            selectedItem.attr.id = targetId;

                            if (targetIdPrefix === SHUTTLE.SELECTED_ID_PREFIX) {
                                selectedItem.attr.isselected = 'true';
                            }

                            if (!document.getElementById(targetId)) {
                                targetList.ojTree('create', selected, 'inside', selectedItem);
                            }

                            if (item.children && item.children.length) {
                                moveChildren($('#' + luxUtils.encodeId(selectedItem.attr.id)), item.children);
                                // This is a parent node, so we reset the (partial).
                                // This is to fix the case where a partial nested parent label (partial)
                                // not getting updated when moving its parent node.
                                self._setParentTitle(item.action.replace(SHUTTLE.SELECTED_ID_PREFIX, ''));
                            }

                            oj.Logger.info('***** moved child', selectedItem);
                        });
                    };

                oj.Logger.info('***** checking for filtered children');

                if (children && children.length) {
                    // iterate over children, if one child has already been
                    // move to the target list, the rest of the children will
                    // not be moved.
                    _.each(children, function(child) {
                        if (_.contains(path, child.action)) {
                            moveTheChildren = false;
                        }
                    }, this);

                    oj.Logger.info('*****', sourceId, 'has filtered children');

                    if (moveTheChildren) {
                        moveChildren($('#' + luxUtils.encodeId(targetId)), children);
                    }
                }
            },

            _isFiltered: function(id, nodeName, filter) {
                return (nodeName.toLowerCase().indexOf(filter.toLowerCase()) > -1);
            },

            _moveAllButtonClick: function(/*event*/) {
                var self = this,
                    availableOptionsList = self._availableOptionsContainer,
                    availableList = availableOptionsList.ojTree('getChildren', -1), // get top level nodes
                    availableListArray = _.map(availableList, function(item) {
                        return item;
                    }, this);

                if (availableList.length === 0) {
                    return;
                }

                self._availableValue = [];

                availableListArray.reverse().forEach(function(item) {
                    self._availableValue.push(item.id);
                }, this);

                self._availableValue = self._availableValue.join(', ');
                this._moveButtonClick();

                this._saveSelections();

                self._setItemCountDisplay();
            },

            _removeAllButtonClick: function(/*event*/) {
                var self = this,
                    selectedOptionsList = self._selectedOptionsContainer,
                    selectedList = selectedOptionsList.ojTree('getChildren', -1), // get top level nodes
                    selectedListArray = _.map(selectedList, function(item) {
                        return item;
                    }, this);

                if (selectedList.length === 0) {
                    return;
                }

                self._selectedValue = [];

                selectedListArray.reverse().forEach(function(item) {
                    self._selectedValue.push(item.id);
                }, this);

                self._selectedValue = self._selectedValue.join(', ');
                this._removeButtonClick();

                this._saveSelections();

                self._setItemCountDisplay();
            },

            _moveUpButtonClick: function() {
                this._moveSort(SHUTTLE.MOVE_UP);
            },

            _moveToTopButtonClick: function() {
                this._moveToEnds(SHUTTLE.MOVE_TOP);
            },

            _moveDownButtonClick: function() {
                this._moveSort(SHUTTLE.MOVE_DOWN);
            },

            _moveToBottomButtonClick: function() {
                this._moveToEnds(SHUTTLE.MOVE_BOTTOM);
            },

            _moveSort: function(direction) {
                var self = this,
                    selectedOptionsList = self._selectedOptionsContainer,
                    selectedId = self._selectedOptionsContainer[0].id,
                    position = SHUTTLE.MOVE_LOCATION_BEFORE,
                    siblingMethod = 'getPrevSibling',
                    selectedNodeArray;

                if (_.isNull(self._selectedValue)) {
                    return;
                }

                selectedNodeArray = self._selectedValue.split(', ');

                if (direction === SHUTTLE.MOVE_DOWN) {
                    selectedNodeArray.reverse();
                    position = SHUTTLE.MOVE_LOCATION_AFTER;
                    siblingMethod = 'getNextSibling';
                }

                selectedNodeArray.forEach(function(item) {
                    var selectedNode = $('#' + luxUtils.encodeId(item)),
                        sibling = selectedOptionsList.ojTree(siblingMethod, selectedNode);

                    if (sibling) {
                        selectedOptionsList.ojTree('move', selectedNode, sibling, position);
                    }

                    if (self.options.filterLists && sibling) {
                        // move the node relative the target destination
                        self._dataTree[selectedId].moveNode(item, sibling[0].id);
                        // update the new node configuration to the selected shadow list
                        self._shadowLists[selectedId] = self._dataTree[selectedId].toJSON();
                    }

                }, this);

                this._saveSelections();
            },

            _moveToEnds: function(direction) {
                var self = this,
                    selectedOptionsList = self._selectedOptionsContainer,
                    selectedId = self._selectedOptionsContainer[0].id,
                    position = SHUTTLE.MOVE_LOCATION_AFTER,
                    selectedNodeArray;

                if (_.isNull(self._selectedValue)) {
                    return;
                }

                selectedNodeArray = self._selectedValue.split(', ');

                if (direction === SHUTTLE.MOVE_TOP) {
                    selectedNodeArray.reverse();
                    position = SHUTTLE.MOVE_LOCATION_BEFORE;
                }

                selectedNodeArray.forEach(function(item) {
                    var selectedNode = $('#' + luxUtils.encodeId(item)),
                        nodes = selectedOptionsList.ojTree('getParent', selectedNode);

                    if (!nodes) { // no parent
                        nodes = selectedOptionsList.ojTree('getChildren', -1);
                        if (direction === SHUTTLE.MOVE_TOP) {
                            nodes = nodes[0];
                        } else {
                            nodes = nodes[nodes.length - 1];
                        }
                    } else { // has parent
                        nodes = selectedOptionsList.ojTree('getChildren', nodes);
                        if (direction === SHUTTLE.MOVE_TOP) {
                            nodes = nodes[0];
                        } else {
                            nodes = nodes[nodes.length - 1];
                        }
                    }

                    selectedOptionsList.ojTree('move', selectedNode, nodes, position);

                    if (self.options.filterLists && nodes) {
                        // move the node relative the target destination
                        self._dataTree[selectedId].moveNode(item, nodes.id);
                        // update the new node configuration to the selected shadow list
                        self._shadowLists[selectedId] = self._dataTree[selectedId].toJSON();
                    }

                }, this);

                this._saveSelections();
            },

            _saveSelections: function() {
                var self = this,
                    selectedId = self._selectedOptionsContainer[0].id,
                    topLevelNodes = self._selectedOptionsContainer.ojTree('getChildren', -1),
                    output = [],
                    getChildren = function(parent, children) {

                        _.each(children, function(item) {
                            var nodeItem = $('#' + luxUtils.encodeId(item.id)),

                                selectedItem = {},
                                childNodes = self._selectedOptionsContainer.ojTree('getChildren', nodeItem);

                            selectedItem.title = nodeItem.find('.oj-tree-title').first().text();
                            selectedItem.attr = self._copyAttributes(nodeItem);
                            selectedItem.attr.id = item.id.replace(SHUTTLE.SELECTED_ID_PREFIX, '');

                            if (childNodes) {
                                selectedItem.children = [];
                                getChildren(selectedItem.children, childNodes);
                            }

                            parent.push(selectedItem);

                        }, self);
                    };

                if (!self.options.onSaveCallback || typeof self.options.onSaveCallback !== 'function') {
                    return;
                }

                if (self.options.filterLists && self._shadowLists) {
                    self.options.onSaveCallback(self._cloneDataList(self._shadowLists[selectedId]));
                    return;
                }

                getChildren(output, topLevelNodes);

                self.options.onSaveCallback(output);
            },

            /**
             * Only with the filter option set. Handles data input to the available list filter
             * @param {object} data - not used
             * @param {object} event - dom event
             * @return {void}
             * @private
             */
            _handleAvailableFilterKeyUp: function(data, event) {
                var self = this,
                    availableId = self._availableOptionsContainer[0].id; // get the id form the DOM

                self._availableFilterText(event.currentTarget.value);
                self._filteredListItems[availableId] = {
                    'listName':       self.options.availableValuesHeading,
                    'headingKey':     'availableValuesHeading',
                    'text':           self._availableFilterText,
                    'contanterClass': SHUTTLE.AVAILABLE_CONTAINER_CLASS,
                    'items':          []};
                //oj.Logger.info('************ Available filter keyup', self._availableFilterText.peek());

                self._handleFilteredList(availableId, self._filteredListItems[availableId].text, self._filteredListItems[availableId].contanterClass);
            },

            /**
             * Only with the filter option set. Handles data input to the selected list filter
             * @param {object} data - not used
             * @param {object} event - dom event
             * @return {void}
             * @private
             */
            _handleSelectedFilterKeyUp: function(data, event) {
                var self = this,
                    selectedId = self._selectedOptionsContainer[0].id; // get the id form the DOM

                self._selectedListFilterText(event.currentTarget.value);
                self._filteredListItems[selectedId] = {
                    'listName':       self.options.selectedValuesHeading,
                    'headingKey':     'selectedValuesHeading',
                    'text':           self._selectedListFilterText,
                    'contanterClass': SHUTTLE.SELECTED_CONTAINER_CLASS,
                    'items':          []
                };
                //oj.Logger.info('************ Selected filter keyup', self._selectedListFilterText.peek());

                self._handleFilteredList(selectedId, self._filteredListItems[selectedId].text, self._filteredListItems[selectedId].contanterClass);
            },

            /**
             * Only with the filter option set. Filters the list based on input from handleSelectedFilterKeyUp and
             * handleAvailableFilterKeyUp functions
             * @param {string} listId - id of list to be filtered
             * @param {string} filterText - text for filter against
             * @param {string} highlightClass - css class to style the highlighting
             * @return {void}
             * @private
             */
            _handleFilteredList: function(listId, filterText, highlightClass) {
                var self = this,
                    el = $('#' + luxUtils.encodeId(listId)),
                    filterList,
                    filter = filterText.peek();

                filterList = self._filterList(self._shadowLists[listId], filter);
                oj.Logger.info('List after filter:', self._shadowLists[listId]);
                // TODO: this could be JET issue. This ojTree call would cause an exception
                // if previous selected node has id with special character. Debugged and found
                // ojTree was trying to re-select the previous removed node. Workaround is to
                // deselect the selection when the selected node is removed (see TODO/LUX-1591 above).
                el.ojTree('option', {'data': filterList});
                el.ojTree('refresh');
                self._highlightFiltered(filter, highlightClass);
                self._filterNotItemsMessage(listId, filterList);
                self._setItemCountDisplay();
            },

            _checkForEmptySourceList: function(listId, filterText) {
                var self = this,
                    filterList,
                    filter;

                filter = filterText && filterText.peek();
                if (!filter) {
                    return;
                }

                filterList = self._filterList(self._shadowLists[listId], filter);
                self._filterNotItemsMessage(listId, filterList);
            },

            _filterNotItemsMessage: function(ListId, filterList) {
                var translations = oj.Translations,
                    msg = translations.getTranslatedString('noItemsMatchMsg'),
                    targetList = $('#' + luxUtils.encodeId(ListId));

                if (!msg || msg === 'noItemsMatchMsg') {
                    msg = SHUTTLE.NO_ITEMS_MATCH_MSG;
                }

                if (filterList.length === 0 && !(document.getElementById(ListId + '-empty'))) {
                    $('<div /> ', {id: ListId + '-empty', class: 'lux-no-filtered-items', text: msg}).appendTo(targetList);
                } else {
                    if (filterList.length > 0 && document.getElementById(ListId + '-empty')) {
                        document.getElementById(ListId + '-empty').remove();
                    }
                }
            },

            _filterMoveWarning: function(listId) {
                var self = this,
                    dialogEl = $('#filterWarningDialog'),
                    filterTextObj = self._filteredListItems[listId],
                    filter = filterTextObj.text.peek(),
                    items = filterTextObj.items,
                    listName  = self.options[filterTextObj.headingKey],
                    filteredOut,
                    msg, title,
                    warningDiv = document.getElementById('filterWarningText'),
                    translations = oj.Translations,
                    filterNode = function(nodes) {
                        return _.reject(nodes, function(node) {

                            if (node.actionLabel.toLowerCase().indexOf(filter.toLowerCase()) < 0) {
                                node.data.listed = true;
                            } else if (node.children && node.children.length) {
                                filterNode(node.children);
                            } else {
                                return node;
                            }
                        });
                    };

                filteredOut = filterNode(items);

                if (filteredOut.length) {
                    msg = translations.getTranslatedString('dialogMessage', {'listName': listName});
                    if (!msg || msg === 'dialogMessage') {
                        msg = translations.applyParameters(SHUTTLE.DIALOG_MESSAGE, {'listName': listName});
                    }
                    title = translations.getTranslatedString('dialogTitle');
                    if (!title || title === 'dialogTitle') {
                        title = SHUTTLE.DIALOG_TITLE;
                    }
                    warningDiv.innerHTML = msg;
                    dialogEl.attr({'title': title});
                    dialogEl.ojDialog('option', 'title', title);
                    dialogEl.ojDialog('refresh');
                    dialogEl.ojDialog('open');
                    if (self.options.filterWarningDialogTimer && _.isNumber(self.options.filterWarningDialogTimer)) {
                        _.delay(function() {
                            $('#filterWarningDialog').ojDialog('close');
                        }, self.options.filterWarningDialogTimer);
                    }
                }
            },

            /**
             * Only with the filter option set. Filters the list.
             * @param {object} nodes
             * @param {string} filter
             * @returns {Array} the filtered list of node items
             * @private
             */
            _filterList: function(nodes, filter) {
                var self = this,
                    result = [], clone, children;

                _.each(nodes, function(node) {
                    // find titles that match the filter
                    if (node.title.toLowerCase().indexOf(filter.toLowerCase()) > -1) {
                        clone = self._cloneNode(node);
                        if (node.children) {
                            children = self._filterList(node.children, filter);
                            if (children.length > 0) {
                                clone.children = children;
                            }
                        }
                        result.push(clone);
                    } else {
                        // check the children of the node, some might
                        // included in the filter
                        if (node.children) {
                            children = self._filterList(node.children, filter);
                            if (children.length > 0) {
                                clone = self._cloneNode(node);
                                clone.children = children;
                                result.push(clone);
                            }
                        }
                    }
                });

                return result;
            },

            _cloneDataList: function(nodes) {
                var self = this,
                    result = [], clone, children;

                _.each(nodes, function(node) {
                    clone = self._cloneNode(node);
                    if (node.children) {
                        children = self._cloneDataList(node.children);
                        if (children.length > 0) {
                            clone.children = children;
                        }
                    }
                    clone.attr.id = clone.attr.id.replace(SHUTTLE.SELECTED_ID_PREFIX, '');
                    result.push(clone);
                });

                return result;
            },

            _cloneNode: function(node) {
                return {'title': node.title, 'attr': $.extend({}, node.attr)};
            },

            _highlightFiltered: function(filter, listSelector) {
                var treeList = document.getElementsByClassName(listSelector)[0],
                    filteredList = treeList.getElementsByClassName('oj-tree-title');

                _.each(filteredList, function(el) {
                    var text = el.innerHTML;

                    if (filter.length && text.toLowerCase().indexOf(filter.toLowerCase()) > -1) {
                        el.classList.add('lux-filter-highlight');
                    } else {
                        el.classList.remove('lux-filter-highlight');
                    }
                });
            },

            _destroy: function() {
                this.element.remove();
            },

            _copyAttributes: function(source) {
                var target = {};

                _.each(source[0].attributes, function(item) {
                    if (_.contains(this._ignoredAttributes, item.name)) {
                        return;
                    }
                    target[item.name] = item.value;

                }, this);

                return target;
            },

            _changePartialLabelForLocale: function() {
                var self = this,
                    selectedOptions = self._selectedOptionsContainer,
                    selectedItemList = selectedOptions.ojTree('getRoot'), // get root
                    selectedId = self._selectedOptionsContainer[0].id,
                    availableId = self._availableOptionsContainer[0].id,
                    availableNodeId,
                    getSelected = function(selected) {
                        _.each(selected, function(item) {
                            var childNodes,
                                availableNode,
                                title,
                                nodeItem = $('#' + luxUtils.encodeId(self.options.selectedListId) + ' #' + luxUtils.encodeId(item.id));

                            if (nodeItem.attr('isselected') && nodeItem.attr('isselected') === 'partial') {
                                // get the associated available list item.
                                availableNodeId = luxUtils.encodeId(item.id.replace(SHUTTLE.SELECTED_ID_PREFIX, ''));
                                availableNode = $('#' + luxUtils.encodeId(self.options.availableListId) + ' #' + availableNodeId);

                                // update available list item partial label.
                                title = availableNode.find('.oj-tree-title').first();
                                title.text(title.text().substring(0, title.text().search(/\(/) - 1) + ' ' + self.options.partialLabel);

                                // now update selected list item partial label.
                                title = nodeItem.find('.oj-tree-title').first();
                                title.text(title.text().substring(0, title.text().search(/\(/) - 1) + ' ' + self.options.partialLabel);

                                if (self.options.filterLists) {
                                    if (self._dataTree[selectedId]) {
                                        self._dataTree[selectedId].changePartialLabelForLocale(item.id, self.options.partialLabel);
                                    }
                                    if (self._dataTree[availableId]) {
                                        self._dataTree[availableId].changePartialLabelForLocale(availableNodeId, self.options.partialLabel);
                                    }
                                }

                                if (self._hasChildren(nodeItem)) {
                                    childNodes = selectedOptions.ojTree('getChildren', nodeItem);
                                    getSelected(childNodes);
                                }
                            }
                        }, this);
                    };

                if (selectedItemList[0].id.length === 0) { // no root so get top level nodes
                    selectedItemList = selectedOptions.ojTree('getChildren', -1); // get top level nodes
                }

                getSelected(selectedItemList);

                if (self.options.filterLists) {
                    self._shadowLists[selectedId] = self._dataTree[selectedId].toJSON();
                    self._shadowLists[availableId] = self._dataTree[availableId].toJSON();
                }
            },

            _refreshButtonLabel: function(buttonLabel, value) {
                var self = this,
                    buttonItem;

                buttonItem = _.find(self._buttonRefreshList, function(item) {
                    return item.label === buttonLabel;
                });

                self['_' + buttonItem.button].ojButton('option', {'label': value});
                self['_' + buttonItem.button].ojButton('refresh');
            },

            _changeLabels: function(key, value) {
                var self = this;

                if (self._observablesMap[key]) {
                    self.options[key] = value;
                    if (key === 'partialLabel') {
                        self._changePartialLabelForLocale();
                    } else {
                        if (key === 'itemCountLabel') {
                            self._setItemCountDisplay();
                            return;
                        }
                        self._observablesMap[key].text(value);
                        if (key.search('ButtonLabel') > 0) {
                            self._refreshButtonLabel(key, value);
                        }
                        if (key.search('DescriptionHeading') > 0) {
                            // update the aria-label attribute on the text input
                            self._changeAriaLabels(key, value);
                        }
                    }
                } else if (self._accessibilityDivs[key]) {
                    self._changeButtonAccessibility(key, value);
                }

                if (self.options.filterLists) {
                    self._changeNoItemsFoundFilterMsg();
                    // update the filter placeholder text for
                    // the locale change
                    if (self._filterPlaceHolderMap[key]) {
                        self._changeFilterPlaceHolder(key, value);
                    }
                }
            },

            _changeNoItemsFoundFilterMsg: function() {
                var self = this,
                    translations = oj.Translations,
                    msg = translations.getTranslatedString('noItemsMatchMsg') || SHUTTLE.NO_ITEMS_MATCH_MSG;

                if (document.getElementById(self.options.availableListId + '-empty')) {
                    document.getElementById(self.options.availableListId + '-empty').textContent = msg;
                }

                if (document.getElementById(self.options.selectedListId + '-empty')) {
                    document.getElementById(self.options.selectedListId + '-empty').textContent = msg;
                }
            },

            _changeButtonAccessibility: function(key, value) {
                var self = this;

                $('#' + luxUtils.encodeId(self._accessibilityDivs[key])).text(value);
            },

            _changeAriaLabels: function(key, value) {
                var self = this,
                    inputField = $(self._observablesMap[key]).attr('for');

                inputField = $('#' + luxUtils.encodeId(inputField));
                inputField.attr('aria-label', value);
            },

            /**
             * _reloadLists: Private function that reloads one or the other lists.
             * @param {string} key - list option to be reloaded.
             * @param {string} value - represents an array of JSON objects
             * @return {*}
             */
            _reloadLists: function _reloadLists(key, value) {
                var self = this;

                if (key === 'availableValues') {
                    self.options.availableValues = value;
                    self._availableOptionsContainer.ojTree('option', {data: {data: self.options.availableValues || self._emptyAvailableOptions}});
                    self._setItemCountDisplay();

                    if (self.options.selectedValues) {
                        self._setSelectedValuesAttributes();
                    } else {
                        self._selectedOptionsContainer.ojTree('option', {data: {data: self._getSelectedValues()}});
                        self._setItemCountDisplay();
                    }
                }
                if (key === 'selectedValues') {
                    self.options.selectedValues = value;
                    self._selectedOptionsContainer.ojTree('option', {data: {data: self.options.selectedValues || self._getSelectedValues()}});
                    self._setItemCountDisplay();
                }

                // update labels for the locale
                self._changePartialLabelForLocale();

                if (self.options.filterLists) {
                    self._initializeFilteredLists();
                    self._saveSelections();
                    if (self._availableFilterText && self._availableFilterText.peek()) {
                        self._handleAvailableFilterKeyUp(null, {currentTarget: {value: self._availableFilterText()}});
                    }
                    if (self._selectedListFilterText && self._selectedListFilterText.peek()) {
                        self._handleSelectedFilterKeyUp(null, {currentTarget: {value: self._selectedListFilterText()}});
                    }
                }
            },

            _changeFilterPlaceHolder: function(key, value) {
                var self = this;

                $(self._observablesMap[self._filterPlaceHolderMap[key]]).ojInputText({'placeholder': value});
            },

            /**
             * _setOption: evaluates the options and calls the
             *             appropriate function to handle change
             *             event.
             * @param {string} key - key of the changed option
             * @param {string} value - new value of the option
             * @return {*}
             */
            _setOption: function(key, value) {
                var self = this;

                if (key === 'allowDnD') {
                    self.options.allowDnD = value;
                    self._dndOption();
                } else {
                    if (key === 'availableValues' || key === 'selectedValues') {
                        self._reloadLists(key, value);
                    }

                    self._changeLabels(key, value);
                }
            },

            /**
             * _setOptions: iterates over all changed options and passes
             *              key values to setOption for evaluation.
             *
             * @param {array} options - array of options
             * @return {*}
             */
            _setOptions: function(options) {
                var self = this;

                $.each(options, function(key, value) {
                    self._setOption(key, value);
                });

            },

            _emptyAvailableOptions: function(node, fn) {
                var data = [];

                fn(data);
            }

        });

    }());

});
