/**
 * Copyright (c) 2016 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implements a toolbar widget to use with LUX data grid.
 * @author Daniel Hoang
 */
/**
 * The luxToolbarGrid widget
 * @module luxToolbarGrid
 * @jqbase
 * @requires ojet
 * @requires jquery
 * @requires lodash
 * @requires knockout
 * @requires jqueryui
 * @requires oj-lux-toolbar
 * @requires luxCustomBindings
 */
define(['ojs/ojcore',
    'jquery',
    'lodash',
    'knockout',
    'text!./view.tmpl.html',
    './viewModel',
    'jqueryui-amd/widget',
    'ojs/ojdialog',
    'ojs/ojcomposite',
    'oj-lux/oj-lux-toolbar/loader',
    'luxCustomBindings'
],

function(oj, $, _, ko, template, ViewModel) {
    'use strict';

    $.widget('GBU.luxToolbarGrid',
        /** @lends module:luxToolbarGrid */
        {
            version: '5.1.0',

            /**
                 * Options
                 * <dl>
                 *     <dt>gridId</dt>
                 *     <dd>
                 *         Specify a grid element ID that is required for the toolbar to bind with.
                 *     </dd>
                 *     <dt>showTopBorder</dt>
                 *     <dd>
                 *         Boolean value indicating whether to show the border above the toolbar. Default is false.
                 *     </dd>
                 *     <dt>showBottomBorder</dt>
                 *     <dd>
                 *         Boolean value indicating whether to show the border below the toolbar. Default is false.
                 *     </dd>
                 *     <dt>translations</dt>
                 *     <dd>
                 *         Option for translation string overrides.
                 *     </dd>
                 *     <dt>customize</dt>
                 *     <dd>
                 *         <div class='row-fluid'>
                 *             <div class='span6'>
                 *                 <ul>
                 *                     <li>
                 *                         <i>Description</i>:
                 *                         (Optional) Callback that application can provide to customize the toolbar. Toolbar will trigger
                 *                         <code class="prettyprint">customize</code> callback before the toolbar widget is rendered.
                 *                     </li>
                 *                     <li><i>Arguments</i>:
                 *                         <ul>
                 *                             <li><b>event</b>: jQuery event object</li>
                 *                             <li><b>data</b>: Object containing the following toolbar customizable objects:
                 *                                 <ul>
                 *                                     <li><b>settings</b>: Object containing toolbar settings
                 *                                         <ul>
                 *                                             <li><b>defaultColumnFilterType</b>: 'columns' or 'text'. Default column filter type is 'text'.</li>
                 *                                             <li><b>maxColumnSortCriteria</b>: Maximum number of the column sort selects generated in the sort dialog. Default is 3 (max number is 9)</li>
                 *                                         </ul>
                 *                                     </li>
                 *                                     <li><b>items</b>: Object containing the pre-defined toolbar items
                 *                                         <ul>
                 *                                             <li><b>viewMenuButton</b>: the 'View' menu button, also requires viewMenu.</li>
                 *                                             <li><b>viewMenu</b>: the 'View' menu component, also requires viewMenuItems.</li>
                 *                                             <li><b>viewMenuItems</b>: Object containing the 'View' menu items
                 *                                                 <ul>
                 *                                                     <li><b>viewMenuItems_detach</b>: 'Detach' menu item</li>
                 *                                                     <li><b>viewMenuItems_sort</b>: 'Sort...' menu item</li>
                 *                                                     <li><b>viewMenuItems_preferences</b>: 'Preferences...' menu item</li>
                 *                                                 </ul>
                 *                                             </li>
                 *                                             <li><b>separator</b>: separator component for the toolbar</li>
                 *                                             <li><b>addRowButton</b>: 'Add Row(s)' button</li>
                 *                                             <li><b>duplicateRowButton</b>: 'Duplicate Row(s)' button</li>
                 *                                             <li><b>editRowButton</b>: 'Edit Row' button</li>
                 *                                             <li><b>viewRowButton</b>:'View Row' button</li>
                 *                                             <li><b>deleteRowButton</b>: 'Delete Row(s)' button</li>
                 *                                             <li><b>separator</b>: separator component for the toolbar</li>
                 *                                             <li><b>refreshButton</b>: 'Refresh' button</li>
                 *                                             <li><b>exportMenuButton</b>: the 'Export' menu button, also requires exportMenu.</li>
                 *                                             <li><b>exportMenu</b>: the 'Export' menu component, also requires exportMenuItems.</li>
                 *                                             <li><b>exportMenuItems</b>: Object containing the 'Export' menu items
                 *                                                 <ul>
                 *                                                     <li><b>exportMenuItems_csv</b>: 'CSV' export menu item</li>
                 *                                                     <li><b>exportMenuItems_excel</b>: 'Excel' export menu item</li>
                 *                                                 </ul>
                 *                                             </li>
                 *                                             <li><b>detachToggleButton</b>:'Detach' toggle button</li>
                 *                                             <li><b>separator</b>: separator component for the toolbar</li>
                 *                                             <li><b>filterTypeRadioButton</b>: 'Columns'/'Text' radio button</li>
                 *                                             <li><b>filterInputText</b>: Input text field for 'Text' filter type</li>
                 *                                         </ul>
                 *                                     </li>
                 *                                     <li><b>appRequiredActions</b>: Object containing the action handlers that application can provide to handle application-specific push button actions (e.g. add, view, edit, refresh buttons)
                 *                                         <ul>
                 *                                             <li><b>addRowButton</b>: a callback function to handle when 'Add Row(s)' button is clicked. This is mainly to change 'Add Row(s)' pre-defined functionality into a 'Create Row' functionality.</li>
                 *                                             <li><b>editRowButton</b>: a callback function to handle when 'Edit Row' button is clicked</li>
                 *                                             <li><b>viewRowButton</b>: a callback function to handle when 'View Row' button is clicked</li>
                 *                                             <li><b>refreshButton</b>: a callback function to handle when 'Refresh Row' button is clicked</li>
                 *                                         </ul>
                 *                                     </li>
                 *                                     <li><b>appDisabledStates</b>: Object containing the boolean-type observables that the application can provide to override the pre-defined push button disabled states
                 *                                         <ul>
                 *                                             <li><b>addRowButton</b>: a boolean-type observable to control the disabled state of 'Add Row(s)' button</li>
                 *                                             <li><b>duplicateRowButton</b>: a boolean-type observable to control the disabled state of 'Duplicate Row(s)' button</li>
                 *                                             <li><b>editRowButton</b>: a boolean-type observable to control the disabled state of 'Edit Row' button</li>
                 *                                             <li><b>viewRowButton</b>: a boolean-type observable to control the disabled state of 'View Row' button</li>
                 *                                             <li><b>deleteRowButton</b>: a boolean-type observable to control the disabled state of 'Delete Row(s)' button</li>
                 *                                             <li><b>refreshButton</b>: a boolean-type observable to control the disabled state of 'Refresh' button</li>
                 *                                         </ul>
                 *                                     </li>
                 *                                 </ul>
                 *                              </li>
                 *                         </ul>
                 *                      </li>
                 *                 </ul>
                 *             </div>
                 *         </div>
                 *     </dd>
                 *     <dt>menuItemSelect</dt>
                 *     <dd>
                 *         <div class='row-fluid'>
                 *             <div class='span6'>
                 *                 <ul>
                 *                     <li>
                 *                         <i>Description</i>:
                 *                         (Optional) Callback that application can provide if it adds custom menu items to the pre-defined toolbar menus.
                 *                         Toolbar will trigger <code class="prettyprint">menuItemSelect</code> callback when any custom menu
                 *                         item in the pre-defined toolbar menus is selected.
                 *                     </li>
                 *                     <li><i>Arguments</i>:
                 *                         <ul>
                 *                             <li><b>event</b>: jQuery event object</li>
                 *                             <li><b>ui</b>: Object containing the selected menu <code class="prettyprint">item</code>
                 *                             </li>
                 *                         </ul>
                 *                      </li>
                 *                 </ul>
                 *             </div>
                 *         </div>

                 *     </dd>
                 * </dl>
                 */
            options: {
                showTopBorder:    false,
                showBottomBorder: false,
                gridId:           undefined,
                translations:     undefined,
                customize:        $.noop,
                menuItemSelect:   function(event, ui) {
                    throw new ReferenceError('No valid selection handler for customized toolbar menu item "' + ui.item[0].id + '"');
                }
            },

            /*
                 * The widget factory automatically fires the _create() and _init() methods during initialization, in that order.
                 * _create() will be called one time for each widget instance, _init() will be called each time the widget is called
                 * without arguments. All widgets have the following properties that we may use here:
                 * document: A jQuery object containing the document that the widget's element is within. Useful if you need to interact with widgets within iframes.
                 * element: A jQuery object containing the element used to instantiate the widget
                 * options: An object containing the options currently being used by the widget. On instantiation, any options provided
                 * by the user will automatically be merged with any default values defined in widget.prototype.options.
                 * uuid: unique integer identifier for the widget
                 */
            _create: function() {

                if (!this.options.gridId) {
                    throw new ReferenceError('Grid toolbar requires a gridId!');
                }

                // _create will automatically run the first time this widget is called.
                // Put the initial widget set-up code here, then we can access the element
                // on which the widget was called via this.element. The options defined above
                // can be accessed via this.options.

                // Set up widget (e.g. create element, apply themeing, bind events, etc.)
                var data = {},
                    self = this;

                    // Create widget's customizable data objects.
                data.items = ViewModel.createPredefinedToolbarItems();
                data.settings = ViewModel.createToolbarSettings();
                data.appRequiredActions = ViewModel.createAppSpecificActions();
                data.appDisabledStates = ViewModel.createAppSpecificDisabledStates();

                // Give user a chance to customize the toolbar ...
                self._trigger('customize', null, data);

                // Filter out those items (View/Edit/Refresh) that app does not provide action handlers.
                data.items = data.items.filter(function(item) {
                    if (item.viewRowButton && !data.appRequiredActions.viewRowButton) {
                        return (false);
                    }
                    if (item.editRowButton && !data.appRequiredActions.editRowButton) {
                        return (false);
                    }
                    if (item.refreshButton && !data.appRequiredActions.refreshButton) {
                        return (false);
                    }
                    return (true);
                });

                // Save data objects for view model to use.
                self.options.data = data;

                // Since toolbar widget depends on a grid instance, we would initialize
                // and render the toolbar when page is ready (that is, grid widget should already
                // been instantiated when page is ready). For more information, see
                // https://api.jquery.com/ready/
                $(function() {
                    var isCCA, jqwidgetContainer, $gridElement,
                        viewTemplate = $(template),
                        rootView = viewTemplate.get(0),
                        toolbarView = viewTemplate.find('#lux-grid-toolbar-wrapper').get(0);

                        // application may pass in a reference to a grid's element ID
                    isCCA = false;
                    $gridElement = $('#' + self.options.gridId);
                    // Check if we can find a grid widget reference, if coming from a CCA.
                    jqwidgetContainer = $gridElement.find('.jqwidget-container');
                    isCCA = jqwidgetContainer.length > 0;

                    // Set the reference to the grid object
                    self.options.grid = isCCA ? jqwidgetContainer : $gridElement;

                    self.viewModel = new ViewModel(self);

                    // Create the view from template
                    self.element.append(rootView);
                    // Apply the model binding to the toolbar view
                    ko.applyBindings(self.viewModel, toolbarView);

                    self.viewModel.init();
                });
            },

            /**
                 * Refreshes the toolbar widget component (this method basically calls the ojToolbar
                 * <a href="http://jet.us.oracle.com/oj-components/jsdocs/oj.ojToolbar.html#refresh">refresh()</a> method).
                 * @method module:luxToolbarGrid#refresh
                 * @public
                 */
            refresh: function() {
                this.viewModel.refresh();
            },

            /*
                 * The public destroy() method cleans up all common data, events, etc. and then delegates out to _destroy()
                 * for custom, widget-specific, cleanup.
                 */
            _destroy: function() {
                this.viewModel.destroy();
                this.viewModel = undefined;
            },

            _setOption: function(key, value) {
                this._super(key, value);
            },

            _setOptions: function(options) {
                var self = this;
                $.each(options, function(key, value) {
                    self._setOption(key, value);
                });
            }
        });
}
);
