/*
 * Copyright (c) 2016 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implements the view model for the toolbar widget.
 * @author Daniel Hoang
 */
define(['ojs/ojcore',
    'jquery',
    'lodash',
    'knockout',
    'ojL10n!./nls/translations',
    'ojs/ojdialog',
    'ojs/ojradioset',
    'ojs/ojcomposite',
    'ojs/ojtoolbar',
    'oj-lux/oj-lux-toolbar/loader',
    'luxCustomBindings'
],

function(oj, $, _, ko, defaultTranslations) {
    'use strict';

    var translations = _.merge({}, defaultTranslations);

    // private constants
    var TOOLBAR_DEFAULT_ELEMENT_ID = 'lux-grid-toolbar';
    var TOOLBAR = {
        DEFAULT_ELEMENT_ID:      TOOLBAR_DEFAULT_ELEMENT_ID,
        ATTACH_WRAPPER_ID:       TOOLBAR_DEFAULT_ELEMENT_ID + '-attach-wrapper',
        DETACH_WRAPPER_ID:       TOOLBAR_DEFAULT_ELEMENT_ID + '-detach-wrapper',
        DETACH_DIALOG_ID:        TOOLBAR_DEFAULT_ELEMENT_ID + '-detach-dialog',
        DETACH_DIALOG_TITLE:     'Detach Dialog for Grid Table',
        DETACH_DIALOG_ANCHOR_ID: TOOLBAR_DEFAULT_ELEMENT_ID + '-detach-dialog-anchor',
        GRID_WRAPPER_ID:         TOOLBAR_DEFAULT_ELEMENT_ID + '-grid-wrapper',

        BUTTON_VIEW_MENU_ID:    TOOLBAR_DEFAULT_ELEMENT_ID + '-view-menu-button',
        BUTTON_VIEW_MENU_LABEL: translations.viewLabel,
        BUTTON_ADD_ROW_ID:      TOOLBAR_DEFAULT_ELEMENT_ID + '-add-row',
        BUTTON_ADD_ROW_LABEL:   translations.addRowLabel,
        BUTTON_DUP_ROW_ID:      TOOLBAR_DEFAULT_ELEMENT_ID + '-dup-row',
        BUTTON_DUP_ROW_LABEL:   translations.duplicateRowLabel,
        BUTTON_EDIT_ROW_ID:     TOOLBAR_DEFAULT_ELEMENT_ID + '-edit-row',
        BUTTON_EDIT_ROW_LABEL:  translations.editRowLabel,
        BUTTON_VIEW_ROW_ID:     TOOLBAR_DEFAULT_ELEMENT_ID + '-view-row',
        BUTTON_VIEW_ROW_LABEL:  translations.viewRowLabel,
        BUTTON_DEL_ROW_ID:      TOOLBAR_DEFAULT_ELEMENT_ID + '-del-row',
        BUTTON_DEL_ROW_LABEL:   translations.deleteRowLabel,
        BUTTON_REFRESH_ID:      TOOLBAR_DEFAULT_ELEMENT_ID + '-refresh',
        BUTTON_REFRESH_LABEL:   translations.refreshLabel,

        BUTTON_EXPORT_MENU_ID:          TOOLBAR_DEFAULT_ELEMENT_ID + '-export-menu-button',
        BUTTON_EXPORT_MENU_LABEL:       translations.exportLabel,
        BUTTON_DETACH_ID:               TOOLBAR_DEFAULT_ELEMENT_ID + '-detach',
        BUTTON_DETACH_LABEL:            translations.detachLabel,
        BUTTON_ATTACH_LABEL:            translations.attachLabel,
        RADIO_FILTER_ID:                TOOLBAR_DEFAULT_ELEMENT_ID + '-filter',
        RADIO_FILTER_LABEL:             translations.filterLabel,
        INPUT_TEXT_FILTER_ID:           TOOLBAR_DEFAULT_ELEMENT_ID + '-filter-text',
        INPUT_TEXT_FILTER_PLACE_HOLDER: translations.filterTextPlaceHolderLabel
    };

    var VIEWMENU_DEFAULT_ELEMENT_ID = TOOLBAR_DEFAULT_ELEMENT_ID + '-view-menu';
    var VIEWMENU = {
        DEFAULT_ELEMENT_ID:   VIEWMENU_DEFAULT_ELEMENT_ID,
        MANAGE_COLUMNS_ID:    VIEWMENU_DEFAULT_ELEMENT_ID + '-manage-columns',
        MANAGE_COLUMNS_LABEL: translations.manageColumnsLabel,
        DETACH_ID:            VIEWMENU_DEFAULT_ELEMENT_ID + '-detach',
        DETACH_LABEL:         translations.detachLabel,
        ATTACH_LABEL:         translations.attachLabel,
        SORT_ID:              VIEWMENU_DEFAULT_ELEMENT_ID + '-sort',
        SORT_LABEL:           translations.sortLabel,
        PREFERENCES_ID:       VIEWMENU_DEFAULT_ELEMENT_ID + '-preferences',
        PREFERENCES_LABEL:    translations.preferencesLabel
    };

    var VIEWMENU_DIALOG = {
        MANAGE_COLUMNS_ID:                VIEWMENU.MANAGE_COLUMNS_ID + '-dialog',
        SORT_ID:                          VIEWMENU.SORT_ID + '-dialog',
        SORT_COLUMN_ID:                   VIEWMENU.SORT_ID + '-column',
        SORT_ORDER_ID:                    VIEWMENU.SORT_ID + '-order',
        SORT_ORDER_ASC_ID:                VIEWMENU.SORT_ID + '-order-asc',
        SORT_ORDER_DSC_ID:                VIEWMENU.SORT_ID + '-order-dsc',
        PREFERENCES_ID:                   VIEWMENU.PREFERENCES_ID + '-dialog',
        PREFERENCES_ADD_ROW_ID:           VIEWMENU.PREFERENCES_ID + '-add-row',
        PREFERENCES_ADD_ROW_BEFORE_ID:    VIEWMENU.PREFERENCES_ID + '-add-row-before',
        PREFERENCES_ADD_ROW_AFTER_ID:     VIEWMENU.PREFERENCES_ID + '-add-row-after',
        PREFERENCES_COLUMN_FILTER_ID:     VIEWMENU.PREFERENCES_ID + '-column-filter',
        PREFERENCES_COLUMN_FILTER_ANY_ID: VIEWMENU.PREFERENCES_ID + '-column-filter-any',
        PREFERENCES_COLUMN_FILTER_ALL_ID: VIEWMENU.PREFERENCES_ID + '-column-filter-all'
    };

    var EXPORTMENU_DEFAULT_ELEMENT_ID = TOOLBAR_DEFAULT_ELEMENT_ID + '-export-menu';
    var EXPORTMENU = {
        DEFAULT_ELEMENT_ID: EXPORTMENU_DEFAULT_ELEMENT_ID,
        CSV_ID:             EXPORTMENU_DEFAULT_ELEMENT_ID + '-csv',
        CSV_LABEL:          translations.csvLabel,
        EXCEL_ID:           EXPORTMENU_DEFAULT_ELEMENT_ID + '-excel',
        EXCEL_LABEL:        translations.excelLabel
    };

    var GRID = {
        TEXT_FILTER_TYPE:         'text',
        TEXT_FILTER_TYPE_LABEL:   translations.textLabel,
        COLUMN_FILTER_TYPE:       'columns',
        COLUMN_FILTER_TYPE_LABEL: translations.columnsLabel
    };

        // Expose a list of our pre-defined toolbar items as keys -> names mapping.
    var PREDEFINED_KEYS = {
        separator:                 'separator',
        viewMenuButton:            'viewMenuButton',
        viewMenu:                  'viewMenu',
        viewMenuItems:             'viewMenuItems',
        viewMenuItems_detach:      'detach',
        // viewMenuItems_manageColumns: 'manageColumns',
        viewMenuItems_sort:        'sort',
        viewMenuItems_preferences: 'preferences',
        addRowButton:              'addRowButton',
        duplicateRowButton:        'duplicateRowButton',
        editRowButton:             'editRowButton',
        viewRowButton:             'viewRowButton',
        deleteRowButton:           'deleteRowButton',
        refreshButton:             'refreshButton',
        exportMenuButton:          'exportMenuButton',
        exportMenu:                'exportMenu',
        exportMenuItems:           'exportMenuItems',
        exportMenuItems_csv:       'csv',
        exportMenuItems_excel:     'excel',
        detachToggleButton:        'detachToggleButton',
        filterTypeRadioButton:     'filterTypeRadioButton',
        filterInputText:           'filterInputText'
    };

        // This _toolbarItem is to get around the ES5 limitation that
        // it does not have the [computed property] as in ES6.
    var _toolbarItem = function(key, value) {
        var obj = {};
        obj[key] = value ? value : {};
        return (obj);
    };

    var _createPredefinedToolbarItems = function() {
        return [
            _toolbarItem(PREDEFINED_KEYS.viewMenuButton),
            _toolbarItem(PREDEFINED_KEYS.viewMenu),
            _toolbarItem(PREDEFINED_KEYS.viewMenuItems, [
                _toolbarItem(PREDEFINED_KEYS.viewMenuItems_detach),
                _toolbarItem(PREDEFINED_KEYS.viewMenuItems_sort),
                _toolbarItem(PREDEFINED_KEYS.viewMenuItems_preferences)
            ]),
            _toolbarItem(PREDEFINED_KEYS.separator),
            _toolbarItem(PREDEFINED_KEYS.addRowButton),
            _toolbarItem(PREDEFINED_KEYS.duplicateRowButton),
            _toolbarItem(PREDEFINED_KEYS.editRowButton),
            _toolbarItem(PREDEFINED_KEYS.viewRowButton),
            _toolbarItem(PREDEFINED_KEYS.deleteRowButton),
            _toolbarItem(PREDEFINED_KEYS.separator),
            _toolbarItem(PREDEFINED_KEYS.refreshButton),
            _toolbarItem(PREDEFINED_KEYS.exportMenuButton),
            _toolbarItem(PREDEFINED_KEYS.exportMenu),
            _toolbarItem(PREDEFINED_KEYS.exportMenuItems, [
                _toolbarItem(PREDEFINED_KEYS.exportMenuItems_csv),
                _toolbarItem(PREDEFINED_KEYS.exportMenuItems_excel)
            ]),
            _toolbarItem(PREDEFINED_KEYS.detachToggleButton),
            _toolbarItem(PREDEFINED_KEYS.separator),
            _toolbarItem(PREDEFINED_KEYS.filterTypeRadioButton),
            _toolbarItem(PREDEFINED_KEYS.filterInputText)
        ];
    };

        // Expose a list of support pre-defined actions that application can override.
    var _createAppSpecificActions = function() {
        return ({
            editRowButton: undefined,
            viewRowButton: undefined,
            refreshButton: undefined
        });
    };

        // List of application specific disabled states (Observable states defined by application)
    var _createAppSpecificDisabledStates = function() {
        return ({});
    };

        // List of toolbar default preference settings, and application can customize.
    var _createToolbarSettings = function() {
        return ({
            defaultColumnFilterType: 'text',
            maxColumnSortCriteria:   3
        });
    };

        /*
         * Implementation of a view model for the grid toolbar.
         */
    var viewModel = function ViewModel(widgetObj) {
        var _gridToolbarBuilder, _gridToolbarMenuItemsBuilder,
            _defineToolbarItems, _defineToolbarActions,
            _createBindingContext, _resetFilterType,
            _initFilterType, _detachHandler, _gridToolbarDisabledStates,
            _gridToolbarStates, _gridToolbarActions,
            _gridToolbarItems, _refreshSortableColumn, _toolbarSettings,
            self = this;

            // widget reference for the view model to use
        self.widgetObj = widgetObj;

        _toolbarSettings = self.widgetObj.options.data.settings;

        // Compute the actual button disabled states based on the application disabled state and toolbar button disabled state.
        // Application disabled states take precedence over the toolbar button disabled states.
        _gridToolbarDisabledStates = {
            addRowButton: ko.pureComputed(function() {
                var isGridEditable = self.grid.grid('option', 'editable'),
                    tbDisabled = _gridToolbarStates.addRowButton_disabled(),
                    appSpecificDisabledStates = self.widgetObj.options.data.appDisabledStates,
                    appDisabled = appSpecificDisabledStates.addRowButton !== undefined ? appSpecificDisabledStates.addRowButton() : false;
                self.refresh();
                return (!isGridEditable || appDisabled || tbDisabled);
            }),
            duplicateRowButton: ko.pureComputed(function() {
                var isGridEditable = self.grid.grid('option', 'editable'),
                    tbDisabled = _gridToolbarStates.duplicateRowButton_disabled(),
                    appSpecificDisabledStates = self.widgetObj.options.data.appDisabledStates,
                    appDisabled = appSpecificDisabledStates.duplicateRowButton !== undefined ? appSpecificDisabledStates.duplicateRowButton() : false;
                self.refresh();
                return (!isGridEditable || appDisabled || tbDisabled);
            }),
            editRowButton: ko.pureComputed(function() {
                var isGridEditable = self.grid.grid('option', 'editable'),
                    tbDisabled = _gridToolbarStates.editRowButton_disabled(),
                    appSpecificDisabledStates = self.widgetObj.options.data.appDisabledStates,
                    appDisabled = appSpecificDisabledStates.editRowButton !== undefined ? appSpecificDisabledStates.editRowButton() : false;
                self.refresh();
                return (!isGridEditable || appDisabled || tbDisabled);
            }),
            viewRowButton: ko.pureComputed(function() {
                var tbDisabled = _gridToolbarStates.viewRowButton_disabled(),
                    appSpecificDisabledStates = self.widgetObj.options.data.appDisabledStates;
                var appDisabled = appSpecificDisabledStates.viewRowButton !== undefined ? appSpecificDisabledStates.viewRowButton() : false;
                self.refresh();
                return (appDisabled || tbDisabled);
            }),
            deleteRowButton: ko.pureComputed(function() {
                var isGridEditable = self.grid.grid('option', 'editable'),
                    tbDisabled = _gridToolbarStates.deleteRowButton_disabled(),
                    appSpecificDisabledStates = self.widgetObj.options.data.appDisabledStates,
                    appDisabled = appSpecificDisabledStates.deleteRowButton !== undefined ? appSpecificDisabledStates.deleteRowButton() : false;
                self.refresh();
                return (!isGridEditable || appDisabled || tbDisabled);
            }),
            refreshButton: ko.pureComputed(function() {
                var tbDisabled = _gridToolbarStates.refreshButton_disabled(),
                    appSpecificDisabledStates = self.widgetObj.options.data.appDisabledStates;
                var appDisabled = appSpecificDisabledStates.pushButtonRefresh !== undefined ? appSpecificDisabledStates.pushButtonRefresh() : false;
                self.refresh();
                return (appDisabled || tbDisabled);
            })
        };

        // local context to map the toolbar item states
        _gridToolbarStates = {
            detachDialogTitle:         ko.observable(TOOLBAR.DETACH_DIALOG_TITLE),
            viewMenuItem_selected:     ko.observable(''),
            viewMenuItem_detachLabel:  ko.observable(VIEWMENU.DETACH_LABEL),
            viewMenuItem_sortCriteria: {
                first: {
                    label:           'Sort First By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                second: {
                    label:           'Sort Second By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                third: {
                    label:           'Sort Third By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                fourth: {
                    label:           'Sort Fourth By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                fifth: {
                    label:           'Sort Fifth By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                sixth: {
                    label:           'Sort Sixth By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                seventh: {
                    label:           'Sort Seventh By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                eighth: {
                    label:           'Sort Eighth By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                },
                ninth: {
                    label:           'Sort Ninth By',
                    sortableColumns: ko.observableArray([]),
                    column:          ko.observableArray([]),
                    order:           ko.observable('asc')
                }
            },
            viewMenuItem_addRowPosition:    ko.observable('after'),
            viewMenuItem_columnFilterMatch: ko.observable('all'),
            addRowButton_disabled:          ko.observable(false),
            duplicateRowButton_disabled:    ko.observable(true),
            editRowButton_disabled:         ko.observable(true),
            viewRowButton_disabled:         ko.observable(true),
            deleteRowButton_disabled:       ko.observable(true),
            refreshButton_disabled:         ko.observable(false),
            detachToggleButton:             ko.observableArray([]),
            detachToggleButtonLabel:        ko.observable(TOOLBAR.BUTTON_DETACH_LABEL),
            filterTypeRadioButton:          ko.observable(_toolbarSettings.defaultColumnFilterType),
            filterInputTextVisible:         ko.observable(_toolbarSettings.defaultColumnFilterType === GRID.TEXT_FILTER_TYPE)
        };

        /*
             * Create a root binding context so we can reference from the view template bindings.
            */
        _createBindingContext = function() {

            var _createToolbarContext = function() {
                    // Toolbar binding context
                    self.showTopBorder = self.widgetObj.options.showTopBorder;
                    self.showBottomBorder = self.widgetObj.options.showBottomBorder;
                    self.grid = self.widgetObj.options.grid;

                    // ID context for other elements
                    self.attachWrapperId = TOOLBAR.ATTACH_WRAPPER_ID;
                    self.detachWrapperId = TOOLBAR.DETACH_WRAPPER_ID;
                    self.detachDialogTitle = _gridToolbarStates.detachDialogTitle;
                    self.detachDialogId = TOOLBAR.DETACH_DIALOG_ID;
                    self.detachDialogAnchorId = TOOLBAR.DETACH_DIALOG_ANCHOR_ID;
                    self.gridWrapperId = TOOLBAR.GRID_WRAPPER_ID;

                    self.viewMenuId = VIEWMENU.DEFAULT_ELEMENT_ID;

                    self.cancelLabel = 'Cancel';
                    self.applyLabel = 'OK';
                },
                _createViewMenuDetachContext = function() {
                    self.viewMenuDetachId = VIEWMENU.DETACH_ID;

                    // method context for dialog operations
                    self.closeDetachDialogEvent = function() {
                        // This is called when the Detach dialog is closed,
                        // so we re-attach the detached elements back ...

                        // Get the handle to the root detach element
                        var gridToolbarDetach = $('#' + self.detachWrapperId);

                        // Notify the JET that they are going to be detached from the DOM so they can save state
                        oj.Components.subtreeDetached(gridToolbarDetach.get(0));

                        // attach the detached element back to the previous place ...
                        gridToolbarDetach.appendTo('#' + self.attachWrapperId);

                        // Notify the JET that they are attached to the DOM again
                        oj.Components.subtreeAttached(gridToolbarDetach.get(0));

                        _gridToolbarStates.viewMenuItem_detachLabel(VIEWMENU.DETACH_LABEL);
                        _gridToolbarStates.detachToggleButtonLabel(TOOLBAR.BUTTON_DETACH_LABEL);
                        _gridToolbarStates.detachToggleButton([]);
                    };
                },
                _createViewMenuSortContext = function() {
                    self.viewMenuSortId = VIEWMENU.SORT_ID;
                    self.viewMenuSortDialogId = VIEWMENU_DIALOG.SORT_ID;
                    self.viewMenuSortColumnId = VIEWMENU_DIALOG.SORT_COLUMN_ID;
                    self.viewMenuSortOrderId = VIEWMENU_DIALOG.SORT_ORDER_ID;
                    self.viewMenuSortOrderAscId = VIEWMENU_DIALOG.SORT_ORDER_ASC_ID;
                    self.viewMenuSortOrderDscId = VIEWMENU_DIALOG.SORT_ORDER_DSC_ID;
                    self.viewMenuPreferencesDialogId = VIEWMENU_DIALOG.PREFERENCES_ID;

                    // This is the master list of the grid columns that are visible and sortable
                    // according to the grid config and view. List populated in init() phase.
                    self.sortableColumns = [];

                    // List of avaiable column selector criteria
                    self.sortCriteria = ko.observable([]);

                    self.ascendingLabel = 'Ascending'; // labelSortAsc
                    self.descendingLabel = 'Descending'; // labelSortDsc

                    // Just close the sort dialog, no changes applied.
                    self.cancelSortDialogButtonClick = function() {
                        $('#' + self.viewMenuSortDialogId).ojDialog('close');
                    };
                    // Applied the sort selection changes, and close the sort dialog.
                    self.applySortDialogButtonClick = function() {
                        var dataIndex, order, sortBy = [];

                        $('#' + self.viewMenuSortDialogId).ojDialog('close');

                        // Gather the column selection and the sort order selection
                        self.sortCriteria().forEach(function(criteria) {
                            dataIndex = criteria.column()[0];
                            if (dataIndex) {
                                order = criteria.order();
                                sortBy.push({
                                    dataIndex: dataIndex,
                                    order:     order
                                });
                            }
                        });

                        // Apply the changes to the grid.
                        self.grid.grid('option', {
                            sortBy: sortBy
                        });
                    };
                    // This is to update the sort column selection and order selection
                    // before displaying the sort dialog.
                    self.beforeOpenSortDialogEvent = function(event, ui) {
                        var i, gridViewConfig, gridSortBy, criteria;

                        // reset our sortCriteria
                        self.sortCriteria().forEach(function(criteria) {
                            criteria.column([]);
                            criteria.order('asc');
                        });

                        // sync our current sortCriteria with the data grid view
                        gridViewConfig = self.grid.grid('getViewConfig');
                        gridSortBy = gridViewConfig.sortBy;
                        i = 0;
                        for (i = 0; i < gridSortBy.length; i++) {
                            // we only process the grid column sorts up to number of
                            // column selects displayed in the dialog.
                            if (i < self.sortCriteria().length) {
                                criteria = gridSortBy[i];
                                self.sortCriteria()[i].column([criteria.dataIndex]);
                                self.sortCriteria()[i].order(criteria.order);
                                // This is needed for the updated column selection to display.
                                _refreshSortableColumn(i);
                            }
                        }
                    };
                    // This is to update the column list before the select dropdown expands.
                    self.sortableColumnsBeforeExpand = function(event, ui) {
                        var selectId = event.target.id,
                            selectIndex = selectId.substring(self.viewMenuSortColumnId.length, selectId.length);
                            // Refresh to exclude the column names already selected in other column selections.
                        _refreshSortableColumn(selectIndex);
                    };
                },
                _createViewMenuPreferencesContext = function() {
                    self.viewMenuPreferencesId = VIEWMENU.PREFERENCES_ID;
                    self.viewMenuPreferencesDialogId = VIEWMENU_DIALOG.PREFERENCES_ID;
                    self.viewMenuPreferencesAddRowId = VIEWMENU_DIALOG.PREFERENCES_ADD_ROW_ID;
                    self.viewMenuPreferencesAddRowBeforeId = VIEWMENU_DIALOG.PREFERENCES_ADD_ROW_BEFORE_ID;
                    self.viewMenuPreferencesAddRowAfterId = VIEWMENU_DIALOG.PREFERENCES_ADD_ROW_AFTER_ID;

                    self.addRowPosition = ko.observable(_gridToolbarStates.viewMenuItem_addRowPosition());
                    self.viewMenuPreferencesAddRowLabel = 'Add Blank Rows';
                    self.addRowBeforeLabel = 'Before';
                    self.addRowAfterLabel = 'After';

                    self.viewMenuPreferencesColumnFilterId = VIEWMENU_DIALOG.PREFERENCES_COLUMN_FILTER_ID;
                    self.viewMenuPreferencesColumnFilterAnyId = VIEWMENU_DIALOG.PREFERENCES_COLUMN_FILTER_ANY_ID;
                    self.viewMenuPreferencesColumnFilterAllId = VIEWMENU_DIALOG.PREFERENCES_COLUMN_FILTER_ALL_ID;

                    self.columnFilterMatch = ko.observable(_gridToolbarStates.viewMenuItem_columnFilterMatch());
                    self.viewMenuPreferencesColumnFilterLabel = 'Match Column Filters';
                    self.columnFilterAnyLabel = 'Any';
                    self.columnFilterAllLabel = 'All';

                    // This is to update the preferences before displaying the dialog
                    self.beforeOpenPreferencesDialogEvent = function(event, ui) {
                        self.addRowPosition(_gridToolbarStates.viewMenuItem_addRowPosition());
                        self.columnFilterMatch(_gridToolbarStates.viewMenuItem_columnFilterMatch());
                    };
                    // Just close the preference dialog, no changes applied.
                    self.cancelPreferencesDialogButtonClick = function() {
                        $('#' + self.viewMenuPreferencesDialogId).ojDialog('close');
                    };
                    self.applyPreferencesDialogButtonClick = function() {
                        $('#' + self.viewMenuPreferencesDialogId).ojDialog('close');
                        _gridToolbarStates.viewMenuItem_addRowPosition(self.addRowPosition());
                        _gridToolbarStates.viewMenuItem_columnFilterMatch(self.columnFilterMatch());
                    };
                },
                _createViewMenuManageColumnsContext = function() {
                    // self.viewMenuManageColumnsId = VIEWMENU.MANAGE_COLUMNS_ID;
                    self.viewMenuManageColumnsDialogId = VIEWMENU_DIALOG.MANAGE_COLUMNS_ID;

                    self.closeManageColumnsDialogEvent = function() {
                        oj.Logger.info('Manage columns dialog closed');
                    };
                },
                _createExportMenuContext = function() {
                    self.exportMenuId = EXPORTMENU.DEFAULT_ELEMENT_ID;
                    self.exportMenuCsvId = EXPORTMENU.CSV_ID;
                    self.exportMenuExcelId = EXPORTMENU.EXCEL_ID;
                },
                _applyConsumerTranslations = function() {
                    if (self.widgetObj.options.translations) {
                        // merge the default with consumer's translation strings.
                        translations = _.merge({}, translations, self.widgetObj.options.translations);

                        // update the following translatable labels
                        TOOLBAR.BUTTON_VIEW_MENU_LABEL = translations.viewLabel;
                        TOOLBAR.BUTTON_ADD_ROW_LABEL = translations.addRowLabel;
                        TOOLBAR.BUTTON_DUP_ROW_LABEL = translations.duplicateRowLabel;
                        TOOLBAR.BUTTON_EDIT_ROW_LABEL = translations.editRowLabel;
                        TOOLBAR.BUTTON_VIEW_ROW_LABEL = translations.viewRowLabel;
                        TOOLBAR.BUTTON_DEL_ROW_LABEL = translations.deleteRowLabel;
                        TOOLBAR.BUTTON_REFRESH_LABEL = translations.refreshLabel;
                        TOOLBAR.BUTTON_EXPORT_MENU_LABEL = translations.exportLabel;
                        TOOLBAR.BUTTON_DETACH_LABEL = translations.detachLabel;
                        TOOLBAR.BUTTON_ATTACH_LABEL = translations.attachLabel;
                        TOOLBAR.RADIO_FILTER_LABEL = translations.filterLabel;
                        TOOLBAR.INPUT_TEXT_FILTER_PLACE_HOLDER = translations.filterTextPlaceHolderLabel;

                        VIEWMENU.MANAGE_COLUMNS_LABEL = translations.manageColumnsLabel;
                        VIEWMENU.DETACH_LABEL = translations.detachLabel;
                        VIEWMENU.ATTACH_LABEL = translations.attachLabel;
                        VIEWMENU.SORT_LABEL = translations.sortLabel;
                        VIEWMENU.PREFERENCES_LABEL = translations.preferencesLabel;

                        EXPORTMENU.CSV_LABEL = translations.csvLabel;
                        EXPORTMENU.EXCEL_LABEL = translations.excelLabel;

                        GRID.TEXT_FILTER_TYPE_LABEL = translations.textLabel;
                        GRID.COLUMN_FILTER_TYPE_LABEL = translations.columnsLabel;
                    }
                };

            _createToolbarContext();
            _createViewMenuDetachContext();
            _createViewMenuSortContext();
            _createViewMenuPreferencesContext();
            _createViewMenuManageColumnsContext();
            _createExportMenuContext();
            _applyConsumerTranslations();
        };

        // Establish a binding context for the view
        _createBindingContext();

        // This method is to refresh the column name options of given column select so
        // not to include the column name already selected in other column selects.
        _refreshSortableColumn = function(criteriaIndex) {
            var criteria, dataIndex, colValue, selectId, sortColumns, excludeColumnValues;

            // Get the list of all column name selected.
            excludeColumnValues = [];
            self.sortCriteria().forEach(function(criteria) {
                colValue = criteria.column();
                if (colValue.length && colValue[0]) {
                    excludeColumnValues.push(colValue[0]);
                }
            });

            criteria = self.sortCriteria()[criteriaIndex];

            colValue = criteria.column();
            dataIndex = colValue[0];
            // Figure out the available column name options
            sortColumns = _.filter(self.sortableColumns, function(column) {
                // options should include the column name it selects, and all other
                // available column names, excluding the column names selected in
                // other column selects.
                return (column.value === dataIndex || _.indexOf(excludeColumnValues, column.value) < 0);
            });
            // Update the column select's options
            selectId = self.viewMenuSortColumnId + criteriaIndex;
            $('#' + selectId).ojSelect({'options': sortColumns});
        };

        _detachHandler = function(event, ui) {
            var gridWrapper, gridTable, gridToolbarDetach,
                detach = (ui.value.length === 0);

            if (detach) {
                self.grid.grid('option', {'inModal': false});
                $('#' + self.detachDialogId).ojDialog('close');
                self.grid.grid('resize');
            } else {
                // First move the grid table into our grid hosting div element.
                // NOTE: this will move the grid table into our grid hosting element permanently!
                gridWrapper = $('#' + self.gridWrapperId);
                // Check if grid table has not been moved yet
                if (gridWrapper.children().length === 0) {
                    // move grid table into our hosting element
                    gridTable = self.grid;
                    oj.Components.subtreeDetached(gridTable.get(0));
                    gridTable.appendTo(gridWrapper);
                    oj.Components.subtreeAttached(gridTable.get(0));
                }

                // Now detach our grid-toolbar div element ...
                gridToolbarDetach = $('#' + self.detachWrapperId);

                // Notify the JET that they are going to be detached from the DOM so they can save state
                oj.Components.subtreeDetached(gridToolbarDetach.get(0));

                // move the grid-toolbar detach element to the detach dialog anchor
                gridToolbarDetach.appendTo('#' + self.detachDialogAnchorId);

                // Notify the JET that they are attached to the DOM again
                oj.Components.subtreeAttached(gridToolbarDetach.get(0));

                // Change toggle button label to 'Attach'
                _gridToolbarStates.detachToggleButtonLabel(TOOLBAR.BUTTON_ATTACH_LABEL);
                _gridToolbarStates.viewMenuItem_detachLabel(VIEWMENU.ATTACH_LABEL);

                // and open the detach dialog
                $('#' + self.detachDialogId).ojDialog('open');
                self.grid.grid('option', {'inModal': true});
                self.grid.grid('resize');
            }
        };

        // Method to pre-define the toolbar item action handlers.
        _defineToolbarActions = function() {
            // map the toolbar item actions
            var toolbarActions = {};
            var appSpecificActions = self.widgetObj.options.data.appRequiredActions;

            toolbarActions[PREDEFINED_KEYS.viewMenuItems] = function(event, ui) {
                var toggleState, dialogId,
                    itemId = ui.item[0].id;
                switch (itemId) {
                case self.viewMenuDetachId:
                    toggleState = _gridToolbarStates.detachToggleButton();
                    _gridToolbarStates.detachToggleButton(toggleState.length === 0 ? ['detach'] : []);
                    break;
                case self.viewMenuSortId:
                case self.viewMenuPreferencesId:
                    dialogId = itemId + '-dialog';
                    $('#' + dialogId).ojDialog('open');
                    break;
                default:
                    // ask widget to forward the event to application to handle
                    self.widgetObj._trigger('menuItemSelect', event, ui);
                }
            };
            toolbarActions[PREDEFINED_KEYS.addRowButton] = function(event) {
                if (appSpecificActions.addRowButton) {
                    appSpecificActions.addRowButton(event);
                } else {
                    var addAfter = _gridToolbarStates.viewMenuItem_addRowPosition() === 'after';
                    self.grid.grid('newRow', {addAfter: addAfter});
                }
            };
            toolbarActions[PREDEFINED_KEYS.duplicateRowButton] = function(event) {
                self.grid.grid('copyRows');
                self.grid.grid('pasteRows');
            };
            toolbarActions[PREDEFINED_KEYS.editRowButton] = function(event) {
                appSpecificActions.editRowButton(event);
            };
            toolbarActions[PREDEFINED_KEYS.viewRowButton] = function(event) {
                appSpecificActions.viewRowButton(event);
            };
            toolbarActions[PREDEFINED_KEYS.deleteRowButton] = function(event) {
                self.grid.grid('deleteRows');
            };
            toolbarActions[PREDEFINED_KEYS.refreshButton] = function(event) {
                appSpecificActions.refreshButton(event);
            };
            toolbarActions[PREDEFINED_KEYS.exportMenuItems] = function(event, ui) {
                var options,
                    menuItemId = ui.item[0].id;

                switch (menuItemId) {
                case EXPORTMENU.CSV_ID:
                    options = {
                        fileName: 'exportToCSV.csv',
                        exportTo: 'csv'
                    };
                    self.grid.grid('exportData', options);
                    break;
                case EXPORTMENU.EXCEL_ID:
                    options = {
                        fileName: 'exportToExcel.xls',
                        exportTo: 'excel'
                    };
                    self.grid.grid('exportData', options);
                    break;
                default:
                    // ask widget to forward the event to application to handle
                    self.widgetObj._trigger('menuItemSelect', event, ui);
                }
            };
            toolbarActions[PREDEFINED_KEYS.detachToggleButton] = _detachHandler;
            toolbarActions[PREDEFINED_KEYS.filterTypeRadioButton] = function(event, ui) {
                _gridToolbarStates.filterInputTextVisible(ui.value === GRID.TEXT_FILTER_TYPE);
                _resetFilterType();
                _initFilterType();
            };

            return (toolbarActions);
        };
        _gridToolbarActions = _defineToolbarActions();

        // Method to initialize filter type selection
        _initFilterType = function() {
            var layout,
                filterType = _gridToolbarStates.filterTypeRadioButton();

            if (filterType === GRID.TEXT_FILTER_TYPE) {
                // wire up the input box to the grid filter
                self.textFilterSearch = _.debounce(function() {
                    self.grid.grid('searchRows', $('#' + TOOLBAR.INPUT_TEXT_FILTER_ID).val());
                }, 200);
                self.textFilterTimeoutId = _.delay(function() { // LUX-272, $('#findInput').length was '0', so wait for it to load
                    $('#' + TOOLBAR.INPUT_TEXT_FILTER_ID).on('keyup', self.textFilterSearch);
                    $('#' + TOOLBAR.INPUT_TEXT_FILTER_ID).focus();
                }, 100);

            } else {
                // Turn on columns filter type
                layout = self.grid.grid('option', 'layout');
                if (layout && layout.rowsets) {
                    self.grid.grid('option', {
                        search: {
                            showAll:      false,
                            showParents:  false,
                            columnFilter: true
                        },
                        // enable the filter row and insert it before the last row.
                        layout: {
                            rowsets: [{
                                id:   layout.rowsets[layout.rowsets.length - 1].id,
                                type: 'flex'
                            }]
                        }
                    });
                }
            }
        };

        // Method to reset the grid filter type
        _resetFilterType = function() {
            var layout;

            // clear text filter method
            self.grid.grid('searchRows', '');
            self.grid.grid('clearColumnFilter');
            $('#' + TOOLBAR.INPUT_TEXT_FILTER_ID).val('');
            if (self.textFilterTimeoutId) {
                clearTimeout(self.textFilterTimeoutId);
                self.textFilterTimeoutId = undefined;
                $('#' + TOOLBAR.INPUT_TEXT_FILTER_ID).off('keyup');
            }
            if (self.textFilterSearch) {
                self.textFilterSearch.cancel();
                self.textFilterSearch = undefined;
            }

            layout = self.grid.grid('option', 'layout');
            if (layout && layout.rowsets) {
                self.grid.grid('option', {
                    search: {
                        columnFilter: false
                    },
                    layout: {
                        rowsets: [{
                            // remove the filter row that was inserted above the last row.
                            id:   layout.rowsets[layout.rowsets.length - 1].id,
                            type: 'flex'
                        }]
                    }
                });
            }
        };

        // Method to pre-define the toolbar item objects specifically for use with the LUX grid.
        _defineToolbarItems = function() {
            var itemFactory = {};
            itemFactory[PREDEFINED_KEYS.separator] = {type: 'separator'};
            itemFactory[PREDEFINED_KEYS.viewMenuButton] = {
                component: 'button', type:      'menu', id:        TOOLBAR.BUTTON_VIEW_MENU_ID, label:     TOOLBAR.BUTTON_VIEW_MENU_LABEL,
                menu:      '#' + VIEWMENU.DEFAULT_ELEMENT_ID
            };
            itemFactory[PREDEFINED_KEYS.viewMenu] = {
                component: 'menu', id:        VIEWMENU.DEFAULT_ELEMENT_ID, label:     TOOLBAR.BUTTON_VIEW_MENU_LABEL,
                select:    _gridToolbarActions.viewMenuItems
            };
            itemFactory[PREDEFINED_KEYS.viewMenuItems] = {
                detach: {
                    type:  'item', id:    self.viewMenuDetachId,
                    label: _gridToolbarStates.detachToggleButtonLabel,
                    // NOTE: cannot use icon: here since we use observable label.
                    href:  '#'
                },
                //                        manageColumns: {
                //                            type: 'item', id: self.viewMenuManageColumnsId, label: VIEWMENU.MANAGE_COLUMNS_LABEL,
                //                            href: '#'
                //                        },
                sort: {
                    type:  'item', id:    self.viewMenuSortId, label: VIEWMENU.SORT_LABEL,
                    href:  '#'
                },
                preferences: {
                    type:  'item', id:    self.viewMenuPreferencesId, label: VIEWMENU.PREFERENCES_LABEL,
                    href:  '#'
                }
            };
            itemFactory[PREDEFINED_KEYS.addRowButton] = {
                component: 'button', type:      'push', id:        TOOLBAR.BUTTON_ADD_ROW_ID, label:     TOOLBAR.BUTTON_ADD_ROW_LABEL,
                icons:     {start: 'lux-oj-alta-icon-plus oj-fwk-icon'},
                disabled:  _gridToolbarDisabledStates.addRowButton,
                event:     {
                    click: _gridToolbarActions.addRowButton
                }
            };
            itemFactory[PREDEFINED_KEYS.duplicateRowButton] = {
                component: 'button', type:      'push', id:        TOOLBAR.BUTTON_DUP_ROW_ID, label:     TOOLBAR.BUTTON_DUP_ROW_LABEL,
                icons:     {start: 'lux-alta-duplicate-icon oj-icon'},
                disabled:  _gridToolbarDisabledStates.duplicateRowButton,
                event:     {
                    click: _gridToolbarActions.duplicateRowButton
                }
            };
            itemFactory[PREDEFINED_KEYS.editRowButton] = {
                component: 'button', type:      'push', id:        TOOLBAR.BUTTON_EDIT_ROW_ID, label:     TOOLBAR.BUTTON_EDIT_ROW_LABEL,
                icons:     {start: 'lux-alta-edit-icon oj-icon'},
                disabled:  _gridToolbarDisabledStates.editRowButton,
                event:     {
                    click: _gridToolbarActions.editRowButton
                }
            };
            itemFactory[PREDEFINED_KEYS.viewRowButton] = {
                component: 'button', type:      'push', id:        TOOLBAR.BUTTON_VIEW_ROW_ID, label:     TOOLBAR.BUTTON_VIEW_ROW_LABEL,
                icons:     {start: 'lux-alta-view-icon oj-icon'},
                disabled:  _gridToolbarDisabledStates.viewRowButton,
                event:     {
                    click: _gridToolbarActions.viewRowButton
                }
            };
            itemFactory[PREDEFINED_KEYS.deleteRowButton] = {
                component: 'button', type:      'push', id:        TOOLBAR.BUTTON_DEL_ROW_ID, label:     TOOLBAR.BUTTON_DEL_ROW_LABEL,
                icons:     {start: 'lux-oj-alta-icon-cross oj-fwk-icon'},
                disabled:  _gridToolbarDisabledStates.deleteRowButton,
                event:     {
                    click: _gridToolbarActions.deleteRowButton
                }
            };
            itemFactory[PREDEFINED_KEYS.refreshButton] = {
                component: 'button', type:      'push', id:        TOOLBAR.BUTTON_REFRESH_ID, label:     TOOLBAR.BUTTON_REFRESH_LABEL,
                icons:     {start: 'lux-alta-refresh-icon oj-icon'},
                disabled:  _gridToolbarDisabledStates.refreshButton,
                event:     {
                    click: _gridToolbarActions.refreshButton
                }
            };
            itemFactory[PREDEFINED_KEYS.exportMenuButton] = {
                component: 'button', type:      'menu', id:        TOOLBAR.BUTTON_EXPORT_MENU_ID, label:     TOOLBAR.BUTTON_EXPORT_MENU_LABEL,
                menu:      '#' + EXPORTMENU.DEFAULT_ELEMENT_ID
            };
            itemFactory[PREDEFINED_KEYS.exportMenu] = {
                component: 'menu', id:        EXPORTMENU.DEFAULT_ELEMENT_ID, label:     TOOLBAR.BUTTON_EXPORT_MENU_LABEL,
                select:    _gridToolbarActions.exportMenuItems
            };
            itemFactory[PREDEFINED_KEYS.exportMenuItems] = {
                csv: {
                    type:  'item', id:    self.exportMenuCsvId, label: EXPORTMENU.CSV_LABEL,
                    href:  '#'
                },
                excel: {
                    type:  'item', id:    self.exportMenuExcelId, label: EXPORTMENU.EXCEL_LABEL,
                    href:  '#'
                }
            };
            itemFactory[PREDEFINED_KEYS.detachToggleButton] = {
                component:    'buttonset', type:         'checkbox', id:           TOOLBAR.BUTTON_DETACH_ID, label:        TOOLBAR.BUTTON_DETACH_LABEL,
                checked:      _gridToolbarStates.detachToggleButton,
                optionChange: _gridToolbarActions.detachToggleButton,
                checkbox:     {
                    id:    'detach', value: 'detach', label: _gridToolbarStates.detachToggleButtonLabel
                }
            };
            itemFactory[PREDEFINED_KEYS.filterTypeRadioButton] = {
                component:    'buttonset', type:         'radiogroup', id:           TOOLBAR.RADIO_FILTER_ID, staticLabel:  TOOLBAR.RADIO_FILTER_LABEL,
                checked:      _gridToolbarStates.filterTypeRadioButton,
                optionChange: _gridToolbarActions.filterTypeRadioButton,
                chroming:     'full',
                radios:       [
                    {id: GRID.COLUMN_FILTER_TYPE, name: 'filterType', label: GRID.COLUMN_FILTER_TYPE_LABEL},
                    {id: GRID.TEXT_FILTER_TYPE, name: 'filterType', label: GRID.TEXT_FILTER_TYPE_LABEL}
                ]
            };
            itemFactory[PREDEFINED_KEYS.filterInputText] = {
                component:   'input', type:        'text', id:          TOOLBAR.INPUT_TEXT_FILTER_ID,
                placeHolder: TOOLBAR.INPUT_TEXT_FILTER_PLACE_HOLDER,
                visible:     _gridToolbarStates.filterInputTextVisible
            };

            return (itemFactory);
        };
        _gridToolbarItems = _defineToolbarItems();

        // Method to build an array of the toolbar menu item objects from menu item specification.
        _gridToolbarMenuItemsBuilder = function(menuItemsRef, menuItems) {
            var key, val, appVal,
                _menuItems = menuItemsRef;

            if (menuItems) {
                _menuItems = [];
                menuItems.forEach(function(item) {
                    key = Object.keys(item)[0];
                    appVal = item[key];

                    val = menuItemsRef[key];
                    if (val) {
                        if (!_.isEmpty(appVal)) {
                            val = _.merge({}, menuItemsRef[key], appVal);
                        }
                    } else {
                        val = appVal;
                    }
                    _menuItems.push(val);
                });
            }

            return (_menuItems);
        };

        // Method to build an array of the toolbar item objects.
        // If user does not provide any toolbar item specification, return a default
        // pre-defined list of toolbar items. Otherwise, build and return
        // a list of toolbar items according to the spec.
        _gridToolbarBuilder = function() {
            var key, val, appVal, toolbarItems,
                viewMenu, exportMenu, items;

            items = self.widgetObj.options.data.items;

            toolbarItems = [];

            // Construct the toolbar items from the item list passed in from widget
            items.forEach(function(item) {
                key = Object.keys(item)[0];
                appVal = item[key];

                switch (key) {
                // Process overridden predefined view menu items, and bind to the view menu
                case 'viewMenuItems':
                    if (viewMenu) {
                        viewMenu.menuItems = _gridToolbarMenuItemsBuilder(_gridToolbarItems.viewMenuItems, appVal);
                    }
                    break;
                    // Process overridden predefined export menu items, and bind to the export menu
                case 'exportMenuItems':
                    if (exportMenu) {
                        exportMenu.menuItems = _gridToolbarMenuItemsBuilder(_gridToolbarItems.exportMenuItems, appVal);
                    }
                    break;
                    // Process overridden predefined menu items
                default:
                    val = _gridToolbarItems[key];
                    if (val) {
                        // Merge the values if application wants to override the predefined attribute
                        if (!_.isEmpty(appVal)) {
                            val = _.merge({}, _gridToolbarItems[key], appVal);
                        }
                    } else {
                        // not a predefined attribute, so just add it
                        val = appVal;
                    }
                    toolbarItems.push(val);
                }

                // Keep track of the menu component to bind the menuItems later.
                if (val.id && val.id === VIEWMENU.DEFAULT_ELEMENT_ID) {
                    viewMenu = val;
                }
                if (val.id && val.id === EXPORTMENU.DEFAULT_ELEMENT_ID) {
                    exportMenu = val;
                }
            });

            return toolbarItems;
        };

        // view model to initialize the grid-toolbar states and action handlers.
        // Mainly for widget to call during its init phase.
        self.init = function() {
            var gridColumns, gridViewConfig, columns, sortCriteria, maxCriteria,
                self = this;

            _resetFilterType();
            _initFilterType();
            _gridToolbarStates.detachDialogTitle(self.grid.grid('option', 'locale').gridLabel ?
                self.grid.grid('option', 'locale').gridLabel :
                TOOLBAR.DETACH_DIALOG_TITLE);

            // Update 'row action' push button states according to the grid row events.
            self.grid.on('grid-rows-checked',
                function(event, args) {
                    var disabled = args.checkedRecords.length === 0;
                    _gridToolbarStates.duplicateRowButton_disabled(disabled);
                    _gridToolbarStates.deleteRowButton_disabled(disabled);

                    disabled = !args.checkedRecords.length || args.checkedRecords.length > 1;
                    _gridToolbarStates.editRowButton_disabled(disabled);
                    _gridToolbarStates.viewRowButton_disabled(disabled);
                }
            );
            self.grid.on('grid-rows-deleted',
                function(event, delRows) {
                    _gridToolbarStates.duplicateRowButton_disabled(true);
                    _gridToolbarStates.deleteRowButton_disabled(true);
                    _gridToolbarStates.editRowButton_disabled(true);
                    _gridToolbarStates.viewRowButton_disabled(true);
                }
            );

            // add grid data exporter
            self.grid.grid('option', {
                dataExporter: function(grid, options) {
                    grid.view.exportToCSV(options);
                }
            });

            // We need the merged data from these two grid data,
            // in order to extract sortable and visible columns.
            gridColumns = self.grid.grid('option', 'columns');
            gridViewConfig = self.grid.grid('getViewConfig');
            columns = _.merge({}, gridColumns, gridViewConfig.columns);

            // Create a list of sortable and viewable columns
            Object.keys(columns).forEach(function(key) {
                var column,
                    value = columns[key];
                if (key !== 'empty-column' && value.sortable) {
                    column = {
                        label: value.title,
                        value: value.dataIndex
                    };
                    if (value.visible) {
                        // Create the option items for the sort select dropdown.
                        self.sortableColumns.push(column);
                    }
                }
            });

            // Initialize the sort dialog's column select with the sortable column list.
            self.sortCriteria().forEach(function(criteria) {
                criteria.sortableColumns(self.sortableColumns);
            });

            // Initialize number of column selector criteria based on user input or max number of sortable columns.
            sortCriteria = [];
            maxCriteria = (_toolbarSettings.maxColumnSortCriteria <= 0 ||
                    _toolbarSettings.maxColumnSortCriteria > 9) ? 9 : _toolbarSettings.maxColumnSortCriteria;
            maxCriteria = Math.min(maxCriteria, self.sortableColumns.length);
            Object.keys(_gridToolbarStates.viewMenuItem_sortCriteria).forEach(function(key, index) {
                if (index < maxCriteria) {
                    sortCriteria.push(_gridToolbarStates.viewMenuItem_sortCriteria[key]);
                }
            });
            self.sortCriteria(sortCriteria);

            self.columnFilterSubscription = _gridToolbarStates.viewMenuItem_columnFilterMatch.subscribe(function(val) {
                var columnFilterMatchAll = val === 'all';
                self.grid.grid('option', {
                    search: {
                        columnFilterMatchAllColumns: columnFilterMatchAll
                    }
                });
            });

            self.refresh = function() {
                if (self.luxToolbarElement) {
                    self.luxToolbarElement.refreshToolbarElement();
                }
            };

            // LUX-910 - Accessible markup for grid toolbar
            _.delay(function() {
                try {
                    var gridId = self.grid.grid('getId');
                    self.luxToolbarElement = self.widgetObj.element[0].getElementsByTagName('oj-lux-toolbar')[0];
                    var toolbarElement = self.luxToolbarElement.getToolbarElement();
                    // TODO: once we move to oj-toolbar, we don't need to convert to jquery object here.
                    self.$toolbar = $(toolbarElement);
                    // Add to the toolbar container the markup for accessibility
                    self.$toolbar.attr('aria-controls', gridId);
                    self.$toolbar.attr('aria-labelledby', gridId + '-description-grid-label');
                    // Filter controls
                    self.grid.find('.grid-descriptions').append(
                        '<span id=\'' + gridId + '-description-filter-field\'>' + translations.textFilterLabel + '</span>');
                    self.$toolbar.find('#' + TOOLBAR.INPUT_TEXT_FILTER_ID).attr('aria-labelledby', gridId + '-description-filter-field');
                } catch (e) {
                    oj.Logger.warn('The toolbarGrid threw an error: ', e);
                }
                // TODO/FIXME (markus 2017-10-21) Understand why a delay of 100ms is required.
            }, 100);
        };

        // view model to reset the grid-toolbar states and action handlers.
        // Mainly for widget to call during its destroy phase.
        self.destroy = function() {
            var self = this;
            _resetFilterType();
            self.grid.off('grid-rows-checked');
            self.grid.off('grid-rows-deleted');
            // Clear columns filter
            if (self.columnFilterSubscription) {
                self.columnFilterSubscription.dispose();
                self.columnFilterSubscription = undefined;
            }
        };

        // Build the toolbar items to pass to the LUX toolbar component (declared in the view)
        self.toolbarItems = _gridToolbarBuilder();

        // return the grid-toolbar view model
        return (self);
    };

        // Method to create list of predefined toolbar items
    viewModel.createPredefinedToolbarItems = _createPredefinedToolbarItems;

    // Method to create list of toolbar default preference settings
    viewModel.createToolbarSettings = _createToolbarSettings;

    // Method to create list of actions application required to implement
    viewModel.createAppSpecificActions = _createAppSpecificActions;

    // Method to create list of disabled states application can provide
    viewModel.createAppSpecificDisabledStates = _createAppSpecificDisabledStates;

    return (viewModel);
}
);
