/**
 * Copyright (c) 2018 Oracle and/or its affiliates. All rights reserved.
 * @author Daniel Hoang
 */
define([
    'ojs/ojcore',
    'knockout',
    'lodash',
    'jquery',
    'jqueryui-amd/widget'
],
function(oj, ko, _, $) {
    'use strict';

    var PROXY_WIDGET_PREFIX = '_widget';
    var PROXY_WIDGET_EVENT_PREFIX = '_widgetEvent_';
    var Logger = oj.Logger;

    /*
     * This utility can be utilized for converting current lux jQueryUI widget to CCA.
     */
    var AbstractCCAWidget = function(context, metadata) {
        var busyContext, properties, options, metaEvents, events;
        var self = this;
        self.eventListeners = {};
        self.element = context.element;
        self.$element = $(self.element);
        self.metadata = metadata;
        self.widgetReady = ko.observable(false);

        // Set up busy context
        busyContext = oj.Context.getContext(context.element).getBusyContext();
        self.setupComplete = busyContext.addBusyState({'descriptions': 'lux-cca-widget startup processing'});

        // Extract unique number from the unique string. This is used to generate component's
        // element IDs.
        self.uniqueNumber = parseInt(context.unique.replace(/[^0-9\.]/g, ''), 10);

        // internal method to generate element ID from a given prefix string
        self._uniqueId = function(idPrefix) {
            return (idPrefix + self.uniqueNumber);
        };

        properties = JSON.parse(self.metadata).properties;
        options = {};
        for (var key in properties) {
            // We check if consumer passing in any properties, and to create a corresponding mapping
            // of widget's options.
            // Abstract class will not handle the 'translations' option, since it may be named
            // differently in widget (e.g. grid uses locale instead of translations).
            // Subclass will need the handle the translations option, if any. See oj-lux-grid for
            // example code.
            if (key !== 'translations' && !_.isUndefined(context.properties[key])) {
                options[key] = context.properties[key];
            }
        }
        self._widgetOptions = options;

        // Create a mapping of widget event name to a proxied CCA event name
        metaEvents = JSON.parse(self.metadata).events;
        events = {};
        for (var eventName in metaEvents) {
            events[eventName] = PROXY_WIDGET_EVENT_PREFIX + eventName;
        }
        self._widgetEvents = events;

        // signal that all the setup work is done
        self.setupComplete();

        // Add "before" advice before calling the implemented life-cycle connected method.
        // This connected() method will be invoked by the CCA component's life-cycle connected()
        // after the View is first inserted into the DOM and then each time the composite is reconnected
        // to the DOM after being disconnected. Since the composite needs to add/remove event listeners,
        // we will use connected/disconnected methods, as recommended by JET ojComposite.
        self.connected = _.wrap(self.connected, function(connected, ctx) {
            // Perform widget proxy setting here
            Logger.info(self.widgetContainerId + '@connected');
            // If this is a first time connected() is invoked, we need to perform
            // the widget proxy setup.
            if (!self.$widget) {
                self.defineMethods();
                self.defineEvents();
            } else {
                // Otherwise, connected() is invoked again after being disconnected(), so we just
                // re-setup proxy events here.
                // See http://jet.us.oracle.com/5.1.0/jsdocs/oj.Composite.html for the
                // connected/disconnected lifecycle.
                self.onWidgetProxyEvents();
            }
            // and then call the implemented connected method
            connected.call(self, ctx);
        });

        // Add "before" advice before calling the implemented life-cycle disconnected method.
        // The disconnected() method will be invoked by the CCA component's life-cycle disconnected()
        // when this composite component is disconnected from the DOM.
        // We will perform the clean-up tasks here.
        self.disconnected = _.wrap(self.disconnected, function(disconnected, ctx) {
            Logger.info(self.widgetContainerId + '@disconnected');

            // Remove widget's proxy event listeners here
            self.offWidgetProxyEvents();

            // and then call the implemented disconnected method
            disconnected.call(self, ctx);
        });

        // Add "before" advice before calling the implemented life-cycle bindingsApplied method.
        // The bindingsApplied() method will be invoked by the CCA component's life-cycle bindingsApplied()
        // after the bindings are applied on this View. This is where we can get to the grid widget element from
        // within the widget container.
        self.bindingsApplied = _.wrap(self.bindingsApplied, function(bindingsApplied, ctx) {
            var selector;

            // locate the widget element and its reference here
            if (self.widgetContainerId) {
                Logger.info(self.widgetContainerId + '@bindingsApplied');
                selector = self.element.querySelector('#' + self.widgetContainerId);
                if (selector) {
                    self.$widget = self.$element.find(selector);
                    // We now setting up widget's events and proxy them to CCA custom events.
                    self.onWidgetProxyEvents();
                } else {
                    throw new Error('Cannot find widget container ID ' + self.widgetContainerId);
                }
            } else {
                throw new Error('No widget container element ID registered. Does subclass call registerWidgetContainerId()?');
            }

            // Call the implemented bindingsApplied method
            bindingsApplied.call(self, ctx);

            // Instantiate the widget here
            self.widgetInstantiation();
        });

        // Add "before" advice before calling the implemented propertyChanged method.
        // The propertyChanged() can be implemented on the ViewModel and will be invoked when properties are
        // updated before the [property]Changed event is fired.
        self.propertyChanged = _.wrap(self.propertyChanged, function(propertyChanged, ctx) {
            // So we assume the property changed is related to a corresponding widget's option.
            var obj = {};
            obj[ctx.property] = ctx.value;
            try {
                // so we first try to proxy the widget option change
                Logger.info('propertyChanged -> widgetOptionChanged: ' + obj);
                self.$widget[self.widgetName]('option', obj);
            } catch (e) {
                Logger.warn('Error proxying widget option: ' + ctx.property);
            }
            // then do what sub-class wants for its implementation.
            propertyChanged.call(self, ctx);
        });
    };

    AbstractCCAWidget.prototype = {
        // define a default life-cycle method in case sub-class does not implement one.
        connected: function() {
        },
        // define a default life-cycle method in case sub-class does not implement one.
        disconnected: function() {
        },
        // define a default life-cycle method in case sub-class does not implement one.
        bindingsApplied: function() {
        },
        // define a default method propertyChanged method in case sub-class does not implement one.
        propertyChanged: function() {
        },
        registerWidgetContainerId: function(widgetContainerId) {
            this.widgetContainerId = widgetContainerId;
        },
        widgetInstantiation: function() {
            this.$widget[this.widgetName](this._widgetOptions);
            this.widgetReady(true);
        },
        defineMethods: function() {
            var self = this;
            var methods = JSON.parse(self.metadata).methods;
            _.each(methods, function(value, key) {
                // We only proxy methods to widget element if it is marked with PROXY_WIDGET_PREFIX.
                if (value.internalName && value.internalName.startsWith(PROXY_WIDGET_PREFIX)) {
                    self[value.internalName] = function() {
                        // arguments object is array like object not Array, so convert it it array.
                        var args = Array.prototype.slice.call(arguments);
                        args.unshift(key);
                        return self.$widget[self.widgetName].apply(self.$widget, args);
                    };
                }
            });
        },
        getWidgetEventName: function(eventName) {
            return (eventName);
        },
        defineEvents: function() {
            var self = this;
            var events = JSON.parse(self.metadata).events;
            _.each(events, function(eventDetails, eventName) {
                var internalEventName = PROXY_WIDGET_EVENT_PREFIX + eventName;
                self[internalEventName] = function() {
                    var metaDetails = eventDetails;
                    var args = Array.prototype.slice.call(arguments);
                    var detail = {};
                    Object.keys(metaDetails.detail).forEach(function(key, index) {
                        if (index <= args.length) {
                            detail[key] = args[index];
                        }
                    });

                    var eventParams = {
                        'bubbles':    metaDetails.bubbles !== undefined ? metaDetails.bubbles : false,
                        'cancelable': metaDetails.cancelable !== undefined ? metaDetails.cancelable : false,
                        'detail':     detail
                    };

                    //Raise the custom event
                    self.element.dispatchEvent(new CustomEvent(eventName, eventParams));
                };
            });
        },
        onWidgetProxyEvents: function() {
            var self = this;
            if (self.$widget) {
                // We are now setting up widget's event listeners to proxy them to CCA custom events.
                _.each(self._widgetEvents, function(customEvent, widgetEvent) {
                    self.$widget.on(self.getWidgetEventName(widgetEvent), self[customEvent]);
                });
            } else {
                Logger.warn('No widget found to create proxy events.');
            }
        },
        offWidgetProxyEvents: function() {
            var self = this;
            if (self.$widget) {
                // We are now tearing down widget's proxied event listeners.
                _.each(self._widgetEvents, function(customEvent, widgetEvent) {
                    self.$widget.off(self.getWidgetEventName(widgetEvent), self[customEvent]);
                });
            } else {
                Logger.warn('No widget found to remove proxy events.');
            }
        },
        setWidgetOption: function(name, value) {
            var self = this;
            self._widgetOptions[name] = value;
        }
    };
    return AbstractCCAWidget;
});
