/** @module shared */
/**
 * Shared 'localizer' module.
 *
 * Currently supports a few relatively safe methods.
 *
 * Given a property in the server side file 'label\_strings.properties',
 * 'global\_ok'.
 * The string for that key can be loaded with the following call:
 *
 *     define(['localizer'], function (localizer) {
 *         localizer.getString('label.global_ok'); // returns 'Primavera'
 *     });
 *
 * If that string is a format string, such as "Undo {0} {1}",
 * it can be retrieved and formatted using the following:
 *
 *     define(['localizer'], function (localizer) {
 *         localizer.getFormattedString('message.global_undo_change', 'Value1', 'Value2'); // returns 'Undo Value1 Value2'
 *     });
 *
 * Or formatted only using the following:
 *
 *     define(['localizer'], function (localizer) {
 *         localizer.formatString('../url/myquery?arg1={0}&arg2={1}', 'Value1', 'Value2'); // returns '../url/myquery?arg1=Value1&arg2=Value2'
 *     });
 *
 * No other methods are supported at this time.
 *
 * @class localizer
 */

// will need to convert to using a custom plugin
// but for now, at least, we can put our strings in here
// also note, this assumes that we will eventually be able to put
//  locale in the cookie or localStorage
//  this seems like a reasonable, safe, and fair assumption
define('localizer', ['text!i18n/bundles/resource_strings.json',
    'gridapi.underscore', 'gridapi.logger'], function (stringJSON, _, logger) {
    'use strict';

    var that = {};

    var stringMap = JSON.parse(stringJSON);

    var __noSuchKey__ = 'KEY NOT FOUND';

    /**
     * Returns the string corresponding to the given key.
     *
     * @param {String} the key to load
     * @method getString
     */
    that.getString = function (key) {
        var keys = key && key.split('.');

        if (!keys) {
            logger.warn('Key: \'' + key + '\' not found.');
            return __noSuchKey__;
        }

        var node = stringMap[keys[0]];

        if (!node) {
            logger.warn('Key: \'' + key + '\' not found.');
            return __noSuchKey__;
        }

        node = node[keys[1]];

        if (!node) {
            logger.warn('Key: \'' + key + '\' not found.');
            return __noSuchKey__;
        }

        return node;
    };

    that.formatString = function formatString (node/*, arg1, arg2, ...*/) {
        var params = _.toArray(arguments);

        return node.replace(/\{([0-9]+)\}/g, function (match, g1) {
            var value = params[parseInt(g1, 10) + 1];
            return value || '';
        });
    };

    that.getFormattedString = function (key/*, arg1, arg2, ...*/) {
        var node = this.getString(key);

        if (node === __noSuchKey__){
            return node;
        }

        var params = _.toArray(arguments);
        params[0] = node;

        return that.formatString.apply(null, params);
    };

    return that;

});
