/**
 * @module intl-bigNumber-converter
 * @requires knockout
 * @requires jquery
 * @requires lodash
 * @requires mathjs
 * @requires ojcore
 * @requires luxBignumber
 * @requires oi18n
 * @requires oi18n/NumberConverter
 * @requires oi18n/Utils
 * @requires ojs/ojvalidation-number
 * @requires ojs/ojvalidation-base
*/
define([
    'ojs/ojcore',
    'knockout',
    'jquery',
    'lodash',
    'luxBignumber',
    'oi18n/oi18n',
    'oi18n/NumberConverter',
    'oi18n/Utils',
    'ojs/ojvalidation-number',
    'ojs/ojvalidation-base'
], function(oj, ko, $, _, luxBignum, ojld) {
    'use strict';

    var lux = {}; // create LUX name space
    /**
     * @description Constructs a lux.NumberRangeValidator that ensures the value provided is within a given range
     * @alias module:lux.NumberRangeValidator
     * @method module:lux.NumberRangeValidator#NumberRangeValidator
     * @constructor
     * @param {Object=} options an object literal used to provide the following properties
     * @param {number=} options.min - the minimum number value of the entered value.
     * @param {number=} options.max - the maximum number value of the entered value.
     * @param {Object=} options.hint - an optional object literal of hints to be used.
     * <p>The hint strings (e.g., hint.min) are  passed as the 'pattern' parameter to
     * [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
     * that documentation, if you are using a reserved character, you need to escape it with
     * a dollar character ('$').
     * </p>
     * @param {string=} options.hint.max - a hint used to indicate the allowed maximum. When not present,
     * the default hint is the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.hint.max</code>.<p>
     *
     * Tokens: <br/>
     * {max} - the maximum<p>
     * Usage: <br/>
     * Enter a number less than or equal to {max}
     * @param {string=} options.hint.min - a hint used to indicate the allowed minimum. When not present,
     * the default hint is the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.hint.min</code>.<p>
     * Tokens: <br/>
     * {min} the minimum <p>
     * Usage: <br/>
     * Enter a number greater than or equal to {min}</li>
     * @param {string=} options.hint.inRange - a hint used to indicate the allowed range. When not
     * present, the default hint is the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.hint.inRange</code>.<p>
     * Tokens:<br/>
     * {min} the minimum<br/>
     * {max} the maximum<p>
     * Usage: <br/>
     * Enter a number between {min} and {max}
     * @param {string=} options.hint.exact - a hint used to indicate the allowed value.
     * This is used when min and max are non-null and are equal to each other.
     * When not present, the default hint is the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.hint.exact</code>.<p>
     * Tokens:<br/>
     * {num} the number allowed<br/>
     * Usage: <br/>
     * Enter the number {num}
     * @since 3.0.0
     * @param {Object=} options.messageDetail - an optional object literal of custom error messages to
     * be used.
     * <p>The messageDetail strings (e.g., messageDetail.rangeUnderflow) are  passed as the 'pattern'
     * parameter to [oj.Translations.html#applyParameters]{@link oj.Translations}. As stated in
     * that documentation, if you are using a reserved character, you need to escape it with
     * a dollar character ('$').
     * </p>
     * @param {string=} options.messageDetail.rangeUnderflow - the detail error message to be used when
     * input value is less than the set minimum value. When not present, the default detail message is
     * the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.messageDetail.rangeUnderflow</code>.<p>
     * Tokens:<br/>
     * {value} - value entered by the user<br/>
     * {min} - the minimum allowed value<p>
     * Usage: <br/>
     * The number must be greater than or equal to {min}.
     * @param {string=} options.messageDetail.rangeOverflow - the detail error message to be used when
     * input value exceeds the maximum value set. When not present, the default detail message is
     * the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.messageDetail.rangeOverflow</code>.<p>
     * Tokens:<br/>
     * {value} - value entered by the user<br/>
     * {max} - the maximum allowed value<p>
     * Usage: <br/>
     * The number must be less than or equal to {max}.
     * @param {string=} options.messageDetail.exact - the detail error message to be used when the
     * input value is not between min and max when min and max are both non-null and equal.
     *  When not present, the default detail message is the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.messageDetail.exact</code>.<p>
     * Tokens:<br/>
     * {num} - the allowed value<p>
     * Usage: <br/>
     * The number must be {num}.
     * @since 3.0.0
     * @param {Object=} options.messageSummary - optional object literal of custom error summary message
     * to be used.
     * @param {string=} options.messageSummary.rangeUnderflow - the summary of the error message when
     * input value is less than the set minimum value. When not present, the default message summary is
     * the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.messageSummary.rangeUnderflow</code>.
     * @param {string=} options.messageSummary.rangeOverflow - the summary of the error message when
     * input value exceeds the maximum value set.  When not present, the default message summary is
     * the resource defined with the key
     * <code class="prettyprint">oj-validator.range.number.messageSummary.rangeOverflow</code>.
     * @export
     * @constructor
     *
     */
    // BEGIN VALIDATION CODE *******************************************
    lux.NumberRangeValidator = function _NumberRangeValidator(options) {
        this.Init(options);
    };

    oj.Object.createSubclass(lux.NumberRangeValidator, oj.NumberRangeValidator, 'lux.NumberRangeValidator');

    /**
     * @memberof lux.NumberRangeValidator
     * Initializes validator instance with the set options
     * @param {Object} options - json object containing parameters for initializing the converter
     *
     * @throws {Error} when luxBigNumber or oi18nResource is not passed in the options
     */
    lux.NumberRangeValidator.prototype.Init = function (options) {
        lux.NumberRangeValidator.superclass.Init.call(this);

        if (options) {
            this._min = options.min;
            this._max = options.max;
            this._math = options.bigNum;
            this._oi18nResources = options.oi18nResources;
            this._converter = oj.IntlConverterUtils.getConverterInstance(options.converter);
            this._hint = options.hint || {};
            this._customMessageSummary = options.messageSummary || {};
            this._customMessageDetail = options.messageDetail || {};

            if (_.isUndefined(this._math)) {
                throw new Error('Required mathjs interface not passed to validator.');
            }

            if (_.isUndefined(this._oi18nResources)) {
                throw new Error('Required oi18nResources LocaleElements not passed to validator.');
            }

        }
    };

    /**
     * Validates the minimum + maximum conditions
     * @memberof lux.NumberRangeValidator
     * @param {string|number} value that is being validated
     * @returns {string} original if validation was successful
     *
     * @throws {Error} when value is out of range
     */
    lux.NumberRangeValidator.prototype.validate = function(value) {
        var self = this,
            math = self._math,
            string = value ? value.toString() : value,
            customMessageSummary = self._customMessageSummary,
            customMessageDetail = self._customMessageDetail,
            messageDetailRangeOverflow = customMessageDetail.rangeOverflow,
            messageDetailRangeUnderflow = customMessageDetail.rangeUnderflow,
            messageDetailExact = customMessageDetail.exact,
            messageSummaryRangeOverflow = customMessageSummary.rangeOverflow,
            messageSummaryRangeUnderflow = customMessageSummary.rangeUnderflow,
            min = self._min !== undefined ? math.bignumber(self._min) : null,
            max = self._max !== undefined ? math.bignumber(self._max) : null,
            minStr = min && self._converter ? self._converter.format(self._min) : self._min,
            maxStr = max && self._converter ? self._converter.format(self._max) : self._max,
            normalized = string,
            params = {'value': value},
            summary, detail,
            translations = oj.Translations;

        if (!value || normalized.length < 1) {
            return true;
        }

        normalized = math.bignumber(normalized);

        if (min !== null && max !== null) {
            //range
            if ((math.largerEq(normalized, min) && math.smallerEq(normalized, max)) || math.larger(min, max)) {
                return string;
            }
        } else {
            //only min
            if (min !== null) {
                if (math.largerEq(normalized, min)) {
                    return string;
                }
            } else {
                if (max === null || math.smallerEq(normalized, max)) {
                    return string;
                }
            }
        }

        if (max !== null && min !== null && math.equal(min, max)) {
            params.num = maxStr;
            detail = messageDetailExact ?
                translations.applyParameters(messageDetailExact, params) :
                translations.getTranslatedString('oj-validator.range.number.messageDetail.exact', params);
            if (min && max) {
                if ((math.smaller(normalized, min))) {
                    params.min = minStr;
                    params.max = maxStr;
                    summary = translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeUnderflow');
                    detail = translations.getTranslatedString('oj-validator.range.number.messageDetail.rangeUnderflow', params);
                    throw new oj.ValidatorError(summary, detail);
                } else if ((math.larger(normalized, max))) {
                    params.max = maxStr;
                    summary = translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeOverflow');
                    detail = translations.getTranslatedString('oj-validator.range.number.messageDetail.rangeOverflow', params);
                    throw new oj.ValidatorError(summary, detail);
                }
            }
        } else
        if (min) {
            if ((math.smaller(normalized, min))) {
                params.min = minStr;
                summary = messageSummaryRangeUnderflow ?
                    messageSummaryRangeUnderflow :
                    translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeUnderflow');
                detail = messageDetailRangeUnderflow ?
                    translations.applyParameters(messageDetailRangeUnderflow, params) :
                    translations.getTranslatedString('oj-validator.range.number.messageDetail.rangeUnderflow', params);
                throw new oj.ValidatorError(summary, detail);
            }
        }
        if (max) {
            if ((math.larger(normalized, max))) {
                params.max = maxStr;
                summary = messageSummaryRangeOverflow ?
                    messageSummaryRangeOverflow :
                    translations.getTranslatedString('oj-validator.range.number.messageSummary.rangeOverflow');
                detail = messageDetailRangeOverflow ?
                    translations.applyParameters(messageDetailRangeOverflow, params) :
                    translations.getTranslatedString('oj-validator.range.number.messageDetail.rangeOverflow', params);
                throw new oj.ValidatorError(summary, detail);
            }
        }
    };

    /**
     * @memberof lux.NumberRangeValidator
     * A hint message may be like "Enter a value between {min} and {max}"
     * or "Enter a number greater than or equal to {min}"
     * @returns {String|null} a hint message or null if no hint is available in the options.
     */
    lux.NumberRangeValidator.prototype.getHint = function() {
        var self = this,
            hint = null,
            math = this._math,
            hints = this._hint,
            hintInRange = hints.inRange,
            hintExact = hints.exact,
            hintMinimum = hints.min,
            hintMaximum = hints.max,
            translations = oj.Translations,
            min = self._min !== undefined ? math.bignumber(self._min) : null,
            max = self._max !== undefined ? math.bignumber(self._max) : null,
            minStr = min && self._converter ? self._converter.format(self._min) : self._min,
            maxStr = max && self._converter ? self._converter.format(self._max) : self._max;

        // if both min and max are specified, the hint may say something like "Enter a value
        // between {min} and {max}".
        if (min !== null && max !== null) {
            if (!math.equal(min, max)) {
                // if hintInRange is specified (validator's hint.inRange option is set),
                // use that string, else use the default.
                hint = hintInRange ?
                    translations.applyParameters(hintInRange, {'min': minStr, 'max': maxStr}) :
                    translations.getTranslatedString('oj-validator.range.number.hint.inRange',
                        {'min': minStr, 'max': maxStr});
            } else {
                // if hintExact is specified (validator's hint.exact option is set),
                // use that string, else use the default.
                hint = hintExact ?
                    translations.applyParameters(hintExact, {'num': minStr}) :
                    translations.getTranslatedString('oj-validator.range.number.hint.exact',
                        {'num': minStr});
            }
        }
        // else if min is specified, the hint may say something like "Enter a value
        // greater than or equal to {min}".
        else if (min !== null) {
            // if hintMinimum is specified (validator's hint.min option is set),
            // use that string, else use the default.
            hint = hintMinimum ?
                translations.applyParameters(hintMinimum, {'min': minStr}) :
                translations.getTranslatedString('oj-validator.range.number.hint.min', {'min': minStr});
        }
        // else if max is specified, the hint may say something like "Enter a value
        // less than or equal to {max}".
        else if (max !== null) {
            // if hintMaximum is specified (validator's hint.max option is set),
            // use that string, else use the default.
            hint = hintMaximum ?
                translations.applyParameters(hintMaximum, {'max': maxStr}) :
                translations.getTranslatedString('oj-validator.range.number.hint.max', {'max': maxStr});
        }

        oj.Logger.info('************ validator hint: ', hint);
        return hint;
    };
    //END VALIDATION CODE ******************************************

    /**
     * @augments oj.IntlNumberConverter
     * @name lux.Intl-bignumber-converter
     * @alias module:intlBigNumberConverter
     * @constructor
     *
     * @description Constructs an immutable instance and initializes it with the options provided. When initialized
     * with no options, the default options for the current locale are assumed. The converters by
     * default use the current page locale (returned by oj.Config.getLocale()). There are several ways
     * to initialize the converter.
     *
     * <ul>
     * <li>Using options defined by the ECMA 402 Specification, these would be the properties style,
     * currency, currencyDisplay, minimumIntegerDigits, minimumFractionDigits, maximumFractionDigits,
     * useGrouping. NOTE: minimumSignificantDigits and maximumSignificantDigits are not supported.</li>
     * <li>Using a custom decimal, currency or percent format pattern. specified using the 'pattern' property</li>
     * <li>Using the decimalFormat option to define a compact pattern, such as "1M" and "1 million".</li>
     * <li>Using the roundingMode and roundDuringParse options to round the number HALF_UP, HALF_DOWN, or HALF_EVEN.</li>
     * </ul>
     * <p>
     *
     * The converter provides leniency when parsing user input value to a number in the following ways:<br/>
     *
     * <ul>
     * <li>Prefix and suffix that do not match the pattern, are removed. E.g., when pattern is
     * "#,##0.00%" (suffix is the % character), a value of "abc-123.45xyz", will be leniently parsed to
     * -123.45</li>
     * <li>When a value includes a symbol but the pattern doesn't require it.  E.g., the options are
     * {pattern: "###", currency: 'USD'}, then values ($123), (123) and -123 will be leniently parsed as
     * -123.</li>
     * </ul>
     * <p>
     * Lenient parse can be disabled by setting the property lenientParse to "none". In which case the user input must
     * be an exact match of the expected pattern and all the leniency described above will be disabled.
     * <p>
     * @property {Object=} options - an object literal used to provide optional information to
     * initialize the converter.
     * @property {string=} options.style - sets the style of number formatting. Allowed values are "decimal"
     * (the default), "currency" or "percent". When a number is formatted as a decimal, the decimal
     * character is replaced with the most appropriate symbol for the locale. In English this is a
     * decimal point ("."), while in many locales it is a decimal comma (","). If grouping is enabled the
     * locale dependent grouping separator is also used. These symbols are also used for numbers
     * formatted as currency or a percentage, where appropriate.
     * @property {string=} options.currency - specifies the currency that will be used when formatting the
     * number. The value should be a ISO 4217 alphabetic currency code. If the style is set to currency,
     * it's required that the currency property also be specified. This is because there is no default
     * currency associated with the current locale. The user must always specify the currency code
     * to be shown, otherwise an error will be thrown. The current page locale
     * (returned by oj.Config.getLocale()) determines the formatting elements of the number
     * like grouping separator and decimal separator. The currency code tells us which currency to
     * display in current page locale. JET has translations for currency names.
     * <p>
     * As an example if we want to format 1000.35 EURO and the page locale is "en-US",
     * we pass {style:'currency', currency:'EUR', currencyDisplay:'symbol'} and we will get "€1,000.35"
     * If the page locale is "fr-FR", with the same options, we will get: "1 000,35 €"
     * </p>
     * @property {string=} options.currencyDisplay - if the number is using currency formatting, specifies
     * if the currency will be displayed using its "code" (as an ISO 4217 alphabetic currency code),
     * "symbol" (a localized currency symbol (e.g. $ for US dollars, £ for Great British pounds, and so
     * on), or "name" (a localized currency name. Allowed values are "code", "symbol" and "name".
     * The default is "symbol".
     * @property {string=} options.decimalFormat -
     * specifies the decimal format length to use when style is set to "decimal".
     * Allowed values are : "standard"(default), "short" and "long". 'standard' is equivalent to not
     * specifying the 'decimalFormat' attribute, in that case the locale’s default decimal pattern
     * is used for formatting.
     * <p>
     * The user can also specify 'minimumFractionDigits' and  'maximumFractionDigits' to display.
     * When not present we use the locale's default max and min fraction digits.
     * </p>
     * <p>
     * There is no need to specify the scale; we automatically detect greatest scale that is less or
     * equal than the input number. For example  1000000 is formatted as "1M" or "1 million" and
     * 1234 is formatted, with zero fractional digits, as "1K" or " 1 thousand" for
     * short and long formats respectively. The pattern for the short and long number is locale dependent
     * and uses plural rules for the particular locale.
     * </p>
     * <p>
     * NOTE: Currently this option formats a value (e.g., 2000 -> 2K), but it does not parse a value
     * (e.g., 2K -> 2000), so it can only be used
     * in a readOnly EditableValue because readOnly EditableValue components do not call
     * the converter's parse function.
     * </p>
     * @property {number} options.minimumIntegerDigits - sets the minimum number of digits before the
     * decimal place (known as integer digits). The number is padded with leading zeros if it would not
     * otherwise have enough digits. The value must be an integer between 1 and 21.
     * @property {number} options.minimumFractionDigits - similar to 'minimumIntegerDigits', except it
     * deals with the digits after the decimal place (fractional digits). It must be an integer between
     * 0 and 20. The fractional digits will be padded with trailing zeros if they are less than the minimum.
     * @property {number} options.maximumFractionDigits - follows the same rules as 'minimumFractionDigits',
     * but sets the maximum number of fractional digits that are allowed. The value will be rounded if
     * there are more digits than the maximum specified.
     * @property {boolean} options.useGrouping - when the value is truthy, the locale dependent grouping
     * separator is used when formatting the number. This is often known as the thousands separator,
     * although it is up to the locale where it is placed. The ‘useGrouping’ is set to true by default.
     * @property {string} options.pattern an optional localized pattern, where the characters used in
     * pattern are as defined in the Unicode CLDR for numbers, percent or currency formats. When present
     * this will override the other "options". <p>
     *
     * &nbsp;&nbsp;- When the pattern represents a currency style the 'currency' property is required to
     * be set, as not setting this will throw an error. The 'currencyDisplay' is optional. <br/>Example:
     * {pattern: '¤#,##0', currency: 'USD'}. <p>
     *
     * &nbsp;&nbsp;- It's not mandatory for the pattern to have the special character '¤' (currency sign)
     * be present. When not present, values are treated as a currency value, but are not formatted to
     * show the currency symbol. <br/>Example: {pattern: '#,##0', currency: 'USD'} <p>
     *
     * &nbsp;&nbsp;- When the pattern represents a percent style, the percent special character ('%') needs to be
     * explicitly specified in the pattern, e.g., {pattern: "#,##0%"}. If the pattern does not contain
     * the percent character it's treated as a decimal pattern, unless the style is set to percent,
     * in which case the value is treated as a percent value, but not formatted to show the percent symbol.
     * <br/>Example: {style: 'percent', pattern: "#,##0"}. <p>
     *
     * &nbsp;&nbsp;- A decimal pattern or exponent pattern is specified in the pattern using the CLDR
     * conventions. <br/>Example: {pattern: "#,##0.00"} or {pattern: "0.##E+0"}. <p>
     *
     * NOTE: 'pattern' is provided for backwards compatibility with existing apps that may want the
     * convenience of specifying an explicit format mask. Setting a pattern will override the default
     * locale specific format. <br/>
     *
     * @property {string} options.roundingMode - specifies the rounding behavior.
     * This follows the Java.Math.RoundingMode behavior.
     * Currently we support the options: HALF_UP, HALF_DOWN, and HALF_EVEN
     *
     * @property {boolean} options.roundDuringParse - Specifies whether or not to round during
     * parse. Defaults to false; the number converter rounds during format but not during parse.
     *
     * @property {Object} options.separators - An object with 2 fields: 'decimal' and 'group'.
     * It allows the user to provide custom decimal and grouping separators. It is accepted for both
     * format and parse methods.
     * <br/>
     *
     * @property {string} options.lenientParse - The lenientParse property can be used to enable or disable leninet parsing.
     *  Allowed values: "full" (default), "none".
     * <p style='padding-left: 5px;'>
     * By default the lenient parse is enabled and the leniency rules descibed above will be used. When lenientParse is
     * set to "none" the lenient parse is disabled and the user input must match the expected input otherwise an exception will
     * be thrown.<br/><br/>
     *
     * @example <caption>Create a number converter for currencies</caption>
     * var converterFactory = oj.Validation.converterFactory("bignumber");
     * var options = {style: "currency", currency: "USD", minimumIntegerDigits: 2};
     * converter = converterFactory.createConverter(options);
     * converter.format(9); --> "$09.00" if page locale is 'en-US'
     * converter.format(9); --> "09,00 $US" if page locale is 'fr-FR'<br/>
     *
     * @example <caption>A number converter for percent values using a custom (CLDR) pattern</caption>
     * var converterFactory = oj.Validation.converterFactory("bignumber");
     * var options = {pattern: '#,##0%'};
     * converter = converterFactory.createConverter(options);<br/>
     *
     * @example <caption>To parse a value as percent but format it without displaying the percent character</caption>
     * var options = {style: 'percent', pattern: '#,##0'};<br/>
     *
     * @example <caption>To parse a value as currency using a custom (CLDR) pattern</caption>
     * var options = {pattern: '¤#,##0', currency: 'USD'};
     *
     * @example <caption>The following decimalFormat examples are in en locale.
     * To format a value as short (default for fraction digits is based on the locale)</caption>
     * var options = {style:’decimal’, decimalFormat:’short’};
     * converter = converterFactory.createConverter(options);
     * converter.format(12345);--> 12.354K<br/>
     *
     * @example <caption>To format a value as long (default for fraction digits is based on the locale):</caption>
     * var options = {style:’decimal’, decimalFormat:’long’};
     * converter = converterFactory.createConverter(options);
     * converter.format(12345);--> 12.345 thousand<br/>
     *
     * @example <caption>To format a value as short with minimum fraction digits:</caption>
     * options = { style:’decimal’, decimalFormat:’short’,
     * minimumFractionDigits:4};
     * converter = converterFactory.createConverter(options);
     * converter.format(1234);--> 1.2340K<br/>
     *
     * @example <caption>To format a value as short with maximum fraction digits:</caption>
     * options = { style:’decimal’, decimalFormat:’short’,
     * maximumFractionDigits:0};
     * converter = converterFactory.createConverter(options);
     * converter.format(12345);--> 12K<br/>
     *
     * @example <caption>To format a value as long with minimum and maximum fraction digits:</caption>
     * options = { style:’decimal’, decimalFormat:’long',
     * minimumFractionDigits:2, maximumFractionDigits:4};
     * converter = converterFactory.createConverter(options);
     * converter.format(12000);--> 12.00 thousand<br/>
     *
     * @example <caption>To format a value as short with minimum and maximum fraction digits:</caption>
     * options = { style:’decimal’, decimalFormat:’long',
     * minimumFractionDigits:2, maximumFractionDigits:4};
     * converter = converterFactory.createConverter(options);
     * converter.format(12345678);--> 12.345 million<br/>
     *
     * @example <caption>decimal style default is standard:</caption>
     * options = { style:’decimal’, decimalFormat:’standard’};
     * converter = converterFactory.createConverter(options);
     * converter.format(12345);--> 12,345<br/>
     *
     * @example <caption>decimal round HALF_DOWN:</caption>
     * options = { style:’decimal’,  maximumFractionDigits:2, roundingMode:'HALF_DOWN'};
     * converter = converterFactory.createConverter(options);
     * converter.format(0.225);--> 0.22
     * converter.parse(0.225);-->0.225 //doesn't round during parse by default<br/>
     *
     * @example <caption>decimal round HALF_UP:</caption>
     * options = { style:’decimal’,  maximumFractionDigits:2, roundingMode:'HALF_UP'};
     * converter = converterFactory.createConverter(options);
     * converter.format(0.225);--> 0.23
     * converter.parse(0.225);--> 0.225 //doesn't round during parse by default<br/>
     *
     * @example <caption>decimal round HALF_EVEN:</caption>
     * options = { style:’decimal’,  maximumFractionDigits:2, roundingMode:'HALF_EVEN'};
     * converter = converterFactory.createConverter(options);
     * converter.format(0.225);--> 0.22
     * converter.format(0.235);--> 0.24
     * converter.parse(0.225);--> 0.225 //doesn't round during parse by default
     * converter.parse(0.235);--> 0.235 //doesn't round during parse by default<br/>
     *
     * @example <caption>decimal round HALF_DOWN and roundDuringParse:</caption>
     * options = { style:’decimal’, maximumFractionDigits:2,
     *             roundingMode:'HALF_DOWN', roundDuringParse: true};
     * converter = converterFactory.createConverter(options);
     * converter.format(0.225);--> 0.22
     * converter.parse(0.225);-->0.22<br/>
     *
     * @example <caption>decimal round HALF_UP and roundDuringParse:</caption>
     * options = { style:’decimal’,  maximumFractionDigits:2,
     *             roundingMode:'HALF_UP', roundDuringParse: true};
     * converter = converterFactory.createConverter(options);
     * converter.format(0.225);--> 0.23
     * converter.parse(0.225);--> 0.23<br/>
     *
     * @example <caption>decimal round HALF_EVEN and roundDuringParse:</caption>
     * options = { style:’decimal’,  maximumFractionDigits:2,
     *             roundingMode:'HALF_EVEN', roundDuringParse: true};
     * converter = converterFactory.createConverter(options);
     * converter.format(0.225);--> 0.22
     * converter.format(0.235);--> 0.24
     * converter.parse(0.225);--> 0.22
     * converter.parse(0.235);--> 0.24<br/>
     *
     * @example <caption>Use custom decimal and grouping separators:</caption>
     * options = { style:'decimal', separators: {decimal: ',',  group: '.'}};
     * converter = converterFactory.createConverter(options);
     * var nb = 1234567.89;
     * converter.format(nb);--> 1.234.567,89
     * converter.parse("1.234.567,89");--> 1234567.89
     * <br/>
     *
     * @example <caption>Disable lenient parse:</caption>
     * options = { style:'decimal',  lenientParse: 'none'};
     * converter = converterFactory.createConverter(options);
     * converter.parse("abc-123.45xyz");--> Error: Unparsable number abc-123.45xyz The expected number pattern is #,##0.###
     * <br/>
     */
    lux.IntlBigNumberConverter = function(options) {
        this.Init(options);
    };

    oj.Object.createSubclass(lux.IntlBigNumberConverter, oj.IntlNumberConverter, 'lux.IntlBigNumberConverter');

    /**
     * Initializes the number converter instance with the set options.
     * @param {Object=} options an object literal used to provide an optional information to
     * initialize the converter.<p>
     * @returns {undefined}
     * @export
     */
    lux.IntlBigNumberConverter.prototype.Init = function(options) {
        lux.IntlBigNumberConverter.superclass.Init.call(this, options);
    };

    /**
     * Formats a Number and returns the formatted string, using the options this converter was
     * initialized with.
     *
     * @method module:intl-bigNumber-converter#format
     * @param {string} value to be formatted for display
     * @return {string} the localized and formatted value suitable for display. When the value is
     * formatted as a percent it's multiplied by 100.
     *
     * @throws {Error} a ConverterError both when formatting fails, or if the options provided during
     * initialization cannot be resolved correctly.
     *
     */
    lux.IntlBigNumberConverter.prototype.format = function (value) {
        var formatted = value;

        // if string value assume bignumber and format the value
        // for the correct number of digits after the decimal point
        // rounded.
        if (_.isString(value) && value.length > 0) {
            formatted = this._bigNumberFormat(value);
        }

        formatted = lux.IntlBigNumberConverter.superclass.format.call(this, formatted);

        return formatted;
    };

    lux.IntlBigNumberConverter.prototype._bigNumberFormat = function (value) {
        var math = luxBignum.BigNumberFactory().createBigNumber({number: luxBignum.BigNumberFactory.BIGNUMBER, precision: 128, epsilon: 1e-25}),
            maximumFractionDigits = this._options.maximumFractionDigits,
            minimumFractionDigits = this._options.minimumFractionDigits,
            bigValue = math.bignumber(value),
            precision = bigValue.dp(),
            formattedValue;

        // if maximumFractionalDigits are less than minimumFractionalDigits
        // for purpose of determining the rounding precision, set the maximumFractionalDigits
        // to maximumFractionalDigits.
        if (minimumFractionDigits > maximumFractionDigits) {
            maximumFractionDigits = minimumFractionDigits;
        }

        // check decimal precision.
        if (precision > maximumFractionDigits) {
            // If actual precision is greater than the maximumFractionalDigits force the maximumFractionalDigits.
            // This will round the entered fractionalDigit.
            precision = maximumFractionDigits;
        } if (precision < minimumFractionDigits) {
            // If actual precision is less than the minimumFractionDigits force use the minimumFractionDigits.
            // This will append a 0 the fractionalDigit.
            precision = minimumFractionDigits;
        }

        if (_.isString(value) && value.length === 0) {
            value = '0';
        }

        formattedValue = math.format(math.bignumber(value), {'notation': 'fixed', 'precision': precision});

        return formattedValue;
    };

    /**
     * Normalizes the input removing the currency and separators characters
     *
     * @method module:intl-bigNumber-converter#normalize
     * @description returns the value stripped of currency and comma / space separators characters.
     * @param {string} value to be normalized for display
     * @param {string} locale - locale to normalize to.
     * @return {string}
     *
     * @throws {Error} a ConverterError both when formatting fails, or if the options provided during
     * initialization cannot be resolved correctly.
     *
     */
    lux.IntlBigNumberConverter.prototype.normalize = function (value, locale) {
        var self = this;

        return self._getWrapped().normalize(value, locale);
    };

    // Returns the wrapped number converter implementation object.
    lux.IntlBigNumberConverter.prototype._getWrapped = function () {
        if (!this._wrapped) {
            this._wrapped = oi18n.NumberConverter.getInstance();
        }

        return this._wrapped;
    };

    oj.ConverterFactory.CONVERTER_TYPE_BIGNUMBER = 'bignumber';
    oj.ConverterFactory.VALIDATOR_TYPE_BIGNUMBERRANGE = 'validatebignumber';

    /**
     * A factory implementation to create the built-in number converter of type
     * {@link lux.IntlNumberConverter}.
     *
     * @name lux.NumberConverterFactory
     * @class
     *
     * @example <caption>create an instance of the jet datetime converter using the options provided</caption>
     * var ncf = oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_BIGNUMBER);
     * var salaryOptions = {currency: "EUR" , pattern: "¤#,##0.00;(¤#,##0.00)"};
     * var salaryConverter = ncf.createConverter(salaryOptions);
     * @public
     * @since 0.6
     *
     */
    lux.BigNumberConverterFactory = (function () {
        function _createBigNumberConverter(options) {
            return new lux.IntlBigNumberConverter(options);
        }
        return {
            /**
             * Creates an immutable (jet) number converter instance.
             *
             * @param {Object=} options an object literal used to provide optional information to initialize
             * the jet number converter with. For details on what to pass for options, refer to
             * {@link IntlNumberConverter}
             *
             * @return {lux.IntlNumberConverter}
             * @memberOf oj.NumberConverterFactory
             * @public
             */
            'createConverter': function(options) {
                return _createBigNumberConverter(options);
            }
        };
    }()); // notice immediate invocation of anonymous function

    /** Register the default factory provider function */
    oj.Validation.converterFactory(oj.ConverterFactory.CONVERTER_TYPE_BIGNUMBER, lux.BigNumberConverterFactory);

    /**
     * a factory method to create an instance of a built-in numberRange validator of type
     * {@link lux.NumberRangeValidator}.
     *
     * @example <caption>create an instance of the numberRange validator using the factory</caption>
     * var lrvf = oj.Validation.validatorFactory(oj.ConverterFactory.VALIDATOR_TYPE_BIGNUMBERRANGE);
     * var options = {hint: {min: 'Enter a value greater than {min}'}, min: 100};
     * var lrValidator = lrvf.createValidator(options);
     *
     * @name lux.NumberRangeValidatorFactory
     * @class
     *
     */
    lux.NumberRangeValidatorFactory = (function () {
        function _createNumberRangeValidator(options) {
            return new lux.NumberRangeValidator(options);
        }
        return {
        /**
         * Creates an immutable validator instance of type {@link lux.NumberRangeValidator} that ensures
         * that the value provided is within a given range.
         *
         * @param {Object=} options an object literal used to provide the minimum, maximum and other
         * optional values. See {@link NumberRangeValidator} for details.<p>
         *
         * @return {oj.NumberRangeValidator}
         * @memberOf oj.NumberRangeValidatorFactory
         * @public
         */
            'createValidator': function(options) {
                return _createNumberRangeValidator(options);
            }
        };
    }()); // notice immediate invocation of anonymous function

    /** Register the default factory provider function */
    oj.Validation.validatorFactory(oj.ConverterFactory.VALIDATOR_TYPE_BIGNUMBERRANGE, lux.NumberRangeValidatorFactory);

    return lux;
});