
/**
 * @module luxFormatters
 */
define('luxCore/formatters/numberFormatter', ['underscore'], function (_) {
    'use strict';

    /**
     * The number formatter allows you to validate and format numbers.
     *
     * The require module will return a function. This function when called with options
     * will return an object with 3 functions: 'format', 'unformat', 'validate'.
     *
     * @example
     *     define(['luxCore.numberFormatter'], function(numberFormatter) //Defined numberFormater
     *
     *     //...
     *
     *     //Usage:
     *     formatter = numberFormatter({
     *         format : '0' //This says to print out a number
     *                      //with no decimal separator, currency symbol,
     *                      //thousands separator and no digits following the decimal
     *     });
     *
     *     formatter.format(0); //Returns '0'
     *
     *     formatter.format(100); //Returns '100'
     *
     *     formatter.format(100.00); //Also returns '100'
     *
     *     //Currency usage:
     *     formatter = numberFormatter({
     *         format : '$0,0.00',
     *         symbol : '$'
     *     });
     *
     *     formatter.format(0); //Returns '$0.00'
     *
     *     formatter.format(10000.01); //Returns '$10,000.01'
     *
     *     //Append suffix to number for presenting hours and days
     *     formatter = numberFormatter({
     *         format : '0h'
     *     });
     *
     *     formatter.format(0); //Returns '0h'
     *
     *     formatter.format(7); //Returns '7h'
     *
     * @class numberFormatter
     * @constructor
     * @alias module:luxFormatters.numberFormatter
     * @param {Object} [opts] The options to use for any future formats using the created formatter.
     */
    var formatter = function(opts) {

        var defaults = {
            /**
             * @property decimalSeparator
             * @type {String}
             * @default '.'
             */
                decimalSeparator:   '.',
                /**
             * @property thousandsSeparator
             * @type {String}
             * @default ','
             */
                thousandsSeparator: ',',
                /**
             * Represents the currency symbol
             *
             * @property symbol
             * @type {String}
             * @default ''
             */
                symbol:             '',
                /**
             * @property negativeSymbol
             * @type {String}
             * @default '-'
             */
                negativeSymbol:     '-',
                /**
             * @property positiveSymbol
             * @type {String}
             * @default '+'
             */
                positiveSymbol:     '+',
                /**
             * This property represents the format the number shall become.
             *
             * It is a small language where:
             *
             * <pre>
             *    0                : <strong>digits</strong>
             *    ,                : <strong>use thousands separator</strong>
             *    .                : <strong>use decimal separator</strong>
             *    $                : <strong>use the currency symbol (aka symbol)</strong>
             *    \-               : <strong>force display of the sign symbol</strong>
             *    # of 0's after . : <strong>use # decimal places</strong>
             * </pre>
             *
             * *Note:* negative symbols will always be displayed when
             *          a number is negative however this will allow for the
             *          placement of the sign and make the positive symbol display.
             *
             * Everything not in this language will print out at its position in the string.
             *
             * @property format
             * @type {String}
             * @default '0'
             */
                format:             '0'
            }, maxPrecision, flags = {
                sign: function(value, curValue, options) {
                //Intentionally not putting a sign on 0
                    if (value > 0) {
                        curValue += options.positiveSymbol;
                    } else if (value < 0) {
                        curValue += options.negativeSymbol;
                    }
                    return curValue;
                },
                digits: function(value, curValue, group, decimal) {
                    if (!group && !decimal) {
                    //First 0 - print all numbers before decimal
                        if (value >= 0) {
                            curValue += parseInt(value, 10);
                        } else {
                            curValue += parseInt(value, 10).toString().substring(1);
                        }
                    } else if (decimal) {
                    //These are the 0s after the decimal separator
                        if (maxPrecision.length > 0){
                            curValue += maxPrecision.slice(0, 1);
                            maxPrecision = maxPrecision.substring(1);
                        } else {
                            curValue += '0';
                        }
                    }

                    return curValue;
                }
            }, globalOptions = _.defaults(opts || {}, defaults);

        return {
            /**
             * Format string/number value
             *
             * The below options are the default values.
             *
             *      options : {
             *          decimalSeparator : '.',
             *          thousandsSeparator : ',',
             *          symbol : '',
             *          negativeSymbol : '-',
             *          positiveSymbol : '+',
             *          format : '0'
             *      }
             *
             * @method module:luxFormatters#format
             * @param {String|Number} value The value to format
             * @param {Object} opts The options to use for this one function call
             * @return String
             */
            format: function(value, opts) {
                var formattedValue = '', group = false, decimal = false, options, sign = false;

                if (_.isNull(value) || value === '') {
                    return '';
                }

                if (!_.isUndefined(opts)) {
                    //We got new options - overwrite default ones;
                    options = _.defaults(opts, defaults);
                } else {
                    options = globalOptions;
                }

                if (_.isString(value)) {
                    value = parseFloat(value);
                }

                if (!_.isNumber(value)) {
                    throw new Error('Must be a number');
                }

                maxPrecision = value.toPrecision(15).replace(/(-)?\d+\./, '');

                if (!_.isNull(options.format)) {
                    //Use string format to parse
                    _.each(options.format, function(ch) {
                        switch (ch) {
                        case '-':
                            //show sign
                            formattedValue = flags.sign(value, formattedValue, options);
                            sign = true;
                            break;
                        case '0' :
                            //print out all digits
                            formattedValue = flags.digits(value, formattedValue, group, decimal);
                            break;
                        case '.' :
                            //print decimal separator
                            formattedValue += options.decimalSeparator;
                            decimal = true;
                            break;
                        case '$' :
                            //print out currency symbol
                            formattedValue += options.symbol;
                            break;
                        case ',' :
                            //group all digits with thousands separator
                            group = true;
                            break;
                        default :
                            formattedValue += ch;
                        }
                    });

                    if (group) {
                        formattedValue = formattedValue.replace(/(\d)(?=(\d{3})+(?!\d))/g, '$1' + options.thousandsSeparator);
                    }

                    if (value < 0 && !sign) {
                        formattedValue = formattedValue.replace(/(\d)/, options.negativeSymbol + '$1');
                    }
                }

                return formattedValue;
            },
            /**
             * Unformats a string.
             *
             * @method module:luxFormatters#unformat
             * @param {String} valueStr The string to format to a number.
             * @param {Object} opts The options to use for this unformat operation.
             *
             * @return {Number} The number that the string represents.
             */
            unformat: function(valueStr, opts) {
                var options, newString;
                if (_.isUndefined(opts)) {
                    options = globalOptions;
                } else {
                    options = _.defaults(opts, defaults);
                }

                newString = valueStr.replace(new RegExp(options.thousandsSeparator, 'g'), '')
                    .replace(options.decimalSeparator, '.')
                    .replace(/[^\d\.\-\+]/g, '');

                return parseFloat(newString);
            },
            /**
             * Validate string value
             *
             * The below options are the default values.
             *
             *      options : {
             *          decimalSeparator : '.',
             *          thousandsSeparator : ',',
             *          symbol : '',
             *          negativeSymbol : '-'
             *      }
             *
             * @method module:luxFormatters#validate
             * @param {string} value
             * @param {object} options
             * @return boolean
             */
            validate: function(value, opts) {
                var multipleDecimalRegex, options;
                if (_.isNumber(value)){
                    return true;
                }
                if (_.isUndefined(opts)) {
                    options = globalOptions;
                } else {
                    options = _.defaults(opts, defaults);
                }

                if (options.decimalSeparator === '.') {
                    multipleDecimalRegex = /.*\..*\..*/;
                } else {
                    multipleDecimalRegex = new RegExp('.*' + options.decimalSeparator + '.*' + options.decimalSeparator + '.*');
                }

                if (isNaN(value) || value.match(multipleDecimalRegex)) {
                    return false;
                }
                return true;
            }
        };
    };

    return formatter;

});

define('luxCore/formatters/dateFormatter', ['underscore'], function(_) {
    'use strict';

    /**
     * *Note*: The date formatter uses [Java Simple Date format flags](http://docs.oracle.com/javase/6/docs/api/java/text/SimpleDateFormat.html).
     *
     * @example
     *     define(['luxCore.dateFormatter'], function(dateFormatter) { //Define date formatter
     *
     *     //...
     *
     *     var timeFormatter = dateFormatter({ format : 'hh:mm a' }),
     *         dateFormatter = dateFormatter({ format : 'yyyy-MM-dd'});
     *
     *     timeFormatter.format(new Date('10/11/12 1:02:03 PM')); //Returns '01:02 PM';
     *
     *     dateFormatter.format(new Date('10/11/12 1:02:03 PM')); //Returns '2012-10-11';
     *
     * @class dateFormatter
     * @constructor
     * @alias module:luxFormatters.dateFormatter
     * @param {Object} options The options for the remaining uses of the formatter
     * @param {Object} lang    The language to use for localization.
     */
    var formatter = function(options, lang) {

        var language,
            pad = function(num, length) {
                num = num.toString();
                length = length || 2;

                while (num.length < length) {
                    num = '0' + num;
                }

                return num;
            }, defaultOpts = {
                format: 'MMM-dd-yyyy'
            },
            defaultLanguage = {
                am: {
                    lower: 'am',
                    upper: 'AM'
                },
                pm: {
                    lower: 'pm',
                    upper: 'PM'
                },
                month: {
                    abbrev: [
                        'Jan',
                        'Feb',
                        'Mar',
                        'Apr',
                        'May',
                        'Jun',
                        'Jul',
                        'Aug',
                        'Sep',
                        'Oct',
                        'Nov',
                        'Dec'
                    ],
                    full: [
                        'January',
                        'February',
                        'March',
                        'April',
                        'May',
                        'June',
                        'July',
                        'August',
                        'September',
                        'October',
                        'November',
                        'December'
                    ]
                },
                day: {
                    abbrev: [
                        'Sun',
                        'Mon',
                        'Tue',
                        'Wed',
                        'Thu',
                        'Fri',
                        'Sat'
                    ],
                    full: [
                        'Sunday',
                        'Monday',
                        'Tuesday',
                        'Wednesday',
                        'Thursday',
                        'Friday',
                        'Saturday'
                    ]
                }
            }, flags = {
                G: function(date, isUTC) {
                    if (isUTC && date.getUTCFullYear() >= 0) {
                        return 'AD';
                    } else if (isUTC) {
                        return 'BC';
                    } else if (date.getFullYear() >= 0) {
                        return 'AD';
                    } else {
                        return 'BC';
                    }
                },
                z: function(date, isUTC) {
                    function convertTimezoneOffsetToHours() {
                        var offset = date.getTimezoneOffset();
                        if (offset < 0) {
                            return '+' + pad(Math.abs(offset) / (60)) + ':00';
                        } else {
                            return '-' + pad(Math.abs(offset) / (60)) + ':00';
                        }
                    }

                    return isUTC ? 'Z' : 'GMT' + convertTimezoneOffsetToHours();
                },
                Z: function(date, isUTC) {
                    function convertTimezoneOffsetToHours() {
                        var offset = date.getTimezoneOffset();
                        if (offset < 0) {
                            return '+' + pad(Math.abs(offset) / (60)) + '00';
                        } else {
                            return '-' + pad(Math.abs(offset) / (60)) + '00';
                        }
                    }

                    return isUTC ? 'Z' : convertTimezoneOffsetToHours();
                },
                //Days
                dd: function(date, isUTC) { //Days with leading 0s 01-31
                    return isUTC ? pad(date.getUTCDate()) : pad(date.getDate());
                },
                d: function(date, isUTC) { //Days without leading zeros
                    return isUTC ? date.getUTCDate() : date.getDate();
                },
                EEE: function(date, isUTC) { // Abbreviated day Mon-Sun
                    return isUTC ? language.day.abbrev[date.getUTCDay()] : language.day.abbrev[date.getDay()];
                },
                EEEE: function (date, isUTC) { //Full day
                    return isUTC ? language.day.full[date.getUTCDay()] : language.day.full[date.getDay()];
                },
                //Months
                M: function (date, isUTC) { //Nonpadded month number
                    return isUTC ? date.getUTCMonth() : date.getMonth();
                },
                MM: function (date, isUTC) { //Padded month number
                    return pad(isUTC ? date.getUTCMonth() + 1 : date.getMonth() + 1);
                },
                MMM: function (date, isUTC) { //Abbreviated month
                    return language.month.abbrev[isUTC ? date.getUTCMonth() : date.getMonth()];
                },
                MMMM: function (date, isUTC) { //Full month format
                    return language.month.full[isUTC ? date.getUTCMonth() : date.getMonth()];
                },
                //Year
                yy: function (date, isUTC) { //2 digit years
                    return isUTC ? date.getUTCFullYear().toString().slice(2) : date.getFullYear().toString().slice(2);
                },
                yyyy: function (date, isUTC) { //4 digit year
                    return isUTC ? date.getUTCFullYear() : date.getFullYear();
                },
                //Time
                a: function (date, isUTC) { //Uppercase AM/PM
                    if (isUTC && date.getUTCHours() < 12) {
                        return language.am.upper;
                    } else if (isUTC) {
                        return language.pm.upper;
                    } else if (date.getHours() < 12) {
                        return language.am.upper;
                    } else {
                        return language.pm.upper;
                    }
                },
                h: function (date, isUTC) { //1-12 hours
                    var hour = isUTC ? date.getUTCHours() % 12 : date.getHours() % 12;
                    if (hour === 0) {
                        return '12';
                    } else {
                        return hour;
                    }
                },
                H: function (date, isUTC) { //0-23 hours
                    return isUTC ? date.getUTCHours() : date.getHours();
                },
                hh: function (date, isUTC) { //Padded 12 hours
                    var hour = isUTC ? date.getUTCHours() % 12 : date.getHours() % 12;
                    if (hour === 0) {
                        return '12';
                    } else {
                        return pad(hour);
                    }
                },
                HH: function (date, isUTC) { //Padded 24 hours
                    return pad(isUTC ? date.getUTCHours() : date.getHours());
                },
                mm: function (date, isUTC) { //Padded minutes
                    return pad(isUTC ? date.getUTCMinutes() : date.getMinutes());
                },
                ss: function (date, isUTC) { //Padded seconds
                    return pad(isUTC ? date.getUTCSeconds() : date.getSeconds());
                },
                S: function (date, isUTC) {
                    return pad(isUTC ? date.getUTCMilliseconds() : date.getMilliseconds(), 1);
                },
                SS: function (date, isUTC) {
                    return pad(isUTC ? date.getUTCMilliseconds() : date.getMilliseconds(), 2);
                },
                SSS: function (date, isUTC) {
                    return pad(isUTC ? date.getUTCMilliseconds() : date.getMilliseconds(), 3);
                }
            };

        options = _.defaults(options || {}, defaultOpts);
        language = _.defaults(lang || {}, defaultLanguage);

        return {
            /**
             * @method module:luxFormatters#format
             * @param {Date|String} date The date to format.
             * @param {Boolean}     [isUTC] Whether or not the date is to be parsed as UTC.
             *
             * @return {String} The formatted string.
             */
            format: function (date, isUTC) {
                var token = /d{1,4}|M{1,4}|yy(?:yy)?|h{1,2}|H{1,2}|k{1,2}|m{1,2}|s{1,2}|S{1,3}|a|G|\'.*\'|z|EEE(?:E)?|Z/g,
                    correctedDate, rfc3339FullDate = /^\d{4}-\d{2}-\d{2}$/, mask = options.format;

                isUTC = isUTC || false;

                if (_.isString(date)) {
                    if (date.match(rfc3339FullDate)) {
                        correctedDate = new Date(date);
                        date = new Date(correctedDate.getTime() + (correctedDate.getTimezoneOffset() * 60000));
                    } else {
                        date = new Date(date);
                    }
                }

                return mask.replace(token, function(match) {
                    return flags[match] ? flags[match](date, isUTC) : match.slice(1, match.length - 1);
                });
            },
            /**
             * @method module:luxFormatters#unformat
             * @param {String} date The date string to be parsed and returned as a Date
             *
             * @return {Date} The date the string represents.
             */
            unformat: function (date) {

                if (!_.isString(date)){
                    throw new Error('Date should be a string');
                }

                if (_.isNaN(Date.parse(date))) { //Assuming this is a time
                    return new Date(new Date().toDateString() + ' ' + date);
                } else {
                    return new Date(date);
                }
            },
            /**
             * @method module:luxFormatters#validate
             * @param {Date|String} The value to validate
             *
             * @return {Boolean} Whether or not the parameter is valid.
             */
            validate: function(date) {
                if (_.isNaN(Date.parse(date))){
                    return false;
                }

                return true;
            }
        };
    };

    return formatter;
});

define('luxFormattersAPI', ['luxCore/formatters/numberFormatter', 'luxCore/formatters/dateFormatter'],
    function (nF, dF) {
        'use strict';

        return {
            numberFormatter: nF,
            dateFormatter:   dF
        };
    }
);
