/*global define */
/**
 * Copyright (c) 2013 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implements a tree structure that extends from luxtree and provides some specific additional properties and
 * functions for describing navigational hierarchies
 * @author David Clifton
 */

/**
 * A module representing a navigational hierarchy as a tree
 * @module luxhierarchy
 * @requires jquery
 */
define(['jquery', 'luxtree'],
    function($, luxTree) {

        /**
         * Constructor for the hierarchy node
         * @alias module:luxhierarchy
         * @extends module:luxtree
         * @param parent The parent node
         * @param data (optional) Data to store in the data property
         * @param id (optional) ID
         */
        var luxHierarchyNode = function(parent, data, id) {
            luxTree.call(this, parent, data, id);
            this.actionLabel = '';
            this.action = '';
            this.childrenLabel = 'Child Nodes';
            this.additionalActions = [];
        };
        luxHierarchyNode.prototype = new luxTree();
        luxHierarchyNode.prototype.constructor = luxHierarchyNode;

        // Replace the logic for adding a child node to utilize the specialized node
        /**
         * Adds a child node - note that the 'data' is what is expected, not a node itself - a node is returned
         * @param data Data to populate on the node's data property
         * @param id (optional) Value to use for identifier
         * @returns {luxhierarchy} The newly constructed child node
         */
        luxHierarchyNode.prototype.addChild = function(data, id) {

            var child = new luxHierarchyNode(this, data, id);
            this.children.push(child);
            return child;

        };

        /**
         * Copies data from the object passed, provides special handling for the 'children' property
         * @param obj The object from which to initialize this node, copying all its data
         * @returns {*} Value of 'this' after initialization
         */
        luxHierarchyNode.prototype.initializeFromObject = function(obj) {

            for (var prop in obj) {
                if (prop !== 'children') {
                    this[prop] = obj[prop];
                }
            }

            // Handle children
            var children = obj.children;
            if (children) {
                for (var i = 0; i < children.length; i++) {
                    (this.addChild()).initializeFromObject(children[i]);
                }
            }
            return this;

        };

        /**
         * Provides for de-serializing JSON into a tree
         * @param jsonData String containing the JSON data to parse
         * @returns {*} Essentially the value of 'this' which will be the top level node defined in the JSON structure
         */
        luxHierarchyNode.prototype.parseJSON = function(jsonData) {

            return this.initializeFromObject($.parseJSON(jsonData));

        };

        /**
         * Find the just one node with a particular action key
         * @param action The action value to search for on the nodes
         * @returns {*} A matching node or null
         */
        luxHierarchyNode.prototype.findNodeByAction = function(action) {

            return this.find(function(node) {
                if (node.action === action) {
                    return true;
                }
            });

        };

        return luxHierarchyNode;
    });