/*global define */
/**
 * Copyright (c) 2013 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implementation of a tree data structure for use in various components
 * @author David Clifton
 */

/**
 * A module representing a tree data structure
 * @module luxtree
 * @requires underscore
 */
define(['underscore'],
    function(_) {

    /**
     * Generates a unique node id (unique in the context of the tree)
      */
        function generateNodeId(parent, separator) {

            separator = separator || '/';
            if (parent) {
                return [parent.id, parent.children.length].join(separator);
            } else {
                return '0';
            }

        }

        function _traverseUp(context, iterator) {

        // Performance optimization: declare these variables outside while
            var i, node;
            while (context) {
                if (iterator(context) === false) {
                    return;
                }
                for (i = 0; i < context.children.length; i++) {
                    node = context.children[i];
                    if (iterator(node) === false) {
                        return;
                    }
                }
                context = context.parent;
            }
        // No return value: pass back by reference: context

        }

        function _traverseDown(context, iterator) {

            (function walkDown(node) {
                var i, newContext;
                if (iterator(node) === false) {
                    return;
                } else {
                    for (i = 0; i < node.children.length; i++) {
                        newContext = node.children[i];
                        walkDown(newContext);
                    }
                }
            })(context); // Recursive singleton function

        }

        // Tracks nodes visited while executing iterator
        function _traverse(context, iterator, callback) {
            var visited = [];
            var callIterator = function(node) {
                var id = node.id, returned;
                if (_.indexOf(visited, id) === -1) {
                    returned = iterator.call(node, node);
                    visited.push(id);
                    if (returned === false) {
                        return returned;
                    }
                }
            };
            callback(context, callIterator);
        }

        /**
     * The tree constructor
     * @alias module:luxtree
     * @param parent The parent node
     * @param data (optional) Data to store in the data property
     * @param id (optional) ID
     */
        var luxTree = function(parent, data, id) {

        // If data is passed in, great - otherwise, just blank object
            this.data = data || {};

            // If parent is passed, that is our parent, otherwise we are a new root
            this.parent = parent || null;

            // Get a unique id
            this.id = id || generateNodeId(parent);

            // Children is an ordered list for this implementation
            this.children = [];

            // Depth is tracked as an optimization, avoiding the need to execute recursive operations to determine it
            this.depth = parent ? parent.depth + 1 : 0;

        };

        /**
     * Am I the root node?
     * @returns {boolean} Whether this node is root
     */
        luxTree.prototype.isRoot = function() {

            return !this.parent;

        };

        /**
     * Simply find the root of the tree
     * @returns {*} Returns the node which is the root of this node's tree
     */
        luxTree.prototype.findRoot = function() {

            if (this.isRoot()) {
                return this;
            } else {
                var currentNode = this;
                while (!currentNode.isRoot()) {
                    currentNode = currentNode.parent;
                }
                return currentNode;
            }

        };

        /**
     * Gets an array of the parents from root down to the current node's parent
     * @returns {Array} Object array of ancestor nodes
     */
        luxTree.prototype.listParents = function() {

            if (this.isRoot()) {
                return [];
            } else {
                var currentNode = this;
                var nodeArray = [];
                while (!currentNode.isRoot()) {
                    currentNode = currentNode.parent;
                    nodeArray.push(currentNode);
                }
                return nodeArray.reverse();
            }

        };

        /**
     * Adds a child node - note that the 'data' is what is expected, not a node itself - a node is returned
     * @param data Data to populate on the node's data property
     * @param id (optional) Value to use for identifier
     * @returns {luxtree} The newly constructed child node
     */
        luxTree.prototype.addChild = function(data, id) {

            var child = new luxTree(this, data, id);
            this.children.push(child);
            return child;

        };

        /**
     * Removes a specified child node of this node
     * @param child The node to remove
     * @returns {*} The collection of remaining child nodes
     */
        luxTree.prototype.removeChild = function(child) {

            if (typeof child === 'number' && this.children[child]) {
                return this.children.splice(child, 1);
            } else if (child instanceof luxTree) {
                this.children = _.without(this.children, child);
                return this.children;
            }

        };

        /**
     * Removes the node - if this node is the root, no action is performed
     * @returns {*|Array|luxtree.parent.children|Array|luxtree.parent.children|Array|luxtree.parent.children|HTMLElement[]|Array|Function|Array|Function|Array|Function|Array|Function|Array|Function|Array|Function|Array|Function|Array|Function|Array|Function|Array|Function|Array|Array|SourceNode.children|Array|Array|SourceNode.children|Array|Array|SourceNode.children|Array|Array|SourceNode.children|Array|Array|SourceNode.children|Array|Array|SourceNode.children|Array|Array|SourceNode.children} The remaining children of this node's parent
     */
        luxTree.prototype.remove = function() {

            if (!this.isRoot()) {
                this.parent.children = _.without(this.parent.children, this);
                return this.parent.children;
            }

        };

        /**
     * Traverse up the tree until iterator returns false
     * @param iterator An iterator function
     */
        luxTree.prototype.traverseUp = function(iterator) {

            _traverse(this, iterator, _traverseUp);

        };

        /**
     * Traverse down the tree until iterator returns false
     * @param iterator An iterator function
     */
        luxTree.prototype.traverseDown = function(iterator) {

            _traverse(this, iterator, _traverseDown);

        };

        /**
     * Traverses down through the tree applying the finder function to each node.
     * @param {finder} finder - Finder function
     * @returns {null} The matching node or null if none found
     */
        luxTree.prototype.find = function(finder) {

            var match = null;

            // Traverse down using our iterator
            this.traverseDown(function(node) {
                if (finder.call(this, node)) {
                    match = node;
                    return false;
                }
            });
            return match;

        };

        return luxTree;

    });

/**
 * @callback finder
 * @param {obj} node The node to test
 */