/*
 * Copyright (c) 2018 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Setup a default grid configuration, and implement grid config initialization method.
 *
 * @author Daniel Hoang
 */
define([
    'ojs/ojcore',
    'knockout',
    'lodash',
    './Constants',
    'gridapi.columns/iconColumn',
    'gridapi.columns/core/editor',
    'gridapi.columns/core/renderer',
    'gridapi.columns/core/headerRenderer'
], function (
    oj, ko, _,
    Constants,
    iconColumn, editor, renderer, headerRenderer
) {
    'use strict';

    var LuxGridConfig = function () {
        var self = this;

        self._setGridConfig = function(properties) {

            /* This defines the default configuration */
            self.gridConfig = {
                search: {
                    columnFilter: false
                },
                columnFilters:     false,
                editable:          (properties.editable !== undefined) ? ko.unwrap(properties.editable) : false,
                touchInteractions: {
                    enabled: true
                },
                focusOnSelection: false,
                layout:           {
                    width:   (properties.width !== undefined) ? ko.unwrap(properties.width) : Constants.DEFAULT_GRID_WIDTH,
                    height:  (properties.height !== undefined) ? ko.unwrap(properties.height) : Constants.DEFAULT_GRID_HEIGHT,
                    rowsets: [{
                        id:   'rowset-' + self.uniqueNumber,
                        type: 'flex'
                    }],
                    columnGroups: [{
                        type:    'flex',
                        id:      'columnGroup1-' + self.uniqueNumber,
                        columns: []
                    }]
                },
                // http://lux.us.oracle.com:8000/prod/apidocs/module-luxGrid.html#recordKeys
                recordKeys:    (properties.recordKeys !== undefined) ? ko.unwrap(properties.recordKeys) : {id: 'id'},
                selectionMode: (properties.selectionMode !== undefined) ? ko.unwrap(properties.selectionMode) : Constants.DEFAULT_GRID_SELECTION_MODE,
                multiSelect:   (properties.multiSelect !== undefined) ? ko.unwrap(properties.multiSelect) : false,
                columnMenu:    {
                    enabled: false
                }
            };
            self.gridConfig.columns = {};
        };

        self._setGridProperties = function(properties, pageSize) {
            self._setGridConfig(properties);
            if (properties.columnFilters !== undefined && properties.columnFilters === true) {
                if (pageSize !== undefined || self.gridConfig.selectionMode !== Constants.DEFAULT_GRID_SELECTION_MODE) {
                    throw new Error('columnFilters not supported with selector column and/or pagination option enabled.');
                } else {
                    self.gridConfig.columnFilters = ko.unwrap(properties.columnFilters);
                }
            }
        };

        var _isPagingDataSource = function(dataSourceType, pageSize) {
            var isValid = false;
            var isPagingDataSource = dataSourceType.indexOf('oj.CollectionPagingDataSource') > -1 ||
                                        dataSourceType.indexOf('oj.ArrayPagingDataSource') > -1;
            if (isPagingDataSource) {
                if (pageSize === undefined || pageSize <= 0) {
                    throw new Error('Invalid value found for pageSize property. Found: ' + pageSize + '. Expected: pageSize > 0.');
                }
                isValid = true;
            }
            return (isValid);
        };

        var _isTableDataSource = function(dataSourceType, pageSize) {
            var isValid = false;
            var isTableDataSource = dataSourceType.indexOf('oj.CollectionTableDataSource') > -1 ||
                                    dataSourceType.indexOf('oj.ArrayTableDataSource') > -1;
            if (isTableDataSource) {
                if (pageSize !== undefined) {
                    throw new Error('Invalid value found for pageSize property. Found: ' + pageSize + '. Expected: Table data source does not require a pageSize.');
                }
                isValid = true;
            }
            return (isValid);
        };

        self._validateDataSource = function(dataSource, pageSize) {
            if (dataSource) {
                var dataSourceType = dataSource.toString().split(' ')[0];
                return (_isPagingDataSource(dataSourceType, pageSize) || _isTableDataSource(dataSourceType, pageSize));
            }
            throw new Error('Data source is required!');
        };

        // TODO: The grid column big-number extension comes from the grid/bigNumberCCA demo.
        // We need to revisit this and move this code into the grid component instead.
        var _extendBigNumberColumn = function(column) {

            column.headerRenderer = headerRenderer.extend({
            });

            column.renderer = renderer.extend({
                format: function (renderOptions) {
                    return ((renderOptions.column.getFormatter === undefined) ?
                        renderOptions.data :
                        renderOptions.column.getFormatter(renderOptions.record).format(renderOptions.data));
                }
            });

            column.editor = editor.extend({

                validateUserInput: function (/* data */) {
                    var validate = this.bigNumCCA.validate();
                    return validate;
                },

                getEditorInput: function () {
                    return this.getEditor().find('oj-lux-input-bignumber');
                },

                getBigNumCCA: function() {
                    return this.getEditor().find('oj-lux-input-bignumber');
                },

                make: function () {
                    var that = this;
                    that.column.describedBy = ['number'];
                    that['cell-editor-class'] = 'cell-editor-big-number';

                    var html = '<oj-lux-input-bignumber ' +
                                ' id="id"' +
                                ' value={{bigNumberCol}}' +
                                ' max=[[max]]' +
                                ' min=[[min]]' +
                                ' step={{step}}' +
                                ' required="required"' +
                                ' converter={{bigconverter}}' +
                                '>' +
                            '</oj-lux-input-bignumber>';

                    var $editor = $('<div class="cell-editor cell-editor-spinner"></div>');
                    that.fromKeyboard = false;
                    $editor.append(html);
                    return $editor;
                },

                show: function(startString) {
                    if (this.disabled) {
                        startString = null;
                    }
                    this.render(startString);
                    var $editor = this.getEditor();

                    var currentOwns = (this.grid.$grid.attr('aria-owns') || '').split(' ');
                    this.grid.$grid.attr('aria-owns', _.union(currentOwns, [this.getEditorId()]).join(' ').trim());

                    $editor.show();
                },

                render: function() {
                    var that = this;
                    var formatter = (that.column.getFormatter === undefined) ? that.column.formatter : that.column.getFormatter(that.record);
                    var data;

                    if (!this.invalid) {
                        data = this.data;
                    } else {
                        // this.data represents prev invalid data, but it can not be set on lux-input-big-num
                        data = null;
                    }

                    var vm = {
                        id:           'bigNumberEditorId',
                        bigNumberCol: data,
                        min:          that.column.bigNumberMin,
                        max:          that.column.bigNumberMax,
                        step:         that.column.bigNumberStep,
                        required:     false,
                        bigconverter: formatter
                    };

                    this.bigNumCCA = this.getEditor()[0].querySelector('oj-lux-input-bignumber');
                    ko.applyBindings(vm, this.bigNumCCA);

                    setTimeout(function() {
                        var $input = this.getEditor().find('input');
                        $input.get(0).focus();
                        $input.get(0).select();
                        if (this.invalid) {
                            var inputText = this.getEditorInput().find('.lux-input-bignumber-input');
                            // this.data is the invalid data in previous attempt
                            $(inputText).val(this.data);
                            this.bigNumCCA.value = this.data;
                        }
                    }.bind(this), 0);

                },

                getNewData: function() {
                    var valid = this.validateUserInput();
                    var data = '';
                    // NOTE: if data is valid, we take the plain value by normalizing the value,
                    // if data is invalid we save the whatever string user has entered without
                    // any conveter on it
                    if (valid) {
                        data = this.bigNumCCA.normalize();
                    } else {
                        var inputText = this.getEditorInput().find('.lux-input-bignumber-input');
                        data = $(inputText).val();
                        // use raw-data if you can: data = inputText.getAttribute('rawData');
                    }
                    return {
                        value: data,
                        valid: valid
                    };
                },

                destroy: function() {
                    var $editor = this.getEditor();
                    if ($editor.length) {
                        setTimeout(function () {
                            $editor.remove();
                        }, 10);
                    }
                }
            });
        };

        self._setLayoutColumns = function(properties) {
            var dataIndex, column;
            var columnSpec = ko.unwrap(properties.columns);
            var columns = {};
            if (columnSpec) {
                oj.Logger.info('Initialize grid columns per columnSpec ...');
                // Convert array of column specs into an object containing the column definitions.
                columnSpec.forEach(function(spec) {
                    // Just copy the column definition from the spec, excluding the dataIndex property.
                    columns[spec.dataIndex] = _.omit(spec, 'dataIndex');
                });

                // extract the column name and put them into the columnGroups
                for (dataIndex in columns) {
                    column = columns[dataIndex];

                    // Set up the required grid columnGroups
                    self.gridConfig.layout.columnGroups[0].columns.push(dataIndex);

                    // set column.filterable to grid's columnFilters
                    column.filterable = self.gridConfig.columnFilters;
                    column.searchable = self.gridConfig.columnFilters;

                    // Extend column with the required big-number extension
                    if (column.bigNumber) {
                        _extendBigNumberColumn(column);
                    }
                }

                // Set the column definitions
                self.gridConfig.columns = columns;
            }
        };

    };

    // Method to config option for column filters.
    // This config requires gridElement's layout to exist.
    LuxGridConfig.prototype.configColumnFilterOption = function () {
        var self = this;
        var layout = self.gridElement.getProperty('layout');
        if (_.isUndefined(layout)) {
            oj.Logger.error('The LuxGridConfig._configColumnFilterOption() requires grid layout.');
        }
        try {
            if (layout && layout.rowsets) {
                if (self.gridConfig.columnFilters) {
                    // Turn on columns filter type
                    self.gridElement.setProperty('selectRowByDefault', false); // Note that this is a work-around to the LUX-506 fix.
                    self.gridElement.setProperty('search', {
                        showAll:      false,
                        showParents:  false,
                        columnFilter: true
                    });
                    // enable the filter row and insert it before the last row.
                    self.gridElement.setProperty('layout', {
                        rowsets: [{
                            id:   layout.rowsets[layout.rowsets.length - 1].id,
                            type: 'flex'
                        }]
                    });
                } else {
                    // Turn off columns filter type
                    self.gridElement.clearColumnFilter();
                    self.gridElement.setProperty('selectRowByDefault', true); // For restore the default setting (doc says = true) of this option.
                    self.gridElement.setProperty('search', {
                        columnFilter: false
                    });
                    // restore the layout
                    self.gridElement.setProperty('layout', {
                        rowsets: [{
                            id:   layout.rowsets[layout.rowsets.length - 1].id,
                            type: 'flex'
                        }]
                    });
                }
            }
        } catch (e) {
            oj.Logger.warn('The LuxGridConfig._configColumnFilterOption() threw an error: ', e);
        }
    };

    // Private method to config column link option
    LuxGridConfig.prototype.configColumnLinkOption = function() {
        var self = this;
        // Get the observable grid column property
        var _columns = self.columns();
        Object.keys(_columns).forEach(function(dataIndex) {
            var _column = _columns[dataIndex];

            // link icon support
            if (_column.showIcon) {
                var properties = self.properties;
                _column.renderer = iconColumn.renderer.extend({
                    getValueMarkup: function(renderOptions) {
                        var that = this;
                        if (_column.dataDisplay !== 'icon') {
                            if (_column.displayType === 'link') {
                                return that.wrapWithLink(renderOptions.formattedValue, renderOptions);
                            }
                        }
                        return renderOptions.formattedValue;
                    },
                    getLinkAttributes: function (/*options*/) {
                        return '';
                    },
                    getIconMarkup: function (renderOptions) {
                        var iconSpec;
                        var that = this;

                        // Passed in what are needed for custom code to use to return an icon markup
                        // for this cell (at column/row).
                        if (typeof properties.getColumnIconMarkupCallback === 'function') {
                            iconSpec = properties.getColumnIconMarkupCallback(
                                renderOptions.column.dataIndex, renderOptions.record);
                        }
                        var valueToRender = iconSpec ? iconSpec : '';
                        if (_column.displayType === 'link') {
                            return that.wrapWithLink(valueToRender, renderOptions);
                        }
                        return (valueToRender);
                    },
                    wrapWithLink: function(valueToRender, renderOptions) {
                        var that = this;
                        var wrappedValue = '<a  id="' + renderOptions.id + '-link" tabindex="-1" data-trigger="link-clicked" class="cell-descendant cell-link" ';
                        wrappedValue += that.getLinkAttributes(renderOptions) + ' >';
                        wrappedValue += valueToRender + '</a>';
                        return wrappedValue;
                    }
                });
            }
        });
        self.columns(_columns);
    };

    LuxGridConfig.prototype.configInitialize = function(properties) {

        oj.Logger.info('Configuring grid: ' + this.gridElementId);

        // Check for pagination option
        if (properties.pagination) {
            var pagination = properties.pagination;
            if (pagination.pageSize !== undefined) {
                // set pageSize to undefined so we don't render the pagingControl element.
                var pageSize = ko.unwrap(pagination.pageSize);
                this.pageSize((pageSize > 0) ? pageSize : undefined);
            }
        }

        // Set grid properties according to consumer's settings.
        this._setGridProperties(properties, this.pageSize());

        // validate the data source type and requirements
        if (this._validateDataSource(properties.data, this.pageSize())) {
            this.dataSource = ko.unwrap(properties.data);
        }

        this._setLayoutColumns(properties);
    };

    return LuxGridConfig;
});
