/*
 * Copyright (c) 2018 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Composite component event handlers.
 *
 * @author Daniel Hoang
 */
define([
    'ojs/ojcore'
], function (oj) {
    'use strict';

    var LuxGridEvents = function () {
        var self = this;

        // Method to handle cell updated event.
        self.onCellUpdatedHandler = function(event) {
            // just forward this event to comsuming view via our composite events.
            var data = event.detail.data;
            oj.Logger.info('cellUpdated event: ' + data.dataIndex);
            self.emitGridCellUpdatedEvent(data.dataIndex, data.record);
        };

        // Handler for grid-link-clicked event.
        self.onLinkClickedHandler = function(event) {
            var data = event.detail.data;
            if (data && data.dataIndex) {
                oj.Logger.info('grid-link-clicked event: ' + data.dataIndex);
                // just forward this event to comsuming view via our composite events.
                self.emitGridLinkClickedEvent(data.dataIndex, data.record);
            }
        };

        // Mainly to update toolbar item states for row selection changes.
        self.onRowsCheckedHandler = function(event) {
            if (self.toolbarOptions !== undefined) {
                var data = event.detail.data;
                self.gridRowsChecked(data.checkedRecords.length);
            }
        };

        // Method to handle when a single row or multiple rows are selected
        self.onRowsSelectedHandler = function(event) {
            // just forward this event to comsuming view via our composite events.
            var data = event.detail.data;
            oj.Logger.info('rowsSelected event: ' + data.selectedRecords.length + 'record(s) selected');
            self.emitGridRowsSelectedEvent(data.selectedRecords);
        };
    };

    // Component can emit an event to signal consuming view to react to the event.
    // Custom events are defined in metadata under "events" section.

    // Use the global directive to tell JSLint about CustomEvent's assumed existence.
    /* global CustomEvent */

    // This method generates and raise an event when a toolbar action is clicked/selected
    LuxGridEvents.prototype.emitGridToolbarActionEvent = function (actionId, selectedRows) {
        var eventParams = {
            'bubbles':    false,
            'cancelable': false,
            'detail':     {
                'actionId':     actionId,
                'selectedRows': selectedRows
            }
        };
        //Raise the custom event
        this.composite.dispatchEvent(new CustomEvent('toolbarActionEvent', eventParams));
    };

    // This method generates and raise an event when a grid cell value updated
    LuxGridEvents.prototype.emitGridCellUpdatedEvent = function (dataIndex, record) {
        var eventParams = {
            'bubbles':    false,
            'cancelable': false,
            'detail':     {
                'dataIndex': dataIndex,
                'record':    record
            }
        };
        //Raise the custom event
        this.composite.dispatchEvent(new CustomEvent('cellUpdated', eventParams));
    };

    // This method generates and raise an event when a grid cell's link is clicked.
    LuxGridEvents.prototype.emitGridLinkClickedEvent = function (dataIndex, record) {
        var eventParams = {
            'bubbles':    false,
            'cancelable': false,
            'detail':     {
                'dataIndex': dataIndex,
                'record':    record
            }
        };
        //Raise the custom event
        this.composite.dispatchEvent(new CustomEvent('linkClicked', eventParams));
    };

    // This method generates and raise an event when a grid's single row or multiple rows are selected.
    LuxGridEvents.prototype.emitGridRowsSelectedEvent = function (selectedRecords) {
        var eventParams = {
            'bubbles':    false,
            'cancelable': false,
            'detail':     {
                'selectedRecords': selectedRecords
            }
        };
        //Raise the custom event
        this.composite.dispatchEvent(new CustomEvent('rowsSelected', eventParams));
    };

    return LuxGridEvents;
});