/*
 * Copyright (c) 2018 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * Implement grid's toolbar feature.
 *
 * @author Daniel Hoang
 */
define([
    'ojs/ojcore',
    'knockout',
    'lodash',
    'luxToolbarGrid'
], function (
    oj, ko, _
) {
    'use strict';

    var Logger = oj.Logger;

    var LuxGridToolbar = function () {

        var self = this;

        self.toolbarOptions = undefined;

        // Mainly to update toolbar item states for row selection changes.
        self.gridRowsChecked = ko.observable(0);

        self._toolbarProxyCode = function(data, toolbarId, gridId, toolbarSpec) {
            if (!toolbarSpec.customToolbar) {
                throw new Error('customToolbar not defined!');
            } else {
                if (typeof toolbarSpec.customToolbar.elementDisabledState !== 'function') {
                    throw new Error('customToolbar.elementDisabledState not defined!');
                }
                if (typeof toolbarSpec.customToolbar.shouldIncludeElement !== 'function') {
                    throw new Error('customToolbar.shouldIncludeElement not defined!');
                }
            }

            if (!toolbarSpec.buttons) {
                throw new Error('buttons not defined!');
            }

            if (!toolbarSpec.menuItems) {
                Logger.warn('menuItems not defined.');
            }

            // if we got a valid toolbar with its predefined toolbar items
            if (data && data.items) {

                var items = [];
                var params = {
                    toolbarId: toolbarId,
                    gridId:    gridId
                };

                var _actionHandler = function(params) {
                    // just forward this event to comsuming view via our composite events.
                    self.emitGridToolbarActionEvent(params.elementId, self.gridElement.getSelection());
                };

                if (toolbarSpec.menuItems) {
                    var _menuItemSelect = function(event, ui) {
                        var _params = _.clone(params);
                        _params.elementId = ui.item[0].id;
                        _actionHandler(_params);
                    };
                    var _shouldMenuItemDisabled = function(itemId) {
                        return ko.computed(function() {
                            var rowChecked = self.gridRowsChecked();
                            return (toolbarSpec.customToolbar.elementDisabledState(itemId, rowChecked));
                        });
                    };

                    // Create a menu component with a select handler.
                    var menuActionsId = self._uniqueId('lux-menu-actions-id-');
                    var _actionMenu = {
                        actionMenu: {
                            id:        menuActionsId,
                            label:     'Actions',
                            select:    _menuItemSelect,
                            component: 'menu'
                        }
                    };
                    // Create a button and bound it to the menu component to create the menu button.
                    var menuButtonActionsId = self._uniqueId('lux-menu-button-actions-id-');
                    var _actionMenuButton = {
                        actionMenuButton: {
                            id:        menuButtonActionsId,
                            label:     'Actions',
                            component: 'button',
                            type:      'menu',
                            menu:      '#' + menuActionsId
                        }
                    };
                    // add action menu at the beginning of the toolbar.
                    items.push(_actionMenuButton);
                    items.push(_actionMenu);

                    // Process menuItems to add state handler
                    var _actionMenuItems = [];
                    toolbarSpec.menuItems.forEach(function(itemSpec) {
                        itemSpec.id = _.kebabCase(itemSpec.label);
                        // Check client if this menu item should be included (i.e. exclude per role-/process-specific)
                        if (toolbarSpec.customToolbar.shouldIncludeElement(itemSpec.id)) {
                            itemSpec.disabled = _shouldMenuItemDisabled(itemSpec.id);
                            itemSpec.type = 'item';
                            itemSpec.href = '#';
                            _actionMenuItems.push(itemSpec);
                        }
                    });
                    // and link the menu item listing to the menu component
                    items[1].actionMenu.menuItems = _actionMenuItems;

                    // Add a toolbar separator
                    items.push({'separator': {type: 'separator'}});
                }

                var _buttonClick = function(data, event) {
                    var _params = _.clone(params);
                    _params.elementId = event.currentTarget.id;
                    _actionHandler(_params);
                };

                var _shouldButtonDisabled = function(buttonSpec) {
                    return ko.computed(function() {
                        var rowChecked = self.gridRowsChecked();
                        var buttonId = buttonSpec.id;
                        var curState = buttonSpec.disabled ? buttonSpec.disabled() : undefined;
                        var newState = toolbarSpec.customToolbar.elementDisabledState(buttonId, rowChecked);
                        // Toolbar refresh is required when its element state changed.
                        if (curState !== newState) {
                            self._toolbarRefresh();
                        }
                        return (newState);
                    });
                };

                // add custom buttons ...
                toolbarSpec.buttons.forEach(function(buttonSpec) {
                    buttonSpec.id = _.kebabCase(buttonSpec.label);
                    // Check client if this button should be included (i.e. exclude per role-/process-specific)
                    if (toolbarSpec.customToolbar.shouldIncludeElement(buttonSpec.id)) {
                        buttonSpec.event = {click: _buttonClick};
                        buttonSpec.disabled = _shouldButtonDisabled(buttonSpec);
                        if (!buttonSpec.component) {
                            buttonSpec.component = 'button';
                            buttonSpec.type = 'push';
                        }
                        var buttonItem = {};
                        buttonItem[buttonSpec.id] = buttonSpec;
                        items.push(buttonItem);
                    }
                });

                // Update the toolbar items with the changes
                data.items = items;
            }
        };
    };

    LuxGridToolbar.prototype._refreshGrid = function() {
        this._fetchData();
        // Refresh will also clear the rows checked, if any.
        this.gridRowsChecked(0);
    };

    LuxGridToolbar.prototype.toolbarInitialize = function(properties) {
        var self = this;
        if (properties.toolbar && properties.toolbar.spec) {
            Logger.info('Initialize grid toolbar: ' + self.gridToolbarId);
            var config = properties.toolbar;
            if (config.spec && Object.keys(config.spec).length > 0) {
                self.toolbarOptions = {
                    gridId:           self.gridElementId,
                    showTopBorder:    config.border === 'top' || config.border === 'both',
                    showBottomBorder: config.border === 'bottom' || config.border === 'both',
                    customize:        function(event, data) {
                        try {
                            var toolbarSpec = config.spec;
                            toolbarSpec.customToolbar = {
                                elementDisabledState: config.isElementDisabledStateCallback,
                                shouldIncludeElement: config.shouldIncludeElementCallback
                            };
                            self._toolbarProxyCode(data, self.toolbarId, self.gridElementId, toolbarSpec);
                        } catch (e) {
                            Logger.error('Error with toolbarSpec Function(): ' + e.message);
                        }
                    }
                };

                self._toolbarRefresh = function() {
                    if (!self.gridToolbarElement) {
                        if (self.composite && self.composite.className.indexOf('oj-complete') > -1) {
                            self.gridToolbarComponent = document.getElementById(self.gridToolbarId);
                        }
                    }
                    if (self.gridToolbarElement) {
                        try {
                            self.gridToolbarElement.refresh();
                        } catch (e) {/* toolbar may not be realized yet */}
                    }
                };
            } else {
                Logger.error('Toolbar spec specified but no specs found!');
            }
        }
    };

    return LuxGridToolbar;
});
