/*
 * Copyright (c) 2018 Oracle and/or its affiliates. All rights reserved.
 *
 * @fileOverview
 * A view model implements the LUX grid component with common basic functionalities.
 *
 * @author Daniel Hoang
 */
/* eslint max-params: ["error", 20] */
define([
    'ojs/ojcore', 'knockout', 'jquery',
    './js/Config',
    './js/Data',
    './js/Toolbar',
    './js/Events',
    // LUX dependencies
    'luxBindingBridge',
    'oj-lux/oj-lux-grid/loader',
    'oj-lux/oj-lux-toolbar-grid/loader',
    // JET dependencies
    'ojs/ojknockout',
    'ojs/ojpagingcontrol'
], function (
    oj, ko, $,
    Config,
    Data,
    Toolbar,
    Events
) {
    'use strict';

    //The constructor is passed a context object as part of this. This viewModel
    // is therefore dedicated to that instance of the Composite Component. in this
    // normal variant of the lifecycle the initialize lifecycle method is never called.
    function ViewModel(context) {
        // Chain this component model constructor with other extension constructors
        Config.apply(this, arguments);
        Data.apply(this, arguments);
        Toolbar.apply(this, arguments);
        Events.apply(this, arguments);

        var self = this;

        // Save this composite component's element.
        self.composite = context.element;

        // Set up busy context
        var busyContext = oj.Context.getContext(context.element).getBusyContext();
        var options = {'descriptions': 'oj-lux-grid-basic startup processing'};
        self.setupComplete = busyContext.addBusyState(options);

        // Create unique element IDs
        self.uniqueNumber = parseInt(context.unique.replace(/[^0-9\.]/g, ''), 10);
        self._uniqueId = function(idPrefix) {
            return (idPrefix + self.uniqueNumber);
        };
        self.gridToolbarId = self._uniqueId('oj-lux-grid-basic-toolbar-id-');
        self.gridElementId = self._uniqueId('oj-lux-grid-basic-id-');
        self.gridPaginationId = self._uniqueId('oj-lux-grid-basic-pagination-id-');

        self.pageSize = ko.observable();

        //Save the resolved properties for later access
        self.properties = context.properties;

        // Setting up proxies of the instance-level translation strings to the underlying components.
        self.gridProxyStrings = ko.observable(self.properties.translations.ojLuxGrid || {});
        self.pagingControlProxyStrings = ko.observable(self.properties.translations.ojPagingControl || {});

        // signal that all the setup work is done
        self.setupComplete();
    }

    // Extends the component model's prototype with the extension prototypes
    $.extend(ViewModel.prototype, Config.prototype);
    $.extend(ViewModel.prototype, Data.prototype);
    $.extend(ViewModel.prototype, Toolbar.prototype);
    $.extend(ViewModel.prototype, Events.prototype);

    // If a ViewModel is provided for a composite component, the following optional callback methods
    // can be defined on its ViewModel and will be called at each stage of the composite component's
    // lifecycle.

    // This optional method may be implemented on the ViewModel and will be invoked after the View is first inserted into
    // the DOM and then each time the composite is reconnected to the DOM after being disconnected. Note that if the
    // composite needs to add/remove event listeners, we recommend using this and the disconnected methods.
    ViewModel.prototype.connected = function(/*context*/) {
        var self = this;
        oj.Logger.info(self.gridElementId + '@connected');

        self.configInitialize(self.properties);
        self.toolbarInitialize(self.properties);

        // This turns all grid config properties into observables.
        // This is required for the grid view to work.
        Object.keys(self.gridConfig).forEach(function(key) {
            self[key] = ko.observable(self.gridConfig[key]);
        });

        // Load the column link configuration option
        self.configColumnLinkOption();
    };

    // optional bindingsApplied() method may be implemented on the ViewModel and will be invoked after
    // the bindings are applied on this View.
    ViewModel.prototype.bindingsApplied = function(/*context*/) {
        oj.Logger.info(this.gridElementId + '@bindingsApplied');
        var self = this;

        var busyContext = oj.Context.getContext(self.ce).getBusyContext();
        busyContext.whenReady().then(function() {
            // Here we get access to the grid element that should be available by now.
            self.gridElement = document.getElementById(self.gridElementId);
            if (self.gridElement) {

                // This is the grid view's entry-point callback to load data into view.
                self.gridElement.whenReady().then(function() {
                    // Get a reference to the paging-control element.
                    self.gridPaginationControlElement = document.getElementById(self.gridPaginationId);

                    // Load the column filter configuration option here, since it requires a grid layout, which should exist at this point.
                    self.configColumnFilterOption();

                    if (self.dataSource) {
                        // then fetch the data
                        self._fetchData(true);
                    }
                });
            } else {
                oj.Logger.warn('Could not find oj-lux-grid-basic id: ' + self.gridElementId);
            }

        });
    };

    // This optional method may be implemented on the ViewModel and will be invoked when this composite
    // component is disconnected from the DOM.
    ViewModel.prototype.disconnected = function(/*element*/) {
        oj.Logger.info(this.gridElementId + '@disconnected');
        // Time to clean up resources.
        if (this.gridPaginationComputed) {
            this.gridPaginationComputed.dispose();
        }
    };

    // return a constructor that will be invoked to create the ViewModel instance.
    return ViewModel;
});
