define([
    'ojs/ojcore',
    'lodash',
    './utils/NumberUtils'
], function(oj, _, NumberUtils) {
    'use strict';

    var BigNumberMethods = {

        /**
         *    Manages updates to the value property.
         *    @param {string} key - property object key
         *    @param  {object} event - ui event object
         *    @returns {undefined}
         */
        updateValue: function (key, event) {
            var propObj = {}, eventDetailValue,
                timestamp = Date.now();

            // if global valueUpdated is true, the value has been already been updated
            // so reset valueUpdated and return.
            if (this.valueUpdated) {
                this.valueUpdated = false;
                oj.Logger.info('******** value already updated:', event.detail.value);
                return;
            }

            if (_.isNumber(event.detail.value)) {
                // the converter only works with strings so if
                // the change is with numeric input change
                // the value to a string.
                event.detail.value = event.detail.value.toString();
            }

            // this should not happen but it may, if it does we just leave.
            if (_.isNull(event.detail.value) || _.isUndefined(event.detail.value)) {
                return;
            }

            // save the value as a string
            eventDetailValue = event.detail.value.toString();
            // use the converter to normalize the value (strips off currency and group separators)
            event.detail.value = BigNumberMethods.getNormalizedValue.call(this, event);

            oj.Logger.info('**************', timestamp, 'update value:',
                this.properties.value, 'previous value:',
                event.detail.previousValue, event.detail.updatedFrom);

            propObj[key] = event.detail.value;

            this._setButtonStates(this.properties.value);

            // update the value to the input text field.
            this._container.ojInputText('option', 'value', event.detail.value);

            // update the DOM elements aria attributes to reflect the change in the input text field's value.
            document.getElementById(this.inputId).setAttribute('aria-valuenow', propObj.value);
            document.getElementById(this.inputId).setAttribute('aria-valuetext', this.properties.value);

            // At this point the input text field has been updated with
            // the new value.
            if (this.properties.value !== eventDetailValue || event.detail.updatedFrom === 'external') {
                // validate the input
                this._container.ojInputText('validate');
                oj.Logger.info('************* updated the property');
            }

            // update the property.value with the formatted value. This will invoke this function again
            // so set the global valueUpdate flag to true to indicate the property.value has been updated.
            this.valueUpdated = true;
            this.properties.value = this._formatInput(event.detail.value);
            oj.Logger.info('**************', timestamp, 'update value:', this.properties.value);
        },

        /**
         * @description - Based on if up / down button or arrows keys are pressed, normalizes
         *                the event.detail.value either with a simple normalization or normalization
         *                based on the locale characteristics.
         * @param {object} event - contains the value and previousValue of the value property
         * @returns {string} normalized - a updated value
         */
        getNormalizedValue: function(event) {
            var normalized = null;

            // if stepping up or down with up / down buttons or the up / arrows
            if (this.stepping) {
                // simply normalized the value
                normalized = this._normalizeInput(event.detail.value);
            } else {
                // not stepping, so normalize based on the locale characteristics.
                normalized = this._normalizeInput(
                    NumberUtils.normalizeForLocale(event.detail.value,
                        this.oi18nResources.getOi18nLocaleElements()));
            }

            return normalized;
        },

        /**
         *    Updates oj-lux-input-bignumber properties that are not handled by specific listeners.
         *    @param {string} key - property object key
         *    @param {object} CONSTANTS - preset values of oj-lux-input-bignumber
         *    @param  {object} event - ui event object
         *    @returns {undefined}
         */
        updateProperties: function (key, CONSTANTS, event) {
            var obj = {},
                newValue;

            oj.Logger.info('******* ', this.inputId, 'property', key, 'changed', event.detail.value);

            obj[key] = event.detail.value;

            if (key === 'step') {
                newValue = this._normalizeInput(this.properties[key]).toString();
                this.properties[key] = this.properties[key] !== CONSTANTS.DEFAULT_STEP ? this._parseStep(newValue) : newValue;
                return;
            }
            // update the option
            this._container.ojInputText('option', obj);
        },

        /**
         * Applies ojInputText to the HTML input tag hardcoded in the view.
         * @returns {undefined}
         */
        setUpInputTextField: function() {
            //***************************************
            // Create the LUX input number component
            //***************************************
            // in case the input is formatted, need to normalize the value.
            // the converter will take care of formatting the value in the
            // input text field.
            this.props.value = this._normalizeInput(this.props.value);
            // create ojInputText object with props options, the converter included
            // in the properties will format that value per locale
            this._container.ojInputText(this.props);

            // format the properties value with correct 2-way binding of the output.
            this.properties.value = this._formatFieldInput(this.props.value);
            document.getElementById(this.inputId).setAttribute('aria-valuenow', this.props.value);
            document.getElementById(this.inputId).setAttribute('aria-valuetext', this.properties.value);

            // Apply oj option change listener on the ojInputText so when the
            // value of the input text changes the corresponding application
            // observable is updated.
            this._container.on({'ojoptionchange': BigNumberMethods._inputTextChange.bind(this)});
            // add keypress event listener to the container
            this._container[0].addEventListener('keydown', this._keypress.bind(this), false);
            //******************************************
            // END Create the LUX input number component
            //******************************************
        },

        /**
         * Function that fires on input changes to the ojInputText
         * @param {object} event - CCA event containing the new value.
         * @param {object} ui - Ui event
         * @returns {undefined}
         */
        _inputTextChange: function (event, ui) {
            var allowedOptions = {'value': true, 'rawValue': true, 'messagesShown': true, 'converter': true};

            if (!allowedOptions[ui.option]) {
                return;
            }

            if (ui.option === 'rawValue' && event.originalEvent) {
                // **********************************************
                // Save the value of each user key press. Will be
                // used later to determine which step button to
                // disable if the value entered is outside max
                // min range.
                // **********************************************
                // Also used to determine if the input value has
                // changed but the focus on the field has not
                // blurred. In that case the rawValue is used to
                // update properties.value. See
                // luxInputBigNumber.prototype._stepNumber function
                // luxInputBigNumber.prototype._adjustValue function
                // luxInputBigNumber.prototype._maxMinRangeAdjustments function
                this.rawValue = ui.value;
            }
            if (ui.option === 'messagesShown' && this.rawValue) {
                // disable the appropriate step button
                this._setButtonStates(this.rawValue);
            }
            if (ui.option === 'converter' && this.rawValue) {
                // SUPER EDGE CASE. When a value of the fractional digits change and
                // there is no BURR event to save the value and then the maximumFactionalDigit or
                // minimumFractionDigit change (ui.option === 'converter'), the fractional digits value
                // could potentially be lost. If the is a raw value, a 'value' option is created
                // with a originalEvent option to make sure the entered value is preserved.
                ui.value = this.rawValue;
                ui.option = 'value';
                event.originalEvent = true;
            }
            if (ui.option === 'value' && event.originalEvent) {
                BigNumberMethods._updateValueProperty.call(this, event, ui);
            }
        },

        /**
         * Updates the value property
         * @param {object} event - CCA event containing the new value.
         * @param {object} ui - Ui event
         * @returns {undefined}
         * @private
         */
        _updateValueProperty: function(event, ui) {
            // if no value assume blank input, no input
            if (_.isNull(ui.value)) {
                ui.value = '';
                // edge case when only group separator and or decimal
                // separators are entered.
            } else if (ui.value.toString().length === 0) {
                ui.value = '0';
            } else {
                // converter will only accept strings so change to string
                ui.value = ui.value.toString();
                // if no length it means only a decimal separator has been entered
                // so set the value to zero.
                if (this._normalizeInput(ui.value).length === 0) {
                    ui.value = '0';
                }
            }

            this._setButtonStates(this.properties.value);
            if (this.properties.optionInputChange && _.isFunction(this.properties.optionInputChange)) {
                ui.value = this.properties.value;
                this.properties.optionInputChange.call(null, event, ui);
            }

            // null the rawValue so it will not be used again.
            this.rawValue = null;

            // save the value in the CCA's properties. This will cause the "value" property change listener to fire
            // and will format the value in the ojinputtext.
            this.properties.value = ui.value;
            oj.Logger.info('********* updated properties value on input blur:', this.properties.value);
        },

        /**
         * Sets up listener for spinner mouse events.
         * @param {object} CONSTANTS - input bignumber constants
         * @returns {undefined}
         */
        setUpSpinnerButtonMouseEventListeners: function(CONSTANTS) {

            this.decButton.addEventListener('mousedown', BigNumberMethods._mouseActionRepeat.bind(this, this._repeat, CONSTANTS.STEP_DOWN, 'decrement mousedown'), false);
            this.decButton.addEventListener('contextmenu', BigNumberMethods._mouseActionRepeat.bind(this, this._repeat, CONSTANTS.STEP_DOWN, 'decrement contextmenu'), false);
            this.decButton.addEventListener('mouseleave', BigNumberMethods._mouseActionStop.bind(this, this._stop, 'increment mouseleave'), false);
            this.decButton.addEventListener('mouseup', BigNumberMethods._mouseActionStop.bind(this, this._stop, 'increment mouseup'), false);

            this.incButton.addEventListener('mousedown', BigNumberMethods._mouseActionRepeat.bind(this, this._repeat, CONSTANTS.STEP_UP, 'decrement mousedown'), false);
            this.incButton.addEventListener('contextmenu', BigNumberMethods._mouseActionRepeat.bind(this, this._repeat, CONSTANTS.STEP_UP, 'decrement contextmenu'), false);
            this.incButton.addEventListener('mouseleave', BigNumberMethods._mouseActionStop.bind(this, this._stop, 'increment mouseleave'), false);
            this.incButton.addEventListener('mouseup', BigNumberMethods._mouseActionStop.bind(this, this._stop, 'increment mouseup'), false);
        },

        _mouseActionStop: function(stopFunction, message, event) {
            event.preventDefault();
            stopFunction.call(this, event);
            oj.Logger.info('*********', message);
            return false;
        },

        _mouseActionRepeat: function(repeatFunction, direction, message, event) {
            event.preventDefault();
            repeatFunction.call(this, null, direction, event);
            oj.Logger.info('*********', message);
            return false;
        },

        /**
         * Sets up property change listeners. Iterates over the CCA properties and applies change
         * listeners each property.
         * @param {object} context - the CCA context.
         * @returns {undefined}
         */
        setUpPropertyChangeListeners: function(context) {
            _.each(this.props, function(value, key) {
                var eventType = key + 'Changed';

                oj.Logger.info(context.element.id, '********** event', eventType);
                context.element.addEventListener(eventType, this._eventListenerFactory(context, eventType, key));
            }, this);
        },

        /**
         * Wraps the button HTML elements in SPAN and applies JET css classes to replicate the look and feel of
         * oj-input-number spinner buttons.
         * @returns {undefined}
         */
        setUpSpinnerButtons: function() {
            // Apply unique id to button wrapping div.
            this._container.parent().append($('#' + this.inputId + '-buttonset'));
            // Use some jquery DOM manipulation to wrap the input text with the decrement, increment buttons
            // that follows how the current ojInputNumber is laid out.
            this._container.add(this._container.next('#' + this.inputId + '-buttonset')).wrapAll('<span class="oj-inputnumber-wrapper"></span>');
            // adjust JET styles so the step buttons align with the inputtext field.
            this._container.parent().parent().removeClass('oj-inputtext').addClass('oj-inputnumber');
            if (this.properties.value) {
                this._container.ojInputText('validate');
            }

            // make buttons non-tab-able
            document.getElementById(this.inputId + '-dec').tabIndex = -1;
            document.getElementById(this.inputId + '-inc').tabIndex = -1;
            if (this.properties.max) {
                document.getElementById(this.inputId).setAttribute('aria-valuemax', this.properties.max);
            }
            if (this.properties.min) {
                document.getElementById(this.inputId).setAttribute('aria-valuemin', this.properties.min);
            }
        },

        /**
         * Wraps the HTML LABEL element in a the DIV tags to replicate the correct layout for ojInputtext and Label tag
         * @param {object} context - CCA context object
         * @param {object} CONSTANTS - Constant contains HTML template that wraps the LABEL tag.
         * @returns {undefined}
         */
        setUpInputTextLabel: function(context, CONSTANTS) {
            // If there is a id specified for the CCA element and there is a label for that CCA,
            // change the 'for' attribute to the ojInputtext id  and wrap the label with OJET DIV classes
            // to get the correct layout for ojInputtext and Label tag.
            if (!_.isUndefined(context.element.id)) {
                this.thisLabel = document.querySelector('label[for="' + context.element.id + '"]');
                // if there is a label.
                if (this.thisLabel) {
                    this.thisLabel.setAttribute('for', this.inputId);
                    $(this.thisLabel).wrap(CONSTANTS.LABELWRAPPER);
                }
                oj.Logger.info('**** label', this.thisLabel);
            }
        }
    };

    return {
        updateValue:                           BigNumberMethods.updateValue,
        updateProperties:                      BigNumberMethods.updateProperties,
        setUpSpinnerButtonMouseEventListeners: BigNumberMethods.setUpSpinnerButtonMouseEventListeners,
        setUpInputTextField:                   BigNumberMethods.setUpInputTextField,
        setUpPropertyChangeListeners:          BigNumberMethods.setUpPropertyChangeListeners,
        setUpSpinnerButtons:                   BigNumberMethods.setUpSpinnerButtons,
        setUpInputTextLabel:                   BigNumberMethods.setUpInputTextLabel
    };

});