define([
    'ojs/ojcore',
    'lodash'
], function(oj, _) {
    'use strict';

    var NumberUtils = {

        /**
         * Takes a string representing an exponential number and returns
         * the full string number.
         * @param {string} input - string representing the number.
         * @returns {string} string number.
         */
        getNumberFromExponent: function (input) {
            var data = input.split(/[e/E]/),
                sign, stringValue, digits,
                digitArray = [],
                result = '';

            // if there is no exponent return the original value.
            if (data.length === 1) {
                return data[0];
            }

            // get the sign from the input
            sign = data[0] < 0 ? '-' : '';
            // strip off the decimal point
            stringValue = data[0].replace('.', '');
            // get the digit count
            digits = Number(data[1]) + 1;

            // if a negative exponent 'e-'
            if (digits < 0) {
            // start output
                result = sign + '0.';
                // add zeros (0) to the result
                // until there are no more digits
                while (digits++) {
                    result += '0';
                }
                // add zeros to the left of the string value and return.
                return result + stringValue.replace('-', '');
            }

            // get the diff between the exponent digits.
            digits -= stringValue.length;

            // if the digits are gt zero (0).
            if (digits > 0) {
            // add zeros (0) to the results
            // until there are no more digits
                while (digits--) {
                    result += '0';
                }
            }

            // reset digits
            digits = Number(data[1]) + 1;
            // append zeros (if any) to the string value.
            stringValue = stringValue + result;
            // split the string value into an array
            digitArray = stringValue.split('');
            // add the decimal point in the digits place
            digitArray.splice(digits, 0, '.');
            // joint the array into a string and return
            return digitArray.join('');
        },

        /**
         * Determines the actual fractional digits.
         * @param {string} value - the string number to use
         * @returns {string} the number of digits after the decimal point.
         */
        getPrecision: function (value) {
            if (!value) {
                return;
            }

            value = value.replace('-', '');
            return value.indexOf('.') > -1 ? (value.length - value.indexOf('.') - 1) : 0;
        },

        /**
         * Determine if the number string had been entered using the locale decimal separator and if
         * it has, sends back the unaltered value. If the decimal separator is not found all non numeric
         * characters are stripped out and the altered value is returned.
         * @param {string} value - string number
         * @param {object} localeElements - object containing the locale number characteristics.
         * @returns {*}
         */
        normalizeLocale: function(value, localeElements) {
            var reg = new RegExp(/^\d+$/),
                numbers,
                num = value.replace(/ /g, ''),
                localeNumberSystem = oi18n.Utils.getLocaleElementsMainNode(localeElements),
                numberSystem = localeNumberSystem.numbers['symbols-numberSystem-latn'],
                decimalSeparator = numberSystem.decimal,
                parts = num.split(decimalSeparator);

            // if the decimal separator is found in the string, that is it!
            // nothing more to do, just return the value.
            if (parts.length > 1) {
                return value;
            }

            // split the number string into an array of characters
            num = num.split('');
            // filter out the non numbers
            numbers = num.filter(function(element) {
                if (reg.test(element)) {
                    return element;
                }
            });

            // rejoin into a string overwriting the original value
            value = numbers.join('');

            oj.Logger.info('*********** normalized value', value);

            // return normalized value
            return value;
        },

        normalizeForLocale: function(value, localeElements) {
            var localeNumberSystem = oi18n.Utils.getLocaleElementsMainNode(localeElements),
                numberSystem = localeNumberSystem.numbers['symbols-numberSystem-latn'],
                decimalSeparator = numberSystem.decimal;

            //replace . separator with localized decimal
            return value.split('.').join(decimalSeparator);
        }
    };

    return {
        getNumberFromExponent: NumberUtils.getNumberFromExponent,
        getPrecision:          NumberUtils.getPrecision,
        normalizeLocale:       NumberUtils.normalizeLocale,
        normalizeForLocale:    NumberUtils.normalizeForLocale
    };
});