define([
    'ojs/ojcore',
    'knockout',
    'text!./component.json',
    'lodash',
    'ojL10n!./resources/nls/oj-lux-shuttle-translations',
    'AbstractCCAWidget',
    'luxhierarchy'
], function(oj, ko, meta, _, shuttleLabels, AbstractCCAWidget, LuxHierarchy) {
    'use strict';

    function createTranslationHierarchy(jsonTrans, node) {

        _.each(jsonTrans, function(item, key) {
            var newNode, nodeObj;

            // create a parent node
            nodeObj = {
                'action':      key,
                'actionLabel': key,
                'description': item.description
            };

            newNode = node.addChild().parseJSON(JSON.stringify(nodeObj));

            if (item.properties) {
                createTranslationHierarchy(item.properties, newNode);
            }

        });
    }

    var Model = function(context) {
        var self = this;

        self.widgetName = 'luxshuttle';
        AbstractCCAWidget.call(self, context, meta);

        // Create element ID for the widget container element
        var widgetContainerId = self._uniqueId('oj-lux-shuttle-id-');

        // Register the widget container ID
        self.registerWidgetContainerId(widgetContainerId);

        // Initialize grid translation strings
        self.translations = shuttleLabels.luxShuttle;
        // save properties in object level variable.
        self.properties = context.properties;

        if (self.properties.translations) {
            self.properties.translations = _.merge({}, self.translations, self.properties.translations);
        }
    };

    Model.prototype = Object.create(AbstractCCAWidget.prototype);
    Model.prototype.constructor = Model;

    // Overriding abstract class' bindingsApplied() life-cycle in order to handle the grid's locale config option.
    Model.prototype.bindingsApplied = function(/*context*/) {
        var self = this,
            // Pick up per instance's translation overrides. We do this by checking for explicit overrides supplied
            // on the tag instance and apply those. They are the most specific.
            instanceTranslations = (self.properties.translations) ? self.properties.translations : self.translations,
            // traverses down the hierarchy to get the last item based on the path
            getTranslationItem = function getTranslationItem(path, translations) {
                var node = null;

                if (path.length > 0) {
                    // if the component property is not present
                    // in the widget translation input, exit.
                    if (_.isUndefined(translations[path[0]])) {
                        return null;
                    }

                    // if the path length is 1, we have arrived
                    // at the translated property with the value
                    // we want.
                    if (path.length === 1) {
                        return translations[path[0]];
                    } else {
                        // save the hierarchy node.
                        node = translations[path.shift()]; // shift the first element off the array as key to node
                    }
                }

                if (path.length > 0) {
                    // still move nodes to travers down.
                    node = getTranslationItem(path, node);
                }

                return node;
            },
            setOptions = function setOptions(hierarchy) {

                hierarchy.forEach(function(node) {
                    var parents, path, translation;

                    // if parent node
                    if (node.children && node.children.length > 0) {
                        // recursively get child nodes
                        setOptions(node.children);
                    } else { // node is a leaf, no children.
                        // get an array of parent nodes
                        parents = node.listParents();
                        // get an array of parent keys
                        path = parents.map(function(nodeItem) {
                            return nodeItem.action;
                        });
                        // shift off the 'home' root from the path.
                        path.shift();
                        // add the node's key to the path
                        path.push(node.action);
                        // get the node's translated text
                        translation = getTranslationItem(path, instanceTranslations);
                        // if the translation
                        if (translation) {
                            // if the translated value is a ko observable, get the actual value.
                            if (ko.isObservable(translation)) {
                                translation = translation();
                            }
                            // add translation option to shuttle widget
                            self.setWidgetOption(node.action, translation);
                        }
                    }
                }, self);
            },
            translationHierarchy;

        oj.Logger.info('Updating shuttle locale=' + oj.Config.getLocale());

        // if there is a translation JSON structure
        if (!_.isEmpty(instanceTranslations)) {
            // create the root luxHierarchy node
            translationHierarchy = new LuxHierarchy().parseJSON('{"action": "home", "actionLabel": "Home"}');
            // create the translations hierarchy
            createTranslationHierarchy(JSON.parse(self.metadata).properties.translations.properties, translationHierarchy);

            oj.Logger.info('************** instance translations:', instanceTranslations);
            oj.Logger.info('************** translationHierarchy:', translationHierarchy);

            // create the options from the translations hierarchy
            setOptions(translationHierarchy.children);
        }
    };

    return Model;
});