/**
 * Copyright (c) 2014-2017, Oracle and/or its affiliates.
 * All rights reserved.
 */

/*
 DESCRIPTION
 oi18n.Utils provides helper functions for converter objects.
 
 PRIVATE CLASSES
 <list of private classes defined - with one-line descriptions>
 
 NOTES
 <other useful comments, qualifications, etc.>
 
 MODIFIED    (MM/DD/YY)
 gmurr       05/13/14 - Creation
 */


oi18n.Utils = {};
//supported numbering systems
oi18n.Utils.numeringSystems = {
  'latn': "\u0030\u0031\u0032\u0033\u0034\u0035\u0036\u0037\u0038\u0039",
  'arab': "\u0660\u0661\u0662\u0663\u0664\u0665\u0666\u0667\u0668\u0669",
  'thai': "\u0e50\u0e51\u0e52\u0e53\u0e54\u0e55\u0e56\u0e57\u0e58\u0e59"
};

oi18n.Utils.regexTrim = /^\s+|\s+$|\u200f|\u200e/g;
oi18n.Utils.regexTrimNumber = /\s+|\u200f|\u200e/g;
oi18n.Utils.regexTrimRightZeros = /0+$/g;
oi18n.Utils.zeros = ["0", "00", "000"];
//ISO 8601 string accepted values:
//date only: YYYY or YYYY-MM or YYYY-MM-dd or YYYYMM or YYYYMMdd
//time only without timezone: Thh:mm or Thh:mm:ss or Thh:mm:ss.SSS or Thhmm or Thhmmss or Thhmmss.SSS
//time only with timezone: any of the time values above followed by any of the following: Z or +/-hh:mm or extended timezone like @America/Los_Angeles
//date time: any of the date values followed by any of the time values
oi18n.Utils._ISO_DATE_REGEXP = /^\d{4}(?:-?\d{2}(?:-?\d{2})?)?(?:T\d{2}:?\d{2}(?::?\d{2}(?:\.\d{1,3})?)?(Z|[+-]\d{2}:\d{2})?)?$|^T\d{2}:?\d{2}(?::?\d{2}(?:\.\d{1,3})?)?(Z|[+-]\d{2}:\d{2})?$/;

/**
 * Returns the timezone offset between UTC and the local time in Etc/GMT[+-]hh:mm syntax.
 * The offset is positive if the local timezone is behind UTC and negative if
 * it is ahead. The offset range is between Etc/GMT-14 and Etc/GMT+12 (UTC-12 and UTC+14)
 * Examples:
 * 1- The local time is UTC-7 (Pacific Daylight Time):
 * oi18n.Utils. getLocalTimeZoneOffset() will return the string "Etc/GMT+07:00" 
 * 2- The local time is UTC+1 (Central European Standard Time):
 * oi18n.Utils. getLocalTimeZoneOffset() will return the string "Etc/GMT-01:00"  
 * @returns {string}  
 */
oi18n.Utils.getLocalTimeZoneOffset = function () {
  var d = new Date();
  var offset = d.getTimezoneOffset();
  return oi18n.Utils.getTimeStringFromOffset("Etc/GMT", offset, false, false);
};


/*
 * Will return timezone if it exists.
 */
oi18n.Utils._getTimeZone = function (isoString)
{
  if (!isoString || typeof isoString !== "string")
  {
    return null;
  }
  var match = oi18n.Utils._ISO_DATE_REGEXP.exec(isoString);
  //make sure it is iso string
  if (match === null)
  {
    oi18n.Utils._throwInvalidISOString(isoString);
  }
  if (match[1] !== undefined)
    return match[1];
  return null;
};

/*
 * Will return local isoString provided a date.
 */
oi18n.Utils.dateToLocalIso = function (date)
{
  var isoStr = oi18n.Utils.padZeros(date.getFullYear(), 4) + "-" + oi18n.Utils.padZeros((date.getMonth() + 1), 2) + "-" + oi18n.Utils.padZeros(date.getDate(), 2) + "T" +
      oi18n.Utils.padZeros((date.getHours()), 2) + ":" + oi18n.Utils.padZeros((date.getMinutes()), 2) + ":" +
      oi18n.Utils.padZeros((date.getSeconds()), 2);
  if (date.getMilliseconds() > 0) {
    isoStr += "." + oi18n.Utils.trimRightZeros(oi18n.Utils.padZeros(date.getMilliseconds(), 3));
  }
  return isoStr;
};

oi18n.Utils.partsToIsoString = function (parts)
{
  var isoStr = oi18n.Utils.padZeros(parts[0], 4) + "-" + oi18n.Utils.padZeros(parts[1], 2) + "-" + oi18n.Utils.padZeros(parts[2], 2) + "T" +
      oi18n.Utils.padZeros(parts[3], 2) + ":" + oi18n.Utils.padZeros(parts[4], 2) + ":" + oi18n.Utils.padZeros(parts[5], 2);
  if (parts[6] > 0) {
    isoStr += "." + oi18n.Utils.trimRightZeros(oi18n.Utils.padZeros(parts[6], 3));
  }
  return isoStr;
};

oi18n.Utils.isoToLocalDate = function (isoString)
{
  if (!isoString || typeof isoString !== "string")
  {
    return null;
  }
  return this._isoToLocalDateIgnoreTimezone(isoString);
};

oi18n.Utils._isoToLocalDateIgnoreTimezone = function (isoString) {
  var datetime = oi18n.Utils._IsoStrParts(isoString);
  return new Date(datetime[0], datetime[1] - 1, datetime[2], datetime[3], datetime[4], datetime[5], datetime[6]);
};

oi18n.Utils._IsoStrParts = function (isoString) {
  var splitted = isoString.split("T"),
      tIndex = isoString.indexOf("T"),
      today = new Date(), i,
      datetime = [today.getFullYear(), today.getMonth() + 1, today.getDate(), 0, 0, 0, 0];

  if (splitted[0] !== "")
  {
    //contains date portion
    var dateSplitted = splitted[0].split("-");
    for (i = 0; i < dateSplitted.length; i++) {
      datetime[i] = parseInt(dateSplitted[i], 10);
    }
  }

  if (tIndex !== -1) {
    var milliSecSplitted = splitted[1].split("."), //contain millseconds
        timeSplitted = milliSecSplitted[0].split(":"); //contain hours, minutes, seconds

    for (i = 0; i < timeSplitted.length; i++)
    {
      datetime[3 + i] = parseInt(timeSplitted[i], 10);
    }

    if (milliSecSplitted.length === 2 && milliSecSplitted[1])
    {
      datetime[6] = parseInt(oi18n.Utils.zeroPad(milliSecSplitted[1], 3, false), 10);
    }
  }
  return datetime;
};

oi18n.Utils.getISOStrFormatInfo = function (isoStr) {
  var res = {
    'format': null,
    'dateTime': null,
    'timeZone': "",
    'isoStrParts': null
  };
  var exe = oi18n.Utils._ISO_DATE_REGEXP.exec(isoStr);
  if (exe === null) {
    oi18n.Utils._throwInvalidISOString(isoStr);
  }
  if (exe[1] === undefined && exe[2] === undefined) {
    res['format'] = 'local';
    res['dateTime'] = isoStr;
    res['isoStrParts'] = oi18n.Utils._IsoStrParts(res['dateTime']);
    return res;
  }
  res['timeZone'] = (exe[1] !== undefined) ? exe[1] : exe[2];
  if (res['timeZone'] === 'Z')
    res['format'] = 'zulu';
  else
    res['format'] = 'offset';
  res['dateTime'] = isoStr.substring(0, isoStr.indexOf(res['timeZone']));
  res['isoStrParts'] = oi18n.Utils._IsoStrParts(res['dateTime']);
  return res;
};

oi18n.Utils._throwTimeZoneNotSupported = function () {
  var msg, error, errorInfo;
  msg = "time zone is not supported";
  error = new Error(msg);
  errorInfo = {
    'errorCode': 'timeZoneNotSupported'
  };
  error['errorInfo'] = errorInfo;
  throw error;
};

oi18n.Utils._throwInvalidISOString = function (str) {
  var msg, error, errorInfo;
  msg = "The string " + str + " is not a valid ISO 8601 string.";
  error = new Error(msg);
  errorInfo = {
    'errorCode': 'invalidISOString',
    'parameterMap': {
      'isoStr': str
    }
  };
  error['errorInfo'] = errorInfo;
  throw error;
};

oi18n.Utils.trim = function (value) {
  return (value + "").replace(oi18n.Utils.regexTrim, "");
};

oi18n.Utils.trimRightZeros = function (value) {
  return (value + "").replace(oi18n.Utils.regexTrimRightZeros, "");
};


oi18n.Utils.trimNumber = function (value) {
  var s = (value + "").replace(oi18n.Utils.regexTrimNumber, "");
  return s;
};

oi18n.Utils.startsWith = function (value, pattern) {
  return value.indexOf(pattern) === 0;
};

oi18n.Utils.toUpper = function (value) {
  // "he-IL" has non-breaking space in weekday names.
  return value.split("\u00A0").join(" ").toUpperCase();
};

oi18n.Utils.padZeros = function (num, c) {
  var r, s = num + "";
  if (c > 1 && s.length < c) {
    r = (oi18n.Utils.zeros[c - 2] + s);
    return r.substr(r.length - c, c);
  }
  else {
    r = s;
  }
  return r;
};

oi18n.Utils.zeroPad = function (str, count, left) {
  str = "" + str;
  var l;
  for (l = str.length; l < count; l += 1) {
    str = (left ? ("0" + str) : (str + "0"));
  }
  return str;
};

oi18n.Utils.getTimeStringFromOffset = function (prefix, offset, reverseSign, alwaysMinutes) {
  var isNegative = reverseSign ? offset >= 0 : offset < 0;
  offset = Math.abs(offset);
  var hours = (offset / 60) << 0;
  var minutes = offset % 60;
  var sign = isNegative ? "-" : "+";
  if (alwaysMinutes) {
    hours = oi18n.Utils.zeroPad(hours, 2, true);
  }
  var str = prefix + sign + hours;
  if (minutes > 0 || alwaysMinutes) {
    str += ":" + oi18n.Utils.zeroPad(minutes, 2, true);
  }
  return str;
};

//get the numbering system key from the locale's unicode extension.
//Verify that the locale data has a numbers entry for it, if not return latn as default.
oi18n.Utils.getNumberingSystemKey = function (localeElements, locale) {
  if (locale === undefined)
    return 'latn';
  var numberingSystemKey = oi18n.Utils.getNumberingExtension(locale);
  var symbols = "symbols-numberSystem-" + numberingSystemKey;
  if (localeElements['numbers'][symbols] === undefined)
    numberingSystemKey = 'latn';
  return numberingSystemKey;
};

//return the language part
oi18n.Utils.getBCP47Lang = function (tag) {
  var arr = tag.split("-");
  return arr[0];
};

//return the region part. tag is lang or lang-region or lang-script or
//lang-script-region
oi18n.Utils.getBCP47Region = function (tag) {
  var arr = tag.split("-");
  if (arr.length === 3)
    return arr[2];
  if (arr.length === 2) {
    if (arr[1].length === 2)
      return arr[1];
  }
  return '001';
};


//get the unicode numbering system extension.
oi18n.Utils.getNumberingExtension = function (locale) {
  locale = locale || "en-US";
  var idx = locale.indexOf("-u-nu-");
  var numbering = 'latn';
  if (idx !== -1) {
    numbering = locale.substr(idx + 6, 4);
  }
  return numbering;
};

oi18n.Utils.haveSamePropertiesLength = function (obj) {
  var count = 0;
  for (var n in obj)
  {
    count++;
  }
  return count;
};

//cldr locale data start with "main" node.
//return the subnode under main.
oi18n.Utils.getLocaleElementsMainNode = function (bundle) {
  var mainNode = bundle['main'];
  var subnode;
  for (var n in mainNode)
  {
    subnode = n;
    break;
  }
  return mainNode[subnode];
};

//get the locale which is a subnode of "main".
oi18n.Utils.getLocaleElementsMainNodeKey = function (bundle) {
  var mainNode = bundle['main'];
  var subnode;
  for (var n in mainNode)
  {
    subnode = n;
    break;
  }
  return subnode;
};

oi18n.Utils._toBoolean = function (value) {
  if (typeof value === "string") {
    var s = value.toLowerCase().trim();
    switch (s) {
      case "true":
      case "1":
        return true;
      case "false":
      case "0":
        return false;
      default:
        return value;
    }
  }
  return value;
};
//Return a function getOption.
//The getOption function extracts the value of the property named 
//property from the provided options object, converts it to the required type,
// checks whether it is one of a List of allowed values, and fills in a 
// fallback value if necessary.
oi18n.Utils.getGetOption = function (options, getOptionCaller) {
  if (options === undefined) {
    throw new Error('Internal ' + getOptionCaller +
        ' error. Default options missing.');
  }

  var getOption = function getOption(property, type, values, defaultValue) {
    if (options[property] !== undefined) {
      var value = options[property];
      switch (type) {
        case 'boolean':
          value = oi18n.Utils._toBoolean(value);
          break;
        case 'string':
          value = String(value);
          break;
        case 'number':
          value = Number(value);
          break;
        default:
          throw new Error('Internal error. Wrong value type.');
      }
      if (values !== undefined && values.indexOf(value) === -1) {
        var expectedValues = [];
        for (var i = 0; i < values.length; i++) {
          expectedValues.push(values[i]);
        }
        var msg = "The value '" + options[property] +
            "' is out of range for '" + getOptionCaller +
            "' options property '" + property + "'. Valid values: " +
            expectedValues;
        var rangeError = new RangeError(msg);
        var errorInfo = {
          'errorCode': 'optionOutOfRange',
          'parameterMap': {
            'propertyName': property,
            'propertyValue': options[property],
            'propertyValueValid': expectedValues,
            'caller': getOptionCaller
          }
        };
        rangeError['errorInfo'] = errorInfo;
        throw rangeError;
      }

      return value;
    }
    return defaultValue;
  };

  return getOption;
};

/**
 * matches a string to a reference string and returns the start and end indexes
 * of the match in the referensed string. The locale and options arguments let 
 * applications specify the language whose sort order should be used and customize 
 * the behavior of the function.
 * 
 * @param {string} str the reference string 
 * @param {string} pat The string against which the reference string is compared
 * @param {string} locale a BCP 47 language tag
 * @param {Object=} options Optional. An object with the following property:
 * sensitivity: 
 *   Which differences in the strings should lead to non-zero result values. Possible values are:
 *   "base": Only strings that differ in base letters compare as unequal. Examples: a ? b,  a = A.
 *   "accent": Only strings that differ in base letters or accents and other diacritic marks compare as unequal. Examples: a ? b, , a = A.
 *   "case": Only strings that differ in base letters or case compare as unequal. Examples: a ? b, a ? A.
 *   "variant": Strings that differ in base letters, accents and other diacritic marks, or case compare as unequal. 
 *   The default is base.
 * @return {Array|null} an array containing the start and end indexes of the match or null if there is no match.
 */
oi18n.Utils.matchString = function (str, pat, locale, options) {
  if (options === undefined) {
    options = {sensitivity: 'base', usage: 'sort'};
  }
  var getOption = oi18n.Utils.getGetOption(options, "oi18n.Utils.matchString");
  options['usage'] = getOption('usage', 'string', ['sort', 'search'], 'sort');
  options['sensitivity'] = getOption('sensitivity', 'string', ['base', 'accent', 'case', 'variant'], 'base');
  var i, j;
  var len = str.length;
  var patLen = pat.length - 1;
  for (i = 0; i < len; i++) {
    for (j = 0; j < 3; j++) {
      var len2 = len - i;
      len2 = Math.min(len2, patLen + j);
      var str2 = str.substr(i, len2);
      var res = str2.localeCompare(pat, locale, options);
      if (res === 0) {
        var end = i + len2 - 1;
        var ret = [i, end];
        return ret;
      }
    }
  }
  return null;
};

var _DEFAULT_TIME_PORTION = "T00:00:00.000";
var _DATE_TIME_KEYS = {"fullYear": {pos: 0, pad: 4}, "month": {pos: 1, pad: 2}, "date": {pos: 2, pad: 2},
  "hours": {pos: 3, pad: 2}, "minutes": {pos: 4, pad: 2}, "seconds": {pos: 5, pad: 2},
  "milliseconds": {pos: 6, pad: 3}, "timeZone": {pos: 7}};

/**
 * Parses the isoString and returns a JavaScript Date object
 * 
 * @param {string} isoString isoString to parse and to return Date of
 * @return {Date} the parsed JavaScript Date Object
 */
oi18n.Utils.isoToDate = function (isoString)
{
  //note new Date w/ isoString in IE fails so need to use parsing from momentjs support
  return new Date(this._normalizeIsoString(isoString));
};

/**
 * Will return an updated toIsoString using the timePortion from the fromIsoString or from the default 
 * oi18n.Utils.DEFAULT_TIME_PORTION
 * 
 * @private
 * @param {string} fromIsoString isoString that may not be a complete isoString
 * @param {string} toIsoString isoString that may not be a complete isoString
 * @returns {string} modified toIsoString with original date portion and the time portion from the fromIsoString
 * @since 1.1
 */
oi18n.Utils._copyTimeOver = function (fromIsoString, toIsoString)
{
  if (!fromIsoString || !toIsoString)
  {
    throw new Error("Provided invalid arguments");
  }

  //need to only normalize toIsoString, since copying only time from fromIsoString
  toIsoString = this._normalizeIsoString(toIsoString);

  var fromTimeIndex = fromIsoString.indexOf("T"),
      toTimeIndex = toIsoString.indexOf("T"),
      toDatePortion = toIsoString.substring(0, toTimeIndex),
      fromTimePortion = fromTimeIndex !== -1 ? fromIsoString.substring(fromTimeIndex) : _DEFAULT_TIME_PORTION;

  return toDatePortion + fromTimePortion;
};

/**
 * Clears the time portion of the isoString
 * 
 * @private
 * @param {string} isoString isoString that may not be a complete isoString
 * @returns {string} an updated isoString
 * @since 1.1
 */
oi18n.Utils._clearTime = function (isoString)
{
  return this._dateTime(isoString, {"hours": 0, "minutes": 0, "seconds": 0, "milliseconds": 0});
};

/**
 * Will accept an isoString and perform a get operation or a set operation depending on whether param is an Array 
 * or a JSON 
 * 
 * The keys for the get and set operation are defined in _DATE_TIME_KEYS.
 * 
 * Note the handling of month starting with 0 in Date object and being 1 based in isoString will be handled by the function 
 * with the usage of doParseValue. Meaning when you doParseValue and you are getting the value it will automatically 
 * decrement the value and when you are setting the param it will check if the value is of number and if so will 
 * increment it.
 * 
 * @private
 * @param {string} isoString isoString that may not be a complete isoString
 * @param {Array|Object} actionParam if an Array will be a get operation, if a JSON will be a set operation
 * @param {boolean=} doParseValue whether one should parseInt the value during the get request
 * @returns {Object|string} an Object when a get operation and a string when a set operation
 * @since 1.1
 */
oi18n.Utils._dateTime = function (isoString, actionParam, doParseValue)
{
  if (!isoString || !actionParam)
  {
    throw new Error("Invalid argument invocation");
  }

  var pos, value, dtKey,
      retVal = null,
      key = null,
      dateTimeKeys = _DATE_TIME_KEYS,
      UtilsPadZero = this.padZeros,
      isoStringNormalized = this._normalizeIsoString(isoString), //note intentionally normalizing 
      captured = /(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):?(\d{2})?\.?(\d{3})?(.*)?/.exec(isoStringNormalized);

  if (!captured)
  {
    throw new Error("Unable to capture anything");
  }

  captured = captured.slice(1);

  if (Array.isArray(actionParam))
  {
    retVal = {};

    //means an array so perform a get operation
    for (var i = 0, len = actionParam.length; i < len; i++)
    {
      key = actionParam[i];

      if (key in dateTimeKeys)
      {
        pos = dateTimeKeys[key].pos;
        value = captured[pos];

        if (doParseValue && "timeZone" === key)
        {
          throw new Error("Dude you tried to ask timezone to be parsed");
        }

        if (doParseValue)
        {
          var parsed = parseInt(value, 10);
          retVal[key] = pos === 1 ? parsed - 1 : parsed; //since month is 0 based, though about having a callback but month only special
        }
        else
        {
          retVal[key] = value;
        }

      }
    }
  }
  else if ($.isPlainObject(actionParam))
  {

    for (var keys in actionParam)
    {
      dtKey = dateTimeKeys[keys];
      pos = dtKey.pos;
      value = actionParam[keys];

      //special case for month again, 0 based so check if number and if so increment it
      if (pos === 1 && typeof value === "number")
      {
        value++;
      }
      captured[pos] = dtKey.pad ? UtilsPadZero(value, dtKey.pad) : value;
    }
    //"2015-02-02T21:12:30.255Z"
    retVal = captured[0] + "-" + captured[1] + "-" + captured[2] + "T" + captured[3] + ":" + captured[4] + ":" + captured[5] +
        (captured.length > 6 && captured[6] ? ("." + captured[6] + (captured.length === 8 && captured[7] ? captured[7] : "")) : "");
  }

  return retVal;
};

/**
 * So the problem is Jet uses incomplete isoString which causes issues in different browsers. 
 * 
 * For instance for a new Date().toISOString() => 2015-02-02T18:00:37.007Z
 * ojInputDate stores 2015-02-02
 * ojInputTime stores T18:00:37.007Z
 * 
 * yet constructing new Date(val) on above causes different results or errors in different browsers, so 
 * this function is to normalize them. Note it is assumed that the point is creating the Date object from the 
 * normalized isoString. Meaning if both contain only the time portion today's date will appended to it.
 * 
 * Here are the use cases
 * 
 * @private
 * @param {string} isoString isoString that may not be a complete isoString
 * @returns {string} a normalized isoString
 * @since 1.1
 */
oi18n.Utils._normalizeIsoString = function (isoString)
{
  if (!isoString)
  {
    throw new Error("Provided invalid arguments");
  }

  var checkTime = function (timeValue)
  {
    var splitted = timeValue.split(":");
    if (splitted.length > 1)
    {
      return timeValue;
    }
    //need at least hour + minute for proper parsing on browser except IE
    return timeValue + ":00";
  },
      todayIsoString = new Date().toISOString(),
      todayDatePortion = todayIsoString.substring(0, todayIsoString.indexOf("T")),
      timeIndex = isoString.indexOf("T"),
      datePortion = timeIndex === -1 ? isoString : isoString.substring(0, timeIndex),
      timePortion = timeIndex !== -1 ? checkTime(isoString.substring(timeIndex)) : _DEFAULT_TIME_PORTION;

  datePortion = datePortion || todayDatePortion;

  return datePortion + timePortion;
};

/**
 * formats a parameterized string
 * 
 * @param {string} str string to be formatted. EX: "{0} bytes"
 * @param {Array} params array of parameters to be substituted in the string
 * @returns {string} The formatted string
 * @since 4.0.0
 */
oi18n.Utils.formatString = function (str, params)
{
  var i;
  var token;
  var len = params.length;
  
  for (i = 0; i < len; i++)
  {
    token = '{' + i + '}';
    str = str.replace(token, params[i]);
  }
  return str;
};

/**
 * Returns a person's initials
 * 
 * @param {string} firstName first name
 * @param {string} lastName last name or surname
 * @returns {string} uppercase concatenation of first letter of first name and first letter
 * of last name.
 * There are the following special cases:
 * - If the name is Arabic characters, it returns empty string.
 * - If the name is Hindi characters, it returns the first letter of the first name.
 * - If the name is Thai characters, it returns the first letter of the first name.
 * - If the name is Korean characters, it returns the first name.
 * - If the name is Japanese or Chinese characters, it returns the last name.
 *   
 * @since 4.0.0
 */
oi18n.Utils.getInitials = function (firstName, lastName)
{
  //We assume the names are valid. We test the first character only.
  var c = 0; 
  var c1 = 0;
  var u;
  if(firstName !== undefined && firstName.length > 0)
    c = firstName.charCodeAt(0);
  
  // Arabic characters. Return empty string
  if(c >= 0x0600 && c <= 0x06FF)
    return "";
  
  // Hindi characters. Return first letter of the first name
  if(c >= 0x0900 && c <= 0x097F)
    return firstName.charAt(0);
  
  // Thai characters. Return first letter of the first name
  if(c >= 0x0E00 && c <= 0x0E7F)
    return firstName.charAt(0);
  
  // Korean characters. Return first name
  if((c >= 0x1100 && c <= 0x11FF) || (c >= 0x3130 && c <= 0x318F) || 
     (c >= 0xA960 && c <= 0xA97F) || (c >= 0xAC00 && c <= 0xD7FF))
    return firstName;
  
  // Japanese and Chinese characters. Return last name
  if(lastName !== undefined && lastName.length > 0)
    c = lastName.charCodeAt(0);
  if((c >= 0x2E80 && c <= 0x2FDF) || (c >= 0x3000 && c <= 0x312F) ||
     (c >= 0x3190 && c <= 0x31FF) || (c >= 0x3300 && c <= 0x4DBF) ||
     (c >= 0x4E00 && c <= 0x9FFF) || (c >= 0xF900 && c <= 0xFAFF))
     return lastName;
   
  //Handle surrogate characters for Japanese and Chinese characters.  
  if(c >= 0xD800 && c <= 0xDBFF) {
    if(lastName.length < 2)
      return "";
    c1 = lastName.charCodeAt(1);
    //c1 must be in DC00-DFFF range
    if(c1 < 0xDC00 || c1 > 0xDFFF)
      return "";
    //Convert high and low surrogates into unicode scalar.
    u =  ((c - 0xD800) * 0x400) + (c1 - 0xDC00) + 0x10000;
    //test the blocks
    if((u >= 0x1B000 && u <= 0x1B0FF) || (u >= 0x1F200 && u <= 0x1F2FF) ||
       (u >= 0x20000 && u <= 0x2A6DF) || (u >= 0x2A700 && u <= 0x2B73F) ||
       (u >= 0x2B740 && u <= 0x2B81F) || (u >= 0x2B820 && u <= 0x2CEAF) ||
       (u >= 0x2F800 && u <= 0x2FA1F))
      return lastName;
  }
  
  //return default
  c = "";
  c1 = "";
  if(firstName !== undefined && firstName.length > 0)
    c = firstName.charAt(0).toUpperCase();
  if(lastName !== undefined && lastName.length > 0)
    c1 = lastName.charAt(0).toUpperCase();
  return c + c1;
};

